@@ -23,9 +23,7 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 */
 
-#import <Foundation/NSArray.h>
-#import <Foundation/NSException.h>
-#include "AppKit/GMAppKit.h"
+#include "GNUstepGUI/GMAppKit.h"
 
 #ifndef AUTORELEASE
 #define AUTORELEASE(object)	[object autorelease]
@@ -158,7 +156,7 @@ void __dummy_GMAppKit_functionForLinking() {}
   [self setTitlePosition:[unarchiver decodeIntWithName:@"titlePosition"]];
   [self setTitle:[unarchiver decodeStringWithName:@"title"]];
   [self setTitleFont:[unarchiver decodeObjectWithName:@"titleFont"]];
-  [self setContentView:[unarchiver decodeObjectWithName:@"contentView"]];
+  [self setContentView: [unarchiver decodeObjectWithName:@"contentView"]];
 
   return self;
 }
@@ -230,70 +228,78 @@ void __dummy_GMAppKit_functionForLinking() {}
 
 - (void)encodeWithModelArchiver:(GMArchiver*)archiver
 {
-    [archiver encodeInt:[self type] withName:@"type"];
-    [archiver encodeObject:[self font] withName:@"font"];
-    [archiver encodeString:[self stringValue] withName:@"stringValue"];
-    [archiver encodeInt:[self entryType] withName:@"entryType"];
-    [archiver encodeInt:[self alignment] withName:@"alignment"];
-    [archiver encodeBOOL:[self wraps] withName:@"wraps"];
-    [archiver encodeObject:[self image] withName:@"image"];
-    [archiver encodeInt:[self state] withName:@"state"];
-    [archiver encodeBOOL:[self isEnabled] withName:@"isEnabled"];
-    [archiver encodeBOOL:[self isBordered] withName:@"isBordered"];
-    [archiver encodeBOOL:[self isBezeled] withName:@"isBezeled"];
-    [archiver encodeBOOL:[self isEditable] withName:@"isEditable"];
-    [archiver encodeBOOL:[self isSelectable] withName:@"isSelectable"];
-    [archiver encodeBOOL:[self isScrollable] withName:@"isScrollable"];
-    [archiver encodeBOOL:[self isContinuous] withName:@"isContinuous"];
-    [archiver encodeInt:[self sendActionOn:0] withName:@"sendActionMask"];
+  [archiver encodeInt:[self type] withName:@"type"];
+  [archiver encodeObject:[self font] withName:@"font"];
+  [archiver encodeString:[self stringValue] withName:@"stringValue"];
+  [archiver encodeInt:[self entryType] withName:@"entryType"];
+  [archiver encodeInt:[self alignment] withName:@"alignment"];
+  [archiver encodeBOOL:[self wraps] withName:@"wraps"];
+  [archiver encodeObject:[self image] withName:@"image"];
+  [archiver encodeInt:[self state] withName:@"state"];
+  [archiver encodeBOOL:[self isEnabled] withName:@"isEnabled"];
+  [archiver encodeBOOL:[self isBordered] withName:@"isBordered"];
+  [archiver encodeBOOL:[self isBezeled] withName:@"isBezeled"];
+  [archiver encodeBOOL:[self isEditable] withName:@"isEditable"];
+  [archiver encodeBOOL:[self isSelectable] withName:@"isSelectable"];
+  [archiver encodeBOOL:[self isScrollable] withName:@"isScrollable"];
+  [archiver encodeBOOL:[self isContinuous] withName:@"isContinuous"];
+  [archiver encodeInt:[self sendActionOn: 0] withName:@"sendActionMask"];
+  {
+    /* NB: this is not decoded so maybe we could just do without
+       encoding it. :-) */
+    int actionMask = [self sendActionOn: 0];
+    [archiver encodeInt:actionMask withName:@"sendActionMask"];
+    [self sendActionOn: actionMask];
+  }
 }
 
 - (id)initWithModelUnarchiver:(GMUnarchiver*)unarchiver
 {
-    int cellType = [unarchiver decodeIntWithName:@"type"];
-    NSFont* font = [unarchiver decodeObjectWithName:@"font"];
-
-    // this is a tricky object to decode, because a number of its methods
-    // have side-effects; [-setEntryType:] converts the cell to a text-type
-    // cell and sets its font to the system font, so it comes first
-    [self setEntryType:[unarchiver decodeIntWithName:@"entryType"]];
-
-    // now set the font
-    [self setFont:font];
-
-    // both [-setImage:] and [-setStringValue:] convert the cell to an
-    // image or text cell (respectively), so they must be called in the
-    // correct order for the type of cell desired
-    switch (cellType)
-      {
-        case NSTextCellType:
-          [self setImage:[unarchiver decodeObjectWithName:@"image"]];
-          [self setStringValue:
+  int cellType = [unarchiver decodeIntWithName:@"type"];
+  NSFont* font = [unarchiver decodeObjectWithName:@"font"];
+  
+  // this is a tricky object to decode, because a number of its methods
+  // have side-effects; [-setEntryType:] converts the cell to a text-type
+  // cell and sets its font to the system font, so it comes first
+  [self setEntryType:[unarchiver decodeIntWithName:@"entryType"]];
+  
+  // now set the font
+  [self setFont:font];
+  
+  // both [-setImage:] and [-setStringValue:] convert the cell to an
+  // image or text cell (respectively), so they must be called in the
+  // correct order for the type of cell desired
+  switch (cellType)
+    {
+    case NSTextCellType:
+      [self setImage:[unarchiver decodeObjectWithName:@"image"]];
+      [self setStringValue: 
+	      [unarchiver decodeStringWithName:@"stringValue"]];
+      break;
+    case NSImageCellType:
+      [self setStringValue:
               [unarchiver decodeStringWithName:@"stringValue"]];
-          break;
-        case NSImageCellType:
-          [self setStringValue:
-              [unarchiver decodeStringWithName:@"stringValue"]];
-          [self setImage:[unarchiver decodeObjectWithName:@"image"]];
-          break;
-        case NSNullCellType:
-          [self setType: NSNullCellType];
-          break;
-      }
-    [self setAlignment:[unarchiver decodeIntWithName:@"alignment"]];
-    [self setWraps:[unarchiver decodeBOOLWithName:@"wraps"]];
-    [self setState:[unarchiver decodeIntWithName:@"state"]];
-    [self setEnabled:[unarchiver decodeBOOLWithName:@"isEnabled"]];
-    [self setBordered:[unarchiver decodeBOOLWithName:@"isBordered"]];
-    [self setBezeled:[unarchiver decodeBOOLWithName:@"isBezeled"]];
-    [self setEditable:[unarchiver decodeBOOLWithName:@"isEditable"]];
-    [self setSelectable:[unarchiver decodeBOOLWithName:@"isSelectable"]];
-    [self setScrollable:[unarchiver decodeBOOLWithName:@"isScrollable"]];
-    [self setContinuous:[unarchiver decodeBOOLWithName:@"isContinuous"]];
-    /* Temporary commented out so buttons keep on working - new code fixing this under testing */
-    //    [self sendActionOn:[unarchiver decodeIntWithName:@"sendActionMask"]];
-
-    return self;
+      [self setImage:[unarchiver decodeObjectWithName:@"image"]];
+      break;
+    case NSNullCellType:
+      [self setType: NSNullCellType];
+      break;
+    }
+  [self setAlignment:[unarchiver decodeIntWithName:@"alignment"]];
+  [self setWraps:[unarchiver decodeBOOLWithName:@"wraps"]];
+  [self setState:[unarchiver decodeIntWithName:@"state"]];
+  [self setEnabled:[unarchiver decodeBOOLWithName:@"isEnabled"]];
+  [self setBordered:[unarchiver decodeBOOLWithName:@"isBordered"]];
+  [self setBezeled:[unarchiver decodeBOOLWithName:@"isBezeled"]];
+  [self setEditable:[unarchiver decodeBOOLWithName:@"isEditable"]];
+  [self setSelectable:[unarchiver decodeBOOLWithName:@"isSelectable"]];
+  [self setScrollable:[unarchiver decodeBOOLWithName:@"isScrollable"]];
+  [self setContinuous:[unarchiver decodeBOOLWithName:@"isContinuous"]];
+  /* Temporary commented out so buttons keep on working - new code
+   * fixing this under testing */
+  //    [self sendActionOn:[unarchiver decodeIntWithName:@"sendActionMask"]];
+  
+  return self;
 }
 
 @end /* NSCell (GMArchiverMethods) */
@@ -445,6 +451,9 @@ void __dummy_GMAppKit_functionForLinking() {}
     nc = [unarchiver decodeIntWithName:@"numberOfColumns"];
     cell_array = [unarchiver decodeObjectWithName:@"cells"];
     [self renewRows:nr columns:nc];
+#if GNU_GUI_LIBRARY
+    _selectedRow = _selectedColumn = 0;
+#endif
     for (i = 0; (i < [cell_array count]) && (i < nr*nc); i++)
     {
         id	cell = [cell_array objectAtIndex:i];
@@ -579,16 +588,12 @@ void __dummy_GMAppKit_functionForLinking() {}
     [archiver encodeFloat:[self brightnessComponent] withName:@"brightness"];
   }
   else if ([colorSpaceName isEqual:@"NSNamedColorSpace"]) {
-    // TODO: change it when NSColor in GNUstep will have named color lists
-#if 1
-    NSColor* new
-	= [self colorUsingColorSpaceName:@"NSCalibratedRGBColorSpace"];
-    [new encodeWithModelArchiver:archiver];
-#else
-    [unarchiver encodeString:[self catalogNameComponent]
+    [archiver encodeString:[self catalogNameComponent]
 		withName:@"catalogName"];
-    [unarchiver encodeString:[self colorNameComponent] withName:@"colorName"];
-#endif
+    [archiver encodeString:[self colorNameComponent] withName:@"colorName"];
+  }
+  else if ([colorSpaceName isEqual:@"NSPatternColorSpace"]) {
+    [archiver encodeObject: [self patternImage] withName: @"patternImage"];
   }
 }
 
@@ -639,8 +644,27 @@ void __dummy_GMAppKit_functionForLinking() {}
     return [self colorWithCalibratedRed:red green:green blue:blue alpha:alpha];
   }
   else if ([colorSpaceName isEqual:@"NSNamedColorSpace"]) {
-    NSAssert (0, @"Named color spaces not supported yet!");
+    NSString *catalog = [unarchiver decodeStringWithName: @"catalogName"];
+    NSString *colornm = [unarchiver decodeStringWithName: @"colorName"];
+    return [self colorWithCatalogName: catalog colorName: colornm];
   }
+  else if ([colorSpaceName isEqual:@"NSPatternColorSpace"]) {
+    NSImage *image = [unarchiver decodeObjectWithName: @"patternImage"];
+    if (image == nil)
+      {
+	NSLog(@"Internal: No can't decode colorspace %@", colorSpaceName);
+	NSLog(@"          creating generic white color");
+	return [NSColor colorWithDeviceWhite: 1.0  alpha: 1.0];
+      }
+    else
+      return [NSColor colorWithPatternImage: image];
+  }
+  else
+    {
+      NSLog(@"Internal: No decoder for colorspace %@", colorSpaceName);
+      NSLog(@"          creating generic white color");
+      return [NSColor colorWithDeviceWhite: 1.0  alpha: 1.0];
+    }
   return nil;
 }
 
@@ -715,8 +739,11 @@ void __dummy_GMAppKit_functionForLinking() {}
 
 + (id)createObjectForModelUnarchiver:(GMUnarchiver*)unarchiver
 {
-  return [NSFont fontWithName:[unarchiver decodeStringWithName:@"name"]
+  NSFont *f = [NSFont fontWithName:[unarchiver decodeStringWithName:@"name"]
 		 size:[unarchiver decodeFloatWithName:@"size"]];
+  if (!f)
+    f = [NSFont systemFontOfSize: [unarchiver decodeFloatWithName:@"size"]];
+  return f;
 }
 
 - (id)initWithModelUnarchiver:(GMUnarchiver*)unarchiver
@@ -779,15 +806,21 @@ void __dummy_GMAppKit_functionForLinking() {}
     }
   [archiver encodeString:[self keyEquivalent] withName:@"keyEquivalent"];
   if ([self respondsToSelector: @selector(state)])
-    [archiver encodeInt:[self state] withName:@"state"];
+    {
+      [archiver encodeInt:[self state] withName:@"state"];
+    }
   [archiver encodeObject:[self target] withName:@"target"];
   [archiver encodeSelector:[self action] withName:@"action"];
   [archiver encodeInt:[self tag] withName:@"tag"];
   [archiver encodeBOOL:[self isEnabled] withName:@"isEnabled"];
   if ([self respondsToSelector: @selector(changesState)])
-    [archiver encodeBOOL:[self changesState] withName:@"changesState"];
+    {
+      [archiver encodeBOOL:[self changesState] withName:@"changesState"];
+    }
   if ([self respondsToSelector: @selector(submenu)])
-    [archiver encodeObject:[self submenu] withName:@"submenu"];
+    {
+      [archiver encodeObject:[self submenu] withName:@"submenu"];
+    }
   [archiver encodeConditionalObject:[self representedObject]
 	                   withName:@"representedObject"];
 }
@@ -803,6 +836,7 @@ void __dummy_GMAppKit_functionForLinking() {}
   [self setKeyEquivalent:[unarchiver decodeStringWithName:@"keyEquivalent"]];
   [self setState:[unarchiver decodeIntWithName:@"state"]];
   [self setAction:[unarchiver decodeSelectorWithName:@"action"]];
+  [self setTarget: [unarchiver decodeObjectWithName: @"target"]];
   [self setTag:[unarchiver decodeIntWithName:@"tag"]];
   [self setEnabled:[unarchiver decodeBOOLWithName:@"isEnabled"]];
   [self setChangesState:[unarchiver decodeBOOLWithName:@"changesState"]];
@@ -826,11 +860,18 @@ void __dummy_GMAppKit_functionForLinking() {}
 
   /* Safety assignment. */
   [self setMenu: nil];
-  /* Set submenu.  If submenu is nil, this will set target to nil too. */
-  [self setSubmenu: [unarchiver decodeObjectWithName: @"submenu"]];
-  /* Set target.  This does not touch submenu. */
-  [self setTarget: [unarchiver decodeObjectWithName: @"target"]];
 
+  /* Set submenu ... but not if it's nil, because this assignment
+   * always has side effects on action and target and we don't want to mess
+   * them up if we don't have a submenu!  */
+  {
+    id submenu = [unarchiver decodeObjectWithName: @"submenu"];
+    
+    if (submenu != nil)
+      {
+	[self setSubmenu: submenu];
+      }
+  }
   /* Now we fix the submenu from the target if needed: */
 #ifdef GNU_GUI_LIBRARY
   /*
@@ -838,14 +879,18 @@ void __dummy_GMAppKit_functionForLinking() {}
    * was created on OPENSTEP).
    */
   if ([NSStringFromSelector ([self action]) 
-      isEqualToString: @"submenuAction:"])
+			    isEqualToString: @"submenuAction:"])
     {
       if ([self submenu] == nil)
 	{
 	  if ([[self target] isKindOfClass: [NSMenu class]])
-	    [self setSubmenu: [self target]];
+	    {
+	      [self setSubmenu: [self target]];
+	    }
 	  else
-	    NSLog(@"Error decoding gmodel - submenu not an NSMenu");
+	    {
+	      NSLog(@"Error decoding gmodel - submenu not an NSMenu");
+	    }
 	}
     }
 #endif
@@ -915,6 +960,7 @@ void __dummy_GMAppKit_functionForLinking() {}
 
 - (void)encodeWithModelArchiver:(GMArchiver*)archiver
 {
+  [archiver encodeRect:     [self frame]     withName: @"frame"];
   [archiver encodeBOOL:     [self pullsDown] withName: @"pullsDown"];
   [archiver encodeBOOL:     [self isEnabled] withName: @"isEnabled"];
   [archiver encodeInt:      [self tag]       withName: @"tag"];
@@ -923,15 +969,21 @@ void __dummy_GMAppKit_functionForLinking() {}
   [archiver encodeInt:      [self autoresizingMask]
              withName:      @"autoresizingMask"];
   
-  /* OUCH! This code crashes the translator; probably we interfere somehow with
-     the way NSPopUpButton is handled by the NeXT's NIB code. Sorry, the
-     popup buttons cannot be handled by the convertor! */
+#if 0 //def __APPLE__
+  {
+    int i;
+    NSMutableArray *array;
+    array = [NSMutableArray arrayWithCapacity: [self numberOfItems]];
+    for (i = 0; i < [self numberOfItems]; i++)
+      [array addObject: [self itemAtIndex: i]];
+    [archiver encodeArray: array withName: @"itemArray"];
+  }
+#else
   [archiver encodeArray:[self itemArray] withName:@"itemArray"];
-  //[archiver encodeString:[self titleOfSelectedItem] withName:@"selectedItem"];
-  //[archiver encodeString:[self title] withName:@"selectedItem"];
-  //[super encodeWithModelArchiver:archiver];
+#endif
 
-  [archiver encodeRect: [self frame] withName:@"frame"];
+  //[archiver encodeString:[self titleOfSelectedItem] withName:@"selectedItem"];
+  [archiver encodeString:[self title] withName:@"selectedItem"];
 }
 
 + (id)createObjectForModelUnarchiver:(GMUnarchiver*)unarchiver
@@ -958,7 +1010,7 @@ void __dummy_GMAppKit_functionForLinking() {}
   
   decodedItems = [unarchiver decodeArrayWithName: @"itemArray"];
   
-  if (decodedItems) 
+  if (decodedItems && [decodedItems count]) 
     {
       count = [decodedItems count]; 
       for (i = 0; i < count; i++) 
@@ -974,10 +1026,15 @@ void __dummy_GMAppKit_functionForLinking() {}
 	  [myItem setTag:           [item tag]];
 	  [myItem setKeyEquivalent: [item keyEquivalent]];
 	}
+      string = [unarchiver decodeStringWithName: @"selectedItem"];
+      [self selectItemWithTitle: string];
+    }
+  else
+    {
+      /* For old gmodels that didn't support popups */
+      [self addItemWithTitle: @"Item 1"];
+      [self selectItemAtIndex: 0];
     }
-  
-  string = [unarchiver decodeStringWithName: @"selectedItem"];
-  [self selectItemWithTitle: string];
 
   [self setEnabled: [unarchiver decodeBOOLWithName:     @"isEnabled"]];
   [self setTag:     [unarchiver decodeIntWithName:      @"tag"]];
@@ -1127,7 +1184,10 @@ void __dummy_GMAppKit_functionForLinking() {}
   self = [super initWithModelUnarchiver:unarchiver];
 
   superview = [unarchiver decodeObjectWithName:@"superview"];
-  [superview addSubview:self];
+  /* If we are the content view of an NSBox, don't add ourselves to it.
+     This is done later in the NSBox unachiver method */
+  if ([superview isKindOfClass: [NSBox class]] == NO)
+    [superview addSubview:self];
 
   subviews = [unarchiver decodeObjectWithName:@"subviews"];
   for (i = 0, count = [subviews count]; i < count; i++)
@@ -1165,6 +1225,7 @@ void __dummy_GMAppKit_functionForLinking() {}
 {
   NSPoint wnOrigin = [self frame].origin;
   NSRect ctFrame = [[self contentView] frame];
+  unsigned int style;
 
   ctFrame.origin = wnOrigin;
 
@@ -1185,10 +1246,22 @@ void __dummy_GMAppKit_functionForLinking() {}
   [archiver encodeBOOL:[self hidesOnDeactivate]
 	    withName:@"hidesOnDeactivate"];
   [archiver encodeObject:[self backgroundColor] withName:@"backgroundColor"];
-  [archiver encodeUnsignedInt:[self styleMask] withName:@"styleMask"];
-  [archiver encodeUnsignedInt:[self backingType] withName:@"backingType"];
-  [archiver encodeConditionalObject:[self initialFirstResponder] 
-	    withName:@"initialFirstResponder"];
+
+  style = [self styleMask];
+#ifndef GNU_GUI_LIBRARY
+  /* Work around a bug in OpenStep, which doesn't set the
+   * NSTitledWindowMask properly.  If the window is not borderless,
+   * always add the title mask.  */
+  if (style != NSBorderlessWindowMask)
+    {
+      style |= NSTitledWindowMask;
+    }
+#endif
+
+  [archiver encodeUnsignedInt: style  withName: @"styleMask"];
+  [archiver encodeUnsignedInt: [self backingType]  withName: @"backingType"];
+  [archiver encodeConditionalObject: [self initialFirstResponder] 
+	    withName: @"initialFirstResponder"];
 }
 
 + (id)createObjectForModelUnarchiver:(GMUnarchiver*)unarchiver
