@@ -25,12 +25,15 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 */
 
-#include <AppKit/GSFontInfo.h>
+#include <math.h>
+
+#include "GNUstepGUI/GSFontInfo.h"
 #include <Foundation/NSDictionary.h>
 #include <Foundation/NSString.h>
 #include <Foundation/NSArray.h>
 #include <Foundation/NSSet.h>
 #include <Foundation/NSValue.h>
+#include <Foundation/NSException.h>
 
 static Class fontEnumeratorClass = Nil;
 static Class fontInfoClass = Nil;
@@ -61,6 +64,10 @@ static GSFontEnumerator *sharedEnumerator = nil;
 
 + (GSFontEnumerator*) sharedEnumerator
 {
+  NSAssert(fontEnumeratorClass,
+    @"Called with fontEnumeratorClass unset."
+    @" The shared NSApplication instance must be created before methods that"
+    @" need the backend may be called.");
   if (!sharedEnumerator)
     sharedEnumerator = [[fontEnumeratorClass alloc] init];
   return sharedEnumerator;
@@ -88,10 +95,28 @@ static GSFontEnumerator *sharedEnumerator = nil;
   return [allFontFamilies objectForKey: family];
 }
 
+
+- (NSString *) defaultSystemFontName
+{
+  return @"Helvetica";
+}
+
+- (NSString *) defaultBoldSystemFontName
+{
+  return @"Helvetica-Bold";
+}
+
+- (NSString *) defaultFixedPitchFontName
+{
+  return @"Courier";
+}
+
 @end
 
 @interface GSFontInfo (Backend)
--initWithFontName: (NSString *)fontName matrix: (const float *)fmatrix;
+-initWithFontName: (NSString *)fontName
+	   matrix: (const float *)fmatrix
+       screenFont: (BOOL)screenFont;
 @end
 
 @implementation GSFontInfo
@@ -102,10 +127,16 @@ static GSFontEnumerator *sharedEnumerator = nil;
 }
 
 + (GSFontInfo*) fontInfoForFontName: (NSString*)nfontName 
-                             matrix: (const float *)fmatrix;
+                             matrix: (const float *)fmatrix
+			 screenFont: (BOOL)screenFont;
 {
+  NSAssert(fontInfoClass,
+    @"Called with fontInfoClass unset."
+    @" The shared NSApplication instance must be created before methods that"
+    @" need the backend may be called.");
   return AUTORELEASE([[fontInfoClass alloc] initWithFontName: nfontName 
-                                                     matrix: fmatrix]);
+						      matrix: fmatrix
+						  screenFont: screenFont]);
 }
 
 + (int) weightForString: (NSString *)weightString
@@ -116,35 +147,35 @@ static GSFontEnumerator *sharedEnumerator = nil;
   if (dict == nil)
     {
       dict = [NSDictionary dictionaryWithObjectsAndKeys:
-			       [NSNumber numberWithInt: 1], @"ultralight",
-			   [NSNumber numberWithInt: 2], @"thin",
-			   [NSNumber numberWithInt: 3], @"light",
-			   [NSNumber numberWithInt: 3], @"extralight",
-			   [NSNumber numberWithInt: 4], @"book",
-			   [NSNumber numberWithInt: 5], @"regular",
-			   [NSNumber numberWithInt: 5], @"plain",
-			   [NSNumber numberWithInt: 5], @"display",
-			   [NSNumber numberWithInt: 5], @"roman",
-			   [NSNumber numberWithInt: 5], @"semilight",
-			   [NSNumber numberWithInt: 6], @"medium",
-			   [NSNumber numberWithInt: 7], @"demi",
-			   [NSNumber numberWithInt: 7], @"demibold",
-			   [NSNumber numberWithInt: 8], @"semi",
-			   [NSNumber numberWithInt: 8], @"semibold",
-			   [NSNumber numberWithInt: 9], @"bold",
-			   [NSNumber numberWithInt: 10], @"extra",
-			   [NSNumber numberWithInt: 10], @"extrabold",
-			   [NSNumber numberWithInt: 11], @"heavy",
-			   [NSNumber numberWithInt: 11], @"heavyface",
-			   [NSNumber numberWithInt: 12], @"ultrabold",
-			   [NSNumber numberWithInt: 12], @"black",
-			   [NSNumber numberWithInt: 13], @"ultra",
-			   [NSNumber numberWithInt: 13], @"ultrablack",
-			   [NSNumber numberWithInt: 13], @"fat",
-			   [NSNumber numberWithInt: 14], @"extrablack",
-			   [NSNumber numberWithInt: 14], @"obese",
-			   [NSNumber numberWithInt: 14], @"nord",
-			   nil];
+	[NSNumber numberWithInt: 1], @"ultralight",
+	[NSNumber numberWithInt: 2], @"thin",
+	[NSNumber numberWithInt: 3], @"light",
+	[NSNumber numberWithInt: 3], @"extralight",
+	[NSNumber numberWithInt: 4], @"book",
+	[NSNumber numberWithInt: 5], @"regular",
+	[NSNumber numberWithInt: 5], @"plain",
+	[NSNumber numberWithInt: 5], @"display",
+	[NSNumber numberWithInt: 5], @"roman",
+	[NSNumber numberWithInt: 5], @"semilight",
+	[NSNumber numberWithInt: 6], @"medium",
+	[NSNumber numberWithInt: 7], @"demi",
+	[NSNumber numberWithInt: 7], @"demibold",
+	[NSNumber numberWithInt: 8], @"semi",
+	[NSNumber numberWithInt: 8], @"semibold",
+	[NSNumber numberWithInt: 9], @"bold",
+	[NSNumber numberWithInt: 10], @"extra",
+	[NSNumber numberWithInt: 10], @"extrabold",
+	[NSNumber numberWithInt: 11], @"heavy",
+	[NSNumber numberWithInt: 11], @"heavyface",
+	[NSNumber numberWithInt: 12], @"ultrabold",
+	[NSNumber numberWithInt: 12], @"black",
+	[NSNumber numberWithInt: 13], @"ultra",
+	[NSNumber numberWithInt: 13], @"ultrablack",
+	[NSNumber numberWithInt: 13], @"fat",
+	[NSNumber numberWithInt: 14], @"extrablack",
+	[NSNumber numberWithInt: 14], @"obese",
+	[NSNumber numberWithInt: 14], @"nord",
+	nil];
       RETAIN(dict);
     }
 
@@ -166,11 +197,11 @@ static GSFontEnumerator *sharedEnumerator = nil;
   if (arr == nil)
     {
       arr = [NSArray arrayWithObjects: @"", @"ultralight",
-		     @"thin", @"light", @"book", @"regular",
-		     @"medium", @"demibold", @"semibold",
-		     @"bold", @"extrabold", @"heavy",
-		     @"black", @"ultrablack", @"extrablack", 
-		     nil];
+	@"thin", @"light", @"book", @"regular",
+	@"medium", @"demibold", @"semibold",
+	@"bold", @"extrabold", @"heavy",
+	@"black", @"ultrablack", @"extrablack", 
+	nil];
       RETAIN(arr);
     }
 
@@ -180,89 +211,9 @@ static GSFontEnumerator *sharedEnumerator = nil;
     return [arr objectAtIndex: aWeight];
 }
 
-+ (NSStringEncoding) encodingForRegistry: (NSString*)registry 
-				encoding: (NSString*)encoding
-{
-  if ([registry isEqualToString: @"iso8859"])
-    {
-      if ([encoding isEqualToString: @"1"])
-	return NSISOLatin1StringEncoding;
-      else if ([encoding isEqualToString: @"2"])
-	return NSISOLatin2StringEncoding;
-      else if ([encoding isEqualToString: @"3"])
-	return NSISOLatin3StringEncoding;
-      else if ([encoding isEqualToString: @"4"])
-	return NSISOLatin4StringEncoding;
-      else if ([encoding isEqualToString: @"5"])
-	return NSISOCyrillicStringEncoding;
-      else if ([encoding isEqualToString: @"6"])
-	return NSISOArabicStringEncoding;
-      else if ([encoding isEqualToString: @"7"])
-	return NSISOGreekStringEncoding;
-      else if ([encoding isEqualToString: @"8"])
-	return NSISOHebrewStringEncoding;
-      // Other latin encodings are currently not supported
-    }
-  else if ([registry isEqualToString: @"iso10646"])
-    {
-      if ([encoding isEqualToString: @"1"])
-	return NSUnicodeStringEncoding;
-    }
-  else if ([registry isEqualToString: @"microsoft"])
-    {
-      if ([encoding isEqualToString: @"symbol"])
-	return NSSymbolStringEncoding;
-      else if ([encoding isEqualToString: @"cp1250"])
-	return NSWindowsCP1250StringEncoding;
-      else if ([encoding isEqualToString: @"cp1251"])
-	return NSWindowsCP1251StringEncoding;
-      else if ([encoding isEqualToString: @"cp1252"])
-	return NSWindowsCP1252StringEncoding;
-      else if ([encoding isEqualToString: @"cp1253"])
-	return NSWindowsCP1253StringEncoding;
-      else if ([encoding isEqualToString: @"cp1254"])
-	return NSWindowsCP1254StringEncoding;
-    }
-  else if ([registry isEqualToString: @"apple"])
-    {
-      if ([encoding isEqualToString: @"roman"])
-	return NSMacOSRomanStringEncoding;
-    }
-  else if ([registry isEqualToString: @"jisx0201.1976"])
-    {
-      if ([encoding isEqualToString: @"0"])
-	return NSShiftJISStringEncoding;
-    }
-  else if ([registry isEqualToString: @"iso646.1991"])
-    {
-      if ([encoding isEqualToString: @"irv"])
-	return NSASCIIStringEncoding;
-    }
-  else if ([registry isEqualToString: @"koi8"])
-    {
-      if ([encoding isEqualToString: @"r"])
-	return NSKOI8RStringEncoding;
-    }
-  else if ([registry isEqualToString: @"gb2312.1980"])
-    {
-      if ([encoding isEqualToString: @"0"])
-	return NSGB2312StringEncoding;
-    }
-  else if ([registry isEqualToString: @"big5"])
-    {
-      if ([encoding isEqualToString: @"0"])
-        return NSBIG5StringEncoding;
-    }
-
-  return NSASCIIStringEncoding;
-}
-
 - init
 {
   [super init];
-  ASSIGN(fontDictionary, [NSMutableDictionary dictionaryWithCapacity:25]);
-  weight = 0;
-  traits = 0;
   mostCompatibleStringEncoding = NSASCIIStringEncoding;
 
   return self;
@@ -277,7 +228,7 @@ static GSFontEnumerator *sharedEnumerator = nil;
   [super dealloc];
 }
 
-- copyWithZone: (NSZone *)zone
+- (id) copyWithZone: (NSZone *)zone
 {
   GSFontInfo *copy;
   if (NSShouldRetainWithZone(self, zone))
@@ -295,7 +246,7 @@ static GSFontEnumerator *sharedEnumerator = nil;
 
 /* We really want a mutable class for this, but this is quick and easy since
    it's not really a public class anyway */
-- mutableCopyWithZone: (NSZone *)zone
+- (id) mutableCopyWithZone: (NSZone *)zone
 {
   GSFontInfo *copy;
   copy = (GSFontInfo*) NSCopyObject (self, 0, zone);
@@ -313,6 +264,66 @@ static GSFontEnumerator *sharedEnumerator = nil;
 
 - (NSDictionary*) afmDictionary
 {
+  if (fontDictionary == nil)
+    {
+      NSString *weightString;
+
+      fontDictionary = [[NSMutableDictionary alloc] initWithCapacity: 25];
+      
+      [fontDictionary setObject: fontName forKey: NSAFMFontName];
+      if (familyName != nil)
+        {
+	  [fontDictionary setObject: familyName 
+			  forKey: NSAFMFamilyName];
+	}
+      if (ascender != 0.0)
+        {
+	  [fontDictionary setObject: [NSNumber numberWithFloat: ascender] 
+			  forKey: NSAFMAscender];
+	}
+      if (descender != 0.0)
+        {
+	  [fontDictionary setObject: [NSNumber numberWithFloat: descender]
+			  forKey: NSAFMDescender];
+	}
+      if (xHeight != 0.0)
+        {
+	  [fontDictionary setObject: [NSNumber numberWithFloat: xHeight]
+			  forKey: NSAFMXHeight];
+	}
+      if (capHeight != 0.0)
+        {
+	  [fontDictionary setObject: [NSNumber numberWithFloat: capHeight]
+			  forKey: NSAFMCapHeight];
+	}
+      if (italicAngle != 0.0)
+        {
+	  [fontDictionary setObject: [NSNumber numberWithFloat: italicAngle]
+			  forKey: NSAFMItalicAngle];
+	}
+      if (underlinePosition != 0.0)
+        {
+	  [fontDictionary setObject: [NSNumber numberWithFloat: underlinePosition]
+			  forKey: NSAFMUnderlinePosition];
+	}
+      if (underlineThickness != 0.0)
+        {
+	  [fontDictionary setObject: [NSNumber numberWithFloat: underlineThickness]
+			  forKey: NSAFMUnderlineThickness];
+	}
+
+      weightString = [GSFontInfo stringForWeight: weight];      
+      if (weightString != nil)
+        {
+	  [fontDictionary setObject: weightString forKey: NSAFMWeight];
+	}
+      if (encodingScheme != nil)
+        {
+	  [fontDictionary setObject: encodingScheme
+			  forKey: NSAFMEncodingScheme];
+	}
+
+    }
   return fontDictionary;
 }
 
@@ -341,6 +352,11 @@ static GSFontEnumerator *sharedEnumerator = nil;
   return familyName;
 }
 
+- (const float*) matrix
+{
+  return matrix;
+}
+
 - (float) pointSize
 {
   return matrix[0];
@@ -406,10 +422,18 @@ static GSFontEnumerator *sharedEnumerator = nil;
   return xHeight; 
 }
 
-- (float)defaultLineHeightForFont
+- (float) defaultLineHeightForFont
 {
-  // ascent plus descent plus some suitable linegap
-  return [self ascender] - [self descender] + [self pointSize]/ 11.0;
+  /*
+  Generally, roman text is centered in the line disregarding the descender,
+  and the descender height is added above and below. Thus, the line height
+  should be ascender_height+2*descender_height. (Note that descender is
+  negative below the baseline.)
+
+  This calculation should match the baseline calculation in
+  GSHorizontalTypesetter, or text will look odd.
+  */
+  return [self ascender] - 2 * [self descender];
 }
 
 - (NSSize) advancementForGlyph: (NSGlyph)aGlyph
@@ -540,4 +564,12 @@ static GSFontEnumerator *sharedEnumerator = nil;
   return weight;
 }
 
+
+-(void) appendBezierPathWithGlyphs: (NSGlyph *)glyphs
+			     count: (int)count
+		      toBezierPath: (NSBezierPath *)path
+{
+  [self subclassResponsibility: _cmd];
+}
+
 @end
