@@ -24,7 +24,7 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 
 #include <Foundation/NSArray.h>
 #include <Foundation/NSSet.h>
@@ -46,16 +46,19 @@
 #include <Foundation/NSPortNameServer.h>
 #include <Foundation/NSTask.h>
 #include <Foundation/NSObjCRuntime.h>
+#include <Foundation/NSInvocation.h>
 
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSPasteboard.h>
-#include <AppKit/NSMenu.h>
-#include <AppKit/NSPanel.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSCell.h>
-#include <AppKit/NSWorkspace.h>
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSPasteboard.h"
+#include "AppKit/NSMenu.h"
+#include "AppKit/NSPanel.h"
+#include "AppKit/NSWindow.h"
+#include "AppKit/NSWorkspace.h"
+#include "AppKit/NSDocumentController.h"
 
-#include <AppKit/GSServicesManager.h>
+#include "GNUstepGUI/GSServicesManager.h"
+
+static GSServicesManager	*manager = nil;
 
 /*
  *	The GSListener class is for talking to other applications.
@@ -64,7 +67,7 @@
  *	messing with us.  This is responsible for forwarding service
  *      requests and other communications with the outside world.
  */
-@interface      GSListener : NSObject
+@interface      GSListener : NSProxy
 + (id) connectionBecameInvalid: (NSNotification*)notification;
 + (GSListener*) listener;
 + (id) servicesProvider;
@@ -74,25 +77,19 @@
 - (void) release;
 - (id) retain;
 - (id) self;
-- (BOOL) application: (NSApplication*)theApp
-	    openFile: (NSString*)file;
-- (BOOL) application: (NSApplication*)theApp
-   openFileWithoutUI: (NSString*)file;
-- (BOOL) application: (NSApplication*)theApp
-	openTempFile: (NSString*)file;
-- (BOOL) application: (NSApplication*)theApp
-	   printFile: (NSString*)file;
-- (void) performService: (NSString*)name
-	 withPasteboard: (NSPasteboard*)pb
-	       userData: (NSString*)ud
-		  error: (NSString**)e;
 @end
 
 static NSConnection	*listenerConnection = nil;
+static NSMutableArray	*listeners = nil;
 static GSListener	*listener = nil;
 static id		servicesProvider = nil;
 static NSString		*providerName = nil;
 
+/**
+ * Unregisters the service provider registered on the named port.<br />
+ * Applications should use [NSApplication-setServicesProvider:] with a nil
+ * argument instead.
+ */
 void
 NSUnregisterServicesProvider(NSString *name)
 {
@@ -113,6 +110,12 @@ NSUnregisterServicesProvider(NSString *name)
   ASSIGN(providerName, nil);
 }
 
+/**
+ * Registers a services providing object using the specified port name.<br />
+ * Applications should not need to use this, as they can use the
+ * [NSApplication-setServicesProvider:] method instead.  The NSApplication
+ * method will use the name of the application rather than an other port name.
+ */
 void
 NSRegisterServicesProvider(id provider, NSString *name)
 {
@@ -153,10 +156,10 @@ NSRegisterServicesProvider(id provider, NSString *name)
 }
 
 /*
- *	The GSListener class exists as a proxy to forward messages to
- *	service provider objects.  It implements very few methods and
- *	those that it does implement are generally designed to defeat
- *	any attack by a malicious program.
+ * The GSListener class exists as a proxy to forward messages to
+ * service provider objects.  It implements very few methods and
+ * those that it does implement are generally designed to defeat
+ * any attack by a malicious program.
  */
 @implementation GSListener
 
@@ -173,15 +176,39 @@ NSRegisterServicesProvider(id provider, NSString *name)
   return self;
 }
 
++ (void) initialize
+{
+  static BOOL	beenHere = NO;
+
+  if (beenHere == NO)
+    {
+      beenHere = YES;
+      listeners = [NSMutableArray new];
+    }
+}
+
 + (GSListener*) listener
 {
   if (listener == nil)
     {
       listener = (id)NSAllocateObject(self, 0, NSDefaultMallocZone());
+      [listeners addObject: listener];
     }
   return listener;
 }
 
+/**
+ * Needed to permit use of this class as a notification observer,
+ * since the notification system caches method implementations for speed.
+ */
++ (IMP) methodForSelector: (SEL)aSelector
+{
+  if (aSelector == 0)
+    [NSException raise: NSInvalidArgumentException
+                format: @"%@ null selector given", NSStringFromSelector(_cmd)];
+  return get_imp(GSObjCClass(self), aSelector);
+}
+
 + (id) servicesProvider
 {
   return servicesProvider;
@@ -198,6 +225,11 @@ NSRegisterServicesProvider(id provider, NSString *name)
     }
 }
 
+- (id) autorelease
+{
+   return self;
+}
+
 - (Class) class
 {
   return 0;
@@ -207,39 +239,13 @@ NSRegisterServicesProvider(id provider, NSString *name)
 {
 }
 
-/*
- * Obsolete ... prefer forwardInvocation now.
+/**
+ * Selectively forwards those messages which are known to be safe.
  */
-- forward: (SEL)aSel :(arglist_t)frame
-{
-  NSString      *selName = NSStringFromSelector(aSel);
-  id            delegate;
-
-  /*
-   *    If the selector matches the correct form for a services request,
-   *    send the message to the services provider - otherwise raise an
-   *    exception to say the method is not implemented.
-   */
-  if ([selName hasSuffix: @":userData:error:"])
-    return [servicesProvider performv: aSel :frame];
-
-  /*
-   *    If tha applications delegate can handle the message - forward to it.
-   */
-  delegate = [[NSApplication sharedApplication] delegate];
-  if ([delegate respondsToSelector: aSel] == YES)
-    return [delegate performv: aSel :frame];
-
-  [NSException raise: NSGenericException
-	      format: @"method %@ not implemented", selName];
-  return nil;
-}
-
 - (void) forwardInvocation: (NSInvocation*)anInvocation
 {
   SEL		aSel = [anInvocation selector];
   NSString      *selName = NSStringFromSelector(aSel);
-  id            delegate;
 
   /*
    *    If the selector matches the correct form for a services request,
@@ -247,96 +253,130 @@ NSRegisterServicesProvider(id provider, NSString *name)
    */
   if ([selName hasSuffix: @":userData:error:"])
     {
-      [anInvocation invokeWithTarget: servicesProvider];
-      return;
-    }
+      if ([servicesProvider respondsToSelector: aSel] == YES)
+	{
+	  NSPasteboard	*pb;
 
-  /*
-   *    If tha applications delegate can handle the message - forward to it.
-   */
-  delegate = [[NSApplication sharedApplication] delegate];
-  if ([delegate respondsToSelector: aSel] == YES)
+	  /*
+	   * Create a local NSPasteboard object for this pasteboard.
+	   * If we try to use the remote NSPasteboard object, we get
+	   * trouble when setting property lists since the remote
+	   * NSPasteboard fails to serialize the local property
+	   * list objects for sending to gpbs.
+	   */
+	  [anInvocation getArgument: (void*)&pb atIndex: 2];
+	  pb = [NSPasteboard pasteboardWithName: [pb name]];
+	  [anInvocation setArgument: (void*)&pb atIndex: 2];
+
+	  [anInvocation invokeWithTarget: servicesProvider];
+	  return;
+	}
+    }
+  else
     {
-      [anInvocation invokeWithTarget: delegate];
-      return;
-    }
+      id	delegate = [[NSApplication sharedApplication] delegate];
 
+      if ([selName hasPrefix: @"application:"] == YES)
+	{
+	  if ([delegate respondsToSelector: aSel] == YES)
+	    {
+	      [anInvocation invokeWithTarget: delegate];
+	      return;
+	    }
+	  else if ([manager respondsToSelector: aSel] == YES)
+	    {
+	      [anInvocation invokeWithTarget: manager];
+	      return;
+	    }
+	}
+      else
+	{
+	  if ([delegate respondsToSelector: aSel] == YES)
+	    {
+	      NSArray	*messages;
+
+	      messages = [[NSUserDefaults standardUserDefaults] arrayForKey:
+		@"GSPermittedMessages"];
+	      if (messages != nil)
+		{
+		  if ([messages containsObject: selName] == YES)
+		    {
+		      [anInvocation invokeWithTarget: delegate];
+		    }
+		}
+	      else
+		{
+		  [anInvocation invokeWithTarget: delegate];
+		}
+	      return;
+	    }
+	}
+    }
   [NSException raise: NSGenericException
 	      format: @"method %@ not implemented", selName];
 }
 
-
-- (BOOL) application: (NSApplication*)theApp
-	    openFile: (NSString*)file
+/**
+ * Return the appropriate method signature for aSelector, checking
+ * to see if it's a standard service message or standard application
+ * message.<br />
+ * If the message is non-standard, it can be checked against a list
+ * of messages specified by the GSPermittedMessages user default.
+ */
+- (NSMethodSignature*) methodSignatureForSelector: (SEL)aSelector
 {
-  id	del = [NSApp delegate];
+  NSMethodSignature	*sig = nil;
+  NSString      	*selName = NSStringFromSelector(aSelector);
 
-  if ([del respondsToSelector: _cmd])
-    return [del application: theApp openFile: file];
-  return NO;
-}
-
-- (BOOL) application: (NSApplication*)theApp
-   openFileWithoutUI: (NSString*)file
-{
-  id	del = [NSApp delegate];
-
-  if ([del respondsToSelector: _cmd])
-    return [del application: theApp openFileWithoutUI: file];
-  return NO;
-}
-
-- (BOOL) application: (NSApplication*)theApp
-	openTempFile: (NSString*)file
-{
-  id	del = [NSApp delegate];
-
-  if ([del respondsToSelector: _cmd])
-    return [del application: theApp openTempFile: file];
-  return NO;
-}
-
-- (BOOL) application: (NSApplication*)theApp
-	   printFile: (NSString*)file
-{
-  id	del = [NSApp delegate];
-
-  if ([del respondsToSelector: _cmd])
-    return [del application: theApp openFile: file];
-  return NO;
-}
-
-- (void) performService: (NSString*)name
-	 withPasteboard: (NSPasteboard*)pb
-	       userData: (NSString*)ud
-		  error: (NSString**)e
-{
-  id	obj = servicesProvider;
-  SEL	msgSel = NSSelectorFromString(name);
-  IMP	msgImp;
-
-  if (obj != nil && [obj respondsToSelector: msgSel])
+  if ([selName hasSuffix: @":userData:error:"])
     {
-      msgImp = [obj methodForSelector: msgSel];
-      if (msgImp != 0)
+      sig = [servicesProvider methodSignatureForSelector: aSelector];
+    }
+  else
+    {
+      id	delegate = [[NSApplication sharedApplication] delegate];
+
+      if ([selName hasPrefix: @"application:"] == YES)
 	{
-	  (*msgImp)(obj, msgSel, pb, ud, e);
-	  return;
+	  if ([delegate respondsToSelector: aSelector] == YES)
+	    {
+	      sig = [delegate methodSignatureForSelector: aSelector];
+	    }
+	  else
+	    {
+	      sig = [manager methodSignatureForSelector: aSelector];
+	    }
+	}
+      else
+	{
+	  NSArray	*messages;
+
+	  messages = [[NSUserDefaults standardUserDefaults] arrayForKey:
+	    @"GSPermittedMessages"];
+
+	  if (messages != nil)
+	    {
+	      if ([messages containsObject: selName] == YES)
+		{
+		  sig = [delegate methodSignatureForSelector: aSelector];
+		}
+	    }
+	  else
+	    {
+	      sig = [delegate methodSignatureForSelector: aSelector];
+	    }
 	}
     }
+  return sig;
+}
 
-  obj = [[NSApplication sharedApplication] delegate];
-  if (obj != nil && [obj respondsToSelector: msgSel])
+- (BOOL) respondsToSelector: (SEL)aSelector
+{
+  if ([self methodSignatureForSelector: aSelector] != nil)
     {
-      msgImp = [obj methodForSelector: msgSel];
-      if (msgImp != 0)
-	{
-	  (*msgImp)(obj, msgSel, pb, ud, e);
-	  return;
-	}
+      return YES;
     }
-
-  *e = @"No object available to provide service"; 
+  return NO;
 }
 
 - (void) release
@@ -358,7 +398,6 @@ NSRegisterServicesProvider(id provider, NSString *name)
 
 @implementation GSServicesManager
 
-static GSServicesManager	*manager = nil;
 static NSString         *servicesName = @".GNUstepServices";
 static NSString         *disabledName = @".GNUstepDisabled";
 
@@ -386,7 +425,7 @@ static NSString         *disabledName = @".GNUstepDisabled";
 
   str = [NSSearchPathForDirectoriesInDomains(NSUserDirectory,
           NSUserDomainMask, YES) objectAtIndex: 0];
-  str = [str stringByAppendingPathComponent: @"Services"];
+  str = [str stringByAppendingPathComponent: @"Library/Services"];
   path = [str stringByAppendingPathComponent: servicesName];
   manager->_servicesPath = [path copy];
   path = [str stringByAppendingPathComponent: disabledName];
@@ -401,11 +440,11 @@ static NSString         *disabledName = @".GNUstepDisabled";
    *	Check for changes to the services cache every thirty seconds.
    */
   manager->_timer =
-	[NSTimer scheduledTimerWithTimeInterval: 30.0
+	RETAIN([NSTimer scheduledTimerWithTimeInterval: 30.0
 					 target: manager
 				       selector: @selector(loadServices)
 				       userInfo: nil
-					repeats: YES];
+					repeats: YES]);
 
   [manager loadServices];
   return manager;
@@ -420,12 +459,70 @@ static NSString         *disabledName = @".GNUstepDisabled";
   return manager;
 }
 
+- (BOOL) application: (NSApplication*)theApp
+	    openFile: (NSString*)file
+{
+  id	del = [NSApp delegate];
+  BOOL	result = NO;
+
+  if ([del respondsToSelector: _cmd])
+    {
+      result = [del application: theApp openFile: file];
+    }
+  else if ([[NSDocumentController sharedDocumentController]
+    openDocumentWithContentsOfFile: file display: YES] != nil)
+    {
+      [NSApp activateIgnoringOtherApps: YES];
+      result = YES;
+    }
+  return result;
+}
+
+- (BOOL) application: (NSApplication*)theApp
+   openFileWithoutUI: (NSString*)file
+{
+  id	del = [NSApp delegate];
+  BOOL	result = NO;
+
+  if ([del respondsToSelector: _cmd])
+    {
+      result = [del application: theApp openFileWithoutUI: file];
+    }
+  else if ([[NSDocumentController sharedDocumentController]
+    openDocumentWithContentsOfFile: file display: NO] != nil)
+    {
+      result = YES;
+    }
+  return result;
+}
+
+- (BOOL) application: (NSApplication*)theApp
+	openTempFile: (NSString*)file
+{
+  BOOL	result = [self application: theApp openFile: file];
+
+  [[NSFileManager defaultManager] removeFileAtPath: file handler: nil];
+
+  return result;
+}
+
+- (BOOL) application: (NSApplication*)theApp
+	   printFile: (NSString*)file
+{
+  id	del = [NSApp delegate];
+
+  if ([del respondsToSelector: _cmd])
+    return [del application: theApp printFile: file];
+  return NO;
+}
+
 - (void) dealloc
 {
   NSString          *appName;
 
   appName = [[NSProcessInfo processInfo] processName];
   [_timer invalidate];
+  RELEASE(_timer);
   NSUnregisterServicesProvider(appName);
   RELEASE(_languages);
   RELEASE(_returnInfo);
@@ -442,7 +539,7 @@ static NSString         *disabledName = @".GNUstepDisabled";
   [super dealloc];
 }
 
-- (void) doService: (NSCell*)item
+- (void) doService: (NSMenuItem*)item
 {
   NSString      *title = [self item2title: item];
   NSDictionary  *info = [_title2info objectForKey: title];
@@ -454,8 +551,6 @@ static NSString         *disabledName = @".GNUstepDisabled";
   NSWindow      *resp = [[_application keyWindow] firstResponder];
   id            obj = nil;
 
-  NSLog(@"doService: called");
-
   for (i = 0; i <= es; i++)
     {
       NSString  *sendType;
@@ -475,8 +570,9 @@ static NSString         *disabledName = @".GNUstepDisabled";
               NSPasteboard      *pb;
 
               pb = [NSPasteboard pasteboardWithUniqueName];
-	      if ([obj writeSelectionToPasteboard: pb
-					    types: sendTypes] == NO)
+	      if (sendType
+		  && [obj writeSelectionToPasteboard: pb
+					       types: sendTypes] == NO)
 		{
 		  NSRunAlertPanel(nil,
 			@"object failed to write to pasteboard",
@@ -484,7 +580,8 @@ static NSString         *disabledName = @".GNUstepDisabled";
 		}
 	      else if (NSPerformService(title, pb) == YES)
 		{
-		  if ([obj readSelectionFromPasteboard: pb] == NO)
+		  if (returnType
+		      && [obj readSelectionFromPasteboard: pb] == NO)
 		    {
 		      NSRunAlertPanel(nil,
 			    @"object failed to read from pasteboard",
@@ -497,6 +594,14 @@ static NSString         *disabledName = @".GNUstepDisabled";
     }
 }
 
+/**
+ * Return a dictionary of information about registered filter services.
+ */
+- (NSArray*) filters
+{
+  return [_allServices objectForKey: @"ByFilter"];
+}
+
 - (BOOL) hasRegisteredTypes: (NSDictionary*)service
 {
   NSArray       *sendTypes = [service objectForKey: @"NSSendTypes"];
@@ -557,11 +662,11 @@ static NSString         *disabledName = @".GNUstepDisabled";
 }
 
 /*
- *      Use tag in menu cell to identify slot in menu titles array that
+ *      Use tag in menu item to identify slot in menu titles array that
  *      contains the full title of the service.
- *      Return nil if this is not one of our service menu cells.
+ *      Return nil if this is not one of our service menu items.
  */
-- (NSString*) item2title: (NSCell*)item
+- (NSString*) item2title: (NSMenuItem*)item
 {
   unsigned      pos;
 
@@ -600,16 +705,15 @@ static NSString         *disabledName = @".GNUstepDisabled";
 	      if (plist)
 		{
 		  NSMutableSet	*s;
-		  stamp = mod;
 		  changed = YES;
 		  s = (NSMutableSet*)[NSMutableSet setWithArray: plist];
 		  ASSIGN(_allDisabled, s);
 		}
 	    }
+	  /* Track most recent version of file loaded */
+	  ASSIGN(_disabledStamp, mod);
 	}
     }
-  /* Track most recent version of file loaded or last time we checked */
-  ASSIGN(_disabledStamp, stamp);
 
   stamp = [NSDate date];
   if ([mgr fileExistsAtPath: _servicesPath])
@@ -632,15 +736,14 @@ static NSString         *disabledName = @".GNUstepDisabled";
 						    mutableContainers: YES];
 	      if (plist)
 		{
-		  stamp = mod;
 		  ASSIGN(_allServices, plist);
 		  changed = YES;
 		}
 	    }
+	  /* Track most recent version of file loaded */
+	  ASSIGN(_servicesStamp, mod);
 	}
     }
-  /* Track most recent version of file loaded or last time we checked */
-  ASSIGN(_servicesStamp, stamp);
   if (changed)
     {
       [self rebuildServices];
@@ -734,7 +837,6 @@ static NSString         *disabledName = @".GNUstepDisabled";
 {
   if (_servicesMenu != nil)
     {
-      NSArray           *itemArray;
       NSMutableSet      *keyEquivalents;
       unsigned          pos;
       unsigned          loc0;
@@ -742,13 +844,11 @@ static NSString         *disabledName = @".GNUstepDisabled";
       SEL               sel = @selector(doService:);
       NSMenu            *submenu = nil;
 
-      itemArray = [[_servicesMenu itemArray] copy];
-      pos = [itemArray count];
-      while (pos > 0)
-        {
-          [_servicesMenu removeItem: [itemArray objectAtIndex: --pos]];
-        }
-      RELEASE(itemArray);
+      [_servicesMenu setAutoenablesItems: NO];
+      pos = [_servicesMenu numberOfItems];
+      for (; pos; pos--)
+        [_servicesMenu removeItemAtIndex: 0];
+      [_servicesMenu setAutoenablesItems: YES];
 
       keyEquivalents = [NSMutableSet setWithCapacity: 4];
       for (loc0 = pos = 0; pos < [_menuTitles count]; pos++)
@@ -821,7 +921,7 @@ static NSString         *disabledName = @".GNUstepDisabled";
                 }
               else
                 {
-                  menu = (NSMenu*)[item target];
+                  menu = (NSMenu*)[item submenu];
                 }
               if (menu != submenu)
                 {
@@ -999,14 +1099,22 @@ static NSString         *disabledName = @".GNUstepDisabled";
 
   [self loadServices];
   if (_allDisabled == nil)
-    _allDisabled = [[NSMutableSet alloc] initWithCapacity: 1];
+    {
+      _allDisabled = [[NSMutableSet alloc] initWithCapacity: 1];
+    }
   if (enable)
-    [_allDisabled removeObject: item];
+    {
+      [_allDisabled removeObject: item];
+    }
   else
-    [_allDisabled addObject: item];
+    {
+      [_allDisabled addObject: item];
+    }
   d = [NSSerializer serializePropertyList: [_allDisabled allObjects]];
   if ([d writeToFile: _disabledPath atomically: YES] == YES)
-    return 0;
+    {
+      return 0;
+    }
   return -1;
 }
 
@@ -1018,7 +1126,7 @@ static NSString         *disabledName = @".GNUstepDisabled";
   return NO;
 }
 
-- (BOOL) validateMenuItem: (NSCell*)item
+- (BOOL) validateMenuItem: (NSMenuItem*)item
 {
   NSString      *title = [self item2title: item];
   NSDictionary  *info = [_title2info objectForKey: title];
@@ -1030,12 +1138,17 @@ static NSString         *disabledName = @".GNUstepDisabled";
   NSWindow      *resp = [[_application keyWindow] firstResponder];
 
   /*
-   *    If the menu item is not in our map, it must be the cell containing
-   *    a sub-menu - so we see if any cell in the submenu is valid.
+   *    If the menu item is not in our map, it must be the item containing
+   *    a sub-menu - so we see if any item in the submenu is valid.
    */
   if (title == nil)
     {
-      NSMenu    *sub = [item target];
+      NSMenu    *sub;
+
+      if (![item isKindOfClass: [NSMenuItem class]])
+        return NO;
+
+      sub = [item submenu];
 
       if (sub && [sub isKindOfClass: [NSMenu class]])
         {
@@ -1053,7 +1166,7 @@ static NSString         *disabledName = @".GNUstepDisabled";
     }
 
   /*
-   *    The cell corresponds to one of our services - so we check to see if
+   *    The item corresponds to one of our services - so we check to see if
    *    there is anything that can deal with it.
    */
   if (es == 0)
@@ -1131,46 +1244,52 @@ static NSString         *disabledName = @".GNUstepDisabled";
 
       for (i = 0; i < [a count]; i++)
         {
-          NSCell        *cell = [a objectAtIndex: i];
-	  BOOL		wasEnabled = [cell isEnabled];
+          NSMenuItem    *item = [a objectAtIndex: i];
+	  BOOL		wasEnabled = [item isEnabled];
 	  BOOL		shouldBeEnabled;
-	  NSString      *title = [self item2title: cell];
+	  NSString      *title = [self item2title: item];
 
 	  /*
-	   *	If there is no title mapping, this cell must be a
-	   *	submenu - so we check the submenu cells.
+	   *	If there is no title mapping, this item must be a
+	   *	submenu - so we check the submenu items.
+	   *
+	   *	We always enable the submenu item itself. We do this
+	   *	to prevent confusion (if the user is trying to use
+	   *	a disabled item, it's clearer to show that item disabled
+	   *	than to hide it in a disabled submenu), and to encourage
+	   *	the user to explore the interface (it makes it possible
+	   *	to browse the service list at any time).
 	   */
-	  if (title == nil && [[cell target] isKindOfClass: [NSMenu class]])
+	  if (title == nil && [[item submenu] isKindOfClass: [NSMenu class]])
 	    {
-	      NSArray		*sub = [[cell target] itemArray];
-	      int		j;
+	      NSArray		*sub = [[item submenu] itemArray];
+	      unsigned		j;
 
-	      shouldBeEnabled = NO;
+	      shouldBeEnabled = YES;
 	      for (j = 0; j < [sub count]; j++)
 		{
-		  NSCell	*subCell = [sub objectAtIndex: j];
-		  BOOL		subWasEnabled = [subCell isEnabled];
+		  NSMenuItem	*subitem = [sub objectAtIndex: j];
+		  BOOL		subWasEnabled = [subitem isEnabled];
 		  BOOL		subShouldBeEnabled = NO;
 
-		  if ([self validateMenuItem: subCell] == YES)
+		  if ([self validateMenuItem: subitem] == YES)
 		    {
-		      shouldBeEnabled = YES;	/* Enabled menu */
 		      subShouldBeEnabled = YES;
 		    }
 		  if (subWasEnabled != subShouldBeEnabled)
 		    {
-		      [subCell setEnabled: subShouldBeEnabled];
+		      [subitem setEnabled: subShouldBeEnabled];
 //		      [subMenuCells setNeedsDisplayInRect:
 //				[subMenuCells cellFrameAtRow: j]];
 		    }
 		}
 	    }
 	  else
-	    shouldBeEnabled = [self validateMenuItem: cell];
+	    shouldBeEnabled = [self validateMenuItem: item];
 
           if (wasEnabled != shouldBeEnabled)
             {
-              [cell setEnabled: shouldBeEnabled];
+              [item setEnabled: shouldBeEnabled];
 //	      [menuCells setNeedsDisplayInRect: [menuCells cellFrameAtRow: i]];
             }
         }
@@ -1180,11 +1299,35 @@ static NSString         *disabledName = @".GNUstepDisabled";
 @end /* GSServicesManager */
 
 
+/**
+ * <p>Establishes an NSConnection to the application listening at port
+ * (by convention usually the application name), launching appName 
+ * if necessary.  Returns the proxy to the remote application, or nil
+ * on failure.
+ * </p>
+ * <p>The value of port specifies the name of the distributed objects
+ * service to which the connection is to be made.  If this is nil
+ * it will be inferred from appName ... by convention, applications
+ * use their own name (minus any path or extension) for this.
+ * </p>
+ * <p>The value of expire provides a timeout in case the application cannot
+ * be contacted promptly.  If it is omitted, a thirty second timeout is
+ * used.
+ * </p>
+ */
 id
 GSContactApplication(NSString *appName, NSString *port, NSDate *expire)
 {
   id	app;
 
+  if (port == nil)
+    {
+      port = [[appName lastPathComponent] stringByDeletingPathExtension];
+    }
+  if (expire == nil)
+    {
+      expire = [NSDate dateWithTimeIntervalSinceNow: 30.0];
+    }
   if (providerName != nil && [port isEqual: providerName] == YES)
     {
       app = [GSListener listener];	// Contect our own listener.
@@ -1239,6 +1382,56 @@ GSContactApplication(NSString *appName, NSString *port, NSDate *expire)
   return app;
 }
 
+NSDictionary *
+_serviceFromAnyLocalizedTitle(NSString *title)
+{
+  NSDictionary	*allServices;
+  NSEnumerator	*e1;
+  NSDictionary	*service;
+
+  allServices = [manager menuServices];
+  if (allServices == nil)
+    {
+      return nil;
+    }
+
+  if ([allServices objectForKey: title] != nil)
+    {
+      return [allServices objectForKey: title];
+     } 
+  e1 = [allServices objectEnumerator];
+  while ((service = [e1 nextObject]) != nil)
+    {
+      NSDictionary	*menuItems;
+      NSString		*itemName;
+      NSEnumerator	*e2;
+
+      menuItems = [service objectForKey: @"NSMenuItem"];
+      if (menuItems == nil)
+	{
+	  continue;
+	}
+      e2 = [menuItems objectEnumerator];
+      while ((itemName = [e2 nextObject]) != nil)
+	{
+	  if ([itemName isEqualToString: title] == YES)
+	    {
+	      return service;
+	    }
+	}
+    }
+
+  return nil;
+}
+
+/**
+ * <p>Given the name of a serviceItem, and some data in a pasteboard
+ * this function sends the data to the service provider (launching
+ * another application if necessary) and retrieves the result of
+ * the service in the pastebaord.
+ * </p>
+ * Returns YES on success, NO otherwise.
+ */
 BOOL
 NSPerformService(NSString *serviceItem, NSPasteboard *pboard)
 {
@@ -1254,12 +1447,13 @@ NSPerformService(NSString *serviceItem, NSPasteboard *pboard)
   NSString		*userData;
   NSString		*error = nil;
 
-  service = [[manager menuServices] objectForKey: serviceItem]; 
+  service = _serviceFromAnyLocalizedTitle(serviceItem);
   if (service == nil)
     {
       NSRunAlertPanel(nil,
-	[NSString stringWithFormat: @"No service matching '%@'", serviceItem],
-	@"Continue", nil, nil);
+	@"No service matching '%@'",
+	@"Continue", nil, nil,
+	serviceItem);
       return NO;			/* No matching service.	*/
     }
 
@@ -1287,9 +1481,9 @@ NSPerformService(NSString *serviceItem, NSPasteboard *pboard)
   if (provider == nil)
     {
       NSRunAlertPanel(nil,
-	[NSString stringWithFormat:
-	    @"Failed to contact service provider for '%@'", serviceItem],
-	@"Continue", nil, nil);
+	@"Failed to contact service provider for '%@'",
+	@"Continue", nil, nil,
+	serviceItem);
       return NO;
     }
 
@@ -1309,13 +1503,31 @@ NSPerformService(NSString *serviceItem, NSPasteboard *pboard)
 
   /*
    * At last, we ask for the service to be performed.
+   * We create an untyped selector matching the message name we have,
+   * Using that, we get a method signature from the provider, and
+   * take the type information from that to make a fully typed
+   * selector, with which we can create and use an invocation.
    */
   NS_DURING
     {
-      [provider performService: selName
-		withPasteboard: pboard
-		      userData: userData
-			 error: &error];
+      const char	*name = [selName UTF8String];
+      SEL		sel = GSSelectorFromNameAndTypes(name, 0); 
+      NSMethodSignature	*sig = [provider methodSignatureForSelector: sel];
+
+      if (sig != nil)
+	{
+	  NSInvocation	*inv;
+	  NSString	**errPtr = &error;
+
+	  sel = GSSelectorFromNameAndTypes(name, [sig methodType]); 
+	  inv = [NSInvocation invocationWithMethodSignature: sig];
+	  [inv setTarget: provider];
+	  [inv setSelector: sel];
+	  [inv setArgument: (void*)&pboard atIndex: 2];
+	  [inv setArgument: (void*)&userData atIndex: 3];
+	  [inv setArgument: (void*)&errPtr atIndex: 4];
+	  [inv invoke];
+	}
     }
   NS_HANDLER
     {
@@ -1326,29 +1538,49 @@ NSPerformService(NSString *serviceItem, NSPasteboard *pboard)
   if (error != nil)
     {
       NSRunAlertPanel(nil,
-	[NSString stringWithFormat:
-	    @"Failed to contact service provider for '%@'", serviceItem],
-	    @"Continue", nil, nil);
+	@"Failed to contact service provider for '%@': %@",
+	@"Continue", nil, nil,
+	serviceItem, error);
       return NO;
     }
 
   return YES;
 }
 
+/**
+ * <p>Controls whether the item name should be included in the services menu.
+ * </p>
+ * <p>If enabled is YES then the services menu for each application will
+ * include the named item, if enabled is NO then the service will not be
+ * shown in application services menus.
+ * </p>
+ * <p>Returns 0 if the setting is successfuly changed. Non-zero otherwise.
+ * </p>
+ */
 int
 NSSetShowsServicesMenuItem(NSString *name, BOOL enabled)
 {
-
   return [[GSServicesManager manager] setShowsServicesMenuItem: name
-							     to: enabled];
+							    to: enabled];
 }
 
+/**
+ * Returns a flag indicating whether the named service is supposed to be
+ * displayed in application services menus.
+ */
 BOOL
-NSShowsServicesMenuItem(NSString * name)
+NSShowsServicesMenuItem(NSString *name)
 {
   return [[GSServicesManager manager] showsServicesMenuItem: name];
 }
 
+/**
+ * A services providing application may use this to update the list of
+ * services it provides.<br />
+ * In order to update the services advertised, the application must
+ * create a <em>.service</em> bundle and place it in
+ * <code>~/Library/Services</code> before invoking this function.
+ */
 void
 NSUpdateDynamicServices(void)
 {
