@@ -27,20 +27,23 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
+#include <Foundation/NSArray.h>
 #include <Foundation/NSString.h>
 #include <Foundation/NSRange.h>
 #include <Foundation/NSBundle.h>
 #include <Foundation/NSFileManager.h>
-#include <AppKit/NSAttributedString.h>
-#include <AppKit/NSDocumentController.h>
-#include <AppKit/NSParagraphStyle.h>
-#include <AppKit/NSTextAttachment.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSColor.h>
+#include "AppKit/NSAttributedString.h"
+#include "AppKit/NSDocumentController.h"
+#include "AppKit/NSParagraphStyle.h"
+#include "AppKit/NSPasteboard.h"
+#include "AppKit/NSTextAttachment.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSFileWrapper.h"
+#include "AppKit/NSFont.h"
 // For the colour name spaces
-#include <AppKit/NSGraphics.h>
+#include "AppKit/NSGraphics.h"
 
-#include <AppKit/GSTextConverter.h>
+#include "GNUstepGUI/GSTextConverter.h"
 
 /* Cache class pointers to avoid the expensive lookup by string. */ 
 static Class dictionaryClass = nil;
@@ -136,7 +139,7 @@ Class converter_bundles(NSString *format, BOOL producer)
 			    @selector(classForFormat:producer:)])
 		{
 		  converter_class = (Class)[bclass classForFormat: format
-					                 producer: producer];
+						   producer: producer];
 		}
 	      else
 		{
@@ -206,6 +209,29 @@ static Class converter_class(NSString *format, BOOL producer)
 
 @implementation NSAttributedString (AppKit)
 
++ (NSArray *) textFileTypes
+{
+  // FIXME
+  return [self textUnfilteredFileTypes];
+}
+
++ (NSArray *) textPasteboardTypes
+{
+  // FIXME
+  return [self textUnfilteredPasteboardTypes];
+}
+
++ (NSArray *) textUnfilteredFileTypes
+{
+  return [NSArray arrayWithObjects: @"txt",  @"rtf", @"rtfd", @"html", nil];
+}
+
++ (NSArray *) textUnfilteredPasteboardTypes
+{
+  return [NSArray arrayWithObjects: NSStringPboardType, NSRTFPboardType, 
+		  NSRTFDPboardType, NSHTMLPboardType, nil];
+}
+
 + (NSAttributedString *) attributedStringWithAttachment: 
                                             (NSTextAttachment *)attachment
 {
@@ -585,7 +611,8 @@ documentAttributes: (NSDictionary **)dict
 
   new = [converter_class(@"RTFD", NO) 
 			parseFile: wrapper
-			documentAttributes: dict];
+			documentAttributes: dict
+			class: [self class]];
   // We do not return self but the newly created object
   RELEASE (self);
   return RETAIN (new); 
@@ -604,7 +631,8 @@ documentAttributes: (NSDictionary **)dict
 
   new = [converter_class(@"RTFD", NO)
 			parseData: data
-			documentAttributes: dict];
+			documentAttributes: dict
+			class: [self class]];
   // We do not return self but the newly created object
   RELEASE (self);
   return RETAIN (new); 
@@ -623,7 +651,8 @@ documentAttributes: (NSDictionary **)dict
 
   new = [converter_class(@"RTF", NO) 
 			parseData: data
-			documentAttributes: dict];
+			documentAttributes: dict
+			class: [self class]];
   // We do not return self but the newly created object
   RELEASE (self);
   return RETAIN (new); 
@@ -655,15 +684,17 @@ documentAttributes: (NSDictionary **)dict
        documentAttributes: (NSDictionary *)dict
 {
   return [converter_class(@"RTF", YES) 
-			 produceDataFrom: [self attributedSubstringFromRange: range]
+			 produceDataFrom: 
+			   [self attributedSubstringFromRange: range]
 			 documentAttributes: dict];
 }
 
 - (NSData *) RTFDFromRange: (NSRange)range
-       documentAttributes: (NSDictionary *)dict
+	documentAttributes: (NSDictionary *)dict
 {
   return [converter_class(@"RTFD", YES)  
-			 produceDataFrom: [self attributedSubstringFromRange: range]
+			 produceDataFrom: 
+			   [self attributedSubstringFromRange: range]
 			 documentAttributes: dict];
 }
 
@@ -671,7 +702,8 @@ documentAttributes: (NSDictionary **)dict
 			  documentAttributes: (NSDictionary *)dict
 {
   return [converter_class(@"RTFD", YES)
-			 produceFileFrom: [self attributedSubstringFromRange: range]
+			 produceFileFrom: 
+			   [self attributedSubstringFromRange: range]
 			 documentAttributes: dict];
 }
 
@@ -851,6 +883,11 @@ documentAttributes: (NSDictionary **)dict
     }
   // FIXME: Should check for each character if it is supported by the 
   // assigned font
+  /*
+  Note that this needs to be done on a script basis. Per-character checks
+  are difficult to do at all, don't give reasonable results, and would have
+  really poor performance.
+  */
 }
 
 - (void) fixParagraphStyleAttributeInRange: (NSRange)range
@@ -886,10 +923,18 @@ documentAttributes: (NSDictionary **)dict
              the range without the style set.  */
 	  if ((NSMaxRange (found) + 1) < end)
 	    {
-	      /* This skips the range with style not set, and attempts
-		 again to set the style for the whole paragraph taking
-		 the style at the first location with a style set.  */
-	      loc = NSMaxRange (found) + 1;
+	      /* There is a paragraph style for part of the paragraph. Set
+	      this style for the entire paragraph.
+
+	      Since NSMaxRange(found) + 1 is outside the longest effective
+	      range for the nil style, it must be non-nil.
+	      */
+	      style = [self attribute: NSParagraphStyleAttributeName
+			      atIndex: NSMaxRange(found) + 1
+			      effectiveRange: NULL];
+	      [self addAttribute: NSParagraphStyleAttributeName
+		    value: style
+		    range: r];
 	    }
 	  else
 	    {
@@ -898,9 +943,6 @@ documentAttributes: (NSDictionary **)dict
 	      [self addAttribute: NSParagraphStyleAttributeName
 		    value: [NSParagraphStyle defaultParagraphStyle]
 		    range: r];
-
-	      /* Move on to the next paragraph.  */
-	      loc = end;	      
 	    }
 	}
       else
@@ -916,11 +958,10 @@ documentAttributes: (NSDictionary **)dict
 		    value: style
 		    range: found];
 	    }
-
-	  /* Move on to the next paragraph.  */
-	  loc = end;
 	}
       
+      /* Move on to the next paragraph.  */
+      loc = end;
     }
 }
 
@@ -986,6 +1027,7 @@ documentAttributes: (NSDictionary **)dict
         {
 	  [self deleteCharactersInRange: NSMakeRange (range.location, 1)];
 	  range.length--;
+	  end--;
 	}
 
       location = NSMaxRange (range);
@@ -1086,12 +1128,14 @@ documentAttributes: (NSDictionary **)dict
       NSStringEncoding encoding = [[options objectForKey: @"CharacterEncoding"] 
 				      intValue];
       NSDictionary *defaultAttrs = [options objectForKey: @"DefaultAttributes"];
+      NSString *str = [[NSString alloc] initWithData: data 
+					encoding: encoding];
       NSAttributedString *attr;
 
-     attr = [[NSAttributedString alloc] 
-		 initWithString: [NSString initWithData: data 
-					   encoding: encoding]
+      attr = [[NSAttributedString alloc] 
+		 initWithString: str
 		    attributes: defaultAttrs];
+      RELEASE(str);
       [self setAttributedString: attr];
       RELEASE(attr);
 
