@@ -27,8 +27,14 @@
    59 Temple Place - Suite 330, Boston, MA 02111 - 1307, USA.
 */
 
-#include <gnustep/gui/config.h>
-#include <AppKit/AppKit.h>
+#include "AppKit/NSBezierPath.h"
+
+#include "AppKit/NSAffineTransform.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/PSOperators.h"
+#include "GNUstepGUI/GSFontInfo.h"
+
 #include <math.h>
 
 #ifndef PI
@@ -53,7 +59,6 @@ static float default_miter_limit = 10.0;
 
 - (void)_invalidateCache;
 - (void)_recalculateBounds;
-- (void)_doPath;
 
 @end
 
@@ -84,7 +89,7 @@ static float default_miter_limit = 10.0;
 {
   if (self != NSBezierPath_concrete_class)
     {
-      return [NSBezierPath_concrete_class alloc];
+      return [NSBezierPath_concrete_class allocWithZone: z];
     }
   else
     {
@@ -92,7 +97,7 @@ static float default_miter_limit = 10.0;
     }    
 }
 
-+ (id)bezierPath
++ (NSBezierPath *)bezierPath
 {
   return AUTORELEASE ([[NSBezierPath_concrete_class alloc] init]);
 }
@@ -100,59 +105,20 @@ static float default_miter_limit = 10.0;
 + (NSBezierPath *)bezierPathWithRect: (NSRect)aRect
 {
   NSBezierPath *path;
-  NSPoint p;
 	
-  path = [NSBezierPath bezierPath];
-  [path moveToPoint: aRect.origin];
-  p.x = aRect.origin.x + aRect.size.width;
-  p.y = aRect.origin.y;
-  [path lineToPoint: p];
-  p.x = aRect.origin.x + aRect.size.width;
-  p.y = aRect.origin.y + aRect.size.height;
-  [path lineToPoint: p];
-  p.x = aRect.origin.x;
-  p.y = aRect.origin.y + aRect.size.height;
-  [path lineToPoint: p];
-  [path closePath];
+  path = [self bezierPath];
+  [path appendBezierPathWithRect: aRect];
 	
   return path;
 }
 
-+ (NSBezierPath *)bezierPathWithOvalInRect: (NSRect)rect
++ (NSBezierPath *)bezierPathWithOvalInRect: (NSRect)aRect
 {
   NSBezierPath *path;
-  NSPoint p, p1, p2;
-  double originx = rect.origin.x;
-  double originy = rect.origin.y;
-  double width = rect.size.width;
-  double height = rect.size.height;
-  double hdiff = width / 2 * KAPPA;
-  double vdiff = height / 2 * KAPPA;
-  
-  path = [NSBezierPath bezierPath];
-  p = NSMakePoint(originx + width / 2, originy + height);
-  [path moveToPoint: p];
-  
-  p = NSMakePoint(originx, originy + height / 2);
-  p1 = NSMakePoint(originx + width / 2 - hdiff, originy + height);
-  p2 = NSMakePoint(originx, originy + height / 2 + vdiff);
-  [path curveToPoint: p controlPoint1: p1 controlPoint2: p2];
-  
-  p = NSMakePoint(originx + width / 2, originy);
-  p1 = NSMakePoint(originx, originy + height / 2 - vdiff);
-  p2 = NSMakePoint(originx + width / 2 - hdiff, originy);
-  [path curveToPoint: p controlPoint1: p1 controlPoint2: p2];	
-  
-  p = NSMakePoint(originx + width, originy + height / 2);
-  p1 = NSMakePoint(originx + width / 2 + hdiff, originy);
-  p2 = NSMakePoint(originx + width, originy + height / 2 - vdiff);
-  [path curveToPoint: p controlPoint1: p1 controlPoint2: p2];	
-  
-  p = NSMakePoint(originx + width / 2, originy + height);
-  p1 = NSMakePoint(originx + width, originy + height / 2 + vdiff);
-  p2 = NSMakePoint(originx + width / 2 + hdiff, originy + height);
-  [path curveToPoint: p controlPoint1: p1 controlPoint2: p2];	
   
+  path = [self bezierPath];
+  [path appendBezierPathWithOvalInRect: aRect];
+
   return path;
 }
 
@@ -176,20 +142,22 @@ static float default_miter_limit = 10.0;
 
 + (void)strokeLineFromPoint: (NSPoint)point1  toPoint: (NSPoint)point2
 {
-  NSBezierPath *path = [NSBezierPath bezierPath];
+  NSBezierPath *path = [[self alloc] init];
   
   [path moveToPoint: point1];
   [path lineToPoint: point2];
   [path stroke];
+  RELEASE(path);
 }
 
 + (void)drawPackedGlyphs: (const char *)packedGlyphs  atPoint: (NSPoint)aPoint
 {
-  NSBezierPath *path = [NSBezierPath bezierPath];
+  NSBezierPath *path = [[self alloc] init];
   
   [path moveToPoint: aPoint];
   [path appendBezierPathWithPackedGlyphs: packedGlyphs];
   [path stroke];  
+  RELEASE(path);
 }
 
 //
@@ -450,7 +418,19 @@ static float default_miter_limit = 10.0;
 {
   NSZone *myZone = [self zone];
   
-  if ( _dash_pattern == NULL)
+  if ((pattern == NULL) || (count == 0))
+    {
+      if (_dash_pattern != NULL)
+        {
+	  NSZoneFree(myZone, _dash_pattern);
+	  _dash_pattern = NULL;
+	}
+      _dash_count = 0;
+      _dash_phase = 0.0;
+      return;
+    }
+
+  if (_dash_pattern == NULL)
     _dash_pattern = NSZoneMalloc(myZone, count * sizeof(float));
   else
     NSZoneRealloc(myZone, _dash_pattern, count * sizeof(float));
@@ -465,6 +445,8 @@ static float default_miter_limit = 10.0;
 //
 - (void)stroke
 {
+  NSGraphicsContext *ctxt = GSCurrentContext();
+  
   if(_cachesBezierPath) 
     {
       NSRect bounds = [self bounds];
@@ -475,22 +457,24 @@ static float default_miter_limit = 10.0;
         {
 	  _cacheImage = [[NSImage alloc] initWithSize: bounds.size];		
 	  [_cacheImage lockFocus];
-	  PStranslate(-origin.x, -origin.y);
-  	  [self _doPath];
-	  PSstroke();
+	  DPStranslate(ctxt, -origin.x, -origin.y);
+	  [ctxt GSSendBezierPath: self];
+	  DPSstroke(ctxt);
 	  [_cacheImage unlockFocus];						
 	}
       [_cacheImage compositeToPoint: origin operation: NSCompositeCopy];
     } 
   else 
     {
-      [self _doPath];
-      PSstroke();
+      [ctxt GSSendBezierPath: self];
+      DPSstroke(ctxt);
     }
 }
 
 - (void)fill
 {
+  NSGraphicsContext *ctxt = GSCurrentContext();
+
   if(_cachesBezierPath) 
     {
       NSRect bounds = [self bounds];
@@ -501,43 +485,47 @@ static float default_miter_limit = 10.0;
         {
 	  _cacheImage = [[NSImage alloc] initWithSize: bounds.size];		
 	  [_cacheImage lockFocus];
-	  PStranslate(-origin.x, -origin.y);
-	  [self _doPath];
+	  DPStranslate(ctxt, -origin.x, -origin.y);
+	  [ctxt GSSendBezierPath: self];
 	  if([self windingRule] == NSNonZeroWindingRule)
-	    PSfill();
+	    DPSfill(ctxt);
 	  else
-	    PSeofill();
+	    DPSeofill(ctxt);
 	  [_cacheImage unlockFocus];
 	}
       [_cacheImage compositeToPoint: origin operation: NSCompositeCopy];
     } 
   else 
     {
-      [self _doPath];
+      [ctxt GSSendBezierPath: self];
       if([self windingRule] == NSNonZeroWindingRule)
-	PSfill();
+	DPSfill(ctxt);
       else
-	PSeofill();
+	DPSeofill(ctxt);
     }
 }
 
 - (void)addClip
 {
-  [self _doPath];
+  NSGraphicsContext *ctxt = GSCurrentContext();
+  
+  [ctxt GSSendBezierPath: self];
   if([self windingRule] == NSNonZeroWindingRule)
-    PSclip();
+    DPSclip(ctxt);
   else
-    PSeoclip();
+    DPSeoclip(ctxt);
 }
 
 - (void)setClip
 {
-  PSinitclip();
-  [self _doPath];
+  NSGraphicsContext *ctxt = GSCurrentContext();
+  
+  DPSinitclip(ctxt);
+  [ctxt GSSendBezierPath: self];
   if([self windingRule] == NSNonZeroWindingRule)
-    PSclip();
+    DPSclip(ctxt);
   else
-    PSeoclip();
+    DPSeoclip(ctxt);
 }
 
 //
@@ -671,7 +659,7 @@ static float default_miter_limit = 10.0;
 
   if (closed)
     [path closePath];
-  return self;
+  return path;
 }
 
 //
@@ -709,7 +697,6 @@ static float default_miter_limit = 10.0;
   INVALIDATE_CACHE();
 }
 
-
 //
 // Path info
 //
@@ -805,7 +792,6 @@ static float default_miter_limit = 10.0;
   int i, count;
 
   count = [aPath elementCount];
-  
   for (i = 0; i < count; i++)
     {
       type = [aPath elementAtIndex: i associatedPoints: points];
@@ -818,9 +804,9 @@ static float default_miter_limit = 10.0;
 	      [self lineToPoint: points[0]];
 	      break;
 	  case NSCurveToBezierPathElement:
-	      [self curveToPoint: points[0] 
-		    controlPoint1: points[1]
-		    controlPoint2: points[2]];
+	      [self curveToPoint: points[2] 
+		    controlPoint1: points[0]
+		    controlPoint2: points[1]];
 	      break;
 	  case NSClosePathBezierPathElement:
 	      [self closePath];
@@ -831,9 +817,21 @@ static float default_miter_limit = 10.0;
     }
 }
 
-- (void)appendBezierPathWithRect:(NSRect)rect
+- (void)appendBezierPathWithRect:(NSRect)aRect
 {
-  [self appendBezierPath: [isa bezierPathWithRect: rect]];
+  NSPoint p;
+	
+  [self moveToPoint: aRect.origin];
+  p.x = aRect.origin.x + aRect.size.width;
+  p.y = aRect.origin.y;
+  [self lineToPoint: p];
+  p.x = aRect.origin.x + aRect.size.width;
+  p.y = aRect.origin.y + aRect.size.height;
+  [self lineToPoint: p];
+  p.x = aRect.origin.x;
+  p.y = aRect.origin.y + aRect.size.height;
+  [self lineToPoint: p];
+  [self closePath];
 }
 
 - (void)appendBezierPathWithPoints:(NSPoint *)points count:(int)count
@@ -845,20 +843,51 @@ static float default_miter_limit = 10.0;
 
   if ([self isEmpty])
     {
-	[self moveToPoint: points[0]];
+      [self moveToPoint: points[0]];
     }
   else
     {
-	[self lineToPoint: points[0]];
+      [self lineToPoint: points[0]];
     }
   
   for (i = 1; i < count; i++)
-    [self lineToPoint: points[i]];
+    {
+      [self lineToPoint: points[i]];
+    }
 }
 
 - (void) appendBezierPathWithOvalInRect: (NSRect)aRect
 {
-  [self appendBezierPath: [isa bezierPathWithOvalInRect: aRect]];
+  NSPoint p, p1, p2;
+  double originx = aRect.origin.x;
+  double originy = aRect.origin.y;
+  double width = aRect.size.width;
+  double height = aRect.size.height;
+  double hdiff = width / 2 * KAPPA;
+  double vdiff = height / 2 * KAPPA;
+  
+  p = NSMakePoint(originx + width / 2, originy + height);
+  [self moveToPoint: p];
+  
+  p = NSMakePoint(originx, originy + height / 2);
+  p1 = NSMakePoint(originx + width / 2 - hdiff, originy + height);
+  p2 = NSMakePoint(originx, originy + height / 2 + vdiff);
+  [self curveToPoint: p controlPoint1: p1 controlPoint2: p2];
+  
+  p = NSMakePoint(originx + width / 2, originy);
+  p1 = NSMakePoint(originx, originy + height / 2 - vdiff);
+  p2 = NSMakePoint(originx + width / 2 - hdiff, originy);
+  [self curveToPoint: p controlPoint1: p1 controlPoint2: p2];	
+  
+  p = NSMakePoint(originx + width, originy + height / 2);
+  p1 = NSMakePoint(originx + width / 2 + hdiff, originy);
+  p2 = NSMakePoint(originx + width, originy + height / 2 - vdiff);
+  [self curveToPoint: p controlPoint1: p1 controlPoint2: p2];	
+  
+  p = NSMakePoint(originx + width / 2, originy + height);
+  p1 = NSMakePoint(originx + width, originy + height / 2 + vdiff);
+  p2 = NSMakePoint(originx + width / 2 + hdiff, originy + height);
+  [self curveToPoint: p controlPoint1: p1 controlPoint2: p2];	
 }
 
 /* startAngle and endAngle are in degrees, counterclockwise, from the
@@ -999,21 +1028,114 @@ static float default_miter_limit = 10.0;
 				  toPoint: (NSPoint)point2
 				   radius: (float)radius
 {
-  // TODO
+  float x1, y1;
+  float dx1, dy1, dx2, dy2;
+  float l, a1, a2;
+  NSPoint p;
+
+  p = [self currentPoint];
+ 
+  x1 = point1.x;
+  y1 = point1.y;
+  dx1 = p.x - x1;
+  dy1 = p.y - y1;
+  
+  l= dx1*dx1 + dy1*dy1;
+  if (l <= 0)
+    {
+      [self lineToPoint: point1];
+      return;
+    }
+  l = 1/sqrt(l);
+  dx1 *= l;
+  dy1 *= l;
+  
+  dx2 = point2.x - x1;
+  dy2 = point2.y - y1;
+
+  l = dx2*dx2 + dy2*dy2;
+  if (l <= 0)
+    {
+      [self lineToPoint: point1];
+      return;
+    }
+	
+  l = 1/sqrt(l);
+  dx2 *= l; 
+  dy2 *= l;
+  
+  l = dx1*dx2 + dy1*dy2;
+  if (l < -0.999)
+    {
+      [self lineToPoint: point1];
+      return;
+    }
+
+  l = radius/sin(acos(l));
+  p.x = x1 + (dx1 + dx2)*l;
+  p.y = y1 + (dy1 + dy2)*l;
+
+  if (dx1 < -1)
+    a1 = 180;
+  else if (dx1 > 1)
+    a1 = 0;
+  else
+    a1 = acos(dx1)/PI*180;
+  if (dy1 < 0)
+    {   
+      a1 = -a1;
+    }
+
+  if (dx2 < -1)
+    a2 = 180;
+  else if (dx2 > 1)
+    a2 = 0;
+  else
+    a2 = acos(dx2)/PI*180;
+  if (dy2 < 0)
+    {   
+      a2 = -a2;
+    }
+
+  l = dx1*dy2 - dx2*dy1;
+  if (l < 0)
+    {
+      a2 = a2 - 90;
+      a1 = a1 + 90;
+      [self appendBezierPathWithArcWithCenter: p  
+	    radius: radius
+	    startAngle: a1  
+	    endAngle: a2  
+	    clockwise: NO];
+    }
+  else
+    {
+      a2 = a2 + 90;
+      a1 = a1 - 90;
+      [self appendBezierPathWithArcWithCenter: p  
+	    radius: radius
+	    startAngle: a1  
+	    endAngle: a2  
+	    clockwise: YES];
+    }
 }
 
 - (void)appendBezierPathWithGlyph:(NSGlyph)glyph inFont:(NSFont *)font
 {
-  // TODO
+  [[font fontInfo] appendBezierPathWithGlyphs: &glyph
+    count: 1
+    toBezierPath: self];
 }
 
 - (void)appendBezierPathWithGlyphs:(NSGlyph *)glyphs 
 			     count:(int)count
 			    inFont:(NSFont *)font
 {
-  // TODO
+  [[font fontInfo] appendBezierPathWithGlyphs: glyphs
+    count: count
+    toBezierPath: self];
 }
-				 
+
 - (void)appendBezierPathWithPackedGlyphs:(const char *)packedGlyphs
 {
   // TODO
@@ -1022,10 +1144,162 @@ static float default_miter_limit = 10.0;
 //
 // Hit detection  
 // 
+/*
+ * Return the contribution of a path segment from start to end to 
+ * the containsPoint method for point.
+ */
+static 
+int contribution(NSPoint point, float dir, NSPoint start, NSPoint end, BOOL *hit)
+{
+  double t;
+  double len;
+  double a;
+
+  start.x -= point.x;
+  start.y -= point.y;
+  len = sqrt(start.x * start.x +  start.y * start.y);
+  a = atan2(start.y, start.x);
+  start.x = cos(a - dir) * len;
+  start.y = sin(a - dir) * len;
+  end.x -= point.x;
+  end.y -= point.y;
+  len = sqrt(end.x * end.x +  end.y * end.y);
+  a = atan2(end.y, end.x);
+  end.x = cos(a - dir) * len;
+  end.y = sin(a - dir) * len;
+
+  if ((start.y == 0) || (end.y == 0))
+    {
+      *hit = YES;
+      return 0;
+    }
+  *hit = NO;
+
+  // Both point on the same half plain
+  if ((start.y < 0) == (end.y < 0))
+    {
+      return 0;
+    }
+  
+  // Does the line hit the coordinate line on the positive side?
+  t = (0 - start.y) * (end.x - start.x) / (end.y - start.y) + start.x;
+
+  if (t > 0)
+    {
+      if (start.y > 0)
+	  return 1;
+      else 
+	  return -1;
+    }
+  else 
+    {
+      if (t == 0)
+	*hit = YES;
+      return 0;
+    }
+}
+
+- (int) contributionToContains: (NSPoint)point
+{
+  int i;
+  // Full total of contribution
+  int sum = 0;
+  // running total for the current subpath
+  int sub = 0;
+  NSBezierPathElement type;
+  NSPoint p, pts[3];
+  NSPoint first_p, last_p;
+  int count = [self elementCount];
+  BOOL first = YES;
+  float dir = 2 * PI * rand() / RAND_MAX; 
+  BOOL hit;
+
+  for(i = 0; i < count; i++) 
+    {
+      type = [self elementAtIndex: i associatedPoints: pts];
+      
+      switch(type) 
+        {
+	  case NSMoveToBezierPathElement:
+	      // Was the last sub path closed without close?
+	      if (!first && NSEqualPoints(first_p, last_p))
+	        {
+		  sum += sub;
+		}
+	      sub = 0;
+	      first_p = last_p = pts[0];
+	      first = NO;
+	      break;
+	  case NSLineToBezierPathElement:
+	      p = pts[0];
+	      if (first)
+	        {
+		  first_p = last_p = p;
+		  first = NO;
+		}
+	      else
+	        {
+		  sub += contribution(point, dir, last_p, p, &hit);
+		  if (hit)
+		    return 0;
+		  last_p = p;
+		}
+	      break;
+	  case NSCurveToBezierPathElement:
+	      // Not possible as flattend!
+	      break;
+	  case NSClosePathBezierPathElement:
+	      sub +=  contribution(point, dir, last_p, first_p, &hit);
+	      if (hit)
+		return 0;
+	      sum += sub;
+	      sub = 0;
+	      last_p = first_p;
+	      first = YES;
+	      break;
+	  default:
+	      break;
+	}
+    }  
+
+  /*
+   * Check if the last sub path was closed.
+   * It wont do any harm, if there was only one point in the subpath, 
+   * as this will result in a subtotal of 0.
+   */
+  if (!first && NSEqualPoints(first_p, last_p))
+    {
+      sum += sub;
+    }
+  
+  return sum;
+}
+
 - (BOOL)containsPoint:(NSPoint)point
 {
-  [self subclassResponsibility:_cmd];
-  return NO;
+  int sum;
+
+  if(![self elementCount])
+    return NO;
+
+  if (!NSPointInRect(point, [self bounds]))
+    return NO;
+
+  sum = [[self bezierPathByFlatteningPath] contributionToContains: point];
+  if ([self windingRule] == NSNonZeroWindingRule)
+    {
+      if (sum == 0)
+	return NO;
+      else	
+	return YES;
+    }
+  else 
+    {
+      if ((sum % 2) == 0)
+	return NO;
+      else	
+	return YES;
+    }
 }
 
 //
@@ -1293,51 +1567,6 @@ static float default_miter_limit = 10.0;
   _shouldRecalculateBounds = NO;
 }
 
-- (void)_doPath
-{
-  NSGraphicsContext *ctxt = GSCurrentContext();
-  NSBezierPathElement type;
-  NSPoint pts[3];
-  int i, count;
-  float pattern[10];
-  float phase;
-
-  DPSnewpath(ctxt);
-  DPSsetlinewidth(ctxt, [self lineWidth]);
-  DPSsetlinejoin(ctxt, [self lineJoinStyle]);
-  DPSsetlinecap(ctxt, [self lineCapStyle]);
-  DPSsetmiterlimit(ctxt, [self miterLimit]);
-  DPSsetflat(ctxt, [self flatness]);
-
-  [self getLineDash: pattern count: &count phase: &phase];
-  if (count != 0 && count < 10)
-    DPSsetdash(ctxt, pattern, count, phase);
-
-  count = [self elementCount];
-  for(i = 0; i < count; i++) 
-    {
-      type = [self elementAtIndex: i associatedPoints: pts];
-      switch(type) 
-        {
-	  case NSMoveToBezierPathElement:
-	      DPSmoveto(ctxt, pts[0].x, pts[0].y);
-	      break;
-	  case NSLineToBezierPathElement:
-	      DPSlineto(ctxt, pts[0].x, pts[0].y);
-	      break;
-	  case NSCurveToBezierPathElement:
-	      DPScurveto(ctxt, pts[0].x, pts[0].y, 
-			 pts[1].x, pts[1].y, pts[2].x, pts[2].y);
-	      break;
-	  case NSClosePathBezierPathElement:
-	      DPSclosepath(ctxt);
-	      break;
-	  default:
-	      break;
-	}
-    }
-}
-
 @end
 
 
@@ -1357,7 +1586,7 @@ typedef struct _PathElement
 #ifdef GSIArray
 #undef GSIArray
 #endif
-#include <base/GSIArray.h>
+#include <GNUstepBase/GSIArray.h>
 
 @interface GSBezierPath : NSBezierPath
 {
@@ -1511,6 +1740,71 @@ typedef struct _PathElement
   return [super bezierPathByFlatteningPath];
 }
 
+- (void) transformUsingAffineTransform: (NSAffineTransform *)transform
+{
+  NSBezierPathElement type;
+  int i, count;
+  PathElement *elments = (PathElement *)GSIArrayItems(pathElements);
+  SEL transformPointSel = @selector(transformPoint:); 
+  NSPoint (*transformPointImp)(NSAffineTransform*, SEL, NSPoint);
+
+  transformPointImp = (NSPoint (*)(NSAffineTransform*, SEL, NSPoint))
+      [transform methodForSelector: transformPointSel];
+
+  count = GSIArrayCount(pathElements);
+  for(i = 0; i < count; i++) 
+    {
+      type = elments[i].type;
+      switch(type) 
+        {
+	  case NSMoveToBezierPathElement:
+	  case NSLineToBezierPathElement:
+	      elments[i].points[0] = (*transformPointImp)(transform,  
+							  transformPointSel, 
+							  elments[i].points[0]);
+	      break;
+	  case NSCurveToBezierPathElement:
+	      elments[i].points[0] = (*transformPointImp)(transform,  
+							  transformPointSel, 
+							  elments[i].points[0]);
+	      elments[i].points[1] = (*transformPointImp)(transform,  
+							  transformPointSel, 
+							  elments[i].points[1]);
+	      elments[i].points[2] = (*transformPointImp)(transform,  
+							  transformPointSel, 
+							  elments[i].points[2]);
+	      break;
+	  case NSClosePathBezierPathElement:
+	      break;
+	  default:
+	      break;
+	}
+    }
+  INVALIDATE_CACHE();
+}
+
+- (void) appendBezierPath: (NSBezierPath *)aPath
+{
+  PathElement elem;
+  int i, count;
+
+  if (![aPath isKindOfClass: isa])
+    {
+      [super appendBezierPath: aPath];
+      return;
+    }
+
+  flat = flat && ((GSBezierPath*)aPath)->flat;
+  count = [aPath elementCount];
+
+  for (i = 0; i < count; i++)
+    {
+      elem = GSIArrayItemAtIndex(((GSBezierPath*)aPath)->pathElements, i).ext;
+      GSIArrayAddItem(pathElements, (GSIArrayItem)elem);
+    }
+  INVALIDATE_CACHE();
+}
+
 //
 // NSCopying Protocol
 //
@@ -1523,111 +1817,6 @@ typedef struct _PathElement
   return path;
 }
 
-//
-// Hit detection  
-// 
-#define PMAX 10000
-
-- (BOOL)containsPoint:(NSPoint)point
-{
-  NSPoint draftPolygon[PMAX];
-  int pcount = 0;
-  // Coordinates of the current point
-  double cx, cy;
-  // Coordinates of the last point
-  double lx, ly;
-  int i;
-  int Rcross = 0;
-  int Lcross = 0;	
-  NSBezierPathElement bpt;
-  NSPoint p, pts[3];
-  double x, y, t, k = 0.25;
-  int count = [self elementCount];
-
-  if(!count)
-    return NO;
-		
-  if (!NSPointInRect(point, [self bounds]))
-    return NO;
-
-  // FIXME: This does not handle multiple segments!
-  for(i = 0; i < count; i++) 
-    {
-      bpt = [self elementAtIndex: i associatedPoints: pts];
-      
-      if(bpt == NSMoveToBezierPathElement || bpt == NSLineToBezierPathElement) 
-        {
-	  draftPolygon[pcount].x = pts[0].x;
-	  draftPolygon[pcount].y = pts[0].y;
-	  
-	  pcount++;
-	} 
-      else if(bpt == NSCurveToBezierPathElement) 
-        {
-	  if(pcount) 
-	    {
-	      p.x = draftPolygon[pcount -1].x;
-	      p.y = draftPolygon[pcount -1].y;
-	    } 
-	  else 
-	    {
-	      p.x = pts[0].x;
-	      p.y = pts[0].y;
-	    }
-				
-	  for(t = k; t <= 1+k; t += k) 
-	    {
-      	      x = (p.x+t*(-p.x*3+t*(3*p.x-p.x*t)))+
-		  t*(3*pts[0].x+t*(-6*pts[0].x+pts[0].x*3*t))+
-		  t*t*(pts[1].x*3-pts[1].x*3*t)+pts[2].x*t*t*t;
-	      y = (p.y+t*(-p.y*3+t*(3*p.y-p.y*t)))+
-		  t*(3*pts[0].y+t*(-6*pts[0].y+pts[0].y*3*t))+
-		  t*t*(pts[1].y*3-pts[1].y*3*t)+pts[2].y*t*t*t;
-	      
-	      draftPolygon[pcount].x = x;
-	      draftPolygon[pcount].y = y;
-	      pcount++;
-	    }
-	}
-
-      // Simple overflow check
-      if (pcount == PMAX)
-	return NO;
-    }  
-
-  lx = draftPolygon[pcount - 1].x - point.x;
-  ly = draftPolygon[pcount - 1].y - point.y;
-  for(i = 0; i < pcount; i++) 
-    {
-      cx = draftPolygon[i].x - point.x;
-      cy = draftPolygon[i].y - point.y;
-      if(cx == 0 && cy == 0) 
-	// on a vertex
-	return NO;
-					
-      if((cy > 0)  && !(ly > 0)) 
-        {
-	  if (((cx * ly - lx * cy) / (ly - cy)) > 0)
-	    Rcross++;
-	}
-      if((cy < 0 ) && !(ly < 0)) 
-        { 
-	  if (((cx * ly - lx * cy) / (ly - cy)) < 0);
-	    Lcross++;		
-	}
-      lx = cx;
-      ly = cy;
-    }
-
-  if((Rcross % 2) != (Lcross % 2))
-    // On the border
-    return NO;
-  if((Rcross % 2) == 1)
-    return YES;
-  else	
-    return NO;
-}
-
 @end // GSBezierPath
 
 
