@@ -25,15 +25,20 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSArray.h>
+#include <Foundation/NSDebug.h>
 #include <Foundation/NSException.h>
+#include <Foundation/NSKeyedArchiver.h>
 #include <Foundation/NSString.h>
 
-#include <AppKit/NSBox.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSTextFieldCell.h>
+#include "AppKit/NSBox.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSTextFieldCell.h"
+#include "GNUstepGUI/GSDrawFunctions.h"
+
+#include <math.h>
 
 @interface NSBox (Private)
 - (NSRect) calcSizesAllowingNegative: (BOOL)aFlag;
@@ -60,12 +65,10 @@
 {
   [super initWithFrame: frameRect];
 	
-  _cell = [[NSTextFieldCell alloc] initTextCell: @"Title"];
+  _cell = [[NSCell alloc] initTextCell: @"Title"];
   [_cell setAlignment: NSCenterTextAlignment];
   [_cell setBordered: NO];
   [_cell setEditable: NO];
-  [_cell setDrawsBackground: YES];
-  //[_cell setBackgroundColor: [NSColor controlColor]];
   _offsets.width = 5;
   _offsets.height = 5;
   _border_rect = _bounds;
@@ -298,7 +301,7 @@
 	}
       else // _content_view has no subviews
 	{
-	  [_content_view setFrame: NSZeroRect];
+  	  [_content_view setFrame: [self calcSizesAllowingNegative: NO]];
 	}
     }
 
@@ -365,18 +368,25 @@
       [[NSColor controlDarkShadowColor] set];
       NSFrameRect(_border_rect);
       break;
-    case NSBezelBorder: 
-      NSDrawGrayBezel(_border_rect, rect);
+    case NSBezelBorder:
+      [GSDrawFunctions drawDarkBezel: _border_rect : rect];
       break;
     case NSGrooveBorder: 
-      NSDrawGroove(_border_rect, rect);
+      [GSDrawFunctions drawGroove: _border_rect : rect];
       break;
     }
 
   // Draw title
   if (_title_position != NSNoTitle)
     {
-      [_cell setBackgroundColor: color];
+      // If the title is on the border, clip a hole in the later 
+      if ((_border_type != NSNoBorder) &&
+	  ((_title_position == NSAtTop) || 
+	   (_title_position == NSAtBottom)))
+        {
+	  [color set];
+	  NSRectFill(_title_rect);
+	}
       [_cell drawWithFrame: _title_rect inView: self];
     }
 }
@@ -403,33 +413,68 @@
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  [super initWithCoder: aDecoder];
+  self = [super initWithCoder: aDecoder];
 
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_cell];
-  _offsets = [aDecoder decodeSize];
-  [aDecoder decodeValueOfObjCType: @encode(NSBorderType)
-			       at: &_border_type];
-  [aDecoder decodeValueOfObjCType: @encode(NSTitlePosition) 
-			       at: &_title_position];
+  if ([aDecoder allowsKeyedCoding])
+    {
+      NSView *contentView = [aDecoder decodeObjectForKey: @"NSContentView"];
+      NSCell *titleCell = [aDecoder decodeObjectForKey: @"NSTitleCell"];
 
-  // The content view is our only sub_view
-  if ([_sub_views count] == 0)
-    {
-  
-      NSDebugLLog(@"NSBox", @"NSBox: decoding without content view\n");
-      // No content view
-      _content_view = nil;
-      [self calcSizesAllowingNegative: NO];
-    }
-  else 
-    {
-      if ([_sub_views count] != 1)
-	{
-	  NSLog (@"Warning: Encoded NSBox with more than one content view!");
+      [self setContentView: contentView];
+      ASSIGN(_cell, titleCell);
+      if ([aDecoder containsValueForKey: @"NSBoxType"])
+        {
+	  //int boxType = [aDecoder decodeIntForKey: @"NSBoxType"];
 	}
-      _content_view = [_sub_views objectAtIndex: 0];
-      // The following also computes _title_rect and _border_rect.
-      [_content_view setFrame: [self calcSizesAllowingNegative: NO]];
+      if ([aDecoder containsValueForKey: @"NSBorderType"])
+        {
+	  NSBorderType borderType = [aDecoder decodeIntForKey: @"NSBorderType"];
+
+	  [self setBorderType: borderType];
+	}
+      if ([aDecoder containsValueForKey: @"NSTitlePosition"])
+        {
+	  NSTitlePosition titlePosition = [aDecoder decodeIntForKey: 
+							@"NSTitlePosition"];
+
+	  [self setTitlePosition: titlePosition];
+	}
+      if ([aDecoder containsValueForKey: @"NSTransparent"])
+        {
+	  //Bool transparent = [aDecoder decodeBoolForKey: @"NSTransparent"];
+	}
+      if ([aDecoder containsValueForKey: @"NSOffsets"])
+        {
+	  [self setContentViewMargins: [aDecoder decodeSizeForKey: @"NSOffsets"]];
+	}
+    }
+  else
+    {
+	[aDecoder decodeValueOfObjCType: @encode(id) at: &_cell];
+	_offsets = [aDecoder decodeSize];
+	[aDecoder decodeValueOfObjCType: @encode(NSBorderType)
+		                     at: &_border_type];
+	[aDecoder decodeValueOfObjCType: @encode(NSTitlePosition) 
+		                     at: &_title_position];
+
+	// The content view is our only sub_view
+	if ([_sub_views count] == 0)
+	  {
+	    NSDebugLLog(@"NSBox", @"NSBox: decoding without content view\n");
+	    // No content view
+	    _content_view = nil;
+	    [self calcSizesAllowingNegative: NO];
+	  }
+	else 
+	  {
+	    if ([_sub_views count] != 1)
+	      {
+		NSLog (@"Warning: Encoded NSBox with more than one content view!");
+	      }
+	    _content_view = [_sub_views objectAtIndex: 0];
+	    // The following also computes _title_rect and _border_rect.
+	    [_content_view setFrame: [self calcSizesAllowingNegative: NO]];
+	  }
     }
   return self;
 }
@@ -531,6 +576,8 @@
 	NSSize titleSize = [_cell cellSize];
 	NSSize borderSize = _sizeForBorderType (_border_type);
 	float c;
+	float topMargin;
+	float topOffset;
 
 	// Add spacer around title
 	titleSize.width += 6;
@@ -538,26 +585,39 @@
 
 	_border_rect = _bounds;
 
+	topMargin = ceil(titleSize.height / 2);
+	topOffset = titleSize.height - topMargin;
+	
 	// Adjust by the title size
-	_border_rect.size.height -= titleSize.height / 2;
-
+	_border_rect.size.height -= topMargin;
+	
 	// Add the offsets to the border rect
 	r.origin.x = _border_rect.origin.x + _offsets.width + borderSize.width;
-	r.origin.y = _border_rect.origin.y + _offsets.height + borderSize.height;
 	r.size.width = _border_rect.size.width - (2 * _offsets.width)
 	  - (2 * borderSize.width);
-	r.size.height = _border_rect.size.height - (2 * _offsets.height)
-	  - (2 * borderSize.height);
+	
+	if (topOffset > _offsets.height)
+	  {
+	    r.origin.y = _border_rect.origin.y + _offsets.height + borderSize.height;
+	    r.size.height = _border_rect.size.height - _offsets.height
+	      - (2 * borderSize.height) - topOffset;
+	  }
+	else
+	  {
+	    r.origin.y = _border_rect.origin.y + _offsets.height + borderSize.height;
+	    r.size.height = _border_rect.size.height - (2 * _offsets.height)
+	      - (2 * borderSize.height);
+	  }
 
 	// Adjust by the title size
-	r.size.height -= (titleSize.height / 2) + borderSize.height;
+	//	r.size.height -= titleSize.height + borderSize.height;
 
 	// center the title cell
 	c = (_border_rect.size.width - titleSize.width) / 2;
 	if (c < 0) c = 0;
 	_title_rect.origin.x = _border_rect.origin.x + c;
 	_title_rect.origin.y = _border_rect.origin.y + _border_rect.size.height
-	  - (titleSize.height / 2);
+	  - topMargin;
 	_title_rect.size = titleSize;
 
 	break;
@@ -567,6 +627,8 @@
 	NSSize titleSize = [_cell cellSize];
 	NSSize borderSize = _sizeForBorderType (_border_type);
 	float c;
+	float bottomMargin;
+	float bottomOffset;
 
 	// Add spacer around title
 	titleSize.width += 6;
@@ -574,22 +636,37 @@
 
 	_border_rect = _bounds;
 
+	bottomMargin = ceil(titleSize.height / 2);
+	bottomOffset = titleSize.height - bottomMargin;
+
 	// Adjust by the title size
-	_border_rect.origin.y += titleSize.height / 2;
-	_border_rect.size.height -= titleSize.height / 2;
+	_border_rect.origin.y += bottomMargin;
+	_border_rect.size.height -= bottomMargin;
 
 	// Add the offsets to the border rect
 	r.origin.x = _border_rect.origin.x + _offsets.width + borderSize.width;
-	r.origin.y = _border_rect.origin.y + _offsets.height + borderSize.height;
 	r.size.width = _border_rect.size.width - (2 * _offsets.width)
 	  - (2 * borderSize.width);
-	r.size.height = _border_rect.size.height - (2 * _offsets.height)
-	  - (2 * borderSize.height);
+
+	if (bottomOffset > _offsets.height)
+	  {
+	    r.origin.y = _border_rect.origin.y + bottomOffset + borderSize.height;
+	    r.size.height = _border_rect.size.height - _offsets.height
+	      - bottomOffset
+	      - (2 * borderSize.height);
+	  }
+	else
+	  {
+	    r.origin.y = _border_rect.origin.y + _offsets.height + borderSize.height;
+	    r.size.height = _border_rect.size.height - (2 * _offsets.height)
+	      - (2 * borderSize.height);
+	  }
 
 	// Adjust by the title size
+	/*
 	r.origin.y += (titleSize.height / 2) + borderSize.height;
 	r.size.height -= (titleSize.height / 2) + borderSize.height;
-
+	*/
 	// center the title cell
 	c = (_border_rect.size.width - titleSize.width) / 2;
 	if (c < 0) c = 0;
