@@ -27,30 +27,28 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSArchiver.h>
 #include <Foundation/NSArray.h>
 #include <Foundation/NSBundle.h>
 #include <Foundation/NSCoder.h>
 #include <Foundation/NSData.h>
 #include <Foundation/NSDictionary.h>
+#include <Foundation/NSDebug.h>
 #include <Foundation/NSEnumerator.h>
+#include <Foundation/NSException.h>
 #include <Foundation/NSInvocation.h>
 #include <Foundation/NSObjCRuntime.h>
+#include <Foundation/NSPathUtilities.h>
 #include <Foundation/NSFileManager.h>
 #include <Foundation/NSString.h>
 #include <Foundation/NSUserDefaults.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSMenu.h>
-#include <AppKit/NSControl.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSView.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSNibConnector.h>
-#include <AppKit/NSNibLoading.h>
-#include <AppKit/IMLoading.h>
-
-static const int currentVersion = 1;
+#include <Foundation/NSKeyValueCoding.h>
+#include <Foundation/NSNotification.h>
+#include "AppKit/NSNibConnector.h"
+#include "AppKit/NSNibLoading.h"
+#include "GNUstepGUI/GSNibTemplates.h"
+#include "GNUstepGUI/IMLoading.h"
 
 @implementation	NSNibConnector
 
@@ -78,11 +76,20 @@ static const int currentVersion = 1;
 {
 }
 
-- (id) initWithCoder: (NSCoder*)aCoder
+- (id) initWithCoder: (NSCoder*)aDecoder
 {
-  [aCoder decodeValueOfObjCType: @encode(id) at: &_src];
-  [aCoder decodeValueOfObjCType: @encode(id) at: &_dst];
-  [aCoder decodeValueOfObjCType: @encode(id) at: &_tag];
+  if ([aDecoder allowsKeyedCoding])
+    {
+      ASSIGN(_src, [aDecoder decodeObjectForKey: @"NSSource"]);
+      ASSIGN(_dst, [aDecoder decodeObjectForKey: @"NSDestination"]);
+      ASSIGN(_tag, [aDecoder decodeObjectForKey: @"NSLabel"]);
+    }
+  else
+    {
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_src];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_dst];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_tag];
+    }
   return self;
 }
 
@@ -147,8 +154,9 @@ static const int currentVersion = 1;
       NSString	*selName;
       SEL	sel;
 
-      selName = [NSString stringWithFormat: @"set%@:",
-	[_tag capitalizedString]];
+      selName = [NSString stringWithFormat: @"set%@%@:",
+			  [[_tag substringToIndex: 1] uppercaseString],
+			  [_tag substringFromIndex: 1]];
       sel = NSSelectorFromString(selName);
 	      
       if (sel && [_src respondsToSelector: sel])
@@ -157,47 +165,88 @@ static const int currentVersion = 1;
 	}
       else
 	{
+	  const char	*nam = [_tag cString];
+	  const char	*type;
+	  unsigned int	size;
+	  unsigned int	offset;
+
 	  /*
-	   * Use the GNUstep additional function to set the instance variable
-	   * directly.
-	   * FIXME - need some way to do this for libFoundation and Foundation
-	   * based systems.
+	   * Use the GNUstep additional function to set the instance
+	   * variable directly.
+	   * FIXME - need some way to do this for libFoundation and
+	   * Foundation based systems.
 	   */
-	  GSSetInstanceVariable(_src, _tag, (void*)&_dst); 
+	  if (GSObjCFindVariable(_src, nam, &type, &size, &offset))
+	    {
+	      GSObjCSetVariable(_src, offset, size, (void*)&_dst); 
+	    }
 	}
     }
 }
 @end
 
-
+/*
+ * This private class is used to collect the nib items while the 
+ * .gorm file is being unarchived.  This is done to allow only
+ * the top level items to be retained in a clean way.  The reason it's
+ * being done this way is because old .gorm files don't have any
+ * array within the nameTable which indicates the objects which are
+ * considered top level, so there is no clean and generic way to determine
+ * this.   Basically the top level items are any instances of or instances
+ * of subclasses of NSMenu, NSWindow, or any controller class.
+ * It's the last one that's hairy.  Controller classes are
+ * represented in .gorm files by the GSNibItem class, but once they transform
+ * into the actual class instance it's not easy to tell if it should be 
+ * retained or not since there are a lot of other things stored in the nameTable
+ * as well.  GJC
+ */
+@interface _GSNibItemCollector : NSObject
+{
+  NSMutableArray *items;
+}
+- (void) handleNotification: (NSNotification *)notification;
+- (NSMutableArray *)items;
+@end
+
+@implementation _GSNibItemCollector
+- (void) handleNotification: (NSNotification *)notification;
+{
+  id obj = [notification object];
+  [items addObject: obj];
+}
+
+- init
+{
+  if((self = [super init]) != nil)
+    {
+      NSNotificationCenter *nc = [NSNotificationCenter defaultCenter];
+
+      // add myself as an observer and initialize the items array.
+      [nc addObserver: self
+	  selector: @selector(handleNotification:)
+	  name: @"__GSInternalNibItemAddedNotification"
+	  object: nil];
+      items = [[NSMutableArray alloc] init];
+    }
+  return self;
+}
+
+- (void) dealloc
+{
+  [[NSNotificationCenter defaultCenter] removeObserver: self];
+  RELEASE(items);
+  [super dealloc];
+}
+
+- (NSMutableArray *)items
+{
+  return items;
+}
+@end
+
 
 @implementation NSBundle (NSBundleAdditions)
 
-- (NSString*) pathForImageResource: (NSString*)name
-{
-  NSString	*ext = [name pathExtension];
-  NSString	*path = nil;
-
-  if ((ext == nil) || [ext isEqualToString:@""])
-    {
-      NSArray	*types = [NSImage imageUnfilteredFileTypes];
-      unsigned	c = [types count];
-      unsigned	i;
-
-      for (i = 0; path == nil && i < c; i++)
-	{
-	  ext = [types objectAtIndex: i];
-	  path = [self pathForResource: name ofType: ext];
-	}
-    }
-  else
-    {
-      name = [name stringByDeletingPathExtension];
-      path = [self pathForResource: name ofType: ext];
-    }
-  return path;
-}
-
 static 
 Class gmodel_class(void)
 {
@@ -237,6 +286,7 @@ Class gmodel_class(void)
   BOOL		loaded = NO;
   NSUnarchiver	*unarchiver = nil;
   NSString      *ext = [fileName pathExtension];
+  id             nibitems = nil;
 
   if ([ext isEqual: @"nib"])
     {
@@ -266,44 +316,66 @@ Class gmodel_class(void)
 		      owner: [context objectForKey: @"NSOwner"]];
     } 
 
-  NSLog(@"Loading Nib `%@'...\n", fileName);
+  NSDebugLog(@"Loading Nib `%@'...\n", fileName);
   NS_DURING
     {
-      NSData	*data = [NSData dataWithContentsOfFile: fileName];
+      NSFileManager	*mgr = [NSFileManager defaultManager];
+      BOOL              isDir = NO;
 
-      if (data != nil)
+      if([mgr fileExistsAtPath: fileName isDirectory: &isDir])
 	{
-	  unarchiver = [[NSUnarchiver alloc] initForReadingWithData: data];
-	  if (unarchiver != nil)
+	  NSData	*data = nil;
+	  
+	  // if the data is in a directory, then load from objects.gorm in the directory
+	  if(isDir == NO)
 	    {
-	      id	obj;
+	      data = [NSData dataWithContentsOfFile: fileName];
+	      NSDebugLog(@"Loaded data from file...");
+	    }
+	  else
+	    {
+	      NSString *newFileName = [fileName stringByAppendingPathComponent: @"objects.gorm"];
+	      data = [NSData dataWithContentsOfFile: newFileName];
+	      NSDebugLog(@"Loaded data from %@...",newFileName);
+	    }
 
-	      [unarchiver setObjectZone: zone];
-	      obj = [unarchiver decodeObject];
-	      if (obj != nil)
+	  if (data != nil)
+	    {
+	      unarchiver = [[NSUnarchiver alloc] initForReadingWithData: data];
+	      if (unarchiver != nil)
 		{
-		  if ([obj isKindOfClass: [GSNibContainer class]])
+		  id obj;
+		  
+		  nibitems = [[_GSNibItemCollector alloc] init];
+		  NSDebugLog(@"Invoking unarchiver");
+		  [unarchiver setObjectZone: zone];
+		  obj = [unarchiver decodeObject];
+		  if (obj != nil)
 		    {
-		      [obj awakeWithContext: context];
-		      /*
-		       *Ok - it's all done now - just retain the nib container
-		       *so that it will not be released when the unarchiver
-		       *is released, and the nib contents will persist.
-		       */
-		      RETAIN(obj);
-		      loaded = YES;
-		    }
-		  else
-		    {
-		      NSLog(@"Nib '%@' without container object!", fileName);
+		      NSArray *items = [nibitems items];
+		      if ([obj isKindOfClass: [GSNibContainer class]])
+			{
+			  NSDebugLog(@"Calling awakeWithContext");
+			    
+			  [obj awakeWithContext: context
+			       topLevelItems: items];
+			  loaded = YES;
+			}
+		      else
+			{
+			  NSLog(@"Nib '%@' without container object!", fileName);
+			}
 		    }
+		  RELEASE(nibitems);
+		  RELEASE(unarchiver);
 		}
-	      RELEASE(unarchiver);
 	    }
 	}
     }
   NS_HANDLER
     {
+      NSLog(@"Exception occured while loading model: %@",[localException reason]);
+      TEST_RELEASE(nibitems);
       TEST_RELEASE(unarchiver);
     }
   NS_ENDHANDLER
@@ -349,6 +421,10 @@ Class gmodel_class(void)
 
   ext = [fileName pathExtension];
   fileName = [fileName stringByDeletingPathExtension];
+  if ([ext isEqualToString: @"nib"] == YES)
+    {
+      ext = @"";
+    }
 
   /*
    * Build an array of resource paths that differs from the normal order -
@@ -388,14 +464,10 @@ Class gmodel_class(void)
 	  path = [rootPath stringByAppendingPathExtension: @"gorm"];
 	  if ([mgr isReadableFileAtPath: path] == NO)
 	    {
-	      path = [rootPath stringByAppendingPathExtension: @"nib"];
+	      path = [rootPath stringByAppendingPathExtension: @"gmodel"];
 	      if ([mgr isReadableFileAtPath: path] == NO)
 		{
-		  path = [rootPath stringByAppendingPathExtension: @"gmodel"];
-		  if ([mgr isReadableFileAtPath: path] == NO)
-		    {
-		      continue;
-		    }
+		  continue;
 		}
 	    }
 	  return path;
@@ -431,278 +503,4 @@ Class gmodel_class(void)
     }
 }
 @end
-
-
-
-@interface NSApplication (GSNibContainer)
-- (void)_deactivateVisibleWindow: (NSWindow *)win;
-@end
-
-@implementation NSApplication (GSNibContainer)
-/* Since awakeWithContext often gets called before the the app becomes
-   active, [win -orderFront:] requests get ignored, so we add the window
-   to the inactive list, so it gets sent an -orderFront when the app
-   becomes active. */
-- (void) _deactivateVisibleWindow: (NSWindow *)win
-{
-  if (_inactive)
-    [_inactive addObject: win];
-}
-@end
-
-/*
- *	The GSNibContainer class manages the internals of a nib file.
- */
-@implementation GSNibContainer
-
-- (void) awakeWithContext: (NSDictionary*)context
-{
-  if (_isAwake == NO)
-    {
-      NSEnumerator	*enumerator;
-      NSNibConnector	*connection;
-      NSString		*key;
-      NSArray		*visible;
-      NSMenu		*menu;
-
-      _isAwake = YES;
-      /*
-       *	Add local entries into name table.
-       */
-      if ([context count] > 0)
-	{
-	  [nameTable addEntriesFromDictionary: context];
-	}
-
-      /*
-       *	Now establish all connections by taking the names
-       *	stored in the connection objects, and replaciong them
-       *	with the corresponding values from the name table
-       *	before telling the connections to establish themselves.
-       */
-      enumerator = [connections objectEnumerator];
-      while ((connection = [enumerator nextObject]) != nil)
-	{
-	  id	val;
-
-	  val = [nameTable objectForKey: [connection source]];
-	  [connection setSource: val];
-	  val = [nameTable objectForKey: [connection destination]];
-	  [connection setDestination: val];
-	  [connection establishConnection];
-	}
-
-      /*
-       * Now tell all the objects that they have been loaded from
-       * a nib.
-       */
-      enumerator = [nameTable keyEnumerator];
-      while ((key = [enumerator nextObject]) != nil)
-	{
-	  if ([context objectForKey: key] == nil)
-	    {
-	      id	o;
-
-	      o = [nameTable objectForKey: key];
-	      if ([o respondsToSelector: @selector(awakeFromNib)])
-		{
-		  [o awakeFromNib];
-		}
-	    }
-	}
-    
-      /*
-       * See if there are objects that should be made visible.
-       */
-      visible = [nameTable objectForKey: @"NSVisible"];
-      if (visible != nil
-	&& [visible isKindOfClass: [NSArray class]] == YES)
-	{
-	  unsigned	pos = [visible count];
-
-	  while (pos-- > 0)
-	    {
-	      NSWindow *win = [visible objectAtIndex: pos];
-	      if ([NSApp isActive])
-		[win orderFront: self];
-	      else
-		[NSApp _deactivateVisibleWindow: win];
-	    }
-	}
-
-      /*
-       * See if there is a main menu to be set.
-       */
-      menu = [nameTable objectForKey: @"NSMenu"];
-      if (menu != nil && [menu isKindOfClass: [NSMenu class]] == YES)
-	{
-	  [NSApp setMainMenu: menu];
-	}
-
-      /*
-       * Now remove any objects added from the context dictionary.
-       */
-      if ([context count] > 0)
-	{
-	  [nameTable removeObjectsForKeys: [context allKeys]];
-	}
-    }
-}
-
-- (NSMutableArray*) connections
-{
-  return connections;
-}
-
-- (void) dealloc
-{
-  RELEASE(nameTable);
-  RELEASE(connections);
-  [super dealloc];
-}
-
-- (void) encodeWithCoder: (NSCoder*)aCoder
-{
-  [aCoder encodeObject: nameTable];
-  [aCoder encodeObject: connections];
-}
-
-- (id) init
-{
-  if ((self = [super init]) != nil)
-    {
-      nameTable = [[NSMutableDictionary alloc] initWithCapacity: 8];
-      connections = [[NSMutableArray alloc] initWithCapacity: 8];
-    }
-  return self;
-}
-
-- (id) initWithCoder: (NSCoder*)aCoder
-{
-  [aCoder decodeValueOfObjCType: @encode(id) at: &nameTable];
-  [aCoder decodeValueOfObjCType: @encode(id) at: &connections];
-  return self;
-}
-
-- (NSMutableDictionary*) nameTable
-{
-  return nameTable;
-}
-
-@end
-
-@implementation	GSNibItem
-+ (void) initialize
-{
-  if (self == [GSNibItem class])
-    {
-      [self setVersion: currentVersion];
-    }
-}
-
-- (void) dealloc
-{
-  RELEASE(theClass);
-  [super dealloc];
-}
-
-- (void) encodeWithCoder: (NSCoder*)aCoder
-{
-  [aCoder encodeObject: theClass];
-  [aCoder encodeRect: theFrame];
-  [aCoder encodeValueOfObjCType: @encode(unsigned int) 
-	  at: &autoresizingMask];
-}
-
-- (id) initWithCoder: (NSCoder*)aCoder
-{
-  int version = [aCoder versionForClassName: 
-			  NSStringFromClass([self class])];
-
-  if (version == 1)
-    {
-      id		obj;
-      Class		cls;
-      unsigned int      mask;
-      
-      [aCoder decodeValueOfObjCType: @encode(id) at: &theClass];
-      theFrame = [aCoder decodeRect];
-      [aCoder decodeValueOfObjCType: @encode(unsigned int) 
-	      at: &mask];
-      
-      cls = NSClassFromString(theClass);
-      if (cls == nil)
-	{
-	  [NSException raise: NSInternalInconsistencyException
-		       format: @"Unable to find class '%@'", theClass];
-	}
-      
-      obj = [cls allocWithZone: [self zone]];
-      if (theFrame.size.height > 0 && theFrame.size.width > 0)
-	obj = [obj initWithFrame: theFrame];
-      else
-	obj = [obj init];
-
-      if ([obj respondsToSelector: @selector(setAutoresizingMask:)])
-	{
-	  [obj setAutoresizingMask: mask];
-	}
-      
-      RELEASE(self);
-      return obj;
-    }
-  else if (version == 0)
-    {
-      id		obj;
-      Class		cls;
-      
-      [aCoder decodeValueOfObjCType: @encode(id) at: &theClass];
-      theFrame = [aCoder decodeRect];
-      
-      cls = NSClassFromString(theClass);
-      if (cls == nil)
-	{
-	  [NSException raise: NSInternalInconsistencyException
-		       format: @"Unable to find class '%@'", theClass];
-	}
-      
-      obj = [cls allocWithZone: [self zone]];
-      if (theFrame.size.height > 0 && theFrame.size.width > 0)
-	obj = [obj initWithFrame: theFrame];
-      else
-	obj = [obj init];
-      
-      RELEASE(self);
-      return obj;
-    }
-  else
-    {
-      NSLog(@"no initWithCoder for this version");
-      RELEASE(self);
-      return nil;
-    }
-}
-
-@end
-
-
-@implementation	GSCustomView
-
-+ (void) initialize
-{
-  if (self == [GSCustomView class])
-    {
-      [self setVersion: currentVersion];
-    }
-}
-
-- (void) encodeWithCoder: (NSCoder*)aCoder
-{
-  [super encodeWithCoder: aCoder];
-}
-
-- (id) initWithCoder: (NSCoder*)aCoder
-{
-  return [super initWithCoder: aCoder];
-}
-@end
+// end of NSBundleAdditions
