@@ -26,17 +26,17 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 
-#include <AppKit/NSButton.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSButtonCell.h>
-#include <AppKit/NSApplication.h>
+#include "AppKit/NSButton.h"
+#include "AppKit/NSWindow.h"
+#include "AppKit/NSButtonCell.h"
+#include "AppKit/NSApplication.h"
 
 //
 // class variables
 //
-id _nsbuttonCellClass = nil;
+static id buttonCellClass = nil;
 
 //
 // NSButton implementation
@@ -60,12 +60,12 @@ id _nsbuttonCellClass = nil;
 //
 + (Class) cellClass
 {
-  return _nsbuttonCellClass;
+  return buttonCellClass;
 }
 
 + (void) setCellClass: (Class)classId
 {
-  _nsbuttonCellClass = classId;
+  buttonCellClass = classId;
 }
 
 //
@@ -86,6 +86,16 @@ id _nsbuttonCellClass = nil;
   [self setNeedsDisplay: YES];
 }
 
+- (void)setHighlightsBy:(int)aType
+{
+  [_cell setHighlightsBy:aType];
+}
+
+- (void)setShowsStateBy:(int)aType
+{
+  [_cell setShowsStateBy:aType];
+}
+
 //
 // Setting the State
 //
@@ -107,7 +117,7 @@ id _nsbuttonCellClass = nil;
 - (void) setState: (int)value
 {
   [_cell setState: value];
-  [self display];
+  [self setNeedsDisplay: YES];
 }
 
 - (int) state
@@ -128,7 +138,7 @@ id _nsbuttonCellClass = nil;
 - (void)setNextState
 {
   [_cell setNextState];
-  [self display];
+  [self setNeedsDisplay: YES];
 }
 
 //
@@ -317,11 +327,6 @@ id _nsbuttonCellClass = nil;
 //
 // Determining the first responder
 //
-- (BOOL) acceptsFirstResponder
-{
-  return [self isEnabled];
-}
-
 - (BOOL) becomeFirstResponder
 {
   [_window disableKeyEquivalentForDefaultButtonCell];
@@ -417,18 +422,4 @@ id _nsbuttonCellClass = nil;
   return [_cell sound];
 }
 
-//
-// NSCoding protocol
-//
-- (void) encodeWithCoder: (NSCoder *)aCoder
-{
-  [super encodeWithCoder: aCoder];
-}
-
-- (id) initWithCoder: (NSCoder *)aDecoder
-{
-  [super initWithCoder: aDecoder];
-
-  return self;
-}
 @end
