@@ -30,23 +30,27 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSLock.h>
 #include <Foundation/NSArray.h>
 #include <Foundation/NSString.h>
 #include <Foundation/NSException.h>
+#include <Foundation/NSDebug.h>
+#include <Foundation/NSValue.h>
+#include <GNUstepBase/GSCategories.h>
 
-#include <AppKit/AppKitExceptions.h>
-#include <AppKit/NSButtonCell.h>
-#include <AppKit/NSButton.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSEvent.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/PSOperators.h>
+#include "AppKit/AppKitExceptions.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSButtonCell.h"
+#include "AppKit/NSButton.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSEvent.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSSound.h"
+#include "AppKit/NSWindow.h"
+#include "GNUstepGUI/GSDrawFunctions.h"
 
 @implementation NSButtonCell
 
@@ -93,8 +97,6 @@
 {
   [super initImageCell: anImage];
 
-  _contents = nil;
-
   return [self _init];
 }
 
@@ -111,6 +113,7 @@
   RELEASE(_altImage);
   RELEASE(_keyEquivalent);
   RELEASE(_keyEquivalentFont);
+  RELEASE(_sound);
 
   [super dealloc];
 }
@@ -139,13 +142,29 @@
 	value = 1;
       break;
     case NSChangeGrayCell:
+      if (_showAltStateMask & NSChangeGrayCellMask)
+	value = 1;
+      break;
+    case NSCellLightsByGray:
       if (_highlightsByMask & NSChangeGrayCellMask)
 	value = 1;
       break;
     case NSChangeBackgroundCell:
+      if (_showAltStateMask & NSChangeBackgroundCellMask)
+	value = 1;
+      break;
+    case NSCellLightsByBackground:
       if (_highlightsByMask & NSChangeBackgroundCellMask)
 	value = 1;
       break;
+    case NSCellChangesContents:
+      if (_showAltStateMask & NSContentsCellMask)
+	value = 1;
+      break;
+    case NSCellLightsByContents:
+      if (_highlightsByMask & NSContentsCellMask)
+	value = 1;
+      break;
     default:
       value = [super cellAttribute: aParameter];
       break;
@@ -166,24 +185,22 @@
       break;
     case NSChangeGrayCell:
       if (value)
-	_highlightsByMask |= NSChangeGrayCellMask;
+	_showAltStateMask |= NSChangeGrayCellMask;
       else
-	_highlightsByMask &= ~NSChangeGrayCellMask;
+	_showAltStateMask &= ~NSChangeGrayCellMask;
       break;
     case NSChangeBackgroundCell:
       if (value)
-	_highlightsByMask |= NSChangeBackgroundCellMask;
+	_showAltStateMask |= NSChangeBackgroundCellMask;
       else
-	_highlightsByMask &= ~NSChangeBackgroundCellMask;
-      break;
-      /*
-    case NSCellLightsByContents: 
-      _cell. = value;
+	_showAltStateMask &= ~NSChangeBackgroundCellMask;
       break;
     case NSCellChangesContents:
-      _cell. = value;
+      if (value)
+	_showAltStateMask |= NSContentsCellMask;
+      else
+	_showAltStateMask &= ~NSContentsCellMask;
       break;
-      */
     case NSCellLightsByGray:
       if (value)
 	_highlightsByMask |= NSChangeGrayCellMask;
@@ -196,6 +213,12 @@
       else
 	_highlightsByMask &= ~NSChangeBackgroundCellMask;
       break;
+    case NSCellLightsByContents:
+      if (value)
+	_highlightsByMask |= NSContentsCellMask;
+      else
+	_highlightsByMask &= ~NSContentsCellMask;
+      break;
     default:
       [super setCellAttribute: aParameter to: value];
     }
@@ -203,8 +226,15 @@
 
 - (void) setFont: (NSFont*)fontObject		
 {
-  // TODO Should change the size of the key equivalent font
+  int size;
+
   [super setFont: fontObject];
+  if ((_keyEquivalentFont != nil) && (fontObject != nil) && 
+      ((size = [fontObject pointSize]) != [_keyEquivalentFont pointSize]))
+    {
+      [self setKeyEquivalentFont: [_keyEquivalentFont fontName] 
+	    size: size];       
+    }
 }
 
 - (void) setTitle: (NSString*)aString
@@ -217,14 +247,26 @@
   ASSIGNCOPY(_altContents, aString);
 
   if (_control_view)
-    if ([_control_view isKindOfClass: [NSControl class]])
-      [(NSControl*)_control_view updateCell: self];
+    {
+      if ([_control_view isKindOfClass: [NSControl class]])
+	{
+	  [(NSControl*)_control_view updateCell: self];
+	}
+    }
 }
 
 - (NSAttributedString *)attributedAlternateTitle
 {
   // TODO
-  return nil;
+  NSDictionary *dict;
+  NSAttributedString *attrStr;
+
+  dict = [self _nonAutoreleasedTypingAttributes];
+  attrStr = [[NSAttributedString alloc] initWithString: _altContents 
+					attributes: dict];
+  RELEASE(dict);
+
+  return AUTORELEASE(attrStr);
 }
 
 - (void)setAttributedAlternateTitle:(NSAttributedString *)aString
@@ -371,13 +413,7 @@
 
 - (BOOL) isOpaque
 {
-  /*
-   * MacOS-X says we should return !transparent && [self isBordered], 
-   * but that's wrong in our case, since if there is no border, 
-   * we draw the interior of the cell to fill completely the bounds.  
-   * They are likely to draw differently.
-   */
-  return !_buttoncell_is_transparent;
+  return !_buttoncell_is_transparent && _cell.is_bordered;
 }
 
 - (NSBezelStyle)bezelStyle
@@ -392,13 +428,18 @@
 
 - (BOOL)showsBorderOnlyWhileMouseInside
 {
-    return _shows_border_only_while_mouse_inside;
+  return _shows_border_only_while_mouse_inside;
 }
 
 - (void)setShowsBorderOnlyWhileMouseInside:(BOOL)show
 {
-  // FIXME: Switch mouse tracking on
+  if (_shows_border_only_while_mouse_inside == show)
+    {
+      return;
+    }
+
   _shows_border_only_while_mouse_inside = show;
+  // FIXME Switch mouse tracking on
 }
 
 - (NSGradientType)gradientType
@@ -445,7 +486,7 @@
 
   switch (buttonType)
     {
-      case NSMomentaryLight: 
+      case NSMomentaryLightButton: 
 	[self setHighlightsBy: NSChangeBackgroundCellMask];
 	[self setShowsStateBy: NSNoCellMask];
 	[self setImageDimsWhenDisabled: YES];
@@ -478,8 +519,8 @@
       case NSSwitchButton: 
 	[self setHighlightsBy: NSContentsCellMask];
 	[self setShowsStateBy: NSContentsCellMask];
-	[self setImage: [NSImage imageNamed: @"common_SwitchOff"]];
-	[self setAlternateImage: [NSImage imageNamed: @"common_SwitchOn"]];
+	[self setImage: [NSImage imageNamed: @"NSSwitch"]];
+	[self setAlternateImage: [NSImage imageNamed: @"NSHighlightedSwitch"]];
 	[self setImagePosition: NSImageLeft];
 	[self setAlignment: NSLeftTextAlignment];
 	[self setBordered: NO];
@@ -489,8 +530,8 @@
       case NSRadioButton: 
 	[self setHighlightsBy: NSContentsCellMask];
 	[self setShowsStateBy: NSContentsCellMask];
-	[self setImage: [NSImage imageNamed: @"common_RadioOff"]];
-	[self setAlternateImage: [NSImage imageNamed: @"common_RadioOn"]];
+	[self setImage: [NSImage imageNamed: @"NSRadioButton"]];
+	[self setAlternateImage: [NSImage imageNamed: @"NSHighlightedRadioButton"]];
 	[self setImagePosition: NSImageLeft];
 	[self setAlignment: NSLeftTextAlignment];
 	[self setBordered: NO];
@@ -535,8 +576,37 @@
   return _cell.state;
 }
 
-// FIXME: The spec says that the stringValue and setStringValue methods should 
-// also be redefined. But this does not fit to the way we uses this for the title.
+- (void) setObjectValue: (id)object 
+{
+  if (object == nil)
+    {
+      [self setState: NSOffState];
+    }
+  else if ([object respondsToSelector: @selector(intValue)])
+    {
+      [self setState: [object intValue]];
+    }
+  else
+    {
+      [self setState: NSOnState];
+    }
+}
+
+- (id) objectValue
+{
+  if (_cell.state == NSOffState)
+    {	
+      return [NSNumber numberWithBool: NO];
+    }
+  else if (_cell.state == NSOnState)
+    {	
+      return [NSNumber numberWithBool: YES];
+    }
+  else // NSMixedState
+    {	
+      return [NSNumber numberWithInt: -1];
+    }
+}
 
 /*
  * Displaying
@@ -566,19 +636,18 @@
     return;
 
   // draw the border if needed
-  if (_cell.is_bordered)
+  if ((_cell.is_bordered) && 
+      (!_shows_border_only_while_mouse_inside || _mouse_inside))
     {
-      [controlView lockFocus];
       // FIXME Should check the bezel and gradient style
       if (_cell.is_highlighted && (_highlightsByMask & NSPushInCellMask))
         {
-          NSDrawGrayBezel(cellFrame, NSZeroRect);
+	  [GSDrawFunctions drawGrayBezel: cellFrame : NSZeroRect];
         }
       else
         {
-          NSDrawButton(cellFrame, NSZeroRect);
+	  [GSDrawFunctions drawButton: cellFrame : NSZeroRect];
         }
-      [controlView unlockFocus];
     }
 
   [self drawInteriorWithFrame: cellFrame inView: controlView];
@@ -589,7 +658,7 @@
   unsigned	mask;
   NSImage	*imageToDisplay;
   NSRect	imageRect;
-  NSString	*titleToDisplay;
+  NSAttributedString	*titleToDisplay;
   NSRect	titleRect;
   NSSize	imageSize = {0, 0};
   NSSize        titleSize = {0, 0};
@@ -604,7 +673,6 @@
   _control_view = controlView;
 
   cellFrame = [self drawingRectForBounds: cellFrame];
-  [controlView lockFocus];
 
   if (_cell.is_highlighted)
     {
@@ -624,19 +692,31 @@
       cellFrame = NSOffsetRect(cellFrame, 1., flippedView ? 1. : -1.);
     }
 
-  /* Determine the background color. */
-  if (mask & (NSChangeGrayCellMask | NSChangeBackgroundCellMask))
-    {
-      backgroundColor = [NSColor selectedControlColor];
-    }
-
-  if (backgroundColor == nil)
-    backgroundColor = [NSColor controlBackgroundColor];
-
-  /* Draw the cell's background color.  We always draw the background,
+  /* Draw the cell's background color.  
+     We draw when there is a border or when highlightsByMask
+     is NSChangeBackgroundCellMask or NSChangeGrayCellMask,
      as required by our nextstep-like look and feel.  */
-  [backgroundColor set];
-  NSRectFill (cellFrame);
+  if (_cell.is_bordered 
+      || (_highlightsByMask & NSChangeBackgroundCellMask)
+      || (_highlightsByMask & NSChangeGrayCellMask))
+    {
+      /* Determine the background color. */
+      if (mask & (NSChangeGrayCellMask | NSChangeBackgroundCellMask))
+        {
+          backgroundColor = [NSColor selectedControlColor];
+        }
+      else if (_cell.is_bordered) 
+        {
+          backgroundColor = [NSColor controlBackgroundColor];
+        }
+      
+      if (backgroundColor != nil) 
+        {
+          [backgroundColor set];
+          NSRectFill (cellFrame);
+        }
+      
+    }
 
   /*
    * Determine the image and the title that will be
@@ -649,31 +729,29 @@
     {
       imageToDisplay = _altImage;
       if (!imageToDisplay)
-	imageToDisplay = _cell_image;
-      titleToDisplay = _altContents;
-      if (titleToDisplay == nil || [titleToDisplay isEqual: @""])
-        titleToDisplay = _contents;
+        {
+	  imageToDisplay = _cell_image;
+	}
+      titleToDisplay = [self attributedAlternateTitle];
+      if (titleToDisplay == nil || [titleToDisplay length] == 0)
+        {
+	  titleToDisplay = [self attributedTitle];
+	}
     }
   else
     {
       imageToDisplay = _cell_image;
-      titleToDisplay = _contents;
+      titleToDisplay = [self attributedTitle];
     }
 
   if (imageToDisplay)
     {
-      /* FIXME - the following is a hack!  Because we don't seem to be
-	 getting alpha composing of images right, we use this hack of
-	 hard-setting manually the background color of the image to
-	 the wanted background color ... this should go away when
-	 alpha composing of images works 100%.  */
-      [imageToDisplay setBackgroundColor: backgroundColor];
       imageSize = [imageToDisplay size];
     }
 
   if (titleToDisplay && (ipos == NSImageAbove || ipos == NSImageBelow))
     {
-      titleSize = [self _sizeText: titleToDisplay];
+      titleSize = [titleToDisplay size];
     }
 
   if (flippedView == YES)
@@ -713,10 +791,7 @@
 	titleRect = imageRect;
 	titleRect.origin.x += imageSize.width + xDist;
 	titleRect.size.width = cellFrame.size.width - imageSize.width - xDist;
-	if (_cell.is_bordered || _cell.is_bezeled) 
-	  {
-	    titleRect.size.width -= 3;
-	  }
+        titleRect.size.width -= 3;
 	break;
 
       case NSImageRight: 
@@ -733,11 +808,8 @@
 	titleRect.origin = cellFrame.origin;
 	titleRect.size.width = cellFrame.size.width - imageSize.width - xDist;
 	titleRect.size.height = cellFrame.size.height;
-	if (_cell.is_bordered || _cell.is_bezeled) 
-	  {
-	    titleRect.origin.x += 3;
-	    titleRect.size.width -= 3;
-	  }
+	titleRect.origin.x += 3;
+	titleRect.size.width -= 3;
 	break;
 
       case NSImageAbove: 
@@ -760,14 +832,13 @@
 
 	if (_cell.is_bordered || _cell.is_bezeled) 
 	  {
+	    imageRect.origin.x += 3;
 	    imageRect.size.width -= 6;
-	    imageRect.origin.x   += 3;
-	    titleRect.size.width -= 6;
-	    titleRect.origin.x   += 3;
 	    imageRect.size.height -= 1;
-	    titleRect.size.height -= 1;
-	    titleRect.origin.y    += 1;
 	  }
+	titleRect.origin.x += 3;
+	titleRect.origin.y += 4;
+	titleRect.size.width -= 6;
 	break;
 
       case NSImageBelow: 
@@ -792,12 +863,12 @@
 	  {
 	    imageRect.size.width -= 6;
 	    imageRect.origin.x   += 3;
-	    titleRect.size.width -= 6;
-	    titleRect.origin.x   += 3;
 	    imageRect.size.height -= 1;
 	    imageRect.origin.y    += 1;
-	    titleRect.size.height -= 1;
 	  }
+	titleRect.size.width -= 6;
+	titleRect.origin.x   += 3;
+	titleRect.size.height -= 4;
 	break;
 
       case NSImageOverlaps: 
@@ -826,19 +897,14 @@
     }
   if (titleToDisplay != nil)
     {
-      [self _drawText: titleToDisplay inFrame: titleRect];
+      [self _drawAttributedText: titleToDisplay inFrame: titleRect];
     }
 
   if (_cell.shows_first_responder
       && [[controlView window] firstResponder] == controlView)
     {
-      if (_cell.is_bordered || _cell.is_bezeled)
 	NSDottedFrameRect(cellFrame);
-      else
-	NSDottedFrameRect(titleRect);
     }
-
-  [controlView unlockFocus];
 }
 
 - (NSSize) cellSize 
@@ -847,7 +913,7 @@
   NSSize borderSize;
   unsigned	mask;
   NSImage	*imageToDisplay;
-  NSString	*titleToDisplay;
+  NSAttributedString	*titleToDisplay;
   NSSize	imageSize;
   NSSize	titleSize;
   
@@ -855,35 +921,49 @@
    * The following code must be kept in sync with -drawInteriorWithFrame
    */ 
   if (_cell.is_highlighted)
-    mask = _highlightsByMask;
+    {
+      mask = _highlightsByMask;
+    }
   else if (_cell.state)
-    mask = _showAltStateMask;
+    {
+      mask = _showAltStateMask;
+    }
   else
-    mask = NSNoCellMask;
-
+    {
+      mask = NSNoCellMask;
+    }
+  
   if (mask & NSContentsCellMask)
     {
       imageToDisplay = _altImage;
       if (!imageToDisplay)
-	imageToDisplay = _cell_image;
-      titleToDisplay = _altContents;
-      if (titleToDisplay == nil || [titleToDisplay isEqual: @""])
-	titleToDisplay = _contents;
+	{
+	  imageToDisplay = _cell_image;
+	}
+      titleToDisplay = [self attributedAlternateTitle];
+      if (titleToDisplay == nil || [titleToDisplay length] == 0)
+        {
+	  titleToDisplay = [self attributedTitle];
+	}
     }
   else
     {
       imageToDisplay = _cell_image;
-      titleToDisplay = _contents;
+      titleToDisplay = [self attributedTitle];
     }
   
   if (imageToDisplay)
-    imageSize = [imageToDisplay size];
-  else 
-    imageSize = NSZeroSize;
-  
+    {
+      imageSize = [imageToDisplay size];
+    }
+  else
+    { 
+      imageSize = NSZeroSize;
+    }
+
   if (titleToDisplay != nil)
     {
-      titleSize = [self _sizeText: titleToDisplay];
+      titleSize = [titleToDisplay size];
     }
   else
     {
@@ -944,6 +1024,10 @@
     {
       borderSize.height += 1;
       borderSize.width  += 3;
+
+      /* Add some more space because it looks better.  */
+      borderSize.height += 2;
+      borderSize.width  += 3;
     }
   
   // Add border size
@@ -970,32 +1054,39 @@
 			 theRect.size.height - 3.);
     }
   else
-    return theRect;
+    {
+      return theRect;
+    }
 }
 
-- (void)setSound:(NSSound *)aSound
+- (void) setSound: (NSSound *)aSound
 {
-  _sound = aSound;
+  ASSIGN(_sound, aSound);
 }
 
-- (NSSound *)sound
+- (NSSound *) sound
 {
   return _sound;
 }
 
-- (void)mouseEntered:(NSEvent *)event
+- (void) mouseEntered: (NSEvent *)event
 {
   _mouse_inside = YES;
+  [(NSView *)[event userData] setNeedsDisplay: YES];
 }
 
-- (void)mouseExited:(NSEvent *)event
+- (void) mouseExited: (NSEvent *)event
 {
   _mouse_inside = NO;
+  [(NSView *)[event userData] setNeedsDisplay: YES];
 }
 
-- (void)performClick:(id)sender
+- (void) performClick: (id)sender
 {
-  // TODO Like super plus playing the sound
+  if (_sound != nil)
+    {
+      [_sound play];
+    }
   [super performClick: sender];
 }
 
@@ -1037,8 +1128,6 @@
   BOOL tmp;
   [super encodeWithCoder: aCoder];
 
-  NSDebugLog(@"NSButtonCell: start encoding\n");
-
   [aCoder encodeObject: _keyEquivalent];
   [aCoder encodeObject: _keyEquivalentFont];
   [aCoder encodeObject: _altContents];
@@ -1053,32 +1142,66 @@
   [aCoder encodeValueOfObjCType: @encode(unsigned int)
 			     at: &_showAltStateMask];
 
-  NSDebugLog(@"NSButtonCell: finish encoding\n");
 }
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  // FIXME: Add new ivars
-  BOOL tmp;
-  [super initWithCoder: aDecoder];
+  self = [super initWithCoder: aDecoder];
 
-  NSDebugLog(@"NSButtonCell: start decoding\n");
+  if ([aDecoder allowsKeyedCoding])
+    {
+      NSString *alternateContents = [aDecoder decodeObjectForKey: @"NSAlternateContents"];
+      NSImage *alternateImage = [aDecoder decodeObjectForKey: @"NSAlternateImage"];
+      //NSControl *control = [aDecoder decodeObjectForKey: @"NSControlView"];
+      NSString *key = [aDecoder decodeObjectForKey: @"NSKeyEquivalent"];
+      int bFlags;
+      int bFlags2;
+      int delay = 0;
+      int interval = 0;
+      
+      [self setAlternateImage: alternateImage];
+      [self setAlternateTitle: alternateContents];
+      [self setKeyEquivalent: key];
 
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_keyEquivalent];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_keyEquivalentFont];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_altContents];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_altImage];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &tmp];
-  _buttoncell_is_transparent = tmp;
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int)
-			       at: &_keyEquivalentModifierMask];
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int)
-			       at: &_highlightsByMask];
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int)
-			       at: &_showAltStateMask];
-
-  NSDebugLog(@"NSButtonCell: finish decoding\n");
+      if ([aDecoder containsValueForKey: @"NSButtonFlags"])
+        {
+	  bFlags = [aDecoder decodeIntForKey: @"NSButtonFlags"];
+	  // FIXME
+	}
+      if ([aDecoder containsValueForKey: @"NSButtonFlags2"])
+        {
+	  bFlags2 = [aDecoder decodeIntForKey: @"NSButtonFlags2"];
+	  // FIXME
+	}
 
+      if ([aDecoder containsValueForKey: @"NSPeriodicDelay"])
+        {
+	  delay = [aDecoder decodeIntForKey: @"NSPeriodicDelay"];
+	}
+      if ([aDecoder containsValueForKey: @"NSPeriodicInterval"])
+        {
+	  interval = [aDecoder decodeIntForKey: @"NSPeriodicInterval"];
+	}
+      [self setPeriodicDelay: delay interval: interval];
+    }
+  else
+    {
+      // FIXME: Add new ivars
+      BOOL tmp;
+      
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_keyEquivalent];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_keyEquivalentFont];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_altContents];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_altImage];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &tmp];
+      _buttoncell_is_transparent = tmp;
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int)
+		                   at: &_keyEquivalentModifierMask];
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int)
+		                   at: &_highlightsByMask];
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int)
+		                   at: &_showAltStateMask];
+    }
   return self;
 }
 
