@@ -31,27 +31,35 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSString.h>
 #include <Foundation/NSGeometry.h>
 #include <Foundation/NSException.h>
 #include <Foundation/NSValue.h>
+#include <Foundation/NSNotification.h>
+#include <Foundation/NSDebug.h>
+#include <Foundation/NSFormatter.h>
+#include <Foundation/NSRunLoop.h>
+#include <GNUstepBase/GSCategories.h>
 
-#include <AppKit/AppKitExceptions.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSView.h>
-#include <AppKit/NSControl.h>
-#include <AppKit/NSCell.h>
-#include <AppKit/NSCursor.h>
-#include <AppKit/NSEvent.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSParagraphStyle.h>
-#include <AppKit/NSMenu.h>
-#include <AppKit/PSOperators.h>
+#include "AppKit/AppKitExceptions.h"
+#include "AppKit/NSAttributedString.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSControl.h"
+#include "AppKit/NSCell.h"
+#include "AppKit/NSClipView.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSCursor.h"
+#include "AppKit/NSEvent.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSMenu.h"
+#include "AppKit/NSParagraphStyle.h"
+#include "AppKit/NSTextView.h"
+#include "AppKit/NSTextContainer.h"
+#include "AppKit/NSView.h"
+#include "AppKit/NSWindow.h"
 
 static Class	colorClass;
 static Class	cellClass;
@@ -145,9 +153,9 @@ static NSColor	*shadowCol;
   //_cell.is_bezeled = NO;
   //_cell.is_scrollable = NO;
   //_cell.is_selectable = NO;
-  //_cell.is_continuous = NO;
   //_cell.state = 0;
   _action_mask = NSLeftMouseUpMask;
+  _menu = [isa defaultMenu];
 
   return self;
 }
@@ -171,8 +179,8 @@ static NSColor	*shadowCol;
   //_cell.is_bezeled = NO;
   //_cell.is_scrollable = NO;
   //_cell.is_selectable = NO;
-  //_cell.is_continuous = NO;
   _action_mask = NSLeftMouseUpMask;
+  _menu = [isa defaultMenu];
 
   return self;
 }
@@ -212,37 +220,40 @@ static NSColor	*shadowCol;
 
 - (double) doubleValue
 {
-  if (_cell.contents_is_attributed_string == NO)
+  if ((_cell.has_valid_object_value == YES) &&
+      ([_objectValue respondsToSelector: @selector(doubleValue)]))
     {
-      return [_contents doubleValue];
+      return [_objectValue doubleValue];
     }
   else
     {
-      return [[(NSAttributedString *)_contents string] doubleValue];
+      return [[self stringValue] doubleValue];
     }
 }
 
 - (float) floatValue
 {
-  if (_cell.contents_is_attributed_string == NO)
+  if ((_cell.has_valid_object_value == YES) &&
+      ([_objectValue respondsToSelector: @selector(floatValue)]))
     {
-      return [_contents floatValue];
+      return [_objectValue floatValue];
     }
   else
     {
-      return [[(NSAttributedString *)_contents string] floatValue];      
+      return [[self stringValue] floatValue];
     }
 }
 
 - (int) intValue
 {
-  if (_cell.contents_is_attributed_string == NO)
+  if ((_cell.has_valid_object_value == YES) &&
+      ([_objectValue respondsToSelector: @selector(intValue)]))
     {
-      return [_contents intValue];
+      return [_objectValue intValue];
     }
   else
     {
-      return [[(NSAttributedString *)_contents string] intValue];
+      return [[self stringValue] intValue];
     }
 }
 
@@ -250,11 +261,12 @@ static NSColor	*shadowCol;
 {
   if (_cell.contents_is_attributed_string == NO)
     {
+      // If we have a formatter this is also the string of the _objectValue
       return _contents;
     }
   else
     {
-      return AUTORELEASE ([[(NSAttributedString *)_contents string] copy]);
+      return [(NSAttributedString *)_contents string];
     }
 }
 
@@ -397,7 +409,7 @@ static NSColor	*shadowCol;
       }
     default:
       {
-	NSDebugLog (@"cell attribute %d not supported", aParameter);
+	NSWarnLog (@"cell attribute %d not supported", aParameter);
 	break;
       }
     }
@@ -513,7 +525,7 @@ static NSColor	*shadowCol;
       }
     default:
       {
-	NSDebugLog (@"cell attribute %d not supported", aParameter);
+	NSWarnLog (@"cell attribute %d not supported", aParameter);
 	break;
       }
     }
@@ -710,6 +722,7 @@ static NSColor	*shadowCol;
 
 - (void) setAlignment: (NSTextAlignment)mode
 {
+  // This does not have any influence on attributed strings
   _cell.text_align = mode;
 }
 
@@ -725,7 +738,13 @@ static NSColor	*shadowCol;
 
 - (void) setFont: (NSFont*)fontObject
 {
+  // This does not have any influence on attributed strings
   ASSIGN (_font, fontObject);
+
+  if (_cell.type != NSTextCellType)
+    {
+      _cell.type = NSTextCellType;
+    }
 }
 
 - (void) setSelectable: (BOOL)flag
@@ -790,10 +809,10 @@ static NSColor	*shadowCol;
       NSDictionary *attributes;
       NSAttributedString *attrStr;
 
-      attributes = AUTORELEASE ([self _nonAutoreleasedTypingAttributes]);
+      attributes = [self _nonAutoreleasedTypingAttributes];
       attrStr = [_formatter attributedStringForObjectValue: _objectValue 
 			    withDefaultAttributes: attributes];
-
+      RELEASE(attributes);
       if (attrStr != nil)
 	{
 	  return attrStr;
@@ -810,10 +829,11 @@ static NSColor	*shadowCol;
       NSDictionary *dict;
       NSAttributedString *attrStr;
 
-      dict = AUTORELEASE ([self _nonAutoreleasedTypingAttributes]);
+      dict = [self _nonAutoreleasedTypingAttributes];
       attrStr = [[NSAttributedString alloc] initWithString: _contents 
 					    attributes: dict];
-      return AUTORELEASE (attrStr);
+      RELEASE(dict);
+      return AUTORELEASE(attrStr);
     }
 }
 
@@ -865,6 +885,16 @@ static NSColor	*shadowCol;
   return textObject;
 }
 
+- (NSString*) title						
+{
+  return [self stringValue];
+}
+
+- (void) setTitle: (NSString*)aString
+{
+  [self setStringValue: aString];
+}
+
 /*
  * Target and Action
  */
@@ -892,13 +922,21 @@ static NSColor	*shadowCol;
 
 - (BOOL) isContinuous
 {
-  return _cell.is_continuous;
+  // Some subclasses should redefine this with NSLeftMouseDraggedMask
+  return (_action_mask & NSPeriodicMask) != 0;
 }
 
 - (void) setContinuous: (BOOL)flag
 {
-  _cell.is_continuous = flag;
-  [self sendActionOn: (NSLeftMouseUpMask|NSPeriodicMask)];
+  // Some subclasses should redefine this with NSLeftMouseDraggedMask
+  if (flag)
+    {
+      _action_mask |= NSPeriodicMask;
+    }
+  else 
+    {
+      _action_mask &= ~NSPeriodicMask;
+    }
 }
 
 - (int) sendActionOn: (int)mask
@@ -1018,8 +1056,11 @@ static NSColor	*shadowCol;
   return [self menu];
 }
 
-/*
- * Comparing to Another NSCell
+/**
+ * Compares the reciever to another to another NSCell.
+ * The argument must be an NSCell sublclass and have
+ * the NSCellType NSTextCellType.  Returns the result
+ * of the comparison of each cell's stringValue.
  */
 - (NSComparisonResult) compare: (id)otherCell
 {
@@ -1034,7 +1075,9 @@ static NSColor	*shadowCol;
       [NSException raise: NSBadComparisonException
 		   format: @"Comparison between non-text cells"];
     }
-  return [_contents compare: ((NSCell*)otherCell)->_contents];
+  /* We shouldn't access instance variables directly as subclasses
+     may override stringValue to retrieve the value from somewhere else.  */
+  return [[self stringValue] compare: [(NSCell*)otherCell stringValue]];
 }
 
 /*
@@ -1059,7 +1102,7 @@ static NSColor	*shadowCol;
 {
   unsigned int location = [aString rangeOfString: @"&"].location;
 
-  [self setStringValue: [aString stringByReplacingString: @"&"
+  [self setTitle: [aString stringByReplacingString: @"&"
 				 withString: @""]];
   // TODO: We should underline this character
   [self setMnemonicLocation: location];
@@ -1068,16 +1111,7 @@ static NSColor	*shadowCol;
 - (NSString*) mnemonic
 {
   unsigned int location = [self mnemonicLocation];
-  NSString *c;
-
-  if (_cell.contents_is_attributed_string)
-    {
-      c = AUTORELEASE ([[(NSAttributedString *)_contents string] copy]);
-    }
-  else
-    {
-      c = _contents;
-    }
+  NSString *c = [self title];
 
   if ((location == NSNotFound) || location >= [c length])
     return @"";
@@ -1123,6 +1157,8 @@ static NSColor	*shadowCol;
     {  
       NSRect   cvBounds = [cv bounds];
       NSWindow *cvWin = [cv window];
+
+      [cv lockFocus];
       
       [self setNextState];
       [self highlight: YES withFrame: cvBounds inView: cv];
@@ -1135,6 +1171,8 @@ static NSColor	*shadowCol;
       [self highlight: NO withFrame: cvBounds inView: cv];
       [cvWin flushWindow];
       
+      [cv unlockFocus];
+
       if (action)
 	{
 	  NS_DURING
@@ -1267,11 +1305,8 @@ static NSColor	*shadowCol;
   BOOL		done;
   BOOL		mouseWentUp;
 
-  NSDebugLog(@"NSCell start tracking\n");
-  NSDebugLog(@"NSCell tracking in rect %f %f %f %f\n",
-		      cellFrame.origin.x, cellFrame.origin.y,
-		      cellFrame.size.width, cellFrame.size.height);
-  NSDebugLog(@"NSCell initial point %f %f\n", point.x, point.y);
+  NSDebugLLog(@"NSCell", @"cell start tracking in rect %@ initial point %f %f",
+	     NSStringFromRect(cellFrame), point.x, point.y);
 
   _mouse_down_flags = [theEvent modifierFlags];
   if (![self startTrackingAt: point inView: controlView])
@@ -1284,14 +1319,14 @@ static NSColor	*shadowCol;
       && [theEvent type] == NSLeftMouseDown)
     [(NSControl*)controlView sendAction: action to: target];
 
-  if (_cell.is_continuous)
+  if (_action_mask & NSPeriodicMask)
     {
       [self getPeriodicDelay: &delay interval: &interval];
       [NSEvent startPeriodicEventsAfterDelay: delay withPeriod: interval];
       event_mask |= NSPeriodicMask;
     }
 
-  NSDebugLog(@"NSCell get mouse events\n");
+  NSDebugLLog(@"NSCell", @"cell get mouse events\n");
   mouseWentUp = NO;
   done = NO;
   while (!done)
@@ -1325,23 +1360,23 @@ static NSColor	*shadowCol;
 	      location = [theEvent locationInWindow];
 	    }
 	  point = [controlView convertPoint: location fromView: nil];
-	  NSDebugLog(@"NSCell location %f %f\n", location.x, location.y);
-	  NSDebugLog(@"NSCell point %f %f\n", point.x, point.y);
+	  NSDebugLLog(@"NSCell", @"location %f %f\n", location.x, location.y);
+	  NSDebugLLog(@"NSCell", @"point %f %f\n", point.x, point.y);
 	}
       else
 	{
 	  periodCount++;
-	  NSDebugLog (@"got a periodic event");
+	  NSDebugLLog (@"NSCell", @"cell got a periodic event");
 	}
 
       if (![controlView mouse: point inRect: cellFrame])
 	{
-	  NSDebugLog(@"NSCell point not in cell frame\n");
+	  NSDebugLLog(@"NSCell", @"point not in cell frame\n");
 
 	  pointIsInCell = NO;	
 	  if (flag == NO) 
 	    {
-	      NSDebugLog(@"NSCell return immediately\n");
+	      NSDebugLLog(@"NSCell", @"cell return immediately\n");
 	      done = YES;
 	    }
 	}
@@ -1354,14 +1389,14 @@ static NSColor	*shadowCol;
 					at: point 	// tracking?
 				    inView: controlView])
 	{
-	  NSDebugLog(@"NSCell stop tracking\n");
+	  NSDebugLLog(@"NSCell", @"cell stop tracking\n");
 	  done = YES;
 	}
       
       // Did the mouse go up?
       if (eventType == NSLeftMouseUp)
 	{
-	  NSDebugLog(@"NSCell mouse went up\n");
+	  NSDebugLLog(@"NSCell", @"cell mouse went up\n");
 	  mouseWentUp = YES;
 	  done = YES;
 	}
@@ -1381,7 +1416,7 @@ static NSColor	*shadowCol;
 	      inView: controlView
 	   mouseIsUp: mouseWentUp];
 
-  if (_cell.is_continuous)
+  if (_action_mask & NSPeriodicMask)
     [NSEvent stopPeriodicEvents];
 
   if (mouseWentUp)
@@ -1394,11 +1429,11 @@ static NSColor	*shadowCol;
   // Return YES only if the mouse went up within the cell
   if (mouseWentUp && (flag || [controlView mouse: point inRect: cellFrame]))
     {
-      NSDebugLog(@"NSCell mouse went up in cell\n");
+      NSDebugLLog(@"NSCell", @"mouse went up in cell\n");
       return YES;
     }
 
-  NSDebugLog(@"NSCell mouse did not go up in cell\n");
+  NSDebugLLog(@"NSCell", @"mouse did not go up in cell\n");
   return NO;				// Otherwise return NO
 }
 
@@ -1468,21 +1503,16 @@ static NSColor	*shadowCol;
     {
       case NSTextCellType:
 	{
-	  if (_cell.contents_is_attributed_string)
+	  NSAttributedString *attrStr;
+
+	  attrStr = [self attributedStringValue];
+	  if ([attrStr length] != 0)
 	    {
-	      s = [(NSAttributedString *)_contents size];
+	      s = [attrStr size];
 	    }
 	  else
 	    {
-	      if ((_contents != nil) 
-		  && ([_contents isEqualToString: @""] == NO))
-		{
-		  s = [self _sizeText: _contents];
-		}
-	      else 
-		{
-		  s = [self _sizeText: @"A"];
-		}
+	      s = [self _sizeText: @"A"];
 	    }
 	}
 	break;
@@ -1563,6 +1593,28 @@ static NSColor	*shadowCol;
     }
 }
 
+- (void) setControlSize: (NSControlSize)controlSize
+{
+  // FIXME
+}
+
+- (NSControlSize) controlSize
+{
+  // FIXME
+  return NSRegularControlSize;
+}
+
+- (void) setControlTint: (NSControlTint)controlTint
+{
+  // FIXME 
+}
+
+- (NSControlTint) controlTint
+{
+  // FIXME
+  return NSDefaultControlTint;
+}
+
 /*
  * Displaying
  */
@@ -1592,21 +1644,14 @@ static NSColor	*shadowCol;
       cellFrame.size.height -= 2;
     }
 
-  [controlView lockFocus];
-
   switch (_cell.type)
     {
       case NSTextCellType:
-	if (_cell.contents_is_attributed_string)
-	  {
-	    [self _drawAttributedText: (NSAttributedString *)_contents  
-		  inFrame: cellFrame];
-	  }
-	else
-	  {
-	    [self _drawText: _contents inFrame: cellFrame];
-	  }
-	 break;
+        {
+	  [self _drawAttributedText: [self attributedStringValue]
+		inFrame: cellFrame];
+	}
+	break;
 
       case NSImageCellType:
 	if (_cell_image)
@@ -1638,13 +1683,10 @@ static NSColor	*shadowCol;
   // NB: We don't do any highlighting to make it easier for subclasses
   // to reuse this code while doing their own custom highlighting and
   // prettyfying
-  [controlView unlockFocus];
 }
 
 - (void) drawWithFrame: (NSRect)cellFrame inView: (NSView*)controlView
 {
-  NSDebugLog (@"NSCell drawWithFrame: inView: ");
-
   // do nothing if cell's frame rect is zero
   if (NSIsEmptyRect(cellFrame) || ![controlView window])
     return;
@@ -1652,16 +1694,12 @@ static NSColor	*shadowCol;
   // draw the border if needed
   if (_cell.is_bordered)
     {
-      [controlView lockFocus];
       [shadowCol set];
       NSFrameRect(cellFrame);
-      [controlView unlockFocus];
     }
   else if (_cell.is_bezeled)
     {
-      [controlView lockFocus];
       NSDrawWhiteBezel(cellFrame, NSZeroRect);
-      [controlView unlockFocus];
     }
 
   [self drawInteriorWithFrame: cellFrame inView: controlView];
@@ -1691,10 +1729,53 @@ static NSColor	*shadowCol;
        * NSCell simply draws border+text/image and makes no highlighting, 
        * for easier subclassing.
        */
-     [self drawWithFrame: cellFrame inView: controlView];
+      if ([self isOpaque] == NO)
+	{
+	  /* FIXME - This looks like potentially generating an
+	   * infinite loop!  The control asking the cell to draw
+	   * itself in the rect, the cell asking the control to draw
+	   * the rect, the control asking the cell to draw itself in
+	   * the rect, the cell ...
+	   *
+	   * I think we should remove it.  The control is responsible
+	   * for using the cell to draw, not vice versa.
+	   */
+	  [controlView displayRect: cellFrame];
+	}
+      [self drawWithFrame: cellFrame inView: controlView];
     }
 }
 
+- (NSColor*) highlightColorWithFrame: (NSRect)cellFrame
+			      inView: (NSView *)controlView
+{
+  return [NSColor selectedControlColor];
+}
+
+- (void) _setupTextWithFrame: (NSRect)aRect
+		      inView: (NSView*)controlView
+		      editor: (NSText*)textObject
+{
+  NSRect titleRect = [self titleRectForBounds: aRect];
+  NSSize maxSize = NSMakeSize(1e8, titleRect.size.height);
+  NSClipView *cv = [[NSClipView alloc] 
+		       initWithFrame: titleRect];
+  NSTextContainer *ct = [(NSTextView*)textObject textContainer];
+
+  [controlView addSubview: cv];
+  RELEASE(cv);
+  [cv setAutoresizesSubviews: NO];
+  [cv setDocumentView: textObject];
+  [textObject setFrame: [cv bounds]];
+  [textObject setHorizontallyResizable: YES];
+  [textObject setVerticallyResizable: NO];
+  [textObject setMaxSize: maxSize];
+  [textObject setMinSize: titleRect.size];
+  [ct setContainerSize: maxSize];
+  [ct setHeightTracksTextView: NO];
+  [ct setWidthTracksTextView: NO];
+}
+
 /*
  * Editing Text
  */
@@ -1707,8 +1788,9 @@ static NSColor	*shadowCol;
   if (!controlView || !textObject || (_cell.type != NSTextCellType))
     return;
 
-  [textObject setFrame: [self titleRectForBounds: aRect]];
-  [controlView addSubview: textObject];  
+  [self _setupTextWithFrame: aRect
+	inView: controlView
+	editor: textObject];
 
   if (_formatter != nil)
     {
@@ -1733,6 +1815,7 @@ static NSColor	*shadowCol;
 	  [textObject setText: [(NSAttributedString *)_contents string]];
 	}
     }
+  [textObject sizeToFit];
   
   [textObject setDelegate: anObject];
   [[controlView window] makeFirstResponder: textObject];
@@ -1741,14 +1824,16 @@ static NSColor	*shadowCol;
     {
       [textObject mouseDown: theEvent];
     }
-
-  [textObject display];
 }
 
 - (void) endEditing: (NSText*)textObject
 {
+  NSClipView *cv;
+
   [textObject setDelegate: nil];
+  cv = (NSClipView*)[textObject superview];
   [textObject removeFromSuperview];
+  [cv removeFromSuperview];
 }
 
 - (void) selectWithFrame: (NSRect)aRect
@@ -1761,21 +1846,38 @@ static NSColor	*shadowCol;
   if (!controlView || !textObject || (_cell.type != NSTextCellType))
     return;
 
-  [textObject setFrame: [self titleRectForBounds: aRect]];
-  [controlView addSubview: textObject];
-  if (_cell.contents_is_attributed_string == NO)
+  [self _setupTextWithFrame: aRect
+	inView: controlView
+	editor: textObject];
+
+  if (_formatter != nil)
     {
-      [textObject setText: _contents];
+      NSString *contents; 
+
+      contents = [_formatter editingStringForObjectValue: _objectValue];
+      if (contents == nil)
+	{
+	  contents = _contents;
+	}
+      [textObject setText: contents];
     }
   else
     {
-      /* FIXME/TODO make sure this is correct. */
-      [textObject setText: [(NSAttributedString *)_contents string]];
+      if (_cell.contents_is_attributed_string == NO)
+	{
+	  [textObject setText: _contents];
+	}
+      else
+	{
+	  /* FIXME/TODO make sure this is correct. */
+	  [textObject setText: [(NSAttributedString *)_contents string]];
+	}
     }
+
+  [textObject sizeToFit];
   [textObject setSelectedRange: NSMakeRange (selStart, selLength)];
   [textObject setDelegate: anObject];
   [[controlView window] makeFirstResponder: textObject];
-  [textObject display];
 }
 
 - (BOOL) sendsActionOnEndEditing 
@@ -1848,7 +1950,11 @@ static NSColor	*shadowCol;
   [aCoder encodeValueOfObjCType: @encode(BOOL) at: &flag];
   flag = _cell.is_selectable;
   [aCoder encodeValueOfObjCType: @encode(BOOL) at: &flag];
-  flag = _cell.is_continuous;
+  // This used to be is_continuous, which has been replaced.
+  /* Ayers 20.03.2003: But we must continue to encode it for backward
+     compatibility or current releases will have undefined behavior when
+     decoding archives (i.e. .gorm files) encoded by this version. */
+  flag = [self isContinuous];
   [aCoder encodeValueOfObjCType: @encode(BOOL) at: &flag];
   flag = _cell.allows_mixed_state;
   [aCoder encodeValueOfObjCType: @encode(BOOL) at: &flag];
@@ -1874,77 +1980,101 @@ static NSColor	*shadowCol;
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  BOOL flag;
-  unsigned int tmp_int;
-
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_contents];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_cell_image];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_font];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_objectValue];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.contents_is_attributed_string = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_highlighted = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_disabled = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_editable = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_rich_text = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.imports_graphics = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.shows_first_responder = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.refuses_first_responder = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.sends_action_on_end_editing = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_bordered = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_bezeled = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_scrollable = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_selectable = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.is_continuous = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.allows_mixed_state = flag;
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
-  _cell.wraps = flag;
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
-  _cell.text_align = tmp_int;
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
-  _cell.type = tmp_int;
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
-  _cell.image_position = tmp_int;
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
-  _cell.entry_type = tmp_int;
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
-  _cell.state = tmp_int;
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) 
-	    at: &_mnemonic_location];
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) 
-	    at: &_mouse_down_flags];
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &_action_mask];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_formatter];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_menu];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_represented_object];
-
-  if (_formatter != nil)
+  if ([aDecoder allowsKeyedCoding])
     {
-      NSString *contents;
+      NSString *contents = [aDecoder decodeObjectForKey: @"NSContents"];
+      NSFont *support = [aDecoder decodeObjectForKey: @"NSSupport"];
+      int cFlags;
+      int cFlags2;
 
-      contents = [_formatter stringForObjectValue: _objectValue];
-      if (contents != nil)
-	{
-	  _cell.has_valid_object_value = YES;
-	  ASSIGN (_contents, contents);
-	  _cell.contents_is_attributed_string = NO;
+      if ([aDecoder containsValueForKey: @"NSCellFlags"])
+        {
+	  cFlags = [aDecoder decodeIntForKey: @"NSCellFlags"];
+	  // FIXME
+	}
+      if ([aDecoder containsValueForKey: @"NSCellFlags2"])
+        {
+	  cFlags2 = [aDecoder decodeIntForKey: @"NSCellFlags2"];
+	  // FIXME
+	}
+
+      self = [self initTextCell: contents];
+      [self setFont: support];
+    }
+  else
+    {
+      BOOL flag;
+      unsigned int tmp_int;
+      
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_contents];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_cell_image];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_font];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_objectValue];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.contents_is_attributed_string = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.is_highlighted = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.is_disabled = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.is_editable = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.is_rich_text = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.imports_graphics = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.shows_first_responder = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.refuses_first_responder = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.sends_action_on_end_editing = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.is_bordered = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.is_bezeled = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.is_scrollable = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.is_selectable = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      // This used to be is_continuous, which has been replaced.
+      //_cell.is_continuous = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.allows_mixed_state = flag;
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &flag];
+      _cell.wraps = flag;
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
+      _cell.text_align = tmp_int;
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
+      _cell.type = tmp_int;
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
+      _cell.image_position = tmp_int;
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
+      _cell.entry_type = tmp_int;
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &tmp_int];
+      _cell.state = tmp_int;
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) 
+		                   at: &_mnemonic_location];
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) 
+		                   at: &_mouse_down_flags];
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &_action_mask];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_formatter];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_menu];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_represented_object];
+      
+      if (_formatter != nil)
+        {
+	  NSString *contents;
+
+	  contents = [_formatter stringForObjectValue: _objectValue];
+	  if (contents != nil)
+	    {
+	      _cell.has_valid_object_value = YES;
+	      ASSIGN (_contents, contents);
+	      _cell.contents_is_attributed_string = NO;
+	    }
 	}
     }
-  
   return self;
 }
 @end
@@ -2012,43 +2142,42 @@ static NSColor	*shadowCol;
   return size;
 }
 
-- (void) _drawAttributedText: (NSAttributedString *)title 
-		     inFrame: (NSRect)cellFrame
+/**
+ * Private internal method to display an attributed string.
+ */
+- (void) _drawAttributedText: (NSAttributedString*)aString 
+		     inFrame: (NSRect)aRect
 {
   NSSize titleSize;
 
-  if (title == nil)
+  if (aString == nil)
     return;
 
-  titleSize = [title size];
+  titleSize = [aString size];
 
-  // Determine y position of text
-
-  /* Important: text should always be vertically centered without
+  /** Important: text should always be vertically centered without
    * considering descender [as if descender did not exist].
    * This is particularly important for single line texts.
    * Please make sure the output remains always correct.
    */
-  cellFrame.origin.y = NSMidY (cellFrame) - titleSize.height/2; 
-  cellFrame.size.height = titleSize.height;
+  aRect.origin.y = NSMidY (aRect) - titleSize.height/2; 
+  aRect.size.height = titleSize.height;
 
-  [title drawInRect: cellFrame];
+  [aString drawInRect: aRect];
 }
 
-- (void) _drawText: (NSString*)title  inFrame: (NSRect)cellFrame
+- (void) _drawText: (NSString*)aString  inFrame: (NSRect)cellFrame
 {
   NSSize titleSize;
   NSDictionary *attributes;
 
-  if (title == nil)
+  if (aString == nil)
     return;
 
   attributes = [self _nonAutoreleasedTypingAttributes];
-  titleSize = [title sizeWithAttributes: attributes];
+  titleSize = [aString sizeWithAttributes: attributes];
 
-  // Determine y position of text
-
-  /* Important: text should always be vertically centered without
+  /** Important: text should always be vertically centered without
    * considering descender [as if descender did not exist].
    * This is particularly important for single line texts.
    * Please make sure the output remains always correct.
@@ -2056,7 +2185,7 @@ static NSColor	*shadowCol;
   cellFrame.origin.y = NSMidY (cellFrame) - titleSize.height/2; 
   cellFrame.size.height = titleSize.height;
 
-  [title drawInRect: cellFrame  withAttributes: attributes];
+  [aString drawInRect: cellFrame  withAttributes: attributes];
   RELEASE (attributes);
 }
 
