@@ -27,15 +27,17 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
-#include <AppKit/NSActionCell.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSColorPanel.h>
-#include <AppKit/NSColorWell.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSPasteboard.h>
-#include <AppKit/NSWindow.h>
+#include "config.h"
+#include "AppKit/NSActionCell.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSColorPanel.h"
+#include "AppKit/NSColorWell.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSDragging.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSPasteboard.h"
+#include "AppKit/NSWindow.h"
+#include <Foundation/NSDebug.h>
 #include <Foundation/NSNotification.h>
 
 static NSString *GSColorWellDidBecomeExclusiveNotification =
@@ -84,6 +86,11 @@ static NSString *GSColorWellDidBecomeExclusiveNotification =
              name: GSColorWellDidBecomeExclusiveNotification
            object: nil];
 
+  [nc addObserver: self
+         selector: @selector(_takeColorFromPanel:)
+             name: NSColorPanelColorChangedNotification
+           object: nil];
+
   _is_active = YES;
 
   [colorPanel setColor: _the_color];
@@ -231,6 +238,8 @@ static NSString *GSColorWellDidBecomeExclusiveNotification =
       // Undo RETAIN by decoder
       TEST_RELEASE(_target);
       [aDecoder decodeValueOfObjCType: @encode(SEL) at: &_action];
+      [self registerForDraggedTypes:
+	[NSArray arrayWithObjects: NSColorPboardType, nil]];
     }
   return self;
 }
@@ -344,6 +353,19 @@ static NSString *GSColorWellDidBecomeExclusiveNotification =
     }
 }
 
+- (void) _takeColorFromPanel: (NSNotification *) notification
+{
+  id sender = [notification object];
+
+  if ([sender respondsToSelector: @selector(color)])
+    {
+      ASSIGN(_the_color, [(id)sender color]);
+      
+      [self sendAction: _action to: _target];
+      [self setNeedsDisplay: YES];
+    }
+}
+
 - (id) target
 {
   return _target;
