@@ -26,25 +26,436 @@
 #include <Foundation/NSNotification.h>
 #include <Foundation/NSString.h>
 #include <Foundation/NSArray.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSButtonCell.h>
-#include <AppKit/NSComboBox.h>
-#include <AppKit/NSComboBoxCell.h>
-#include <AppKit/NSGraphicsContext.h>
-#include <AppKit/NSScreen.h>
+#include <Foundation/NSRunLoop.h>
+#include <Foundation/NSException.h>
+#include <Foundation/NSAutoreleasePool.h>
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSBox.h"
+#include "AppKit/NSBrowser.h"
+#include "AppKit/NSBrowserCell.h"
+#include "AppKit/NSButtonCell.h"
+#include "AppKit/NSComboBox.h"
+#include "AppKit/NSComboBoxCell.h"
+#include "AppKit/NSGraphicsContext.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSMatrix.h"
+#include "AppKit/NSPanel.h"
+#include "AppKit/NSScreen.h"
+#include "AppKit/NSScroller.h"
+
+@interface GSComboWindow : NSPanel
+{
+   NSBrowser	*browser;
+
+@private
+   NSComboBoxCell *_cell;
+   BOOL		_stopped;
+}
+
++ (GSComboWindow *)defaultPopUp;
+
+- (void) positionForCell:(NSComboBoxCell *)aCell
+		    view: (NSView *)popView;
+- (NSSize) popUpCellSizeForPopUp:(NSComboBoxCell *)aCell
+			   width: (float) width;
+- (void) popUpForCell: (NSComboBoxCell *)aCell
+		 view: (NSView *)popView;
+- (void) runModalPopUp;
+- (void) runLoop;
+- (void) reloadData;
+- (void) noteNumberOfItemsChanged;
+- (void) scrollItemAtIndexToTop: (int)index;
+- (void) scrollItemAtIndexToVisible: (int)index;
+- (void) selectItemAtIndex: (int)index;
+- (void) deselectItemAtIndex: (int)index;
+
+@end
 
-#include "GSComboSupport.h"
 
 static NSNotificationCenter *nc;
 
 @interface NSComboBoxCell(_Private_)
-- (NSArray *)_dataSourceObjectValues;
+- (NSString *) _stringValueAtIndex: (int)index;
 - (void) _didClickInRect: (NSRect)cellFrame
 	   ofView: (NSView *)controlView;
 - (void) _didClick: (id)sender;
 - (GSComboWindow *) _popUp;
 @end
 
+
+@implementation GSComboWindow
+
++ (GSComboWindow *) defaultPopUp
+{
+  static GSComboWindow *gsWindow = nil;
+
+  if (!gsWindow)
+    gsWindow = [[self alloc] initWithContentRect: NSMakeRect(0,0,100,100)
+			     styleMask: NSBorderlessWindowMask
+			     backing: NSBackingStoreNonretained //NSBackingStoreBuffered
+			     defer: YES];
+  return gsWindow;
+}
+
+- (id) initWithContentRect: (NSRect)contentRect
+		 styleMask: (unsigned int)aStyle
+		   backing: (NSBackingStoreType)bufferingType
+		     defer: (BOOL)flag
+{
+  NSBox *box;
+   
+  self = [super initWithContentRect: contentRect
+		styleMask: aStyle
+		backing: bufferingType
+		defer: flag];
+  [self setLevel: NSPopUpMenuWindowLevel];
+  [self setWorksWhenModal: YES];
+  [self setBecomesKeyOnlyIfNeeded: YES];
+
+  box = [[NSBox alloc] initWithFrame: contentRect];
+  [box setAutoresizingMask: NSViewWidthSizable | NSViewHeightSizable];
+  [box setBorderType: NSLineBorder];
+  [box setTitlePosition: NSNoTitle];
+  [box setContentViewMargins: NSMakeSize(1,1)];
+  [box sizeToFit];
+  [self setContentView: box];
+  RELEASE(box);
+  browser = [[NSBrowser alloc] initWithFrame: contentRect];
+  [browser setMaxVisibleColumns: 1];
+  [browser setTitled: NO];
+  [browser setHasHorizontalScroller: NO];
+  [browser setTarget: self];
+  [browser setAction: @selector(selectItem:)];
+  [browser setDelegate: self];
+//    [browser setRefusesFirstResponder: YES];
+  [browser setAutoresizingMask: NSViewWidthSizable | NSViewWidthSizable];
+  [browser setAllowsEmptySelection: NO];
+  [browser setAllowsMultipleSelection: NO];
+  [browser setReusesColumns: YES];
+  // Create an empty matrix
+  [browser loadColumnZero];
+  [box setContentView: browser];
+  RELEASE(browser);
+  
+  return self;
+}
+
+- (BOOL) canBecomeKeyWindow { return YES; }
+
+- (void)dealloc
+{
+  // Browser was not retained so don't release it
+  [super dealloc];
+}
+
+- (NSSize) popUpCellSizeForPopUp: (NSComboBoxCell *)aCell 
+			   width: (float) width
+{
+  NSMatrix *matrix = [browser matrixInColumn: 0];
+  NSSize size;
+  NSSize bsize = _sizeForBorderType(NSLineBorder);
+  float itemHeight;
+  float cellSpacing;
+  int num = [aCell numberOfItems];
+  int max = [aCell numberOfVisibleItems];
+
+  if (num > max)
+    num = max;
+
+  itemHeight = [aCell itemHeight];
+  if (itemHeight <= 0)
+    {
+      itemHeight = [matrix cellSize].height;
+    }
+  else 
+    {
+      size.height = itemHeight;
+      if ([aCell hasVerticalScroller])
+        {
+	  size.width = width - [NSScroller scrollerWidth] - bsize.width;
+	}
+      else 
+        {
+	  size.width = width - bsize.width;
+	}
+      [matrix setCellSize: size];
+    }
+
+  size = [aCell intercellSpacing];
+  cellSpacing = size.height;
+  if (cellSpacing <= 0)
+    cellSpacing = [matrix intercellSpacing].height;
+  else 
+    [matrix setIntercellSpacing: size];
+
+  return NSMakeSize(width, 2.0 + bsize.height + (itemHeight + cellSpacing) * num);
+}
+
+- (void) positionForCell: (NSComboBoxCell *)aCell
+		    view: (NSView *)popView
+{
+  NSRect popRect = [popView bounds];
+  NSRect screenFrame;
+  NSRect rect;
+  NSSize size;
+  NSPoint point, oldPoint;
+   
+  size = [self popUpCellSizeForPopUp: aCell width: NSWidth(popRect)];
+  if (size.width == 0 || size.height == 0)
+    return;
+
+  screenFrame = [[[popView window] screen] frame];
+  point = popRect.origin;
+  if ([popView isFlipped])
+    point.y += NSHeight(popRect);
+  point = [popView convertPoint: point toView: nil];
+  point.y -= 1.0;
+  point = [[popView window] convertBaseToScreen: point];
+  point.y -= size.height;
+  if (point.y < 0)
+    {
+      // Off screen, so move it.
+      oldPoint = point;
+      point = popRect.origin;
+      if (![popView isFlipped])
+	  point.y += NSHeight(popRect);
+      point = [popView convertPoint: point toView: nil];
+      point.y += 1.0;
+      point = [[popView window] convertBaseToScreen: point];
+      if (point.y > NSHeight(screenFrame))
+	  point = oldPoint;
+      
+      if (point.y + size.height > NSHeight(screenFrame))
+	  point.y = NSHeight(screenFrame) - size.height;
+    }
+
+  if (point.x + size.width > NSWidth(screenFrame))
+    point.x = NSWidth(screenFrame) - size.width;
+  if (point.x < 0.0)
+    point.x = 0.0;
+
+  rect.size = size;
+  rect.origin.x = point.x;
+  rect.origin.y = point.y;
+  rect = [NSWindow frameRectForContentRect: rect
+		   styleMask: _styleMask];
+  [self setFrame: rect display: NO];
+}
+
+- (void) popUpForCell: (NSComboBoxCell *)aCell
+	   view: (NSView *)popView
+{
+  [self positionForCell: aCell
+	view: popView];
+
+  _cell = aCell;
+  [self reloadData];
+  
+//    [self enableKeyEquivalentForDefaultButtonCell];
+  [self runModalPopUp];
+
+  _cell = nil;
+}
+
+- (void) runModalPopUp
+{
+  NSWindow	*onWindow;
+  NSEvent	*event;
+  NSException	*exception = nil;
+  
+  onWindow = [[_cell controlView] window];
+//  [self setLevel: [onWindow level]];
+//  [self orderWindow: NSWindowAbove relativeTo: [onWindow windowNumber]];
+
+  while ((event = [NSApp nextEventMatchingMask: NSAnyEventMask
+			 untilDate: [NSDate dateWithTimeIntervalSinceNow: 0]
+			 inMode: NSDefaultRunLoopMode
+			 dequeue: NO]))
+    {
+      if ([event type] == NSAppKitDefined ||
+	  [event type] == NSSystemDefined ||
+	  [event type] == NSApplicationDefined ||
+	  [event windowNumber] == [self windowNumber])
+	break;
+      [NSApp nextEventMatchingMask: NSAnyEventMask
+	     untilDate: [NSDate distantFuture]
+	     inMode: NSDefaultRunLoopMode
+	     dequeue: YES];
+    }
+
+  [self makeKeyAndOrderFront: nil];
+
+  NS_DURING
+    [self runLoop];
+  NS_HANDLER
+    exception = localException;
+  NS_ENDHANDLER;
+
+  if (onWindow)
+    {
+      [onWindow makeKeyWindow];
+      [onWindow orderFrontRegardless];
+    }
+
+  if ([self isVisible])
+      [self orderOut:nil];
+
+  if (exception)
+    [exception raise];
+}
+
+- (void) runLoop
+{
+  NSEvent		*event;
+  int			cnt = 0;
+  BOOL			kDown;
+  NSDate *limit = [NSDate distantFuture];
+  CREATE_AUTORELEASE_POOL (pool);
+  
+  _stopped = NO;
+  while (!_stopped)
+    {
+      kDown = NO;
+      cnt++;
+      if (cnt >= 5)
+        {
+	  RELEASE(pool);
+	  IF_NO_GC(pool = [[NSAutoreleasePool alloc] init]);
+	  cnt = 0;
+	}
+      event = [NSApp nextEventMatchingMask: NSAnyEventMask
+		     untilDate: limit
+		     inMode: NSDefaultRunLoopMode
+		     dequeue: NO];
+      if (event)
+        {
+	  if ([event type] == NSAppKitDefined ||
+	      [event type] == NSSystemDefined ||
+	      [event type] == NSApplicationDefined ||
+	      [event windowNumber] == [self windowNumber])
+	    {
+	      event = [NSApp nextEventMatchingMask: NSAnyEventMask
+			     untilDate: limit
+			     inMode: NSDefaultRunLoopMode
+			     dequeue: YES];
+	      [NSApp sendEvent: event];
+	      if ([event type] == NSKeyDown)
+		  kDown = YES;
+	    }
+	  else if ([event type] == NSMouseMoved ||
+		   [event type] == NSLeftMouseDragged ||
+		   [event type] == NSOtherMouseDragged ||
+		   [event type] == NSRightMouseDragged ||
+		   [event type] == NSMouseEntered ||
+		   [event type] == NSMouseExited ||
+		   [event type] == NSCursorUpdate)
+	    {
+	      event = [NSApp nextEventMatchingMask: NSAnyEventMask
+			     untilDate: limit
+			     inMode: NSDefaultRunLoopMode
+			     dequeue: YES];
+	      [NSApp sendEvent:event];
+	    }
+	  else
+	    _stopped = YES;
+	}
+    }
+
+  if (kDown)
+    while ((event = [NSApp nextEventMatchingMask: NSAnyEventMask
+			   untilDate: limit
+			   inMode: NSDefaultRunLoopMode
+			   dequeue: NO]))
+      {
+	if ([event windowNumber] != [self windowNumber])
+	  break;
+	event = [NSApp nextEventMatchingMask: NSAnyEventMask
+		       untilDate: limit
+		       inMode: NSDefaultRunLoopMode
+		       dequeue: YES];
+	[NSApp sendEvent: event];
+	if ([event type] == NSKeyUp)
+	  break;
+      }
+  RELEASE(pool);
+}
+
+- (void) reloadData
+{
+  [browser loadColumnZero];
+  [self selectItemAtIndex: [_cell indexOfSelectedItem]];
+}
+
+- (void) noteNumberOfItemsChanged
+{
+// FIXME: Should only load the additional items
+  [self reloadData];
+}
+
+- (void) scrollItemAtIndexToTop: (int)index
+{
+  NSMatrix *matrix = [browser matrixInColumn: 0];
+  NSRect rect;
+
+  rect = [matrix cellFrameAtRow: index column: 0];
+  [matrix scrollPoint: rect.origin];
+}
+
+- (void) scrollItemAtIndexToVisible: (int)index
+{
+  NSMatrix *matrix = [browser matrixInColumn: 0];
+
+  [matrix scrollCellToVisibleAtRow: index column: 0];
+}
+
+- (void) selectItemAtIndex: (int)index
+{
+  [browser selectRow: index inColumn: 0];
+}
+
+- (void) deselectItemAtIndex: (int)index
+{
+  NSMatrix *matrix = [browser matrixInColumn: 0];
+
+  [matrix deselectSelectedCell];
+}
+
+// Target/Action of Browser
+- (void) selectItem: (id)sender
+{
+  if (_cell)
+    {
+      [_cell selectItemAtIndex: [sender selectedRowInColumn: 0]];
+      [_cell setStringValue: [[sender selectedCell] stringValue]];
+      _stopped = YES;
+    }
+}
+
+// Browser Delegate Methods
+- (int) browser: (NSBrowser *)sender 
+numberOfRowsInColumn: (int)column
+{
+  if (!_cell)
+    return 0;
+
+  return [_cell numberOfItems];
+}
+
+- (void) browser: (NSBrowser *)sender 
+ willDisplayCell: (id)aCell
+	   atRow: (int)row 
+	  column: (int)column
+{
+  if (!_cell)
+    return;
+
+  [aCell setStringValue: [_cell _stringValueAtIndex: row]];
+  [aCell setLeaf: YES];
+}
+
+@end
+
+
 @implementation NSComboBoxCell
 
 //
@@ -67,17 +478,15 @@ static NSNotificationCenter *nc;
   //
   //_dataSource = nil;
   //_buttonCell = nil;
-  _popUpList = [[NSMutableArray alloc] init];
   //_usesDataSource = NO;
   //_completes = NO;
+  _popUpList = [[NSMutableArray alloc] init];
   _hasVerticalScroller = YES;
   _visibleItems = 10;
   _intercellSpacing = NSMakeSize(3.0, 2.0);
   _itemHeight = 16;
   _selectedItem = -1;
 
-  _popRect = NSZeroRect;
-  //_mUpEvent = nil;
   _buttonCell = [[NSButtonCell alloc] initImageCell: 
 					  [NSImage imageNamed: @"NSComboArrow"]];
   [_buttonCell setImagePosition: NSImageOnly];
@@ -93,7 +502,6 @@ static NSNotificationCenter *nc;
 
 - (void) dealloc
 {
-  TEST_RELEASE(_dataSource);
   RELEASE(_buttonCell);
   RELEASE(_popUpList);
   
@@ -128,12 +536,12 @@ static NSNotificationCenter *nc;
 
 - (void) reloadData
 {
-// TODO notify popup
+  [_popup reloadData];
 }
 
 - (void) noteNumberOfItemsChanged
 {
-// TODO notify popup
+  [_popup noteNumberOfItemsChanged];
 }
 
 - (BOOL) usesDataSource { return _usesDataSource; }
@@ -144,12 +552,12 @@ static NSNotificationCenter *nc;
 
 - (void) scrollItemAtIndexToTop: (int)index
 {
-// TODO
+  [_popup scrollItemAtIndexToTop: index];
 }
 
 - (void) scrollItemAtIndexToVisible: (int)index
 {
-// TODO
+  [_popup scrollItemAtIndexToVisible: index];
 }
 
 - (void) selectItemAtIndex: (int)index
@@ -160,7 +568,9 @@ static NSNotificationCenter *nc;
   if (_selectedItem != index)
     {
       _selectedItem = index;
-      // TODO: Notify popup
+      
+      [_popup selectItemAtIndex: index];
+
       [nc postNotificationName: NSComboBoxSelectionDidChangeNotification
 	  object: [self controlView]
 	  userInfo: nil];
@@ -172,7 +582,9 @@ static NSNotificationCenter *nc;
   if (_selectedItem == index)
   {
     _selectedItem = -1;
-    // TODO: Notify popup
+
+      [_popup deselectItemAtIndex: index];
+
     [nc postNotificationName: NSComboBoxSelectionDidChangeNotification
 	object: [self controlView]
 	userInfo: nil];
@@ -216,7 +628,7 @@ static NSNotificationCenter *nc;
   if (!_usesDataSource)
     NSLog(@"Method Invalid: ComboBox does not use dataSource");
   else
-    ASSIGN(_dataSource, aSource);
+    _dataSource = aSource;
 }
 
 - (void) addItemWithObjectValue: (id)object
@@ -319,19 +731,21 @@ static NSNotificationCenter *nc;
 
 - (int) indexOfItemWithObjectValue: (id)object
 {
-   if (_usesDataSource)
-   {
+  if (_usesDataSource)
+    {
       NSLog(@"Method Invalid: ComboBox uses dataSource");
       return 0;
-   }
-   return [_popUpList indexOfObject: object];
+    }
+  return [_popUpList indexOfObject: object];
 }
 
 - (NSArray *)objectValues
 {
   if (_usesDataSource)
-    // FIXME: This should give a warning
-    return [self _dataSourceObjectValues];
+    {
+      NSLog(@"Method Invalid: ComboBox uses dataSource");
+      return nil;
+    }
   return _popUpList;
 }
 
@@ -354,7 +768,7 @@ static NSNotificationCenter *nc;
     }
   else
   {
-    int i;
+    unsigned int i;
 
     for (i = 0; i < [_popUpList count]; i++)
       {
@@ -474,7 +888,7 @@ buttonCellFrameFromRect(NSRect cellRect)
 		  ofView: controlView untilMouseUp: flag];
 
   nEvent = [NSApp currentEvent];
-  if ([theEvent type] == NSLeftMouseDown &&
+  if ([theEvent type] == NSLeftMouseDown  &&
       [nEvent type] == NSLeftMouseUp)
     {
       point = [controlView convertPoint: [theEvent locationInWindow]
@@ -484,25 +898,26 @@ buttonCellFrameFromRect(NSRect cellRect)
 	  point = [controlView convertPoint: [nEvent locationInWindow]
 			       fromView: nil];
 	  if (NSPointInRect(point, buttonCellFrameFromRect(cellFrame)))
+ 	    {
 //      [_buttonCell performClick: self];
-	    [self _didClickInRect: cellFrame ofView: controlView];
+	      [self _didClickInRect: cellFrame ofView: controlView];
+	    }
 	}
     }
-  _mUpEvent = nEvent;
   
   return rValue;
 }
 
 - (void) resetCursorRect: (NSRect)cellFrame inView: (NSView *)controlView
 {
-   [super resetCursorRect: textCellFrameFromRect(cellFrame)
-	  inView: controlView];
+  [super resetCursorRect: textCellFrameFromRect(cellFrame)
+	 inView: controlView];
 }
 
 - (void) setEnabled: (BOOL)flag
 {
-   [_buttonCell setEnabled: flag];
-   [super setEnabled: flag];
+  [_buttonCell setEnabled: flag];
+  [super setEnabled: flag];
 }
 
 // NSCoding
@@ -522,26 +937,51 @@ buttonCellFrameFromRect(NSRect cellRect)
   [coder encodeValueOfObjCType: @encode(int) at: &_selectedItem];
 
   if (_usesDataSource == YES)
-    [coder encodeValueOfObjCType: @encode(id) at: &_dataSource];      
+    [coder encodeConditionalObject: _dataSource];      
 }
 
-- (id) initWithCoder: (NSCoder *)coder
+- (id) initWithCoder: (NSCoder *)aDecoder
 {
-  self = [super initWithCoder: coder];
+  self = [super initWithCoder: aDecoder];
   
-  [coder decodeValueOfObjCType: @encode(id) at: &_buttonCell];
-  [coder decodeValueOfObjCType: @encode(id) at: &_popUpList];
-  [coder decodeValueOfObjCType: @encode(BOOL) at: &_usesDataSource];
-  [coder decodeValueOfObjCType: @encode(BOOL) at: &_hasVerticalScroller];
-  [coder decodeValueOfObjCType: @encode(BOOL) at: &_completes];
-  [coder decodeValueOfObjCType: @encode(BOOL) at: &_usesDataSource];
-  [coder decodeValueOfObjCType: @encode(int) at: &_visibleItems];
-  [coder decodeValueOfObjCType: @encode(NSSize) at: &_intercellSpacing];
-  [coder decodeValueOfObjCType: @encode(float) at: &_itemHeight];
-  [coder decodeValueOfObjCType: @encode(int) at: &_selectedItem];
+  if ([aDecoder allowsKeyedCoding])
+    {
+      //id delegate = [aDecoder decodeObjectForKey: @"NSDelegate"];
+      // FIXME: This does not match the way GNUstep currently implements
+      // the list of popup items.
+      //id table = [aDecoder decodeObjectForKey: @"NSTableView"];
 
-  if (_usesDataSource == YES)
-    [coder decodeValueOfObjCType: @encode(id) at: &_dataSource];      
+      if ([aDecoder containsValueForKey: @"NSHasVerticalScroller"])
+        {
+	  [self setHasVerticalScroller: [aDecoder decodeBoolForKey: 
+						      @"NSHasVerticalScroller"]];
+	}
+      if ([aDecoder containsValueForKey: @"NSVisibleItemCount"])
+        {
+	  [self setNumberOfVisibleItems: [aDecoder decodeIntForKey: 
+						       @"NSVisibleItemCount"]];
+	}
+    }
+  else
+    {
+      BOOL dummy;
+      
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_buttonCell];
+      RETAIN(_buttonCell);
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_popUpList];
+      RETAIN(_popUpList);
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_usesDataSource];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasVerticalScroller];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_completes];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &dummy];
+      [aDecoder decodeValueOfObjCType: @encode(int) at: &_visibleItems];
+      [aDecoder decodeValueOfObjCType: @encode(NSSize) at: &_intercellSpacing];
+      [aDecoder decodeValueOfObjCType: @encode(float) at: &_itemHeight];
+      [aDecoder decodeValueOfObjCType: @encode(int) at: &_selectedItem];
+
+      if (_usesDataSource == YES)
+	[self setDataSource: [aDecoder decodeObject]];      
+    }
 
   return self;
 }
@@ -550,44 +990,34 @@ buttonCellFrameFromRect(NSRect cellRect)
 
 @implementation NSComboBoxCell(_Private_)
 
-- (NSArray *)_dataSourceObjectValues
+- (NSString *)_stringValueAtIndex: (int)index
 {
-   NSMutableArray	*array = nil;
-   id			obj;
-   SEL			selector;
-   int			i,cnt;
-   
-   if (!_dataSource)
-      NSLog(@"No DataSource Specified");
-   else
-   {
-      cnt = [self numberOfItems];
-      if ([[self controlView] isKindOfClass:[NSComboBox class]])
-      {
-	 obj = [self controlView];
-	 selector = @selector(comboBox:objectValueForItemAtIndex:);
-	 if ([_dataSource respondsToSelector:selector])
-	 {
-	    array = [NSMutableArray array];
-	    for (i=0;i<cnt;i++)
-	       [array addObject:[_dataSource comboBox:obj
-					     objectValueForItemAtIndex:i]];
-	 }
-      }
-      else
-      {
-	 obj = self;
-	 selector = @selector(comboBoxCell:indexOfItemWithStringValue:);
-	 if ([_dataSource respondsToSelector:selector])
-	 {
-	    array = [NSMutableArray array];
-	    for (i=0;i<cnt;i++)
-	       [array addObject:[_dataSource comboBoxCell:obj
-					     objectValueForItemAtIndex:i]];
-	 }
-      }
-   }
-   return array;
+  if (!_usesDataSource)
+    {
+      return [[self itemObjectValueAtIndex: index] description];
+    }
+  else
+    {
+      if (!_dataSource)
+        {
+	  NSLog(@"ComboBox: No DataSource Specified");
+	  return nil;
+	}
+      if ([_dataSource respondsToSelector: 
+			   @selector(comboBox:objectValueForItemAtIndex:)])
+        {
+	  return [[_dataSource comboBox: (NSComboBox *)[self controlView] 
+			       objectValueForItemAtIndex: index] description];
+	}
+      else if ([_dataSource respondsToSelector: 
+				@selector(comboBoxCell:objectValueForItemAtIndex:)])
+        {
+	  return [[_dataSource comboBoxCell: self
+			      objectValueForItemAtIndex: index] description];
+	}
+    }
+
+  return nil;
 }
 
 - (void) _didClickInRect: (NSRect)cellFrame
@@ -596,86 +1026,53 @@ buttonCellFrameFromRect(NSRect cellRect)
   // We can not use performClick: on the button cell here as 
   // the button uses only part of the bounds of the control view.
   NSWindow *cvWin = [controlView window];
-      
+
+  [controlView lockFocus];
   [_buttonCell highlight: YES 
 	       withFrame: buttonCellFrameFromRect(cellFrame) 
 	       inView: controlView];
+  [controlView unlockFocus];
   [cvWin flushWindow];
 
-  _popRect = cellFrame;
-
   [self _didClick: self];
-  _popRect = NSZeroRect;
   
+  [controlView lockFocus];
   [_buttonCell highlight: NO 
 	       withFrame: buttonCellFrameFromRect(cellFrame) 
 	       inView: controlView];
+  [controlView unlockFocus];
   [cvWin flushWindow];
+
 }
 
 - (void) _didClick: (id)sender
 {
-  NSSize size;
-  NSPoint point,oldPoint;
-  NSRect screenFrame;
   NSView *popView = [self controlView];
 
-  if (_cell.is_disabled)
+  if ((_cell.is_disabled) || (popView == nil))
     return;
 
-  size = [[self _popUp] popUpCellSizeForPopUp: self];
-  if (size.width == 0 || size.height == 0)
-    return;
-
-  screenFrame = [[[popView window] screen] frame];
-  point = _popRect.origin;
-  if ([popView isFlipped])
-    point.y += NSHeight(_popRect);
-  point = [popView convertPoint: point toView: nil];
-  point.y -= 1.0;
-  point = [[popView window] convertBaseToScreen: point];
-  point.y -= size.height;
-  if (point.y < 0)
-    {
-      // Off screen, so move it.
-      oldPoint = point;
-      point = _popRect.origin;
-      if (![popView isFlipped])
-	  point.y += NSHeight(_popRect);
-      point = [popView convertPoint: point toView: nil];
-      point.y += 1.0;
-      point = [[popView window] convertBaseToScreen: point];
-      if (point.y > NSHeight(screenFrame))
-	  point = oldPoint;
-      
-      if (point.y + size.height > NSHeight(screenFrame))
-	  point.y = NSHeight(screenFrame) - size.height;
-    }
-
-  if (point.x + size.width > NSWidth(screenFrame))
-    point.x = NSWidth(screenFrame) - size.width;
-  if (point.x < 0.0)
-    point.x = 0.0;
-
   [nc postNotificationName: NSComboBoxWillPopUpNotification
       object: popView
       userInfo: nil];
 
-  [[self _popUp] popUpCell: self popUpAt: point width: NSWidth(_popRect)];
+  // HACK Abort the editing, otherwise the selected value is
+  // overwritten by the editor
+  //if ([_control_view isKindOfClass: NSControl])
+  [(NSControl *)_control_view abortEditing];
+
+  _popup = [self _popUp];
+  [_popup popUpForCell: self view: popView];
+  _popup = nil;
 
   [nc postNotificationName: NSComboBoxWillDismissNotification
       object: popView
       userInfo: nil];
 }
 
-- (NSEvent *)_mouseUpEvent
-{
-   return _mUpEvent;
-}
-
 - (GSComboWindow *)_popUp
 {
-   return [GSComboWindow defaultPopUp];
+  return [GSComboWindow defaultPopUp];
 }
 
 @end
