@@ -29,13 +29,15 @@
 
 #include <Foundation/NSArray.h>
 #include <Foundation/NSDebug.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSCursor.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSBitmapImageRep.h>
-#include <AppKit/NSGraphicsContext.h>
-#include <AppKit/DPSOperators.h>
+#include <Foundation/NSKeyedArchiver.h>
+
+#include "AppKit/NSColor.h"
+#include "AppKit/NSCursor.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSBitmapImageRep.h"
+
+#include "GNUstepGUI/GSDisplayServer.h"
 
 // Class variables
 static NSMutableArray *gnustep_gui_cursor_stack;
@@ -61,7 +63,7 @@ static NSMutableDictionary *cursorDict = nil;
       NSCursor_class = self;
       gnustep_gui_cursor_stack = [[NSMutableArray alloc] initWithCapacity: 2];
       gnustep_gui_hidden_until_move = YES;
-      cursorDict = RETAIN([NSMutableDictionary dictionary]);
+      cursorDict = [NSMutableDictionary new]; 
       [[self arrowCursor] push];
     }
 }
@@ -100,9 +102,9 @@ static NSMutableDictionary *cursorDict = nil;
   if (_hot_spot.y >= [rep pixelsHigh])
     _hot_spot.y = [rep pixelsHigh]-1;
 
-  DPSimagecursor(GSCurrentContext(), _hot_spot.x, _hot_spot.y, 
-		 [rep pixelsWide], [rep pixelsHigh],
-		 [rep samplesPerPixel], [rep bitmapData], &c);
+  [GSCurrentServer() imagecursor: _hot_spot 
+		 : [rep pixelsWide] : [rep pixelsHigh]
+		 : [rep samplesPerPixel] : [rep bitmapData] : &c];
   _cid = c;
 }
 
@@ -111,7 +113,7 @@ static NSMutableDictionary *cursorDict = nil;
  */
 + (void) hide
 {
-  DPShidecursor(GSCurrentContext());
+  [GSCurrentServer() hidecursor];
 }
 
 + (void) pop
@@ -141,7 +143,7 @@ static NSMutableDictionary *cursorDict = nil;
 
 + (void) unhide
 {
-  DPSshowcursor(GSCurrentContext());  
+  [GSCurrentServer() showcursor];
 }
 
 /*
@@ -156,7 +158,7 @@ static NSMutableDictionary *cursorDict = nil;
       void *c;
     
       cursor = [[NSCursor_class alloc] initWithImage: nil];
-      DPSstandardcursor(GSCurrentContext(), GSArrowCursor, &c);
+      [GSCurrentServer() standardcursor: GSArrowCursor : &c];
       [cursor _setCid: c];
       [cursorDict setObject: cursor forKey: name];
       RELEASE(cursor);
@@ -178,7 +180,7 @@ static NSMutableDictionary *cursorDict = nil;
       void *c;
     
       cursor = [[NSCursor_class alloc] initWithImage: nil];
-      DPSstandardcursor(GSCurrentContext(), GSIBeamCursor, &c);
+      [GSCurrentServer() standardcursor: GSIBeamCursor : &c];
       [cursor _setCid: c];
       [cursorDict setObject: cursor forKey: name];
       RELEASE(cursor);
@@ -195,8 +197,9 @@ static NSMutableDictionary *cursorDict = nil;
       void *c;
     
       cursor = [[NSCursor_class alloc] initWithImage: nil];
-      DPSstandardcursor(GSCurrentContext(), GSArrowCursor, &c);
-      DPSsetcursorcolor(GSCurrentContext (), 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, c);
+      [GSCurrentServer() standardcursor: GSArrowCursor : &c];
+      [GSCurrentServer() setcursorcolor: [NSColor greenColor] 
+      	                               : [NSColor blackColor] : c];
       [cursor _setCid: c];
       [cursorDict setObject: cursor forKey: name];
       RELEASE(cursor);
@@ -242,19 +245,15 @@ backgroundColorHint:(NSColor *)bg
 	fg = [NSColor blackColor];
       bg = [bg colorUsingColorSpaceName: NSDeviceRGBColorSpace];
       fg = [fg colorUsingColorSpaceName: NSDeviceRGBColorSpace];
-      NSLog(@"fg color is %@", fg);
-      DPSsetcursorcolor(GSCurrentContext (), 
-			[fg redComponent],
-			[fg greenComponent],
-			[fg blueComponent],
-			[bg redComponent],
-			[bg greenComponent],
-			[bg blueComponent],
-			_cid);
+      [GSCurrentServer() setcursorcolor: fg : bg : _cid];
     }
   return cursor;
 }
-
+- (void)dealloc
+{
+  RELEASE (_cursor_image);
+  [super dealloc];
+}
 /*
  * Defining the Cursor
  */
@@ -313,6 +312,7 @@ backgroundColorHint:(NSColor *)bg
 
 - (void) mouseExited: (NSEvent*)theEvent
 {
+  NSDebugLLog(@"NSCursor", @"Cursor mouseExited:");
   if (_is_set_on_mouse_exited == YES)
     {
       [self set];
@@ -344,7 +344,7 @@ backgroundColorHint:(NSColor *)bg
   gnustep_gui_current_cursor = self;
   if (_cid)
     {
-      DPSsetcursorcolor(GSCurrentContext(), -1, 0, 0, 1, 1, 1, _cid);
+      [GSCurrentServer() setcursorcolor: nil : nil : _cid];
     }
 }
 
@@ -372,12 +372,45 @@ backgroundColorHint:(NSColor *)bg
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_is_set_on_mouse_entered];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_is_set_on_mouse_exited];
-  _cursor_image = [aDecoder decodeObject];
-  _hot_spot = [aDecoder decodePoint];
-  [self _computeCid];
+  if ([aDecoder allowsKeyedCoding])
+    {
+      int type = 0;
+      NSPoint hotSpot;
+      
+      if ([aDecoder containsValueForKey: @"NSCursorType"])
+        {
+	  type = [aDecoder decodeIntForKey: @"NSCursorType"];
+	}
 
+      DESTROY(self);
+      // FIXME
+      if (type == 0)
+        {
+	  self = [NSCursor arrowCursor];
+	}
+      else if (type == 1)
+        {
+	  self = [NSCursor IBeamCursor];
+	}
+      else 
+        {
+	  // FIXME
+	  self = [NSCursor arrowCursor];
+	}
+      
+      if ([aDecoder containsValueForKey: @"NSHotSpot"])
+        {
+	  hotSpot = [aDecoder decodePointForKey: @"NSHotSpot"];
+	}
+    }
+  else
+    {
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_is_set_on_mouse_entered];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_is_set_on_mouse_exited];
+      _cursor_image = [aDecoder decodeObject];
+      _hot_spot = [aDecoder decodePoint];
+      [self _computeCid];
+    }
   return self;
 }
 
