@@ -27,22 +27,28 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSFontPanel.h>
-#include <AppKit/NSFontManager.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSSplitView.h>
-#include <AppKit/NSScrollView.h>
-#include <AppKit/NSBrowser.h>
-#include <AppKit/NSBrowserCell.h>
-#include <AppKit/NSTextView.h>
-#include <AppKit/NSTextField.h>
-#include <AppKit/NSTextFieldCell.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSPanel.h>
-#include <AppKit/NSButton.h>
-#include <AppKit/NSBox.h>
+#include "config.h"
+#include <Foundation/NSDebug.h>
+#include <Foundation/NSValue.h>
+#include "AppKit/NSDragging.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSFontPanel.h"
+#include "AppKit/NSFontManager.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSSplitView.h"
+#include "AppKit/NSScrollView.h"
+#include "AppKit/NSBrowser.h"
+#include "AppKit/NSBrowserCell.h"
+#include "AppKit/NSTextView.h"
+#include "AppKit/NSTextField.h"
+#include "AppKit/NSTextFieldCell.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSPanel.h"
+#include "AppKit/NSButton.h"
+#include "AppKit/NSBox.h"
+
+#define _SAVE_PANEL_X_PAD	5
+#define _SAVE_PANEL_Y_PAD	4
 
 static inline void _setFloatValue (NSTextField *field, float size)
 {
@@ -63,25 +69,32 @@ static inline void _setFloatValue (NSTextField *field, float size)
 }
 
 
-NSText *sizeFieldText = nil;
+static NSText *sizeFieldText = nil;
+
+static float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
+			14.0, 16.0, 18.0, 24.0, 36.0, 48.0, 64.0};
 
 /* Implemented in NSBrowser */
 @interface GSBrowserTitleCell : NSTextFieldCell
+{
+}
 @end
 
-float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0, 
-		 14.0, 16.0, 18.0, 24.0, 36.0, 48.0, 64.0};
-
 @interface NSFontPanel (Private)
 - (NSFont*) _fontForSelection: (NSFont*) fontObject;
 
+-(void) _trySelectSize: (float)size;
+
 // Some action methods
 - (void) cancel: (id) sender;
 - (void) _togglePreview: (id) sender;
 - (void) _doPreview;
 - (void) ok: (id) sender;
 
+- (void) _getOriginalSize;
 - (id)_initWithoutGModel;
+
+- (BOOL) _includeFont: (NSString *)fontName  delegate: (id)delegate;
 @end
 
 @implementation NSFontPanel
@@ -93,9 +106,6 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 {
   if (self == [NSFontPanel class])
     {
-      NSDebugLog(@"Initialize NSFontPanel class\n");
-
-      // Initial version
       [self setVersion: 1];
     }
 }
@@ -125,10 +135,11 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
   //  if (![NSBundle loadNibNamed: @"FontPanel" owner: self]);
   [self _initWithoutGModel];
 
-  [self reloadDefaultFontFamilies];
   ASSIGN(_faceList, [NSArray array]);
   _face = -1;
   _family = -1;
+  [self reloadDefaultFontFamilies];
+  [self _getOriginalSize];
 
   return self;
 }
@@ -164,9 +175,36 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 - (void) reloadDefaultFontFamilies
 {
   NSFontManager *fm = [NSFontManager sharedFontManager];
+  id fmDelegate = [fm delegate];
   NSBrowser *familyBrowser = [[self contentView] viewWithTag: NSFPFamilyBrowser];
+  NSArray *fontFamilies = [fm availableFontFamilies];
+  unsigned int i,j;
+
+  DESTROY(_familyList);
+
+  /*
+  Build an array of all families that have a font that will be included.
+  */
+  _familyList = [[NSMutableArray alloc]
+		  initWithCapacity: [fontFamilies count]];
+
+  for (i = 0; i < [fontFamilies count]; i++)
+    {
+      NSArray *familyMembers;
+      familyMembers = [fm availableMembersOfFontFamily:
+			[fontFamilies objectAtIndex: i]];
+
+      for (j = 0; j < [familyMembers count]; j++)
+	{
+	  if ([self _includeFont: [[familyMembers objectAtIndex: j] objectAtIndex: 0]
+			delegate: fmDelegate])
+	    {
+	      [_familyList addObject: [fontFamilies objectAtIndex: i]];
+	      break;
+	    }
+	}
+    }
 
-  ASSIGN(_familyList, [fm availableFontFamilies]);
   // Reload the display. 
   [familyBrowser loadColumnZero];
   // Reselect the current font. (Hopefully still there)
@@ -205,17 +243,15 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
       NSString *family = [fontObject familyName];
       NSString *fontName = [fontObject fontName];
       float size = [fontObject pointSize];
-      NSTextField *sizeField = [[self contentView] viewWithTag: NSFPSizeField];
-      NSBrowser *sizeBrowser = [[self contentView] viewWithTag: NSFPSizeBrowser];
       NSBrowser *familyBrowser = [[self contentView] viewWithTag: NSFPFamilyBrowser];
       NSBrowser *faceBrowser = [[self contentView] viewWithTag: NSFPFaceBrowser];
       NSString *face = @"";
-      int i;
+      unsigned int i;
  
       // Store style information for font
       _traits = [fm traitsOfFont: fontObject];
       _weight = [fm weightOfFont: fontObject];
-
+      
       // Select the row for the font family
       for (i = 0; i < [_familyList count]; i++)
 	{
@@ -246,14 +282,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 	}
 
       // show point size and select the row if there is one
-      _setFloatValue (sizeField, size);
-      for (i = 0; i < sizeof(sizes)/sizeof(float); i++)
-	{
-	  if (size == sizes[i])
-	    {
-	      [sizeBrowser selectRow: i inColumn: 0];
-	    }
-	}
+      [self _trySelectSize: size];
 
       // Use in preview
       [previewArea setFont: fontObject];
@@ -309,8 +338,10 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
   return _accessoryView;
 }
 
+/*
 - (void) setAccessoryView: (NSView*)aView
 {
+
   // FIXME: We have to resize
   // Perhaps we could copy the code from NSSavePanel over to here
   if (_accessoryView != nil)
@@ -321,6 +352,113 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
   ASSIGN(_accessoryView, aView);
   [[self contentView] addSubview: aView];
 }
+*/
+
+- (void) setAccessoryView: (NSView*)aView
+{
+  NSRect accessoryViewFrame, bottomFrame;
+  NSRect tmpRect;
+  NSSize contentSize, contentMinSize;
+  float addedHeight, accessoryWidth;
+
+  if (aView == _accessoryView)
+    return;
+  
+  /* The following code is very tricky.  Please think and test a lot
+     before changing it. */
+
+  /* Remove old accessory view if any */
+  if (_accessoryView != nil)
+    {
+      /* Remove accessory view */
+      accessoryViewFrame = [_accessoryView frame];
+      [_accessoryView removeFromSuperview];
+
+      /* Change the min size before doing the resizing otherwise it
+	 could be a problem. */
+      [self setMinSize: _originalMinSize];
+
+      /* Resize the panel to the height without the accessory view. 
+	 This must be done with the special care of not resizing 
+	 the heights of the other views. */
+      addedHeight = accessoryViewFrame.size.height + (_SAVE_PANEL_Y_PAD * 2);
+      contentSize = [[self contentView] frame].size;
+      contentSize.height -= addedHeight;
+      // Resize without modifying topView and bottomView height.
+      [_topView setAutoresizingMask: NSViewWidthSizable | NSViewMinYMargin];
+      [self setContentSize: contentSize];
+      [_topView setAutoresizingMask: NSViewWidthSizable|NSViewHeightSizable];
+    }
+  
+  /* Resize the panel to its original size.  This resizes freely the
+     heights of the views.  NB: minSize *must* come first */
+  [self setMinSize: _originalMinSize];
+  [self setContentSize: _originalSize];
+  
+  /* Set the new accessory view */
+  _accessoryView = aView;
+  
+  /* If there is a new accessory view, plug it in */
+  if (_accessoryView != nil)
+    {
+      /* Make sure the new accessory view behaves  - its height must be fixed
+       * and its position relative to the bottom of the superview must not
+       * change	- so its position rlative to the top must be changable. */
+      [_accessoryView setAutoresizingMask: NSViewMaxYMargin
+	| ([_accessoryView autoresizingMask] 
+	& ~(NSViewHeightSizable | NSViewMinYMargin))];  
+      
+      /* Compute size taken by the new accessory view */
+      accessoryViewFrame = [_accessoryView frame];
+      addedHeight = accessoryViewFrame.size.height + (_SAVE_PANEL_Y_PAD * 2);
+      accessoryWidth = accessoryViewFrame.size.width + (_SAVE_PANEL_X_PAD * 2);
+
+      /* Resize content size accordingly */
+      contentSize = _originalSize;
+      contentSize.height += addedHeight;
+      if (accessoryWidth > contentSize.width)
+	{
+	  contentSize.width = accessoryWidth;
+	}
+      
+      /* Set new content size without resizing heights of topView, bottomView */
+      // Our views should resize horizontally if needed, but not vertically
+      [_topView setAutoresizingMask: NSViewWidthSizable | NSViewMinYMargin];
+      [self setContentSize: contentSize];
+      // Restore the original autoresizing masks
+      [_topView setAutoresizingMask: NSViewWidthSizable|NSViewHeightSizable];
+
+      /* Compute new min size */
+      contentMinSize = _originalMinSize;
+      contentMinSize.height += addedHeight;
+      // width is more delicate
+      tmpRect = NSMakeRect (0, 0, contentMinSize.width, contentMinSize.height);
+      tmpRect = [NSWindow contentRectForFrameRect: tmpRect 
+			  styleMask: [self styleMask]];
+      if (accessoryWidth > tmpRect.size.width)
+	{
+	  contentMinSize.width += accessoryWidth - tmpRect.size.width;
+	}
+      // Set new min size
+      [self setMinSize: contentMinSize];
+
+      /*
+       * Pack the Views
+       */
+
+      /* BottomView is ready */
+      bottomFrame = [_bottomView frame];
+
+      /* AccessoryView */
+      accessoryViewFrame.origin.x 
+	= (contentSize.width - accessoryViewFrame.size.width) / 2;
+      accessoryViewFrame.origin.y =  NSMaxY (bottomFrame) + _SAVE_PANEL_Y_PAD;
+      [_accessoryView setFrameOrigin: accessoryViewFrame.origin];
+
+      /* Add the accessory view */
+      [[self contentView] addSubview: _accessoryView];
+    }
+}
 
 /*
  * NSCoding protocol
@@ -370,7 +508,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 
 @end
 
-@implementation NSFontPanel (Privat)
+@implementation NSFontPanel (Private)
 
 - (id) _initWithoutGModel
 {
@@ -422,6 +560,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
   // preview and selection
   topArea = [[NSView alloc] initWithFrame: topAreaRect];
   [topArea setAutoresizingMask: (NSViewWidthSizable | NSViewHeightSizable)];
+  _topView = topArea;
 
   splitView = [[NSSplitView alloc] initWithFrame: splitViewRect];
   [splitView setVertical: NO]; 
@@ -461,6 +600,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
   [familyBrowser setTakesTitleFromPreviousColumn: NO];
   [familyBrowser setTarget: self];
   [familyBrowser setDoubleAction: @selector(familySelected:)];
+  [familyBrowser setAction: @selector(_familySelectionChanged:)];
   [familyBrowser setAutoresizingMask: (NSViewWidthSizable
 				       | NSViewMaxXMargin
 				       | NSViewHeightSizable)];
@@ -481,6 +621,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
   [faceBrowser setTakesTitleFromPreviousColumn: NO];
   [faceBrowser setTarget: self];
   [faceBrowser setDoubleAction: @selector(faceSelected:)];
+  [faceBrowser setAction: @selector(_faceSelectionChanged:)];
   [faceBrowser setAutoresizingMask: (NSViewWidthSizable
 				     | NSViewMinXMargin
 				     | NSViewHeightSizable)];
@@ -511,6 +652,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
   [sizeField setAlignment: NSCenterTextAlignment];
   [sizeField setBackgroundColor: [NSColor windowFrameTextColor]];
   [sizeField setAutoresizingMask: NSViewMinXMargin | NSViewMinYMargin];
+  [sizeField setDelegate: self];
   [sizeField setTag: NSFPSizeField];
   [bottomSplit addSubview: sizeField];
   RELEASE(sizeField);
@@ -525,6 +667,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
   [sizeBrowser setTakesTitleFromPreviousColumn: NO];
   [sizeBrowser setTarget: self];
   [sizeBrowser setDoubleAction: @selector(sizeSelected:)];
+  [sizeBrowser setAction: @selector(_sizeSelectionChanged:)];
   [sizeBrowser setAutoresizingMask: NSViewMinXMargin | NSViewHeightSizable];
   [sizeBrowser setTag: NSFPSizeBrowser];
   [bottomSplit addSubview: sizeBrowser];
@@ -542,6 +685,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 
   // action buttons
   bottomArea = [[NSView alloc] initWithFrame: bottomAreaRect];
+  _bottomView = bottomArea;
 
   slash = [[NSBox alloc] initWithFrame: slashRect];
   [slash setBorderType: NSGrooveBorder];
@@ -552,7 +696,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 
   // cancel button
   revertButton = [[NSButton alloc] initWithFrame: revertButtonRect];
-  [revertButton setStringValue: @"Revert"];
+  [revertButton setTitle: @"Revert"];
   [revertButton setAction: @selector(cancel:)];
   [revertButton setTarget: self];
   [revertButton setTag: NSFPRevertButton];
@@ -562,18 +706,20 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 
   // toggle button for preview
   previewButton = [[NSButton alloc] initWithFrame: previewButtonRect];
-  [previewButton setStringValue: @"Preview"];
+  [previewButton setTitle: @"Preview"];
   [previewButton setButtonType: NSOnOffButton];
   [previewButton setAction: @selector(_togglePreview:)];
   [previewButton setTarget: self];
   [previewButton setTag: NSFPPreviewButton];
   [previewButton setAutoresizingMask: NSViewMinXMargin];
+  [previewButton setState: NSOnState];
+  _preview = YES;
   [bottomArea addSubview: previewButton];
   RELEASE(previewButton);
 
   // button to set the font
   setButton = [[NSButton alloc] initWithFrame: setButtonRect];
-  [setButton setStringValue: @"Set"];
+  [setButton setTitle: @"Set"];
   [setButton setAction: @selector(ok:)];
   [setButton setTarget: self];
   [setButton setTag: NSFPSetButton];
@@ -654,7 +800,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 	{
 	  familyName = [_familyList objectAtIndex: _family];
 	}
-      if (_face == -1)
+      if (_face == -1 || ![_faceList count])
 	{
 	  faceName = @"NoFace";
 	}
@@ -691,7 +837,8 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 {
   float		size;
   NSString	*fontName;
-  NSTextField *sizeField = [[self contentView] viewWithTag: NSFPSizeField];
+  NSTextField	*sizeField = [[self contentView] viewWithTag: NSFPSizeField];
+  unsigned	i = [_faceList count];
 
   size = [sizeField floatValue];
   if (size == 0.0)
@@ -707,7 +854,6 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
     }
   if (_face < 0)
     {
-      unsigned	i = [_faceList count];
 
       if (i == 0)
 	{
@@ -718,76 +864,225 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
     }
   else
     {
-      fontName = [[_faceList objectAtIndex: _face] objectAtIndex: 0];
+      /*
+      i really should be > 0 here, except for the very obscure case where
+      the delegate has refused all fonts (so that our family and face lists
+      are completely empty).
+      */
+      if (i)
+	fontName = [[_faceList objectAtIndex: _face] objectAtIndex: 0];
+      else
+	return nil;
     }
   
   // FIXME: We should check if the font is correct
   return [NSFont fontWithName: fontName size: size];
 }
- 
+
+
+-(void) _trySelectSize: (float)size
+{
+  unsigned int i;
+  NSBrowser *sizeBrowser = [[self contentView] viewWithTag: NSFPSizeBrowser];
+  NSTextField *sizeField;
+
+  /* Make sure our sizeField is updated. */
+  sizeField = [[self contentView] viewWithTag: NSFPSizeField];
+  _setFloatValue (sizeField, size);
+        
+  /* Make sure our column is loaded. */ 
+  [sizeBrowser loadColumnZero];
+
+  for (i = 0; i < sizeof(sizes) / sizeof(float); i++)
+    {
+      if (size == sizes[i])
+	{
+          /* select the cell */
+	  [sizeBrowser selectRow: i inColumn: 0];
+	  break;
+	}
+    }
+  if (i == sizeof(sizes) / sizeof(float))
+    {
+      /* TODO: No matching size found in the list. We should deselect
+      everything. */
+    }
+}
+
+- (void) controlTextDidChange: (NSNotification *)n
+{
+  NSTextField *sizeField = [[self contentView] viewWithTag: NSFPSizeField];
+  float size = [sizeField floatValue];
+  [self _trySelectSize: size];
+  [self _doPreview];
+}
+
+/*
+Ask the NSFontManager:s delegate if a font should be included. For speed,
+the delegate is an argument; a repeat-caller can then cache it.
+*/
+- (BOOL) _includeFont: (NSString*)fontName  delegate: (id)fmDelegate
+{
+  if (fmDelegate != nil
+    && [fmDelegate respondsToSelector: @selector(fontManager:willIncludeFont:)])
+    {
+      return [fmDelegate fontManager: [NSFontManager sharedFontManager]
+		     willIncludeFont: fontName];
+    }
+  else
+    return YES;
+}
+
+
+- (void) _getOriginalSize
+{
+  /* Used in setMinSize: */
+  _originalMinSize = [self minSize];
+  /* Used in setContentSize: */
+  _originalSize = [[self contentView] frame].size;
+}
+
 @end
 
 
 @implementation NSFontPanel (NSBrowserDelegate)
 
-- (BOOL) browser: (NSBrowser*)sender 
-       selectRow: (int)row
-	inColumn: (int)column
-{
-  switch ([sender tag])
-    {
-    case NSFPFamilyBrowser:
-      {
-	NSFontManager *fm = [NSFontManager sharedFontManager];
-	NSBrowser *faceBrowser = [[self contentView] viewWithTag: NSFPFaceBrowser];
-	int i;
 
-	ASSIGN(_faceList, [fm availableMembersOfFontFamily: 
-				[_familyList objectAtIndex: row]]);
-	_family = row;
-	
-	// Select a face with similar properties
-	for (i = 0; i < [_faceList count]; i++)
+static int score_difference(int weight1, int traits1,
+			    int weight2, int traits2)
+{
+  int score, t;
+
+  score = (weight1 - weight2);
+  score = 10 * score * score;
+
+  t = traits1 ^ traits2;
+
+  if (t & NSFixedPitchFontMask) score += 1000;
+  if (t & NSCompressedFontMask) score += 150;
+  if (t & NSPosterFontMask) score += 200;
+  if (t & NSSmallCapsFontMask) score += 200;
+  if (t & NSCondensedFontMask) score += 150;
+  if (t & NSExpandedFontMask) score += 150;
+  if (t & NSNarrowFontMask) score += 150;
+  if (t & NSBoldFontMask) score += 20;
+  if (t & NSItalicFontMask) score += 45;
+
+  return score;
+}
+
+
+- (void) _familySelectionChanged: (id)sender
+{
+  NSFontManager *fm = [NSFontManager sharedFontManager];
+  id fmDelegate = [fm delegate];
+
+  NSBrowser *faceBrowser = [[self contentView] viewWithTag: NSFPFaceBrowser];
+  NSBrowser *familyBrowser = [[self contentView] viewWithTag: NSFPFamilyBrowser];
+  int row = [familyBrowser selectedRowInColumn: 0];
+
+  unsigned int i;
+  NSArray *entireFaceList;
+
+
+  entireFaceList = [fm availableMembersOfFontFamily:
+  			[_familyList objectAtIndex:row]];
+
+  DESTROY(_faceList);
+  _faceList = [[NSMutableArray alloc] initWithCapacity: [entireFaceList count]];
+
+  for (i = 0; i < [entireFaceList count]; i++)
+    {
+      id aFace = [entireFaceList objectAtIndex:i];
+      if ([self _includeFont: [aFace objectAtIndex:0]  delegate: fmDelegate])
+	{
+	  [_faceList addObject: aFace];
+	}
+    }
+
+  _family = row;
+
+  // Select a face with the same properties
+  for (i = 0; i < [_faceList count]; i++)
+    {
+      NSArray *font_info = [_faceList objectAtIndex: i];
+
+      if (([[font_info objectAtIndex: 2] intValue] == _weight)
+	  && ([[font_info objectAtIndex: 3] unsignedIntValue] == _traits))
+	break;
+    }
+
+  // Find the face that differs the least from what we want
+  if (i == [_faceList count])
+    {
+      int best, best_score, score;
+
+      best_score = 1e6;
+      best = -1;
+
+      for (i = 0; i < [_faceList count]; i++)
 	{
 	  NSArray *font_info = [_faceList objectAtIndex: i];
-	  
-	  if (([[font_info objectAtIndex: 2]  intValue] == _weight) &&
-	      ([[font_info objectAtIndex: 3]  unsignedIntValue] == _traits))
-	    break;
+	  score = score_difference(_weight, _traits,
+	    [[font_info objectAtIndex: 2] intValue],
+	    [[font_info objectAtIndex: 3] unsignedIntValue]);
+	  if (score < best_score)
+	    {
+	      best = i;
+	      best_score = score;
+	    }
 	}
-	if (i == [_faceList count])
-	  i = 0;
-
-	_face = i;
-	[faceBrowser loadColumnZero];
-	[faceBrowser selectRow: i inColumn: 0];
-	break;
-      }
-    case NSFPFaceBrowser:
-      {
-	NSArray *font_info = [_faceList objectAtIndex: row];
-
-	_face = row;
-	_weight = [[font_info objectAtIndex: 2] intValue];
-	_traits = [[font_info objectAtIndex: 3] unsignedIntValue];
-	break;
-      }
-    case NSFPSizeBrowser:
-    default:
-      {
-	NSTextField *sizeField;
-
-	sizeField = [[self contentView] viewWithTag: NSFPSizeField];
-	_setFloatValue (sizeField, sizes[row]);
-	break;
-      }
+      if (best != -1)
+	i = best;
     }
+
+  if (i == [_faceList count])
+    i = 0;
+
+  _face = i;
+  [faceBrowser loadColumnZero];
+  [faceBrowser selectRow: i inColumn: 0];
+
+  /* Also make sure the size column has some value */
+  {
+    NSTextField *sizeField = [[self contentView] viewWithTag: NSFPSizeField];
+    float size = [sizeField floatValue];
+
+    if (size == 0.0)
+      {
+	[self _trySelectSize: 12.0];
+      }
+  }
+
+  [self _doPreview];
+}
+
+- (void) _faceSelectionChanged: (id)sender
+{
+  NSBrowser *faceBrowser = [[self contentView] viewWithTag: NSFPFaceBrowser];
+  int row = [faceBrowser selectedRowInColumn: 0];
+  NSArray *font_info = [_faceList objectAtIndex: row];
+
+  _face = row;
+  _weight = [[font_info objectAtIndex: 2] intValue];
+  _traits = [[font_info objectAtIndex: 3] unsignedIntValue];
+
+  [self _doPreview];
+}
+
+- (void) _sizeSelectionChanged: (id)sender
+{
+  NSBrowser *sizeBrowser = [[self contentView] viewWithTag: NSFPSizeBrowser];
+  int row = [sizeBrowser selectedRowInColumn: 0];
+  NSTextField *sizeField;
+
+  sizeField = [[self contentView] viewWithTag: NSFPSizeField];
+  _setFloatValue (sizeField, sizes[row]);
   
   [self _doPreview];
-
-  return YES;
 }
 
+
 - (int) browser: (NSBrowser*)sender  numberOfRowsInColumn: (int)column
 {
   switch ([sender tag])
@@ -836,12 +1131,15 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
 	  column: (int)column
 {
   NSString *value = nil;
-  
+
+  if (row < 0)
+    return;
+
   switch ([sender tag])
     {
     case NSFPFamilyBrowser:
       {
-	if ([_familyList count] > row)
+	if ([_familyList count] > (unsigned)row)
 	  {
 	    value = [_familyList objectAtIndex: row];
 	  }
@@ -849,7 +1147,7 @@ float sizes[] = {4.0, 6.0, 8.0, 9.0, 10.0, 11.0, 12.0, 13.0,
       }
     case NSFPFaceBrowser:
       {
-	if ([_faceList count] > row)
+	if ([_faceList count] > (unsigned)row)
 	  {
 	    value = [[_faceList objectAtIndex: row] objectAtIndex: 1];
 	  } 
