@@ -27,13 +27,13 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSNotification.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSFormCell.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSTextFieldCell.h>
+#include "AppKit/NSColor.h"
+#include "AppKit/NSFormCell.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSTextFieldCell.h"
 
 static NSColor	*shadowCol;
 
@@ -214,7 +214,7 @@ static NSColor	*shadowCol;
     }
 }
 
-- (float)titleWidth: (NSSize)size
+- (float)titleWidth: (NSSize)aSize
 {
   if (_formcell_auto_title_width == NO)
     return _displayedTitleWidth;
@@ -222,10 +222,10 @@ static NSColor	*shadowCol;
     {
       NSSize titleSize = [_titleCell cellSize];
 
-      if (size.width > titleSize.width)
+      if (aSize.width > titleSize.width)
 	return titleSize.width;
       else
-	return size.width;
+	return aSize.width;
     }
 }
 
@@ -317,16 +317,12 @@ static NSColor	*shadowCol;
   
   if (_cell.is_bordered)
     {
-      [controlView lockFocus];
       [shadowCol set];
       NSFrameRect(borderedFrame);
-      [controlView unlockFocus];
     }
   else if (_cell.is_bezeled)
     {
-      [controlView lockFocus];
       NSDrawWhiteBezel(borderedFrame, NSZeroRect);
-      [controlView unlockFocus];
     }
 
   //
