@@ -22,7 +22,7 @@
    If not, write to the Free Software Foundation,
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
    */ 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <string.h>
 
 #include <Foundation/NSString.h>
@@ -35,20 +35,50 @@
 #include <Foundation/NSString.h>
 #include <Foundation/NSDebug.h>
 
-#include <AppKit/NSImage.h>
-#include <AppKit/AppKitExceptions.h>
-#include <AppKit/NSBitmapImageRep.h>
-#include <AppKit/NSCachedImageRep.h>
-#include <AppKit/NSView.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSScreen.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSPasteboard.h>
-#include <AppKit/PSOperators.h>
+#include "AppKit/NSImage.h"
+#include "AppKit/AppKitExceptions.h"
+#include "AppKit/NSBitmapImageRep.h"
+#include "AppKit/NSCachedImageRep.h"
+#include "AppKit/NSView.h"
+#include "AppKit/NSWindow.h"
+#include "AppKit/NSScreen.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSPasteboard.h"
+#include "AppKit/NSPrintOperation.h"
+#include "AppKit/PSOperators.h"
+#include "GNUstepGUI/GSDisplayServer.h"
 
-BOOL	NSImageDoesCaching = YES;	/* enable caching	*/
 BOOL	NSImageForceCaching = NO;	/* use on missmatch	*/
 
+@implementation NSBundle (NSImageAdditions)
+
+- (NSString*) pathForImageResource: (NSString*)name
+{
+  NSString	*ext = [name pathExtension];
+  NSString	*path = nil;
+
+  if ((ext == nil) || [ext isEqualToString:@""])
+    {
+      NSArray	*types = [NSImage imageUnfilteredFileTypes];
+      unsigned	c = [types count];
+      unsigned	i;
+
+      for (i = 0; path == nil && i < c; i++)
+	{
+	  ext = [types objectAtIndex: i];
+	  path = [self pathForResource: name ofType: ext];
+	}
+    }
+  else
+    {
+      name = [name stringByDeletingPathExtension];
+      path = [self pathForResource: name ofType: ext];
+    }
+  return path;
+}
+
+@end
+
 @interface	GSRepData : NSObject
 {
 @public
@@ -85,10 +115,10 @@ static NSColor			*clearColor = nil;
 static Class			cachedClass = 0;
 static Class			bitmapClass = 0;
 
-NSArray *iterate_reps_for_types(NSArray *imageReps, SEL method);
+static NSArray *iterate_reps_for_types(NSArray *imageReps, SEL method);
 
 /* Find the GSRepData object holding a representation */
-GSRepData*
+static GSRepData*
 repd_for_rep(NSArray *_reps, NSImageRep *rep)
 {
   NSEnumerator	*enumerator = [_reps objectEnumerator];
@@ -122,16 +152,9 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
 {
   if (self == [NSImage class])
     {
-#ifdef GNUSTEP_BASE_LIBRARY
-      NSString	*path = [NSBundle pathForGNUstepResource: @"nsmapping"
-				  ofType: @"strings"
-				  inDirectory: @"Images"];
-#else
-      NSBundle	*system = [NSBundle bundleWithPath: @GNUSTEP_INSTALL_LIBDIR];
-      NSString	*path = [system pathForResource: @"nsmapping"
-					 ofType: @"strings"
-				    inDirectory: @"Images"];
-#endif
+      NSString *path = [NSBundle pathForLibraryResource: @"nsmapping"
+				                 ofType: @"strings"
+				             inDirectory: @"Images"];
 
       // Initial version
       [self setVersion: 1];
@@ -213,46 +236,23 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
 	{
 	  if (ext)
 	    {
-#ifdef GNUSTEP_BASE_LIBRARY
-	      path = [NSBundle pathForGNUstepResource: the_name
-			       ofType: ext
-			       inDirectory: @"Images"];
-#else
-	      NSBundle *system = [NSBundle bundleWithPath: @GNUSTEP_INSTALL_LIBDIR];
-	  
-	      path = [system pathForResource: the_name
-			     ofType: ext
-			     inDirectory: @"Images"];
-#endif 
+	      path = [NSBundle pathForLibraryResource: the_name
+				               ofType: ext
+				          inDirectory: @"Images"];
 	    }
 	  else 
 	    {
-#ifdef GNUSTEP_BASE_LIBRARY
 	      id o, e;
 
 	      e = [array objectEnumerator];
 	      while ((o = [e nextObject]))
 		{
-		  path = [NSBundle pathForGNUstepResource: the_name
-			       ofType: o
-			       inDirectory: @"Images"];
+		  path = [NSBundle pathForLibraryResource: the_name
+			       	                   ofType: o
+				              inDirectory: @"Images"];
 		  if (path != nil && [path length] != 0)
 		    break;
 		}
-#else
-	      NSBundle *system = [NSBundle bundleWithPath: @GNUSTEP_INSTALL_LIBDIR];
-	      id o, e;
-
-	      e = [array objectEnumerator];
-	      while ((o = [e nextObject]))
-		{
-		  path = [system pathForResource: the_name
-					  ofType: o
-				     inDirectory: @"Images"];
-		  if (path != nil && [path length] != 0)
-		    break;
-		}
-#endif
 	    }
 	}
 
@@ -300,6 +300,7 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
   //_flags.syncLoad = NO;
   _reps = [[NSMutableArray alloc] initWithCapacity: 2];
   ASSIGN(_color, clearColor);
+  _cacheMode = NSImageCacheDefault;
   
   return self;
 }
@@ -357,6 +358,7 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
     }
   self = [self init];
   [self addRepresentation: rep];
+  RELEASE(rep);
   return self;
 }
 
@@ -372,6 +374,7 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
     }
   self = [self init];
   [self addRepresentation: rep];
+  RELEASE(rep);
   return self;
 }
 
@@ -399,9 +402,12 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
     [self addRepresentations: reps]; 
   else
     {
-      NSString* file = [pasteboard propertyListForType: NSFilenamesPboardType];
+      NSArray *array = [pasteboard propertyListForType: NSFilenamesPboardType];
+      NSString* file; 
       
-      if (file != nil || ![self _loadFromFile: file])
+      if ((array == nil) || ([array count] == 0) ||
+	  (file = [array objectAtIndex: 0]) == nil || 
+	  ![self _loadFromFile: file])
         {
 	  RELEASE(self);
 	  return nil;
@@ -454,11 +460,11 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
   return copy;
 }
 
-- (BOOL) setName: (NSString *)string
+- (BOOL) setName: (NSString *)aName
 {
   BOOL retained = NO;
   
-  if (!string || [nameDict objectForKey: string])
+  if (!aName || [nameDict objectForKey: aName])
     return NO;
 
   if (_name && self == [nameDict objectForKey: _name])
@@ -470,7 +476,7 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
       [nameDict removeObjectForKey: _name];
     }
   
-  ASSIGN(_name, string);
+  ASSIGN(_name, aName);
   
   [nameDict setObject: self forKey: _name];
   if (retained)
@@ -581,6 +587,17 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
   return _flags.unboundedCacheDepth;
 }
 
+- (void) setCacheMode: (NSImageCacheMode)mode
+{
+  _cacheMode = mode;
+}
+
+- (NSImageCacheMode) cacheMode
+{
+  return _cacheMode;
+}
+
+
 // Determining How the Image is Drawn 
 - (BOOL) isValid
 {
@@ -606,7 +623,7 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
 
 - (void) recache
 {
-  unsigned i, count;
+  unsigned i;
 
   i = [_reps count];
   while(i--) 
@@ -662,22 +679,20 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
 		 fromRect: (NSRect)aRect
 		operation: (NSCompositingOperation)op
 {
-  NSImageRep *rep;
+  NSImageRep *rep = nil;
 
   NS_DURING
-    {
-      rep = [self _doImageCache];
-      
-      if (NSImageDoesCaching == YES && [rep isKindOfClass: cachedClass])
+    { 
+      if ([GSCurrentContext() isDrawingToScreen] == YES)
+	  rep = [self _doImageCache];
+      if (rep
+	  &&_cacheMode != NSImageCacheNever 
+	  && [rep isKindOfClass: cachedClass])
         {
-	  NSRect rect = [(NSCachedImageRep *)rep rect];
-	  NSGraphicsContext *ctxt = GSCurrentContext();	  
+	  NSRect rect;
 	  float y = aPoint.y;
 
-	  // FIXME: This undos the change done in NSCell, perhaps we can remove both?
-	  if ([[ctxt focusView] isFlipped])
-	    y -= rect.size.height;
-
+	  rect = [(NSCachedImageRep *)rep rect];
 	  NSDebugLLog(@"NSImage", @"composite rect %@ in %@", 
 		      NSStringFromRect(rect), NSStringFromRect(aRect));
 	  // Move the drawing rectangle to the origin of the image rep
@@ -692,13 +707,15 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
       else	
         {
 	  NSRect rect;
-
+          rep = [self bestRepresentationForDevice: nil];
 	  rect = NSMakeRect(aPoint.x, aPoint.y, _size.width, _size.height);
 	  [self drawRepresentation: rep inRect: rect];
 	}
     }
   NS_HANDLER
     {
+      NSLog(@"NSImage: compositeToPoint:fromRect:operation: failed due to %@: %@", 
+	    [localException name], [localException reason]);
       if ([_delegate respondsToSelector: @selector(imageDidNotDraw:inRect:)])
         {
 	  NSImage *image = [_delegate imageDidNotDraw: self inRect: aRect];
@@ -745,21 +762,20 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
 		fromRect: (NSRect)aRect 
 		fraction: (float)aFloat
 {
-  NSImageRep *rep;
+  NSImageRep *rep = nil;
 
   NS_DURING
     {
-      rep = [self _doImageCache];
-
-      if (NSImageDoesCaching == YES && [rep isKindOfClass: cachedClass])
+      if ([GSCurrentContext() isDrawingToScreen] == YES)
+	  rep = [self _doImageCache];
+      if (rep
+	  &&_cacheMode != NSImageCacheNever 
+	  && [rep isKindOfClass: cachedClass])
         {
-	  NSRect rect = [(NSCachedImageRep *)rep rect];
-	  NSGraphicsContext *ctxt = GSCurrentContext();	  
+	  NSRect rect;
 	  float y = aPoint.y;
 
-	  if ([[ctxt focusView] isFlipped])
-	    y -= rect.size.height;
-
+	  rect = [(NSCachedImageRep *)rep rect];
 	  // Move the drawing rectangle to the origin of the image rep
 	  // and intersect the two rects.
 	  aRect.origin.x += rect.origin.x;
@@ -772,12 +788,17 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
         {
 	  NSRect rect;
 
+	  /* FIXME: Here we are supposed to composite directly from the source
+	     but how do you do that? */
+          rep = [self bestRepresentationForDevice: nil];
 	  rect = NSMakeRect(aPoint.x, aPoint.y, _size.width, _size.height);
 	  [self drawRepresentation: rep inRect: rect];
 	}
     }
   NS_HANDLER
     {
+      NSLog(@"NSImage: dissolve failed due to %@: %@", 
+	    [localException name], [localException reason]);
       if ([_delegate respondsToSelector: @selector(imageDidNotDraw:inRect:)])
         {
 	  NSImage *image = [_delegate imageDidNotDraw: self inRect: aRect];
@@ -795,8 +816,16 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
 {
   if (_color != nil)
     {
+      NSRect fillrect = aRect;
       [_color set];
-      NSRectFill(aRect);
+      if ([[NSView focusView] isFlipped])
+	fillrect.origin.y -= _size.height;
+      NSRectFill(fillrect);
+      if ([GSCurrentContext() isDrawingToScreen] == NO)
+	{
+	  /* Reset alpha for image drawing. */
+	  [[NSColor whiteColor] set];
+	}
     }
 
   if (!_flags.scalable) 
@@ -872,20 +901,44 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
 
 - (void) lockFocusOnRepresentation: (NSImageRep *)imageRep
 {
-  if (NSImageDoesCaching == YES)
+  if (_cacheMode != NSImageCacheNever)
     {
       NSWindow	*window;
+      GSRepData *repd;
 
       if (imageRep == nil)
 	imageRep = [self bestRepresentationForDevice: nil];
 
       imageRep = [self _cacheForRep: imageRep];
+      repd = repd_for_rep(_reps, imageRep);
+
       window = [(NSCachedImageRep *)imageRep window];
       _lockedView = [window contentView];
       if (_lockedView == nil)
 	[NSException raise: NSImageCacheException
 		     format: @"Cannot lock focus on nil rep"];
       [_lockedView lockFocus];
+
+      /* Validate cached image */
+      if (repd->bg == nil)
+	{
+	  repd->bg = [_color copy];
+	  [_color set];
+	  
+	  if ([_color alphaComponent] < 1)
+	    {
+	      /* With a Quartz-like alpha model, alpha can't be cleared
+	         with a rectfill, so we need to clear the alpha channel
+		 explictly. (A compositerect with NSCompositeCopy would
+		 be more efficient, but it doesn't seem like it's 
+		 implemented correctly in all backends yet (as of 
+		 2002-08-23). Also, this will work with both the Quartz-
+		 and DPS-model.) */
+	      PScompositerect(0, 0, _size.width, _size.height,
+			      NSCompositeClear);
+	    }
+	  NSRectFill(NSMakeRect(0, 0, _size.width, _size.height));
+	}
     }
 }
 
@@ -898,30 +951,102 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
     }
 }
 
-- (NSImageRep*) bestRepresentationForDevice: (NSDictionary*)deviceDescription
+/* Determine if the device is color or gray scale and find the reps of
+   the same type
+*/
+- (NSMutableArray *) _bestRep: (NSArray *)reps 
+	       withColorMatch: (NSDictionary*)deviceDescription
 {
-  NSArray *reps = [self representations];
+  int colors = 3;
+  NSImageRep* rep;
+  NSMutableArray *breps;
   NSEnumerator *enumerator = [reps objectEnumerator];
-  NSImageRep *rep = nil;  
-  NSImageRep *best = nil;
-
+  NSString *colorSpace = [deviceDescription objectForKey: NSDeviceColorSpaceName];
+  
+  if (colorSpace != nil)
+    colors = NSNumberOfColorComponents(colorSpace);
+  
+  breps = [NSMutableArray array];
   while ((rep = [enumerator nextObject]) != nil)
     {
-      /*
-       * What's the best representation? 
-       * FIXME: At the moment we take the last bitmap we find.
-       * If we can't find a bitmap, we take whatever we can.
-       * NSCursor will only handle returned NSBitmapImageReps!
-       */
-      if ([rep isKindOfClass: bitmapClass])
-	best = rep;
-      else if (best == nil)
-	best = rep;
+      if ([rep colorSpaceName] || abs(NSNumberOfColorComponents([rep colorSpaceName]) - colors) <= 1)
+        [breps addObject: rep];
     }
-  return best;
+  
+  /* If there are no matches, pass all the reps */
+  if ([breps count] == 0)
+    return (NSMutableArray *)reps;
+  return breps;
 }
 
-- (NSArray *) representations
+/* Find reps that match the resolution of the device or return the rep
+   that has the highest resolution */
+- (NSMutableArray *) _bestRep: (NSArray *)reps 
+	  withResolutionMatch: (NSDictionary*)deviceDescription
+{
+  NSImageRep* rep;
+  NSMutableArray *breps;
+  NSSize dres;
+  NSEnumerator *enumerator = [reps objectEnumerator];
+  NSValue *resolution = [deviceDescription objectForKey: NSDeviceResolution];
+
+  if (resolution)
+    dres = [resolution sizeValue];
+  else
+    dres = NSMakeSize(0, 0);
+
+  breps = [NSMutableArray array];
+  while ((rep = [enumerator nextObject]) != nil)
+    {
+      /* FIXME: Not sure about checking resolution */
+      [breps addObject: rep];
+    }
+  
+  /* If there are no matches, pass all the reps */
+  if ([breps count] == 0)
+    return (NSMutableArray *)reps;
+  return breps;
+}
+
+/* Find reps that match the bps of the device or return the rep that
+   has the highest bps */
+- (NSMutableArray *) _bestRep: (NSArray *)reps 
+		 withBpsMatch: (NSDictionary*)deviceDescription
+{
+  NSImageRep* rep, *max_rep;
+  NSMutableArray *breps;
+  NSEnumerator *enumerator = [reps objectEnumerator];
+  int bps = [[deviceDescription objectForKey: NSDeviceBitsPerSample] intValue];
+  int max_bps;
+
+  breps = [NSMutableArray array];
+  max_bps = 0;
+  max_rep = nil;
+  while ((rep = [enumerator nextObject]) != nil)
+    {
+      int rep_bps = 0;
+      if ([rep respondsToSelector: @selector(bitsPerPixel)])
+        rep_bps = [(NSBitmapImageRep *)rep bitsPerPixel];
+      if (rep_bps > max_bps)
+        {
+          max_bps = rep_bps;
+          max_rep = rep;
+        }
+      if (rep_bps == bps)
+        [breps addObject: rep];
+    }
+  
+
+  if ([breps count] == 0 && max_rep != nil)
+    [breps addObject: max_rep];
+
+  /* If there are no matches, pass all the reps */
+  if ([breps count] == 0)
+    return (NSMutableArray *)reps;
+  return breps;
+}
+
+- (NSMutableArray *) _representationsWithCachedImages: (BOOL)flag
 {
   unsigned	count;
 
@@ -941,17 +1066,63 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
   else
     {
       id	repList[count];
-      unsigned	i;
+      unsigned	i, j;
 
       [_reps getObjects: repList];
+      j = 0;
       for (i = 0; i < count; i++) 
 	{
-	  repList[i] = ((GSRepData*)repList[i])->rep;
+          if (flag || ((GSRepData*)repList[i])->original == nil)
+            {
+              repList[j] = ((GSRepData*)repList[i])->rep;
+              j++;
+            }
 	}
-      return [NSArray arrayWithObjects: repList count: count];
+      return [NSArray arrayWithObjects: repList count: j];
     }
 }
 
+- (NSImageRep*) bestRepresentationForDevice: (NSDictionary*)deviceDescription
+{
+  NSMutableArray *reps = [self _representationsWithCachedImages: NO];
+  
+  if (deviceDescription == nil)
+    {
+      if ([GSCurrentContext() isDrawingToScreen] == YES)
+        {
+          int screen = [[[GSCurrentServer() attributes] 
+			  objectForKey: GSScreenNumber] intValue];
+          deviceDescription = [[[NSScreen screens] objectAtIndex: screen] 
+				deviceDescription];
+        }
+      else if ([NSPrintOperation currentOperation])
+        {
+          /* FIXME: We could try to use the current printer, 
+	     but there are many cases where might
+             not be printing (EPS, PDF, etc) to a specific device */
+        }
+    }
+
+  if (_flags.colorMatchPreferred == YES)
+    {
+      reps = [self _bestRep: reps withColorMatch: deviceDescription];
+      reps = [self _bestRep: reps withResolutionMatch: deviceDescription];
+    }
+  else
+    {
+      reps = [self _bestRep: reps withResolutionMatch: deviceDescription];
+      reps = [self _bestRep: reps withColorMatch: deviceDescription];
+    }
+  reps = [self _bestRep: reps withBpsMatch: deviceDescription];
+  /* Pick an arbitrary representation if there is more than one */
+  return [reps lastObject];
+}
+
+- (NSArray *) representations
+{
+  return [self _representationsWithCachedImages: YES];
+}
+
 - (void) setDelegate: anObject
 {
   _delegate = anObject;
@@ -1138,7 +1309,7 @@ repd_for_rep(NSArray *_reps, NSImageRep *rep)
    array of objects.  Add these together, with duplicates
    weeded out.  Used by imageUnfilteredPasteboardTypes,
    imageUnfilteredFileTypes, etc. */
-NSArray *
+static NSArray *
 iterate_reps_for_types(NSArray* imageReps, SEL method)
 {
   NSImageRep	*rep;
@@ -1246,7 +1417,7 @@ iterate_reps_for_types(NSArray* imageReps, SEL method)
 {
   NSImageRep *rep = [self bestRepresentationForDevice: nil];
 
-  if (NSImageDoesCaching == YES)
+  if (_cacheMode != NSImageCacheNever)
     {
       GSRepData *repd;
 
