@@ -25,7 +25,7 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <string.h>
 #include <Foundation/NSArray.h>
 #include <Foundation/NSData.h>
@@ -35,14 +35,14 @@
 #include <Foundation/NSNotification.h>
 #include <Foundation/NSUserDefaults.h>
 #include <Foundation/NSDebug.h>
-#include <AppKit/NSImageRep.h>
-#include <AppKit/NSBitmapImageRep.h>
-#include <AppKit/NSEPSImageRep.h>
-#include <AppKit/NSPasteboard.h>
-#include <AppKit/NSGraphicsContext.h>
-#include <AppKit/NSView.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/DPSOperators.h>
+#include "AppKit/NSImageRep.h"
+#include "AppKit/NSBitmapImageRep.h"
+#include "AppKit/NSEPSImageRep.h"
+#include "AppKit/NSPasteboard.h"
+#include "AppKit/NSGraphicsContext.h"
+#include "AppKit/NSView.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/DPSOperators.h"
 
 static NSMutableArray *imageReps = nil;
 static Class NSImageRep_class = NULL;
@@ -55,8 +55,6 @@ static Class NSImageRep_class = NULL;
      them can load in data from an external source. */
   if (self == [NSImageRep class])
     {
-      id obj;
-
       NSImageRep_class = self;
       imageReps = [[NSMutableArray alloc] initWithCapacity: 2];
       [imageReps addObject: [NSBitmapImageRep class]];
@@ -171,6 +169,7 @@ static Class NSImageRep_class = NULL;
       NSLog(@"Extension missing from image filename - '%@'", filename);
       return nil;
     }
+  ext = [ext lowercaseString];
 
   if (self == NSImageRep_class)
     {
@@ -183,24 +182,20 @@ static Class NSImageRep_class = NULL;
   else
     return nil;
 
-  // This is a GNUstep extension for special file types
-  if ([rep respondsToSelector: @selector(imageRepsWithFile:)])
-    return [rep imageRepsWithFile: filename];
-  else
-    {  
-      NSData* data;
-	  
-      data = [NSData dataWithContentsOfFile: filename];
-      if ([rep respondsToSelector: @selector(imageRepsWithData:)])
-	return [rep imageRepsWithData: data];
-      else if ([rep respondsToSelector: @selector(imageRepWithData:)])
-        {
-	  NSImageRep *imageRep = [rep imageRepWithData: data];
-	      
-	  if (imageRep != nil)
-	    return [NSArray arrayWithObject: imageRep];
-	}
-    }
+  {
+    NSData* data;
+
+    data = [NSData dataWithContentsOfFile: filename];
+    if ([rep respondsToSelector: @selector(imageRepsWithData:)])
+      return [rep imageRepsWithData: data];
+    else if ([rep respondsToSelector: @selector(imageRepWithData:)])
+      {
+	NSImageRep *imageRep = [rep imageRepWithData: data];
+
+	if (imageRep != nil)
+	  return [NSArray arrayWithObject: imageRep];
+      }
+  }
 
   return nil;
 }
@@ -233,13 +228,13 @@ static Class NSImageRep_class = NULL;
     }
   else
     return nil;
-	  
+
   if ([rep respondsToSelector: @selector(imageRepsWithData:)])
     return [rep imageRepsWithData: data];
   else if ([rep respondsToSelector: @selector(imageRepWithData:)])
     {
       NSImageRep *imageRep = [rep imageRepWithData: data];
-	      
+
       if (imageRep != nil)
 	return [NSArray arrayWithObject: imageRep];
     }
@@ -434,6 +429,7 @@ static Class NSImageRep_class = NULL;
 {
   BOOL ok, reset;
   NSGraphicsContext *ctxt;
+  NSAffineTransform *ctm = nil;
 
   if (_size.width == 0 && _size.height == 0)
     return NO;
@@ -445,13 +441,13 @@ static Class NSImageRep_class = NULL;
     {
       if ([[ctxt focusView] isFlipped])
 	aPoint.y -= _size.height;
-      DPSmatrix(ctxt); DPScurrentmatrix(ctxt);
+      ctm = GSCurrentCTM(ctxt);
       DPStranslate(ctxt, aPoint.x, aPoint.y);
       reset = 1;
     }
   ok = [self draw];
   if (reset)
-    DPSsetmatrix(ctxt);
+    GSSetCTM(ctxt, ctm);
   return ok;
 }
 
@@ -460,6 +456,7 @@ static Class NSImageRep_class = NULL;
   NSSize scale;
   BOOL ok;
   NSGraphicsContext *ctxt;
+  NSAffineTransform *ctm;
 
   NSDebugLLog(@"NSImage", @"Drawing in rect (%f %f %f %f)\n", 
 	      NSMinX(aRect), NSMinY(aRect), NSWidth(aRect), NSHeight(aRect));
@@ -471,11 +468,11 @@ static Class NSImageRep_class = NULL;
 		     NSHeight(aRect) / _size.height);
   if ([[ctxt focusView] isFlipped])
     aRect.origin.y -= NSHeight(aRect);
-  DPSmatrix(ctxt); DPScurrentmatrix(ctxt);
+  ctm = GSCurrentCTM(ctxt);
   DPStranslate(ctxt, NSMinX(aRect), NSMinY(aRect));
   DPSscale(ctxt, scale.width, scale.height);
   ok = [self draw];
-  DPSsetmatrix(ctxt);
+  GSSetCTM(ctxt, ctm);
   return ok;
 }
 
