@@ -33,24 +33,26 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 
 #include <Foundation/NSBundle.h>
 #include <Foundation/NSCoder.h>
-#include <AppKit/NSPanel.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSButton.h>
-#include <AppKit/NSMatrix.h>
-#include <AppKit/NSTextField.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSBox.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSScreen.h>
-#include <AppKit/NSScrollView.h>
-#include <AppKit/IMLoading.h>
-#include <AppKit/GMAppKit.h>
+#include "AppKit/NSPanel.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSButton.h"
+#include "AppKit/NSMatrix.h"
+#include "AppKit/NSTextField.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSBox.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSScreen.h"
+#include "AppKit/NSScrollView.h"
+#include "AppKit/NSHelpManager.h"
+
+#include "GNUstepGUI/IMLoading.h"
+#include "GNUstepGUI/GMAppKit.h"
+#include "GNUstepGUI/GMArchiver.h"
 
-#include <AppKit/GMArchiver.h>
 
 #ifdef ALERT_TITLE
 static NSString	*defaultTitle = @"Alert";
@@ -94,8 +96,6 @@ static NSString	*defaultTitle = @" ";
 
 - (BOOL) canBecomeKeyWindow
 {
-  if (_becomesKeyOnlyIfNeeded)
-    return NO;
   return YES;
 }
 
@@ -191,6 +191,70 @@ static NSString	*defaultTitle = @" ";
   return self;
 }
 
+
+- (void) sendEvent: (NSEvent*)theEvent
+{
+/*
+Code shared with [NSWindow -sendEvent:], remember to update both places.
+*/
+  NSView	*v;
+  NSEventType	type;
+
+  type = [theEvent type];
+  if (type == NSLeftMouseDown)
+    {
+      if (_f.visible)
+	{
+	  BOOL	wasKey = _f.is_key;
+
+	  if (_f.has_closed == NO)
+	    {
+	      v = [_contentView hitTest: [theEvent locationInWindow]];
+	      if (_f.is_key == NO)
+		{
+		  /* NSPanel modification: check becomesKeyOnlyIfNeeded. */
+		  if (!_becomesKeyOnlyIfNeeded || [v needsPanelToBecomeKey])
+		    [self makeKeyAndOrderFront: self];
+		}
+	      /* Activate the app *after* making the receiver key, as app
+		 activation tries to make the previous key window key. */
+	      if ([NSApp isActive] == NO 
+		  && (NSWindow*)self != [NSApp iconWindow])
+		{
+		  [NSApp activateIgnoringOtherApps: YES];
+		}
+	      if (_firstResponder != v)
+		{
+		  [self makeFirstResponder: v];
+		}
+	      if (_lastView)
+ 		{
+		  DESTROY(_lastView);
+		}
+	      if (wasKey == YES || [v acceptsFirstMouse: theEvent] == YES)
+		{
+		  if ([NSHelpManager isContextHelpModeActive])
+		    {
+		      [v helpRequested: theEvent];
+		    }
+		  else
+		    {
+		      ASSIGN(_lastView, v);
+		      [v mouseDown: theEvent];
+		    }
+		}
+	      else
+		{
+		  [self mouseDown: theEvent];
+		}
+	    }
+	  _lastPoint = [theEvent locationInWindow];
+	}
+      return;
+    }
+  [super sendEvent: theEvent];
+}
+
 @end /* NSPanel */
 
 
@@ -332,7 +396,7 @@ static GSAlertPanel	*criticalAlertPanel = nil;
 
 @implementation	GSAlertPanel
 
-static const float WTitleHeight = 0.0;      // TODO: Check this value.
+//static const float WTitleHeight = 0.0;      // TODO: Check this value.
 static const float WinMinWidth = 362.0;
 static const float WinMinHeight = 161.0;
 static const float IconSide = 48.0;
@@ -343,9 +407,9 @@ static const float TitleMinRight = 8.0;
 static const float LineHeight = 2.0;
 static const float LineBottom = -66.0;       // from the top of the window.
 static const float LineLeft = 0.0;
-static const float ScrollMinSize = 48.0;     // in either direction.
+//static const float ScrollMinSize = 48.0;     // in either direction.
 static const float MessageHorzMargin = 8.0;  // 5 is too little margin.
-static const float MessageMinHeight = 20.0;
+//static const float MessageMinHeight = 20.0;
 static const float MessageVertMargin = 6.0;  // from the top of the buttons.
 static const float MessageTop = -72;         // from the top of the window;
 static const float ButtonBottom = 8.0;       // from the bottom of the window.
@@ -355,7 +419,6 @@ static const float ButtonMinHeight = 24.0;
 static const float ButtonMinWidth = 72.0;
 
 #define MessageFont [NSFont messageFontOfSize: 14]
-// TODO: Check on NeXTSTEP, I think that the message font size is only 12.0.
 
 + (void) initialize
 {
@@ -408,7 +471,7 @@ makeScrollViewWithRect(NSRect rect)
   NSScrollView	*scroll = [[NSScrollView alloc]initWithFrame: rect];
 
   [scroll setBorderType: NSLineBorder];
-  [scroll setBackgroundColor: [NSColor lightGrayColor]];
+  [scroll setBackgroundColor: [NSColor controlBackgroundColor]];
   [scroll setHasHorizontalScroller: YES];
   [scroll setHasVerticalScroller: YES];
   [scroll setScrollsDynamically: YES];
@@ -459,11 +522,14 @@ setControl(NSView* content, id control, NSString *title)
 
 - (id) _initWithoutGModel
 {
-  NSRect rect;
+  NSRect  rect;
   NSImage *image;
-  NSBox	*box;
-  NSView *content;
-  NSRect r = NSMakeRect(0.0, 0.0, WinMinWidth, WinMinHeight);
+  NSBox	  *box;
+  NSView  *content;
+  NSRect  r = NSMakeRect(0.0, 0.0, WinMinWidth, WinMinHeight);
+  NSFont  *titleFont = [NSFont systemFontOfSize: 18.0];
+  float   titleHeight = [titleFont boundingRectForFont].size.height;
+
 
   self = [self initWithContentRect: r
 	       styleMask: NSTitledWindowMask
@@ -494,10 +560,12 @@ setControl(NSView* content, id control, NSString *title)
   image = [[NSApplication sharedApplication] applicationIconImage];
   [icoButton setImage: image];
   [content addSubview: icoButton];
-  
-  rect.size.height = 20.0; // will be sized to fit anyway.
-  rect.size.width = 80.0;  // will be sized to fit anyway.
-  rect.origin.y = r.origin.y + r.size.height + IconBottom;
+
+  // Title
+  rect.size.height = 0.0; // will be sized to fit anyway.
+  rect.size.width = 0.0;  // will be sized to fit anyway.
+  rect.origin.y = r.origin.y + r.size.height 
+                  + IconBottom + (IconSide - titleHeight)/2;;
   rect.origin.x = TitleLeft;
   titleField = [[NSTextField alloc] initWithFrame: rect];
   [titleField setAutoresizingMask: NSViewMinYMargin];
@@ -506,8 +574,9 @@ setControl(NSView* content, id control, NSString *title)
   [titleField setBezeled: NO];
   [titleField setDrawsBackground: NO];
   [titleField setStringValue: @""];
-  [titleField setFont: [NSFont systemFontOfSize: 18.0]];
-  
+  [titleField setFont: titleFont];
+
+  // Horizontal line
   rect.size.height = LineHeight;
   rect.size.width = r.size.width;
   rect.origin.y = r.origin.y + r.size.height + LineBottom;
@@ -520,8 +589,8 @@ setControl(NSView* content, id control, NSString *title)
   RELEASE(box);
   
   // Then, make the subviews that'll be sized by sizePanelToFit;
-  rect.size.height = 20.0;
-  rect.size.width = 80.0;
+  rect.size.height = 0.0;
+  rect.size.width = 0.0;
   rect.origin.y = 0.0;
   rect.origin.x = 0.0;
   
@@ -536,13 +605,15 @@ setControl(NSView* content, id control, NSString *title)
   */
   [messageField setBezeled: NO];
   [messageField setDrawsBackground: YES];
-  [messageField setBackgroundColor: [NSColor lightGrayColor]];
+  [messageField setBackgroundColor: [NSColor controlBackgroundColor]];
   [messageField setAlignment: NSCenterTextAlignment];
   [messageField setStringValue: @""];
   [messageField setFont: MessageFont];
   
   defButton = [self _makeButtonWithRect: rect];
   [defButton setKeyEquivalent: @"\r"];
+  [defButton setHighlightsBy: NSPushInCellMask | NSChangeGrayCellMask 
+                              | NSContentsCellMask];
   [defButton setImagePosition: NSImageRight];
   [defButton setImage: [NSImage imageNamed: @"common_ret"]];
   [defButton setAlternateImage: [NSImage imageNamed: @"common_retH"]];
@@ -773,7 +844,7 @@ setControl(NSView* content, id control, NSString *title)
 	   * the window has a minimum size, thus may be greated
 	   * than expected.
 	   */
-	  mrect.origin.x = bounds.origin.x + MessageHorzMargin;
+	  mrect.origin.x = (wsize.width - mrect.size.width)/2;
 	  vmargin = bounds.size.height + LineBottom-mrect.size.height;
 	  if (numberOfButtons > 0)
 	    {
@@ -1232,13 +1303,13 @@ NSRunInformationalAlertPanel(
 
 
 void
-NSReleaseAlertPanel(id alertPanel)
+NSReleaseAlertPanel(id panel)
 {
-  if ((alertPanel != standardAlertPanel)
-     && (alertPanel != informationalAlertPanel)
-     && (alertPanel != criticalAlertPanel))
+  if ((panel != standardAlertPanel)
+     && (panel != informationalAlertPanel)
+     && (panel != criticalAlertPanel))
     {
-      RELEASE(alertPanel);
+      RELEASE(panel);
     }
 }
 
