@@ -2,6 +2,8 @@
 
    Copyright (C) 1999 Free Software Foundation, Inc.
    
+   Author: Fred Kiefer <FredKiefer@gmx.de>
+   Date: Jul 2003
    Author:  Michael Hanni <mhanni@sprintmail.com>
    Date: 1999
 
@@ -23,17 +25,15 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <gnustep/gui/config.h>  
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSMatrix.h>
-#include <AppKit/NSMenu.h>
-#include <AppKit/NSPopUpButton.h>
-#include <AppKit/NSPopUpButtonCell.h>
-#include <AppKit/PSOperators.h>
+#include "config.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSMenu.h"
+#include "AppKit/NSMenuView.h"
+#include "AppKit/NSPopUpButton.h"
+#include "AppKit/NSPopUpButtonCell.h"
+#include "AppKit/NSWindow.h"
 
 /* The image to use in a specific popupbutton is
  * _pbc_image[_pbcFlags.pullsDown]; that is, _pbc_image[0] if it is a
@@ -45,12 +45,17 @@ static NSImage *_pbc_image[2];
 {
   if (self == [NSPopUpButtonCell class])
     {
-      [self setVersion: 1];
+      [self setVersion: 2];
       ASSIGN(_pbc_image[0], [NSImage imageNamed: @"common_Nibble"]);
       ASSIGN(_pbc_image[1], [NSImage imageNamed: @"common_3DArrowDown"]);
     }
 }
 
+- (id) init
+{
+  return [self initTextCell: @"" pullsDown: NO];
+}
+
 - (id) initTextCell: (NSString *)stringValue
 {
   return [self initTextCell: stringValue pullsDown: NO];
@@ -59,28 +64,63 @@ static NSImage *_pbc_image[2];
 - (id) initTextCell: (NSString *)stringValue
 	  pullsDown: (BOOL)pullDown
 {
+  NSMenu *menu;
+
   [super initTextCell: stringValue];
 
-  _pbcFlags.pullsDown = pullDown;
+  menu = [[NSMenu alloc] initWithTitle: @""];
+  [self setMenu: menu];
+  RELEASE(menu);
+
+  [self setPullsDown: pullDown];
   _pbcFlags.usesItemFromMenu = YES;
-  _pbcFlags.altersStateOfSelectedItem = YES;
 
   if ([stringValue length] > 0)
     {
       [self addItemWithTitle: stringValue]; 
     }
 
-  _menu = [[NSMenu alloc] initWithTitle: @""];
-  [_menu _setOwnedByPopUp: self];
-
   return self;
 }
 
+- (void) dealloc
+{
+  /* 
+   * We don't use methods here to clean up the selected item, the menu
+   * item and the menu as these methods internally update the menu,
+   * which tries to access the target of the menu item (or of this cell). 
+   * When the popup is relases this target may already have been freed, 
+   * so the local reference to it is invalid and will result in a 
+   * segmentation fault. 
+   */
+  _selectedItem = nil;
+  [super dealloc];
+}
+
 - (void) setMenu: (NSMenu *)menu
 {
-  [_menu _setOwnedByPopUp: nil];    
+  if (_menu == menu)
+    {
+      return;
+    }
+
+  if (_menu != nil)
+    {
+      [_menu _setOwnedByPopUp: nil];
+    }
   ASSIGN(_menu, menu);
-  [_menu _setOwnedByPopUp: self];
+  if (_menu != nil)
+    {
+      [_menu _setOwnedByPopUp: self];
+      /* We need to set the menu view so we trigger the special case 
+       * popupbutton code in super class NSMenuItemCell
+       */
+      [self setMenuView: [_menu menuRepresentation]];
+    }
+  else 
+    {
+      [self setMenuView: nil];
+    }
 }
 
 - (NSMenu *) menu
@@ -88,10 +128,29 @@ static NSImage *_pbc_image[2];
   return _menu;
 }
 
-// Behavior settings
+// Behaviour settings
 - (void) setPullsDown: (BOOL)flag
 {
+  NSMenuItem *item = _menuItem;
+
+  [self setMenuItem: nil];
   _pbcFlags.pullsDown = flag;
+  [self setAltersStateOfSelectedItem: !flag];
+
+  if (!flag)
+    {
+      // pop up
+      [self setArrowPosition: NSPopUpArrowAtCenter];
+      [self setPreferredEdge: NSMinYEdge];
+    }
+  else
+    {
+      // pull down
+      [self setArrowPosition: NSPopUpArrowAtBottom];
+      [self setPreferredEdge: NSMaxYEdge];
+    }
+
+  [self setMenuItem: item];
 }
 
 - (BOOL) pullsDown
@@ -112,7 +171,6 @@ static NSImage *_pbc_image[2];
 // The preferred edge is used for pull down menus and for popups under 
 // severe screen position restrictions.  It indicates what edge of the
 // cell the menu should pop out from.
-
 - (void) setPreferredEdge: (NSRectEdge)edge
 {
   _pbcFlags.preferredEdge = edge;
@@ -128,7 +186,6 @@ static NSImage *_pbc_image[2];
 // a pull-down.  If this is NO, then the menu item set with -setMenuItem: 
 // is always displayed.  This can be useful for a popup button that is an
 // icon button that pops up a menu full of textual items, for example.
-
 - (void) setUsesItemFromMenu: (BOOL)flag
 {
   _pbcFlags.usesItemFromMenu = flag;
@@ -139,12 +196,21 @@ static NSImage *_pbc_image[2];
   return _pbcFlags.usesItemFromMenu;
 }
 
-// This only has an effect for popups (it is ignored for pulldowns).  
-// If YES (the default) then the selected item gets its state set to
-// NSOnState.  If NO the items in the menu are left alone.
-
+// If YES (the default for popups) then the selected item gets its state
+// set to NSOnState. If NO the items in the menu are left alone.
 - (void) setAltersStateOfSelectedItem: (BOOL)flag
 {
+  id <NSMenuItem> selectedItem = [self selectedItem];
+
+  if (flag)
+    {
+      [selectedItem setState: NSOffState];
+    }
+  else
+    {
+      [selectedItem setState: NSOnState];
+    }
+
   _pbcFlags.altersStateOfSelectedItem = flag;
 }
 
@@ -173,8 +239,17 @@ static NSImage *_pbc_image[2];
 
 - (void) insertItemWithTitle: (NSString *)title atIndex: (int)index
 {
-  NSMenuItem *anItem;
-  int count = [_menu numberOfItems];
+  id <NSMenuItem> anItem;
+  int i, count;
+  
+  i = [self indexOfItemWithTitle: title];
+
+  if (-1 != i)
+    {
+      [self removeItemAtIndex: i];
+    }
+
+  count = [_menu numberOfItems];
 
   if (index < 0)
     index = 0;
@@ -190,20 +265,31 @@ static NSImage *_pbc_image[2];
      it to appear on the screen.  */
   [anItem setOnStateImage: nil];
   [anItem setMixedStateImage: nil];
+
+  // FIXME: The documentation is unclear what to set here.
+  //[anItem setAction: [self action]];
+  //[anItem setTarget: [self target]];
 }
 
 - (void) removeItemWithTitle: (NSString *)title
 {
-  [_menu removeItemAtIndex: [_menu indexOfItemWithTitle: title]];
+  [self removeItemAtIndex: [self indexOfItemWithTitle: title]];
 }
 
 - (void) removeItemAtIndex: (int)index
 {
+  if (index == [self indexOfSelectedItem])
+    {
+      [self selectItem: nil];
+    }
+
   [_menu removeItemAtIndex: index];
 }
 
 - (void) removeAllItems
 {
+  [self selectItem: nil];
+
   while ([_menu numberOfItems] > 0)
     {
       [_menu removeItemAtIndex: 0];
@@ -248,7 +334,14 @@ static NSImage *_pbc_image[2];
 
 - (id <NSMenuItem>) itemAtIndex: (int)index
 {
-  return [_menu itemAtIndex: index];
+  if ((index >= 0) && (index < [_menu numberOfItems]))
+    {
+      return [_menu itemAtIndex: index];
+    }
+  else 
+    {
+      return nil;
+    }
 }
 
 - (id <NSMenuItem>) itemWithTitle: (NSString *)title
@@ -265,6 +358,61 @@ static NSImage *_pbc_image[2];
   return [_menu itemAtIndex: end];
 }
 
+- (id <NSCopying>) objectValue
+{
+  return [NSNumber numberWithInt: [self indexOfSelectedItem]];
+}
+
+- (void) setObjectValue: (id)object
+{
+  if ([object respondsTo: @selector(intValue)])
+    {
+      int i = [object intValue];
+	
+      [self selectItemAtIndex: i];
+    }
+}
+
+- (void) setImage: (NSImage *)anImage
+{
+  // Do nothing as the image is determined by the current item
+}
+
+- (void) setMenuItem: (NSMenuItem *)item
+{
+  NSImage *image;
+
+  if (_menuItem == item)
+    return;
+
+  if (_pbcFlags.arrowPosition == NSPopUpArrowAtBottom)
+    {
+      image = _pbc_image[1];
+    }
+  else if (_pbcFlags.arrowPosition == NSPopUpArrowAtCenter)
+    {
+      image = _pbc_image[0];
+    }
+  else
+    {
+      // No image for NSPopUpNoArrow
+      image = nil;
+    }
+
+  if ([_menuItem image] == image)
+    {
+      [_menuItem setImage: nil];
+    }
+
+  //[super setMenuItem: item];
+  ASSIGN (_menuItem, item);
+
+  if ([_menuItem image] == nil)
+    {
+      [_menuItem setImage: image];
+    }
+}
+
 // Dealing with selection
 - (void) selectItem: (id <NSMenuItem>)item
 {
@@ -277,7 +425,6 @@ static NSImage *_pbc_image[2];
 	{
 	  [_selectedItem setState: NSOffState];
 	}
-      [_selectedItem setImage: nil];
     }
 
   _selectedItem = item;
@@ -288,7 +435,6 @@ static NSImage *_pbc_image[2];
         {
 	  [_selectedItem setState: NSOnState];
         }
-      [_selectedItem setImage: _pbc_image[_pbcFlags.pullsDown]];
     }
 
   /* Set the item in the menu */
@@ -298,7 +444,7 @@ static NSImage *_pbc_image[2];
 
 - (void) selectItemAtIndex: (int)index
 {
-  NSMenuItem	*anItem;
+  id <NSMenuItem> anItem;
 
   if (index < 0) 
     anItem = nil;
@@ -310,14 +456,14 @@ static NSImage *_pbc_image[2];
 
 - (void) selectItemWithTitle: (NSString *)title
 {
-  NSMenuItem	*anItem = [self itemWithTitle: title];
+  id <NSMenuItem> anItem = [self itemWithTitle: title];
 
   [self selectItem: anItem];
 }
 
 - (void) setTitle: (NSString *)aString
 {
-  NSMenuItem	*anItem;
+  id <NSMenuItem> anItem;
 
   if (_pbcFlags.pullsDown)
     {
@@ -354,29 +500,45 @@ static NSImage *_pbc_image[2];
 
 - (int) indexOfSelectedItem
 {
-  return [_menu indexOfItem: _selectedItem];
+  return [_menu indexOfItem: [self selectedItem]];
 }
 
 - (void) synchronizeTitleAndSelectedItem
 {
+  int index;
+
   if (!_pbcFlags.usesItemFromMenu)
     return;
 
   if ([_menu numberOfItems] == 0)
     {
-      _selectedItem = nil;
+      index = -1;
     }
   else if (_pbcFlags.pullsDown)
     {
-      [self selectItem: [self itemAtIndex: 0]];
+      index = 0;
     }
   else
     {
-      int index = [[_menu menuRepresentation] highlightedItemIndex];
-      
+      index = [[_menu menuRepresentation] highlightedItemIndex];
+
+      if (index < 0)
+	index = [self indexOfSelectedItem];
       if (index < 0)
 	index = 0;
-      [self selectItemAtIndex: index];
+    }
+
+  if ((index >= 0)  && ([_menu numberOfItems] > index))
+    {
+      NSMenuItem *anItem;
+
+      // This conversion is needed as [setMenuItem:] expects an NSMenuItem
+      anItem = (NSMenuItem *)[_menu itemAtIndex: index];
+      [self setMenuItem: anItem];
+    }
+  else
+    {
+      [self setMenuItem: nil];
     }
 }
 
@@ -403,8 +565,10 @@ static NSImage *_pbc_image[2];
 
 - (NSString *) titleOfSelectedItem
 {
-  if (_selectedItem != nil)
-    return [_selectedItem title];
+  id <NSMenuItem> item = [self selectedItem];
+
+  if (item != nil)
+    return [item title];
   else
     return @"";
 }
@@ -440,10 +604,24 @@ static NSImage *_pbc_image[2];
   
   // Last, display the window
   [[mr window] orderFrontRegardless];
+
+  /* This covers an obscure case where the user subsequently
+     uses the mouse to select an item. We'll never know
+     this was done (and thus cannot dismiss the popUp) without
+     getting this notification */
+  [nc addObserver: controlView
+      selector: @selector(_handleNotification:)
+      name: NSMenuDidSendActionNotification
+      object: _menu];
 }
 
 - (void) dismissPopUp
 {
+  NSNotificationCenter	*nc = [NSNotificationCenter defaultCenter];
+
+  [nc removeObserver: [self controlView]
+      name: NSMenuDidSendActionNotification
+      object: _menu];
   [_menu close];
 }
 
@@ -455,16 +633,48 @@ static NSImage *_pbc_image[2];
   return NO;
 }
 
+// This method is called to simulate programmatically a click
+// [as NSCell's performClick:]
+// This method is not executed upon mouse down; rather, it should
+// simulate what would happen upon mouse down.  It should not start
+// any real mouse tracking.
 - (void) performClickWithFrame: (NSRect)frame
 			inView: (NSView *)controlView
 {
-  // TODO
+  NSWindow *cvWin = [controlView window];
 
-  // This method is called to simulate programmatically a click
-  // [as NSCell's performClick:]
-  // This method is not executed upon mouse down; rather, it should
-  // simulate what would happen upon mouse down.  It should not start
-  // any real mouse tracking.
+  if(_cell.is_disabled == YES)
+    {
+      return;
+    }
+
+  [controlView lockFocus];
+      
+  [self setNextState];
+  [self highlight: YES withFrame: frame inView: controlView];
+  [cvWin flushWindow];
+      
+  // Wait approx 1/10 seconds
+  [[NSRunLoop currentRunLoop] 
+      runUntilDate: [NSDate dateWithTimeIntervalSinceNow: 0.1]];
+      
+  [self highlight: NO withFrame: frame inView: controlView];
+  [cvWin flushWindow];
+      
+  [controlView unlockFocus];
+
+  [self attachPopUpWithFrame: frame inView: controlView];
+}
+
+- (void) performClick: (id)sender
+{
+  NSView *cv = [self controlView];
+
+  if (cv != nil)
+    {  
+      [self performClickWithFrame: [cv bounds]
+	    inView: cv];
+    }
 }
 
 // Arrow position for bezel style and borderless popups.
@@ -473,75 +683,51 @@ static NSImage *_pbc_image[2];
   return _pbcFlags.arrowPosition;
 }
 
-/*
- * Does nothing for now.
- */
 - (void) setArrowPosition: (NSPopUpArrowPosition)position
 {
   _pbcFlags.arrowPosition = position;
 }
 
 /*
- * What would be nice and natural is to make this drawing using the same code 
- * that is used to draw cells in the menu.
- * This looks like a mess to do in this framework.
- *
- * Well, here is an attempt to make this work.
- */
-- (void) drawWithFrame: (NSRect)cellFrame inView: (NSView*)controlView
-{
-  NSMenuItemCell    *aCell;
-
-  // Save last view drawn to
-  if (_control_view != controlView)
-    _control_view = controlView;
-
-  // Transparent buttons never draw 
-  if (_buttoncell_is_transparent)
-    return;
-
-  // Do nothing if cell's frame rect is zero
-  if (NSIsEmptyRect(cellFrame))
-    return;
-
-  // Do nothing if the window is deferred
-  if ([[controlView window] gState] == 0)
-    return;
-
-  /* Get the NSMenuItemCell of the selected item */
-  aCell = [[_menu menuRepresentation] menuItemCellForItemAtIndex: [self indexOfSelectedItem]];
-
-  /* Turn off highlighting so the NSPopUpButton looks right */
-  [aCell setHighlighted: NO];
-  
-  [aCell drawWithFrame: cellFrame inView: controlView];
-
-  /* Draw our own interior so we pick up our dotted frame */
-  [self drawInteriorWithFrame: cellFrame inView: controlView];
-
-  /* Rehighlight item for consistency */
-  [aCell setHighlighted: YES];
-}
-
-/* FIXME: This needs to be removed in favor of allowing the cell to draw 
- * our NSDottedRect.
+ * This drawing uses the same code that is used to draw cells in the menu.
  */
 - (void) drawInteriorWithFrame: (NSRect)cellFrame
-			inView: (NSView*)view
+			inView: (NSView*)controlView
 {
-  // Transparent buttons never draw
-  if (_buttoncell_is_transparent)
-    return;
+  BOOL new = NO;
 
-  [view lockFocus];
+  if ([self menuItem] == nil)
+    {
+      NSMenuItem *anItem;
 
-  cellFrame = [self drawingRectForBounds: cellFrame];
+      /* 
+       * Create a temporary NSMenuItemCell to at least draw our control,
+       * if items array is empty.
+       */
+      anItem = [NSMenuItem new];
+      [anItem setTitle: [self title]];
+      /* We need this menu item because NSMenuItemCell gets its contents 
+       * from the menuItem not from what is set in the cell */
+      [self setMenuItem: anItem];
+      RELEASE(anItem);
+      new = YES;
+    }      
 
-  if (_cell.shows_first_responder
-      && [[view window] firstResponder] == view)
-    NSDottedFrameRect(cellFrame);
+  /* We need to calc our size to get images placed correctly */
+  [self calcSize];
+  [super drawInteriorWithFrame: cellFrame inView: controlView];
 
-  [view unlockFocus]; 
+  if (_cell.shows_first_responder)
+    {
+      cellFrame = [self drawingRectForBounds: cellFrame];
+      NSDottedFrameRect(cellFrame);
+    }
+
+  /* Unset the item to restore balance if a new was created */
+  if (new)
+    {
+      [self setMenuItem: nil];
+    }
 }
 
 /* FIXME: this method needs to be rewritten to be something like 
@@ -611,7 +797,7 @@ static NSImage *_pbc_image[2];
   [super encodeWithCoder: aCoder];
 
   [aCoder encodeObject: _menu];
-  [aCoder encodeConditionalObject: _selectedItem];
+  [aCoder encodeConditionalObject: [self selectedItem]];
   flag = _pbcFlags.pullsDown;
   [aCoder encodeValueOfObjCType: @encode(int) at: &flag];
   flag = _pbcFlags.preferredEdge;
@@ -626,23 +812,90 @@ static NSImage *_pbc_image[2];
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  int flag;
+  NSMenu *menu;
 
   self = [super initWithCoder: aDecoder];
-  _menu = [aDecoder decodeObject];
-  _selectedItem = [aDecoder decodeObject];
-  [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
-  _pbcFlags.pullsDown = flag;
-  [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
-  _pbcFlags.preferredEdge = flag;
-  [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
-  _pbcFlags.usesItemFromMenu = flag;
-  [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
-  _pbcFlags.altersStateOfSelectedItem = flag;
-  [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
-  _pbcFlags.arrowPosition = flag;
 
-  [_menu _setOwnedByPopUp: self];
+  if ([aDecoder allowsKeyedCoding])
+    {
+      if ([aDecoder containsValueForKey: @"NSAltersState"])
+        {
+	  BOOL alters = [aDecoder decodeBoolForKey: @"NSAltersState"];
+	  
+	  [self setAltersStateOfSelectedItem: alters];
+	}
+      if ([aDecoder containsValueForKey: @"NSUsesItemFromMenu"])
+        {
+	  BOOL usesItem = [aDecoder decodeBoolForKey: @"NSUsesItemFromMenu"];
+	  
+	  [self setUsesItemFromMenu: usesItem];
+	}
+      if ([aDecoder containsValueForKey: @"NSArrowPosition"])
+        {
+	  NSPopUpArrowPosition position = [aDecoder decodeIntForKey: 
+							@"NSArrowPosition"];
+	  
+	  [self setArrowPosition: position];
+	}
+      if ([aDecoder containsValueForKey: @"NSPreferredEdge"])
+        {
+	  NSRectEdge edge = [aDecoder decodeIntForKey: @"NSPreferredEdge"];
+	  
+	  [self setPreferredEdge: edge];
+	}
+
+      menu = [aDecoder decodeObjectForKey: @"NSMenu"];
+      [self setMenu: menu];
+    }
+  else
+    {
+      int flag;
+      id<NSMenuItem> selectedItem;
+      int version = [aDecoder versionForClassName: 
+				  @"NSPopUpButtonCell"];
+
+      menu = [aDecoder decodeObject];
+      /* 
+	 FIXME: This same ivar already gets set in NSCell initWithCoder, 
+	 but there it is used directly not via a method call. So here we first 
+	 unset it and than set it again as our setMenu: method tries to optimize 
+	 duplicate calls.
+      */
+      [self setMenu: nil];
+      [self setMenu: menu];
+      selectedItem = [aDecoder decodeObject];
+      [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
+      _pbcFlags.pullsDown = flag;
+      [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
+      _pbcFlags.preferredEdge = flag;
+      [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
+      _pbcFlags.usesItemFromMenu = flag;
+      [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
+      _pbcFlags.altersStateOfSelectedItem = flag;
+      [aDecoder decodeValueOfObjCType: @encode(int) at: &flag];
+      _pbcFlags.arrowPosition = flag;
+      
+      if (version < 2)
+        {
+	  int i;
+	  
+	  // Not the stored format did change but the interpretation of it.
+	  // in version 1 most of the ivars were not used, so their values may
+	  // be arbitray. We overwrite them with valid settings.
+	  [self setPullsDown: _pbcFlags.pullsDown];
+	  _pbcFlags.usesItemFromMenu = YES;
+	  
+	  for (i = 0; i < [_menu numberOfItems]; i++)
+	    {
+	      id <NSMenuItem> anItem = [menu itemAtIndex: i];
+	      
+	      [anItem setOnStateImage: nil];
+	      [anItem setMixedStateImage: nil];
+	    }
+	  [self setEnabled: YES];
+	}
+      [self selectItem: selectedItem];
+    }
   return self;
 }
 
