@@ -30,24 +30,28 @@
 */ 
 
 #include <math.h>
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSString.h>
 #include <Foundation/NSDebug.h>
 #include <Foundation/NSData.h>
-#include <Foundation/NSException.h>
-#include <Foundation/NSThread.h>
 #include <Foundation/NSFileManager.h>
+#include <Foundation/NSException.h>
+#include <Foundation/NSPathUtilities.h>
+#include <Foundation/NSTask.h>
+#include <Foundation/NSThread.h>
+#include <Foundation/NSUserDefaults.h>
 #include <Foundation/NSValue.h>
-#include <AppKit/AppKitExceptions.h>
-#include <AppKit/NSAffineTransform.h>
-#include <AppKit/NSGraphicsContext.h>
-#include <AppKit/NSView.h>
-#include <AppKit/NSPrintPanel.h>
-#include <AppKit/NSPrintInfo.h>
-#include <AppKit/NSPrintOperation.h>
-#include <AppKit/PSOperators.h>
-
-#include <gnustep/base/GSLocale.h>
+#include "AppKit/AppKitExceptions.h"
+#include "AppKit/NSAffineTransform.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSGraphicsContext.h"
+#include "AppKit/NSView.h"
+#include "AppKit/NSPrinter.h"
+#include "AppKit/NSPrintPanel.h"
+#include "AppKit/NSPrintInfo.h"
+#include "AppKit/NSPrintOperation.h"
+#include "AppKit/NSWorkspace.h"
+#include "AppKit/PSOperators.h"
 
 #define NSNUMBER(a) [NSNumber numberWithInt: (a)]
 #define NSFNUMBER(a) [NSNumber numberWithFloat: (a)]
@@ -68,6 +72,7 @@ typedef struct _page_info_t {
   int    nup;                /* Number up pages to print on a sheet */
   double lastWidth, lastHeight;
   NSPrintingOrientation orient;
+  int    pageDirection;      /* NSPrintPageDirection */
 } page_info_t;
 
 @interface NSPrintOperation (Private)
@@ -81,6 +86,11 @@ typedef struct _page_info_t {
 
 @end
 
+@interface NSPrintPanel (Private)
+- (void) _setStatusStringValue: (NSString *)string;
+@end
+
+
 @interface NSView (NSPrintOperation)
 - (void) _displayPageInRect: (NSRect)pageRect
 	        atPlacement: (NSPoint)location
@@ -88,6 +98,10 @@ typedef struct _page_info_t {
 - (void) _endSheet;
 @end
 
+@interface NSView (NPrintOperationPrivate)
+- (void) _cleanupPrinting;
+@end
+
 // Subclass for the regular printing
 @interface GSPrintOperation: NSPrintOperation
 {
@@ -399,7 +413,6 @@ static NSString *NSPrintOperationThreadKey = @"NSPrintOperationThreadKey";
 */
 - (void)destroyContext
 {
-  [_context destroyContext];
   DESTROY(_context);
 }
 
@@ -431,11 +444,12 @@ static NSString *NSPrintOperationThreadKey = @"NSPrintOperationThreadKey";
 //
 // Running a Print Operation
 //
-/** Called by the print operation and it has finished running a printing
+/** Called by the print operation after it has finished running a printing
     operation.
 */
 - (void)cleanUpOperation
 {
+  [[self printPanel] orderOut: self];
   _currentPage = 0;
   [NSPrintOperation setCurrentOperation: nil];
 }
@@ -450,10 +464,13 @@ static NSString *NSPrintOperationThreadKey = @"NSPrintOperationThreadKey";
   return NO;
 }
 
-/* Private method to run the printing operation */
+/* Private method to run the printing operation. Needs to create an
+   autoreleaes pool to make sure the print context is destroyed before
+   returning (which closes the print file.) */
 - (BOOL) _runOperation
 {
   BOOL result;
+  CREATE_AUTORELEASE_POOL(pool);
   NSGraphicsContext *oldContext = [NSGraphicsContext currentContext];
 
   [self createContext];
@@ -479,12 +496,14 @@ static NSString *NSPrintOperationThreadKey = @"NSPrintOperationThreadKey";
     }
   NS_HANDLER
     {
+      [_view _cleanupPrinting];
       [NSGraphicsContext setCurrentContext: oldContext];
       NSRunAlertPanel(@"Error", @"Printing error: %@", 
 		      @"OK", NULL, NULL, localException);
     }
   NS_ENDHANDLER
   [self destroyContext];
+  RELEASE(pool);
   return result;
 }
 
@@ -546,6 +565,8 @@ static NSString *NSPrintOperationThreadKey = @"NSPrintOperationThreadKey";
   id delegate;
   SEL *didRunSelector;
   NSMutableDictionary *dict;
+  void (*didRun)(id, SEL, BOOL, id);
+
   if (success == YES)
     {
       NSPrintPanel *panel = [self printPanel];
@@ -558,9 +579,9 @@ static NSString *NSPrintOperationThreadKey = @"NSPrintOperationThreadKey";
   dict = [_printInfo dictionary];
   didRunSelector = [[dict objectForKey: @"GSModalRunSelector"] pointerValue];
   delegate = [dict objectForKey: @"GSModalRunDelegate"];
-  [delegate performSelector: *didRunSelector
-                   withObject: success
-                   withObject: contextInfo];
+  didRun = (void (*)(id, SEL, BOOL, id))[delegate methodForSelector: 
+						    *didRunSelector];
+  didRun (delegate, *didRunSelector, success, contextInfo);
 }
 
 /** Run a print operation modally with respect to a window.
@@ -646,7 +667,11 @@ static NSString *NSPrintOperationThreadKey = @"NSPrintOperationThreadKey";
   _showPanels = NO;
   [self setPrintInfo: aPrintInfo];
 
-  ASSIGN(_path, @"/tmp/NSTempPrintFile");
+  _path = [NSTemporaryDirectory() stringByAppendingPathComponent: @"GSPrint-"];
+  _path = [_path stringByAppendingString: 
+		   [[NSProcessInfo processInfo] globallyUniqueString]];
+  _path = [_path stringByAppendingPathExtension: @"ps"];
+  RETAIN(_path);
   _pathSet = NO;
   _currentPage = 0;
 
@@ -767,19 +792,36 @@ scaleRect(NSRect rect, double scale)
     {
       info->nupScale = 2.0 / (float)info->nup;
     }
+
+  if ([[dict objectForKey: NSPrintPageDirection] isEqual: @"Columns"])
+    info->pageDirection = 1;
+  else
+    info->pageDirection = 0;
 }
 
 /* Our personnel method to calculate the print rect for the specified page.
    Note, we assume this function is called in order from our first to last
    page. The returned pageRect is in the view's coordinate system
 */
-- (NSRect) _rectForPage: (int)page info: (page_info_t *)info
+- (NSRect) _rectForPage: (int)page info: (page_info_t *)info 
+		  xpage: (int *)xptr
+		  ypage: (int *)yptr
 {
   int xpage, ypage;
   NSRect pageRect;
 
-  xpage = (page - 1) % info->xpages;
-  ypage = (page - 1) % info->ypages;
+  if (info->pageDirection == 1)
+    {
+      xpage = (page - 1) / info->ypages;
+      ypage = (page - 1) % info->ypages;
+    }
+  else
+    {
+      xpage = (page - 1) % info->xpages;
+      ypage = (page - 1) / info->xpages;
+    }
+  *xptr = xpage;
+  *yptr = ypage;
   if (xpage == 0)
     info->lastWidth = 0;
   if (ypage == 0)
@@ -799,7 +841,7 @@ scaleRect(NSRect rect, double scale)
 			last: (int)last 
 			info: (page_info_t *)info
 {
-  int i;
+  int i, xpage, ypage;
   double hlimit, wlimit;
   NSRect pageRect;
   hlimit = [_view heightAdjustLimit];
@@ -807,7 +849,7 @@ scaleRect(NSRect rect, double scale)
   for (i = first; i <= last; i++)
     {
       float newVal, limitVal;
-      pageRect = [self _rectForPage: i info: info];
+      pageRect = [self _rectForPage: i info: info xpage: &xpage ypage: &ypage];
       limitVal = NSMaxY(pageRect) - hlimit * NSHeight(pageRect);
       [_view adjustPageHeightNew: &newVal
 	                     top: NSMinY(pageRect)
@@ -822,8 +864,10 @@ scaleRect(NSRect rect, double scale)
 	                   limit: limitVal];
       if (newVal < NSMaxX(pageRect))
 	pageRect.size.width = MAX(newVal, limitVal) - NSMinX(pageRect);
-      info->lastWidth = NSMaxX(pageRect)*(info->pageScale*info->printScale);
-      info->lastHeight = NSMaxY(pageRect)*(info->pageScale*info->printScale);
+      if (info->pageDirection == 0 || ypage == info->ypages - 1)
+	info->lastWidth = NSMaxX(pageRect)*(info->pageScale*info->printScale);
+      if (info->pageDirection == 1 || xpage == info->xpages - 1)
+	info->lastHeight = NSMaxY(pageRect)*(info->pageScale*info->printScale);
     }
   return pageRect;
 }
@@ -833,7 +877,6 @@ scaleRect(NSRect rect, double scale)
   int i, dir;
   BOOL knowsPageRange, allPages;
   NSRange viewPageRange;
-  NSString *clocale;
   NSMutableDictionary *dict;
   page_info_t info;
   
@@ -858,10 +901,10 @@ scaleRect(NSRect rect, double scale)
     {
       info.first = [[dict objectForKey: NSPrintFirstPage] intValue];
       info.last  = [[dict objectForKey: NSPrintLastPage] intValue];
-      info.first = MAX(info.first, viewPageRange.location);
-      info.first = MIN(info.first, NSMaxRange(viewPageRange) - 1);
+      info.first = MAX(info.first, (int)viewPageRange.location);
+      info.first = MIN(info.first, (int)(NSMaxRange(viewPageRange) - 1));
       info.last = MAX(info.last, info.first);
-      info.last = MIN(info.last, NSMaxRange(viewPageRange) - 1);
+      info.last = MIN(info.last, (int)(NSMaxRange(viewPageRange) - 1));
       viewPageRange = NSMakeRange(info.first, (info.last-info.first)+1);
     }
   info.lastWidth = info.lastHeight = 0;
@@ -894,12 +937,6 @@ scaleRect(NSRect rect, double scale)
 		                   info: &info];
     }
 
-  /* Reset the current locale to a generic C locale so numbers
-     get printed correctly for PostScript (maybe we should only
-     set the numeric locale?). Save the current locale for later. */
-  clocale = GSSetLocale(nil);
-  GSSetLocale(@"C");
-
   /* Print the header information */
   [_view beginDocument];
 
@@ -924,8 +961,9 @@ scaleRect(NSRect rect, double scale)
 			                  last: _currentPage 
 			                  info: &info];
 	}
-        NSDebugLLog(@"NSPrinting", @" current page %d, rect %@", 
-		    _currentPage, NSStringFromRect(pageRect));
+
+      NSDebugLLog(@"NSPrinting", @" current page %d, rect %@", 
+		  _currentPage, NSStringFromRect(pageRect));
       if (NSIsEmptyRect(pageRect))
 	break;
 
@@ -965,7 +1003,6 @@ scaleRect(NSRect rect, double scale)
       [_view _endSheet];
     }
   [_view endDocument];
-  GSSetLocale(clocale);
 
   /* Setup/reset for next time */
   [dict setObject: NSNUMBER(info.last) forKey: NSPrintLastPage];
@@ -1004,7 +1041,6 @@ scaleRect(NSRect rect, double scale)
 	    label: label
 	    bBox: info.sheetBounds
 	    fonts: nil];
-      DPSPrintf(ctxt, "/__GSsheetsaveobject save def\n");
       if (info.orient == NSLandscapeOrientation)
 	{
 	  DPSrotate(ctxt, 90);
@@ -1021,20 +1057,19 @@ scaleRect(NSRect rect, double scale)
     DPSscale(ctxt, scale, scale);
   if ([self isFlipped])
     {
-      NSAffineTransformStruct	ats = { 1, 0, 0, -1, 0, 1 };
+      NSAffineTransformStruct	ats = { 1, 0, 0, -1, 0, NSHeight(_bounds) };
       NSAffineTransform *matrix, *flip;
       flip = [NSAffineTransform new];
       matrix = [NSAffineTransform new];
       [matrix makeIdentityMatrix];
-      [matrix appendTransform: _boundsMatrix];
+      [matrix prependTransform: _boundsMatrix];
       /*
        * The flipping process must result in a coordinate system that
        * exactly overlays the original.	 To do that, we must translate
        * the origin by the height of the view.
        */
       [flip setTransformStruct: ats];
-      flip->matrix.ty = NSHeight(_bounds);
-      [matrix appendTransform: flip];
+      [matrix prependTransform: flip];
       [matrix concat];
       yoffset = NSHeight(_frame) - NSMaxY(pageRect);
     }
@@ -1071,7 +1106,6 @@ scaleRect(NSRect rect, double scale)
   NSPrintOperation *printOp = [NSPrintOperation currentOperation];
   if ([printOp isEPSOperation] == NO)
     DPSPrintf(ctxt, "showpage\n");
-  DPSPrintf(ctxt, "__GSsheetsaveobject restore\n");
   DPSPrintf(ctxt, "%%%%PageTrailer\n");
   DPSPrintf(ctxt, "\n");
 }
@@ -1119,16 +1153,88 @@ scaleRect(NSRect rect, double scale)
   return _context;
 }
 
-- (BOOL)deliverResult
+- (BOOL) _deliverSpooledResult
 {
-  // FIXME
+  int copies;
+  NSDictionary *dict;
+  NSTask *task;
+  NSString *name, *status;
+  NSMutableArray *args;
+  name = [[_printInfo printer] name];
+  status = [NSString stringWithFormat: @"Spooling to printer %@.", name];
+  [_printPanel _setStatusStringValue: status];
 
-/*
+  dict = [_printInfo dictionary];
+  args = [NSMutableArray array];
+  copies = [[dict objectForKey: NSPrintCopies] intValue];
+  if (copies > 1)
+    [args addObject: [NSString stringWithFormat: @"-#%0d", copies]];
+  if ([name isEqual: @"Unknown"] == NO)
+    {
+      [args addObject: @"-P"];
+      [args addObject: name];
+    }
+  [args addObject: _path];
+
+  task = [NSTask new];
+  [task setLaunchPath: @"lpr"];
+  [task setArguments: args];
+  [task launch];
+  [task waitUntilExit];
+  AUTORELEASE(task);
+  return YES;
+}
+
+- (BOOL) deliverResult
+{
+  BOOL success;
+  NSString *job;
+  
+  success = YES;
+  job = [_printInfo jobDisposition];
+  if ([job isEqual: NSPrintPreviewJob])
+    {
+      /* Check to see if there is a GNUstep app that can preview PS files.
+	 It's not likely at this point, so also check for a standards
+	 previewer, like gv.
+      */
+      NSTask *task;
+      NSString *preview;
+      NSWorkspace *ws = [NSWorkspace sharedWorkspace];
+      [_printPanel _setStatusStringValue: @"Opening in previewer..."];
+      preview = [ws getBestAppInRole: @"Viewer" forExtension: @"ps"];
+      if (preview)
+	{
+	  [ws openFile: _path withApplication: preview];
+	}
+      else
+	{
+	  NSUserDefaults *def = [NSUserDefaults standardUserDefaults];
+	  preview = [def objectForKey: @"NSPreviewApp"];
+	  if (preview == nil || [preview length] == 0)
+	    preview = @"gv";
+	  task = [NSTask new];
+	  [task setLaunchPath: preview];
+	  [task setArguments: [NSArray arrayWithObject: _path]];
+	  [task launch];
+	  AUTORELEASE(task);
+	}
+    }
+  else if ([job isEqual: NSPrintSpoolJob])
+    {
+      success = [self _deliverSpooledResult];
+    }
+  else if ([job isEqual: NSPrintFaxJob])
+    {
+    }
+
+  /* We can't remove the temp file because the previewer might still be
+     using it, perhaps the printer is also?
   if (!_pathSet)
     [[NSFileManager defaultManager] removeFileAtPath: _path
 				    handler: nil];
-*/    
-  return YES;
+  */
+  return success;
 }
 
 @end
@@ -1168,13 +1274,6 @@ scaleRect(NSRect rect, double scale)
 
 - (void) _print
 {
-  NSString *clocale;
-  /* Reset the current locale to a generic C locale so numbers
-     get printed correctly for PostScript (maybe we should only
-     set the numeric locale?). Save the current locale for later. */
-  clocale = GSSetLocale(nil);
-  GSSetLocale(@"C");
-  
   /* Save this for the view to look at. Seems like there should
      be a better way to pass it to beginDocument */
   [[_printInfo dictionary] setObject: [NSValue valueWithRect: _rect]
@@ -1183,8 +1282,6 @@ scaleRect(NSRect rect, double scale)
   [_view beginPageInRect: _rect atPlacement: NSMakePoint(0,0)];
   [_view displayRectIgnoringOpacity: _rect];
   [_view endDocument];
-
-  GSSetLocale(clocale);
 }
 
 - (BOOL)isEPSOperation
@@ -1208,6 +1305,7 @@ scaleRect(NSRect rect, double scale)
 - (NSGraphicsContext*)createContext
 {
   NSMutableDictionary *info;
+
   if (_context)
     return _context;
 
@@ -1217,7 +1315,6 @@ scaleRect(NSRect rect, double scale)
   [info setObject: NSGraphicsContextPSFormat
 	   forKey: NSGraphicsContextRepresentationFormatAttributeName];
   _context = RETAIN([NSGraphicsContext graphicsContextWithAttributes: info]);
-
   return _context;
 }
 
@@ -1264,16 +1361,7 @@ scaleRect(NSRect rect, double scale)
 
 - (void) _print
 {
-  NSString *clocale;
-  /* Reset the current locale to a generic C locale so numbers
-     get printed correctly for PostScript (maybe we should only
-     set the numeric locale?). Save the current locale for later. */
-  clocale = GSSetLocale(nil);
-  GSSetLocale(@"C");
-
   [_view displayRectIgnoringOpacity: _rect];
-
-  GSSetLocale(clocale);
 }
 
 - (BOOL)deliverResult
