@@ -26,21 +26,22 @@
 
 */
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSArray.h>
 #include <Foundation/NSBundle.h>
 #include <Foundation/NSDictionary.h>
 #include <Foundation/NSValue.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSForm.h>
-#include <AppKit/NSNibLoading.h>
-#include <AppKit/NSPrinter.h>
-#include <AppKit/NSPrintPanel.h>
-#include <AppKit/NSPrintInfo.h>
-#include <AppKit/NSPrintOperation.h>
-#include <AppKit/NSPopUpButton.h>
-#include <AppKit/NSSavePanel.h>
-#include <AppKit/NSView.h>
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSForm.h"
+#include "AppKit/NSNibLoading.h"
+#include "AppKit/NSPrinter.h"
+#include "AppKit/NSPrintPanel.h"
+#include "AppKit/NSPrintInfo.h"
+#include "AppKit/NSPrintOperation.h"
+#include "AppKit/NSPopUpButton.h"
+#include "AppKit/NSSavePanel.h"
+#include "AppKit/NSView.h"
+#include "GSGuiPrivate.h"
 
 static NSPrintPanel *shared_instance;
 
@@ -101,7 +102,7 @@ static NSPrintPanel *shared_instance;
 		     defer: (BOOL)flag
 		    screen: (NSScreen*)aScreen
 {
-  int i;
+  unsigned int i;
   id control;
   NSArray *subviews, *list;
   NSString *panel;
@@ -115,8 +116,8 @@ static NSPrintPanel *shared_instance;
   if (self == nil)
     return nil;
 
-  panel = [NSBundle pathForGNUstepResource: GSPANELNAME ofType: @"gorm"
-		    inDirectory: nil];
+  panel = [GSGuiBundle() pathForResource: GSPANELNAME ofType: @"gorm"
+		      inDirectory: nil];
   if (panel == nil)
     {
       NSRunAlertPanel(@"Error", @"Could not find print panel resource", 
@@ -156,6 +157,15 @@ static NSPrintPanel *shared_instance;
   return self;
 }
 
+- (void) dealloc
+{
+  RELEASE(_accessoryView);
+  RELEASE(_savePath);
+  RELEASE(_optionPanel);
+  
+  [super dealloc];
+}
+
 //
 // Customizing the Panel 
 //
@@ -179,14 +189,16 @@ static NSPrintPanel *shared_instance;
 /** Display the Print panel in a modal loop. Saves any aquired 
    information in the NSPrintInfo object for the current NSPrintOperation. 
    Returns NSCancelButton if the user clicks the Cancel button or 
-   NSOKButton otherwise.
+   NSOKButton otherwise. Unlike other panels, this one does not order
+   itself out after the modal session is finished. You must do that
+   yourself.
 */
 - (int)runModal
 {
   _picked = NSOKButton;
   [NSApp runModalForWindow: self];
   [_optionPanel orderOut: self];
-  [self orderOut: self];
+  /* Don't order ourselves out, let the NSPrintOperation do that */
   return (_picked == NSCancelButton) ? NSCancelButton :  NSOKButton;
 }
 
@@ -240,10 +252,6 @@ static NSPrintPanel *shared_instance;
   else if (tag == NSPPPreviewButton)
     {
       _picked = NSPPPreviewButton;
-      NSRunAlertPanel(@"Sorry", @"Previewing of print file not implemented", 
-		      @"OK", NULL, NULL);
-      /* Don't stop the modal session */
-      return;
     }
   else if (tag ==NSFaxButton )
     {
@@ -271,7 +279,6 @@ static NSPrintPanel *shared_instance;
     }
   else if (tag == NSPPOptionOKButton)
     {
-      NSLog(@"order out options");
       /* Do nothing. Just stops model on the options panel */
     }
   else
@@ -393,8 +400,9 @@ static NSPrintPanel *shared_instance;
       list = [printer stringListForKey:@"Resolution" inTable: @"PPD"];
       if ([list count])
 	{
-	  int i;
+	  unsigned int i;
 	  NSString *display, *option;
+
 	  for (i = 0; i < [list count]; i++)
 	    {
 	      NSString *key = [list objectAtIndex: i];
@@ -430,14 +438,16 @@ static NSPrintPanel *shared_instance;
     {
       NSString *manual;
       NSArray *list;
+
       manual = [printer stringForKey:@"DefaultManualFeed" inTable: @"PPD"];
       if (manual)
 	[control addItemWithTitle: @"Manual"];
       list = [printer stringListForKey:@"InputSlot" inTable: @"PPD"];
       if ([list count])
 	{
-	  int i;
+	  unsigned int i;
 	  NSString *display, *option;
+
 	  for (i = 0; i < [list count]; i++)
 	    {
 	      NSString *paper = [list objectAtIndex: i];
@@ -623,4 +633,9 @@ static NSPrintPanel *shared_instance;
 
 }
 
+/* Private method for NSPrintOperation */
+- (void) _setStatusStringValue: (NSString *)string
+{
+  [CONTROL(self, NSPPStatusField) setStringValue: string ];
+}
 @end
