@@ -23,15 +23,16 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <AppKit/NSProgressIndicator.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSWindow.h>
+#include <Foundation/NSTimer.h>
+#include "AppKit/NSProgressIndicator.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSWindow.h"
 
 @implementation NSProgressIndicator
 
-NSColor *fillColour = nil;
+static NSColor *fillColour = nil;
 #define maxCount 1
-NSImage *images[maxCount];
+static NSImage *images[maxCount];
 
 + (void) initialize
 {
@@ -301,16 +302,35 @@ NSImage *images[maxCount];
 
 - (id)initWithCoder:(NSCoder *)aDecoder
 {
-   self = [super initWithCoder:aDecoder];
-   [aDecoder decodeValueOfObjCType: @encode(BOOL) at:&_isIndeterminate];
-   [aDecoder decodeValueOfObjCType: @encode(BOOL) at:&_isBezeled];
-   [aDecoder decodeValueOfObjCType: @encode(BOOL) at:&_usesThreadedAnimation];
-   [aDecoder decodeValueOfObjCType: @encode(NSTimeInterval)
-	     at:&_animationDelay];
-   [aDecoder decodeValueOfObjCType: @encode(double) at:&_doubleValue];
-   [aDecoder decodeValueOfObjCType: @encode(double) at:&_minValue];
-   [aDecoder decodeValueOfObjCType: @encode(double) at:&_maxValue];
-   [aDecoder decodeValueOfObjCType: @encode(BOOL) at:&_isVertical];
+  self = [super initWithCoder:aDecoder];
+  if ([aDecoder allowsKeyedCoding])
+    {
+      //id *matrix = [aDecoder decodeObjectForKey: @"NSDrawMatrix"];
+
+      if ([aDecoder containsValueForKey: @"NSMaxValue"])
+        {
+	  int max = [aDecoder decodeIntForKey: @"NSMaxValue"];
+
+	  [self setMaxValue: max];
+	}
+      if ([aDecoder containsValueForKey: @"NSpiFlags"])
+        {
+	  //int flags = [aDecoder decodeIntForKey: @"NSpiFlags"];
+	  // FIXME
+	}
+    }
+  else
+    {
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at:&_isIndeterminate];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at:&_isBezeled];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at:&_usesThreadedAnimation];
+      [aDecoder decodeValueOfObjCType: @encode(NSTimeInterval)
+		                   at:&_animationDelay];
+      [aDecoder decodeValueOfObjCType: @encode(double) at:&_doubleValue];
+      [aDecoder decodeValueOfObjCType: @encode(double) at:&_minValue];
+      [aDecoder decodeValueOfObjCType: @encode(double) at:&_maxValue];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at:&_isVertical];
+    }
    return self;
 }
 
