@@ -1,6 +1,6 @@
 /** <title>NSSavePanel</title>
 
-   <abstract>Standard save panel for saving files</abstract>
+   <abstract>Standard panel for saving files</abstract>
 
    Copyright (C) 1999, 2000 Free Software Foundation, Inc.
 
@@ -31,23 +31,30 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSBox.h>
-#include <AppKit/NSBrowser.h>
-#include <AppKit/NSBrowserCell.h>
-#include <AppKit/NSButton.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSForm.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSImageView.h>
-#include <AppKit/NSMatrix.h>
-#include <AppKit/NSSavePanel.h>
-#include <AppKit/NSOpenPanel.h>
-#include <AppKit/NSTextField.h>
-#include <AppKit/NSWorkspace.h>
+#include <Foundation/NSAutoreleasePool.h>
 #include <Foundation/NSException.h>
 #include <Foundation/NSFileManager.h>
+#include <Foundation/NSNotification.h>
 #include <Foundation/NSPathUtilities.h>
+#include <Foundation/NSUserDefaults.h>
+#include <Foundation/NSURL.h>
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSBox.h"
+#include "AppKit/NSBrowser.h"
+#include "AppKit/NSBrowserCell.h"
+#include "AppKit/NSButton.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSForm.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSImageView.h"
+#include "AppKit/NSMatrix.h"
+#include "AppKit/NSPasteboard.h"
+#include "AppKit/NSDragging.h"
+#include "AppKit/NSSavePanel.h"
+#include "AppKit/NSTextField.h"
+#include "AppKit/NSWorkspace.h"
+
+#include "GSGuiPrivate.h"
 
 #define _SAVE_PANEL_X_PAD	5
 #define _SAVE_PANEL_Y_PAD	4
@@ -57,6 +64,13 @@ static NSFileManager *_fm = nil;
 
 static BOOL _gs_display_reading_progress = NO;
 
+//
+// SavePanel filename compare
+//
+@interface NSString (_gsSavePanel)
+- (NSComparisonResult)_gsSavePanelCompare:(NSString *)other;
+@end
+
 //
 // NSSavePanel private methods
 //
@@ -74,9 +88,47 @@ static BOOL _gs_display_reading_progress = NO;
 - (void) _selectCellName: (NSString *)title;
 - (void) _setupForDirectory: (NSString *)path file: (NSString *)name;
 - (BOOL) _shouldShowExtension: (NSString *)extension isDir: (BOOL *)isDir;
+- (void) _windowResized: (NSNotification*)n;
+- (NSComparisonResult) _compareFilename: (NSString *)n1 with: (NSString *)n2;
 @end /* NSSavePanel (PrivateMethods) */
 
 @implementation NSSavePanel (_PrivateMethods)
+
+- (unsigned int) draggingEntered: (id <NSDraggingInfo>)sender
+{
+  NSPasteboard *pb;
+
+  pb = [sender draggingPasteboard];
+  if ([[pb types] indexOfObject: NSFilenamesPboardType] == NSNotFound)
+    {
+      return NSDragOperationNone;
+    }
+  return NSDragOperationAll;
+}
+    
+- (BOOL) performDragOperation: (id<NSDraggingInfo>)sender
+{
+  NSArray	*types;
+  NSPasteboard	*dragPb;
+
+  dragPb = [sender draggingPasteboard];
+  types = [dragPb types];
+  if ([types containsObject: NSFilenamesPboardType] == YES)
+    {
+      NSArray	*names = [dragPb propertyListForType: NSFilenamesPboardType];
+      NSString	*file = [names lastObject];
+
+      [self setDirectory: file];
+      return YES;
+    }
+  return NO;
+}
+
+- (BOOL) prepareForDragOperation: (id<NSDraggingInfo>)sender
+{
+  return YES;
+}
+
 -(id) _initWithoutGModel
 {
   NSBox *bar;
@@ -85,6 +137,12 @@ static BOOL _gs_display_reading_progress = NO;
   NSRect r;
   id lastKeyView;
 
+  // Track window resizing so we can change number of browser columns.
+  [[NSNotificationCenter defaultCenter] addObserver: self
+    selector: @selector(_windowResized:)
+    name: NSWindowDidResizeNotification
+    object: self];
+
   //
   // WARNING: We create the panel sized (308, 317), which is the 
   // minimum size we want it to have.  Then, we resize it at the 
@@ -94,7 +152,9 @@ static BOOL _gs_display_reading_progress = NO;
 	 styleMask: (NSTitledWindowMask | NSResizableWindowMask) 
 	 backing: 2 defer: YES];
   [self setMinSize: [self frame].size];
-  [[self contentView] setBounds: NSMakeRect (0, 0, 308, 317)];
+
+  r = NSMakeRect (0, 0, 308, 317);
+  [[self contentView] setBounds: r];
   
   r = NSMakeRect (0, 64, 308, 245);
   _topView = [[NSView alloc] initWithFrame: r]; 
@@ -122,6 +182,7 @@ static BOOL _gs_display_reading_progress = NO;
   [_browser setTag: NSFileHandlingPanelBrowser];
   [_browser setAction:@selector(_selectText:)];
   [_browser setTarget:self];
+  [_browser setMinColumnWidth: 140];
   [_topView addSubview: _browser];
   [_browser release];
 
@@ -265,6 +326,9 @@ static BOOL _gs_display_reading_progress = NO;
   [self setInitialFirstResponder: _form];
   [super setTitle: @""];
 
+  [self registerForDraggedTypes: [NSArray arrayWithObjects:
+	NSFilenamesPboardType, nil]];
+
   return self;
 }
 
@@ -312,7 +376,7 @@ static BOOL _gs_display_reading_progress = NO;
   NSBrowserCell *selectedCell;
   BOOL           isLeaf;
 
-  if(column == -1)
+  if (column == -1)
     return;
 
   matrix = [_browser matrixInColumn:column];
@@ -328,7 +392,7 @@ static BOOL _gs_display_reading_progress = NO;
     }
   else
     {
-      if([[[_form cellAtIndex: 0] stringValue] length] > 0)
+      if ([[[_form cellAtIndex: 0] stringValue] length] > 0)
 	{
 	  [_okButton setEnabled:YES];
 	  [self _selectCellName:[[_form cellAtIndex: 0] stringValue]];
@@ -351,53 +415,82 @@ static BOOL _gs_display_reading_progress = NO;
   NSArray            *cells;
   NSMatrix           *matrix;
   NSComparisonResult  result;
-  NSRange             range;
   int                 i, titleLength, cellLength, numberOfCells;
 
   matrix = [_browser matrixInColumn:[_browser lastColumn]];
-  if([matrix selectedCell])
+  if ([matrix selectedCell])
     return;
 
   titleLength = [title length];
-  if(!titleLength)
+  if (!titleLength)
     return;
 
-  range.location = 0;
-  range.length = titleLength;
-
   cells = [matrix cells];
   numberOfCells = [cells count];
 
-  for(i = 0; i < numberOfCells; i++)
+  for (i = 0; i < numberOfCells; i++)
     {
       cellString = [[matrix cellAtRow:i column:0] stringValue];
 
       cellLength = [cellString length];
-      if(cellLength != titleLength)
+      if (cellLength != titleLength)
 	continue;
 
-      result = [cellString compare:title options:0 range:range];
+      result = [self _compareFilename:cellString with:title];
 
-      if(result == NSOrderedSame)
+      if (result == NSOrderedSame)
 	{
 	  [matrix selectCellAtRow:i column:0];
 	  [matrix scrollCellToVisibleAtRow:i column:0];
 	  [_okButton setEnabled:YES];
 	  return;
 	}
-      else if(result == NSOrderedDescending)
+      else if (result == NSOrderedDescending)
 	break;
     }
 }
 
+- (BOOL) _browser: (NSBrowser*)sender
+selectCellWithString: (NSString*)title
+	inColumn: (int)column
+{
+  NSMatrix *m;
+  BOOL isLeaf;
+  NSString *path;
+
+  m = [sender matrixInColumn: column];
+  isLeaf = [[m selectedCell] isLeaf];
+  path = [sender pathToColumn: column];
+
+  if (isLeaf)
+    {
+      ASSIGN (_directory, path);
+      ASSIGN (_fullFileName, [path stringByAppendingPathComponent: title]);
+    }
+  else
+    {
+      ASSIGN (_directory, [path stringByAppendingPathComponent: title]);
+      ASSIGN (_fullFileName, nil);
+    }
+
+  [self _selectTextInColumn:column];
+  
+  return YES;
+}
+
 - (void) _setupForDirectory: (NSString *)path file: (NSString *)filename
 {
-  if (path == nil || filename == nil)
-    [NSException raise: NSInvalidArgumentException
-		format: @"NSSavePanel runModalForDirectory:file: "
-		 @"does not accept nil arguments."];
-
-  ASSIGN (_directory, path);
+  if (path == nil)
+    {
+      if (_directory == nil)
+	ASSIGN(_directory, [_fm currentDirectoryPath]);
+    }
+  else
+    {
+      ASSIGN (_directory, path);
+    }
+  if (filename == nil)
+    filename = @"";
   ASSIGN (_fullFileName, [path stringByAppendingPathComponent: filename]);
   [_browser setPath: _fullFileName];
 
@@ -413,7 +506,7 @@ static BOOL _gs_display_reading_progress = NO;
    */
   _OKButtonPressed = NO;
 
-  [self browser: _browser
+  [self _browser: _browser
 	selectCellWithString: [[_browser selectedCell] stringValue] 
 	inColumn: [_browser selectedColumn]];
 }
@@ -421,8 +514,16 @@ static BOOL _gs_display_reading_progress = NO;
 - (BOOL) _shouldShowExtension: (NSString *)extension
 			isDir: (BOOL *)isDir;
 {
-  if (_treatsFilePackagesAsDirectories == NO && *isDir == YES &&
-	  ![extension isEqualToString: @""])
+  if (*isDir == NO)
+    {
+      if ([extension isEqualToString: _requiredFileType] == NO)
+	return NO;
+    }
+  else if ([extension length] == 0)
+    {
+      /* Automatic YES */
+    }
+  else if (_treatsFilePackagesAsDirectories == NO)
     {
       if ([extension isEqualToString: _requiredFileType] == YES)
 	  *isDir = NO;
@@ -431,11 +532,49 @@ static BOOL _gs_display_reading_progress = NO;
   return YES;
 }
 
+- (void) _windowResized: (NSNotification*)n
+{
+  [_browser setMaxVisibleColumns: [_browser frame].size.width / 140];
+}
+
+- (NSComparisonResult) _compareFilename: (NSString *)n1 with: (NSString *)n2
+{
+  if (_delegateHasCompareFilter)
+    {
+      return [_delegate panel: self
+              compareFilename: n1 
+                         with: n2 
+                caseSensitive: YES];
+    }
+  else
+    {
+      return [n1 _gsSavePanelCompare: n2];
+    }
+}
+
 @end /* NSSavePanel (PrivateMethods) */
 
 //
 // NSSavePanel methods
 //
+/** 
+    <p>Implements a panel that allows the user to save a file.
+    </p>
+    <p>
+    There is only one save panel per application and this panel is obtained
+    by calling the +savePanel class method. From here, you should set the
+    required file extension using -setRequiredFileType:
+    When ready to show the panel, use the
+    -runModal, or a similar method to show the panel in a modal
+    session. Other methods allow you to set the initial directory and
+    initially choosen file. The method will return one of NSOKButton
+    or NSCancelButton depending on which button the user pressed.
+    </p>
+    <p>
+    Use the -filename method to retrieve the name of the
+    file the user choose.
+    </p>
+ */
 @implementation NSSavePanel
 
 + (void) initialize
@@ -454,6 +593,7 @@ static BOOL _gs_display_reading_progress = NO;
     }
 }
 
+/** Returns the shared NSSavePanel instance */
 + (id) savePanel
 {
   if (_gs_gui_save_panel == nil)
@@ -467,8 +607,9 @@ static BOOL _gs_display_reading_progress = NO;
 }
 //
 
--(void) dealloc
+- (void) dealloc
 {
+  [[NSNotificationCenter defaultCenter] removeObserver: self];
   TEST_RELEASE (_fullFileName);
   TEST_RELEASE (_directory);  
   TEST_RELEASE (_requiredFileType);
@@ -497,6 +638,14 @@ static BOOL _gs_display_reading_progress = NO;
   return self;
 }
 
+/** Sets an accessory view which is shown near the bottom of the
+    panel. The panel is automatically expanded with enough room to
+    show the extra view. You can use this extra view to customize
+    various characteristics of the file selection mechanism. For instance
+    you could add a popup button which allows the user to select the
+    format that the file is saved in (e.g. rtf or txt). See
+    also -validateVisibleColumns .
+*/
 - (void) setAccessoryView: (NSView*)aView
 {
   NSRect accessoryViewFrame, bottomFrame;
@@ -603,6 +752,12 @@ static BOOL _gs_display_reading_progress = NO;
     }
 }
 
+/**
+ * Sets the title of the NSSavePanel to title. By default, 
+ * 'Save' is the title string. If you adapt the NSSavePanel 
+ * for other uses, its title should reflect the user action 
+ * that brings it to the screen.
+ */
 - (void) setTitle: (NSString*)title
 {
   [_titleField setStringValue: title];
@@ -612,27 +767,39 @@ static BOOL _gs_display_reading_progress = NO;
   [_titleField sizeToFit];
 }
 
+/** Returns the title of the save panel */
 - (NSString*) title
 {
   return [_titleField stringValue];
 }
 
+/**
+ * Returns the prompt of the Save panel field that holds 
+ * the current pathname or file name. By default this 
+ * prompt is 'Name: '.
+ */
 - (void) setPrompt: (NSString*)prompt
 {
   [[_form cellAtIndex: 0] setTitle: prompt];
   [_form setNeedsDisplay: YES];
 }
 
+/** Returns the prompt used in the current path field.  */
 - (NSString*) prompt
 {
   return [[_form cellAtIndex: 0] title];
 }
 
+/** Returns the accesory view (if any). */
 - (NSView*) accessoryView
 {
   return _accessoryView;
 }
 
+/**
+ * Sets the current path name in the Save panel's browser. 
+ * The path argument must be an absolute path name.
+ */
 - (void) setDirectory: (NSString*)path
 {
   NSString *standardizedPath = [path stringByStandardizingPath];
@@ -648,50 +815,83 @@ static BOOL _gs_display_reading_progress = NO;
     }
 }
 
+/**
+ * Specifies the type, a file name extension to be appended to 
+ * any selected files that don't already have that extension;
+ * The argument type should not include the period that begins 
+ * the extension.  Invoke this method each time the Save panel 
+ * is used for another file type within the application.
+ */
 - (void) setRequiredFileType: (NSString*)fileType
 {
   ASSIGN(_requiredFileType, fileType);
 }
 
+/** Returns the required file type. The default is no required file type.  */
 - (NSString*) requiredFileType
 {
   return _requiredFileType;
 }
 
+/** Returns YES if file packages are shown as directories. The default
+    is NO.  */
 - (BOOL) treatsFilePackagesAsDirectories
 {
   return _treatsFilePackagesAsDirectories;
 }
 
+/**
+ * Sets the NSSavePanel's behavior for displaying file packages 
+ * (for example, MyApp.app) to the user.  If flag is YES, the 
+ * user is shown files and subdirectories within a file 
+ * package.  If NO, the NSSavePanel shows each file package as 
+ * a file, thereby giving no indication that it is a directory.
+ */
 - (void) setTreatsFilePackagesAsDirectories: (BOOL)flag
 {
   _treatsFilePackagesAsDirectories = flag;
 }
 
+/**
+ * Validates and possibly reloads the browser columns that are visible 
+ * in the Save panel by causing the delegate method 
+ * -panel:shouldShowFilename: to be invoked. One situation in 
+ * which this method would find use is whey you want the 
+ * browser to show only files with certain extensions based on the 
+ * selection made in an accessory-view pop-up list.  When the 
+ * user changes the selection, you would invoke this method to
+ * revalidate the visible columns. 
+ */
 - (void) validateVisibleColumns
 {
   [_browser validateVisibleColumns];
 }
 
+/**
+ * Shows the save panel for the user. This method invokes
+ * -runModalForDirectory:file: with empty strings for the filename.
+ * Returns NSOKButton (if the user clicks the OK button) or
+ * NSCancelButton (if the user clicks the Cancel button).
+ */
 - (int) runModal
 {
-  if (_directory)
-    return [self runModalForDirectory: _directory 
-				 file: @""];
-  else
-    return [self runModalForDirectory: [_fm currentDirectoryPath] 
-				 file: @""];
+  return [self runModalForDirectory: nil file: @""];
 }
 
+/**
+ * Initializes the panel to the directory specified by path and,
+ * optionally, the file specified by filename, then displays it and
+ * begins its modal event loop; path and filename can be empty
+ * strings.  The method invokes [NSApplication:-runModalForWindow:]
+ * method with self as the argument.  Returns NSOKButton (if the user
+ * clicks the OK button) or NSCancelButton (if the user clicks the
+ * Cancel button). If path is nil then the panel displays the last
+ * selected directory or as a last resort, the current working directory.
+ */
 - (int) runModalForDirectory: (NSString*)path file: (NSString*)filename
 {
   [self _setupForDirectory: path file: filename];
-  [NSApp runModalForWindow: self];
-
-  if (_OKButtonPressed)
-    return NSOKButton;
-  else 
-    return NSCancelButton;
+  return [NSApp runModalForWindow: self];
 }
 
 - (int) runModalForDirectory: (NSString *)path
@@ -699,13 +899,8 @@ static BOOL _gs_display_reading_progress = NO;
 	    relativeToWindow: (NSWindow*)window
 {
   [self _setupForDirectory: path file: filename];
-  [NSApp runModalForWindow: self
-	 relativeToWindow: window];
-
-  if (_OKButtonPressed)
-    return NSOKButton;
-  else 
-    return NSCancelButton;
+  return [NSApp runModalForWindow: self
+		relativeToWindow: window];
 }
 
 - (void) beginSheetForDirectory: (NSString *)path
@@ -723,6 +918,11 @@ static BOOL _gs_display_reading_progress = NO;
 	 contextInfo: contextInfo];
 }
 
+/**
+ * Returns the directory choosen by the user.  Do not invoke directory
+ * within a modal loop because the information that these methods
+ * fetch is updated only upon return.
+ */
 - (NSString*) directory
 {
   if (_directory)
@@ -731,6 +931,11 @@ static BOOL _gs_display_reading_progress = NO;
     return @"";
 }
 
+/**
+ * Returns the absolute filename choosen by the user.  Do not invoke
+ * filename within a modal loop because the information that these
+ * methods fetch is updated only upon return.
+ */
 - (NSString*) filename
 {
   if (_fullFileName == nil)
@@ -758,7 +963,7 @@ static BOOL _gs_display_reading_progress = NO;
 {
   _fullFileName = nil;
   _directory = nil;
-  [NSApp stopModal];
+  [NSApp stopModalWithCode: NSCancelButton];
   [self close];
 }
 
@@ -767,6 +972,7 @@ static BOOL _gs_display_reading_progress = NO;
   NSMatrix      *matrix;
   NSBrowserCell *selectedCell;
   NSString      *filename;
+  BOOL		isDir = NO;
 
   matrix = [_browser matrixInColumn: [_browser lastColumn]];
   selectedCell = [matrix selectedCell];
@@ -804,18 +1010,58 @@ static BOOL _gs_display_reading_progress = NO;
     }
 
   ASSIGN (_directory, [_browser pathToColumn:[_browser lastColumn]]);
-  ASSIGN (_fullFileName, [_directory stringByAppendingPathComponent:
-				       [[_form cellAtIndex: 0] stringValue]]);
+  filename = [[_form cellAtIndex: 0] stringValue];
+  if ([filename isAbsolutePath] == YES)
+    {
+      ASSIGN (_fullFileName, filename);
+    }
+  else
+    {
+      ASSIGN (_fullFileName, [_directory stringByAppendingPathComponent:
+				       filename]);
+    }
 
+  filename = [_fullFileName stringByDeletingLastPathComponent];
+  if ([_fm fileExistsAtPath: filename isDirectory: &isDir] == NO)
+    {
+      int	result;
+
+      result = NSRunAlertPanel(_(@"Save"),
+	_(@"The directory '%@' does not exist, do you want to create it?"),
+	_(@"Yes"), _(@"No"), nil,
+	filename
+	);
+
+      if (result == NSAlertDefaultReturn)
+	{
+	  if ([_fm createDirectoryAtPath: filename attributes: nil] == NO)
+	    {
+	      NSRunAlertPanel(_(@"Save"),
+		_(@"The directory '%@' could not be created."),
+		_(@"Dismiss"), nil, nil,
+		filename
+		);
+	      return;
+	    }
+	}
+    }
+  else if (isDir == NO)
+    {
+      NSRunAlertPanel(_(@"Save"),
+	_(@"The path '%@' is not a directory."),
+	_(@"Dismiss"), nil, nil,
+	filename
+	);
+      return;
+    }
   if ([_fm fileExistsAtPath: [self filename] isDirectory: NULL])
     {
       int result;
 
-      //FIXME -- localize
-      result = NSRunAlertPanel(@"Save",
-			       @"The file '%@' in '%@' exists. Replace it?",
-			       @"Replace", @"Cancel", nil,
-			       [[_form cellAtIndex: 0] stringValue],
+      result = NSRunAlertPanel(_(@"Save"),
+			       _(@"The file '%@' in '%@' exists. Replace it?"),
+			       _(@"Replace"), _(@"Cancel"), nil,
+			       [[self filename] lastPathComponent],
 			       _directory);
 
       if (result != NSAlertDefaultReturn)
@@ -827,7 +1073,7 @@ static BOOL _gs_display_reading_progress = NO;
       return;
 
   _OKButtonPressed = YES;
-  [NSApp stopModal];
+  [NSApp stopModalWithCode: NSOKButton];
   [self close];
 }
 
@@ -881,6 +1127,28 @@ static BOOL _gs_display_reading_progress = NO;
   [super setDelegate: aDelegate];
 }
 
+- (void) setCanSelectHiddenExtension: (BOOL) flag
+{
+  // FIXME
+}
+
+- (BOOL) isExtensionHidden
+{
+  // FIXME
+  return NO;
+}
+
+- (void) setExtensionHidden: (BOOL) flag
+{
+  // FIXME
+}
+
+- (BOOL) isExpanded
+{
+  // FIXME
+  return NO;
+}
+
 //
 // NSCoding protocol
 //
@@ -900,10 +1168,6 @@ static BOOL _gs_display_reading_progress = NO;
 //
 // SavePanel filename compare
 //
-@interface NSString (_gsSavePanel)
-- (NSComparisonResult)_gsSavePanelCompare:(NSString *)other;
-@end
-
 @implementation NSString (_gsSavePanel)
 - (NSComparisonResult)_gsSavePanelCompare:(NSString *)other
 {
@@ -986,25 +1250,39 @@ static BOOL _gs_display_reading_progress = NO;
 // NSSavePanel browser delegate methods
 //
 @interface NSSavePanel (_BrowserDelegate)
-- (void) browser: (id)sender
+- (void) browser: (NSBrowser*)sender
 createRowsForColumn: (int)column
         inMatrix: (NSMatrix*)matrix;
 
 - (BOOL) browser: (NSBrowser*)sender
    isColumnValid: (int)column;
 
-- (BOOL) browser: (NSBrowser*)sender
-selectCellWithString: (NSString*)title
-	inColumn: (int)column;
-
 - (void) browser: (NSBrowser*)sender
  willDisplayCell: (id)cell
            atRow: (int)row
           column: (int)column;
 @end 
 
+struct NSSavePanel_struct
+{
+  @defs (NSSavePanel)
+};
+
+static int compareFilenames (id elem1, id elem2, void *context)
+{
+  /* TODO - use IMP optimization here.  */
+  struct NSSavePanel_struct *s = (struct NSSavePanel_struct *)context;
+  NSSavePanel *self = (NSSavePanel *)context;
+
+  return (int)[s->_delegate panel: self
+		compareFilename: elem1
+		with: elem2
+		caseSensitive: YES];
+}
+
+
 @implementation NSSavePanel (_BrowserDelegate)
-- (void) browser: (id)sender
+- (void) browser: (NSBrowser*)sender
 createRowsForColumn: (int)column
 	inMatrix: (NSMatrix*)matrix
 {
@@ -1029,13 +1307,16 @@ createRowsForColumn: (int)column
   /* Remove hidden files.  */
   {
     NSString *h;
-    NSArray *hiddenFiles;
+    NSArray *hiddenFiles = nil;
     BOOL gsSavePanelHideDotFiles;
-
-    /* We need to remove files listed in the xxx/.hidden file.  */
-    h = [path stringByAppendingPathComponent: @".hidden"];
-    h = [NSString stringWithContentsOfFile: h];
-    hiddenFiles = [h componentsSeparatedByString: @"\n"];
+    
+    if ([files containsObject: @".hidden"] == YES)
+      {
+	/* We need to remove files listed in the xxx/.hidden file.  */
+	h = [path stringByAppendingPathComponent: @".hidden"];
+	h = [NSString stringWithContentsOfFile: h];
+	hiddenFiles = [h componentsSeparatedByString: @"\n"];
+      }
 
     /* We need to remove files starting with `.' (dot), but only if
        the user asked for it in the defaults.  Perhaps we could add a
@@ -1061,7 +1342,7 @@ createRowsForColumn: (int)column
 	  {
 	    [mutableFiles removeObjectsInArray: hiddenFiles];
 	  }
-  
+	
 	if (gsSavePanelHideDotFiles)
 	  {
 	    /* Don't use i which is unsigned.  */
@@ -1070,7 +1351,7 @@ createRowsForColumn: (int)column
 	    while (j >= 0)
 	      {
 		NSString *file = (NSString *)[mutableFiles objectAtIndex: j];
-
+		
 		if ([file hasPrefix: @"."])
 		  {
 		    /* NSLog (@"Removing dot file %@", file); */
@@ -1101,7 +1382,7 @@ createRowsForColumn: (int)column
       progressString = [@"Reading Directory " stringByAppendingString: path];
       [super setTitle: progressString];
       // Is the following really safe? 
-      [GSCurrentContext() flush];
+      [GSCurrentContext() flushGraphics];
     }
 
   //TODO: Sort after creation of matrix so we do not sort 
@@ -1109,16 +1390,8 @@ createRowsForColumn: (int)column
   // Sort list of files to display
   if (_delegateHasCompareFilter == YES)
     {
-      int compare(id elem1, id elem2, void *context)
-      {
-	/* TODO - use IMP optimization here.  */
-	return (int)[_delegate panel: self
-		     compareFilename: elem1
-				with: elem2
-		       caseSensitive: YES];
-      }
-      files = [files sortedArrayUsingFunction: compare 
-		     context: nil];
+      files = [files sortedArrayUsingFunction: compareFilenames 
+		     context: self];
     }
   else
     files = [files sortedArrayUsingSelector: @selector(_gsSavePanelCompare:)];
@@ -1132,7 +1405,7 @@ createRowsForColumn: (int)column
 	  reached_frac++;
 	  progressString = [progressString stringByAppendingString: @"."];
 	  [super setTitle: progressString];
-	  [GSCurrentContext() flush];
+	  [GSCurrentContext() flushGraphics];
 	}
       // Now the real code
       file = [files objectAtIndex: i];
@@ -1183,7 +1456,7 @@ createRowsForColumn: (int)column
   if (display_progress)
     {
       [super setTitle: @""];
-      [GSCurrentContext() flush];
+      [GSCurrentContext() flushGraphics];
     }
 
   RELEASE (pool);
@@ -1209,34 +1482,6 @@ createRowsForColumn: (int)column
   return YES;
 }
 
-- (BOOL) browser: (NSBrowser*)sender
-selectCellWithString: (NSString*)title
-	inColumn: (int)column
-{
-  NSMatrix *m;
-  BOOL isLeaf;
-  NSString *path;
-
-  m = [sender matrixInColumn: column];
-  isLeaf = [[m selectedCell] isLeaf];
-  path = [sender pathToColumn: column];
-
-  if (isLeaf)
-    {
-      ASSIGN (_directory, path);
-      ASSIGN (_fullFileName, [path stringByAppendingPathComponent: title]);
-    }
-  else
-    {
-      ASSIGN (_directory, [path stringByAppendingPathComponent: title]);
-      ASSIGN (_fullFileName, nil);
-    }
-
-  [self _selectTextInColumn:column];
-  
-  return YES;
-}
-
 - (void) browser: (NSBrowser*)sender
  willDisplayCell: (id)cell
 	   atRow: (int)row
@@ -1263,14 +1508,22 @@ selectCellWithString: (NSString*)title
   NSComparisonResult  result;
   NSRange             range;
 
-  matrix = [_browser matrixInColumn:[_browser lastColumn]];
   s = [[[aNotification userInfo] objectForKey:@"NSFieldEditor"] string];
 
+  /*
+   * If the user typed in an absolute path, display it.
+   */
+  if ([s isAbsolutePath] == YES)
+    {
+      [self setDirectory: s];
+    }
+
   sLength = [s length];
   range.location = 0;
   range.length = sLength;
 
-  if(sLength == 0)
+  matrix = [_browser matrixInColumn:[_browser lastColumn]];
+  if (sLength == 0)
     {
       [matrix deselectAllCells];
       [_okButton setEnabled:NO];
@@ -1282,31 +1535,31 @@ selectCellWithString: (NSString*)title
   selectedRow = [matrix selectedRow];
   cells = [matrix cells];
 
-  if(selectedString)
+  if (selectedString)
     {
       result = [s compare:selectedString options:0 range:range];
 
-      if(result == NSOrderedSame)
+      if (result == NSOrderedSame)
 	return;
     }
   else
     result = NSOrderedDescending;
 
-  if(result == NSOrderedDescending)
+  if (result == NSOrderedDescending)
     {
       int numberOfCells = [cells count];
 
-      for(i = selectedRow+1; i < numberOfCells; i++)
+      for (i = selectedRow+1; i < numberOfCells; i++)
 	{
 	  selectedString = [[matrix cellAtRow:i column:0] stringValue];
 
 	  cellLength = [selectedString length];
-	  if(cellLength != sLength)
+	  if (cellLength != sLength)
 	    continue;
 
 	  result = [selectedString compare:s options:0 range:range];
 
-	  if(result == NSOrderedSame)
+	  if (result == NSOrderedSame)
 	    {
 	      [matrix deselectAllCells];
 	      [matrix selectCellAtRow:i column:0];
@@ -1318,17 +1571,17 @@ selectCellWithString: (NSString*)title
     }
   else
     {
-      for(i = selectedRow; i >= 0; --i)
+      for (i = selectedRow; i >= 0; --i)
 	{
 	  selectedString = [[matrix cellAtRow:i column:0] stringValue];
 
 	  cellLength = [selectedString length];
-	  if(cellLength != sLength)
+	  if (cellLength != sLength)
 	    continue;
 
 	  result = [selectedString compare:s options:0 range:range];
 
-	  if(result == NSOrderedSame)
+	  if (result == NSOrderedSame)
 	    {
 	      [matrix deselectAllCells];
 	      [matrix selectCellAtRow:i column:0];
