@@ -28,17 +28,18 @@
    Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111 USA.
 */
 
+#include <Foundation/NSDebug.h>
 #include <Foundation/NSException.h>
-#include <AppKit/NSScroller.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSCell.h>
-#include <AppKit/NSClipView.h>
-#include <AppKit/NSScrollView.h>
-#include <AppKit/NSRulerView.h>
-#include <AppKit/NSTableHeaderView.h>
-#include <AppKit/NSTableView.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/PSOperators.h>
+#include "AppKit/NSScroller.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSCell.h"
+#include "AppKit/NSClipView.h"
+#include "AppKit/NSScrollView.h"
+#include "AppKit/NSRulerView.h"
+#include "AppKit/NSTableHeaderView.h"
+#include "AppKit/NSTableView.h"
+#include "AppKit/NSWindow.h"
+#include "AppKit/PSOperators.h"
 
 @implementation NSScrollView
 
@@ -55,10 +56,9 @@ static float scrollerWidth;
 {
   if (self == [NSScrollView class])
     {
-      NSDebugLog(@"Initialize NSScrollView class\n");
       [self setRulerViewClass: [NSRulerView class]];
       scrollerWidth = [NSScroller scrollerWidth];
-      [self setVersion: 1];
+      [self setVersion: 2];
     }
 }
 
@@ -141,7 +141,7 @@ static float scrollerWidth;
   _hPageScroll = 10;
   _vLineScroll = 10;
   _vPageScroll = 10;
-  _borderType = NSBezelBorder;
+  _borderType = NSNoBorder;
   _scrollsDynamically = YES;
   [self tile];
 
@@ -290,7 +290,7 @@ static float scrollerWidth;
 - (void) scrollWheel: (NSEvent *)theEvent
 {
   NSRect	clipViewBounds;
-  float		deltaY = [theEvent deltaY];
+  float		delta = [theEvent deltaY];
   float 	amount;
   NSPoint	point;
 
@@ -303,47 +303,137 @@ static float scrollerWidth;
       clipViewBounds = [_contentView bounds];
     }
   point = clipViewBounds.origin;
-  if (([theEvent modifierFlags] & NSAlternateKeyMask) == NSAlternateKeyMask)
+
+
+  if (_hasHorizScroller == YES
+    && ([theEvent modifierFlags] & NSShiftKeyMask) == NSShiftKeyMask)
     {
-      amount = - (clipViewBounds.size.height - _vPageScroll) * deltaY;
+      if (([theEvent modifierFlags] & NSAlternateKeyMask) == NSAlternateKeyMask)
+	{
+	  amount = - (clipViewBounds.size.width - _hPageScroll) * delta;
+	}
+      else
+	{
+	  amount = - _hLineScroll * delta;
+	}
+      NSDebugLLog (@"NSScrollView", 
+	@"increment/decrement: amount = %f, horizontal", amount);
+
+      point.x = clipViewBounds.origin.x + amount;
     }
   else
     {
-      amount = - _vLineScroll * deltaY;
+      if (([theEvent modifierFlags] & NSAlternateKeyMask) == NSAlternateKeyMask)
+	{
+	  amount = - (clipViewBounds.size.height - _vPageScroll) * delta;
+	}
+      else
+	{
+	  amount = - _vLineScroll * delta;
+	}
+
+      if (_contentView != nil && !_contentView->_rFlags.flipped_view)
+	{
+	  /* If view is flipped reverse the scroll direction */
+	  amount = -amount;
+	}
+      NSDebugLLog (@"NSScrollView", 
+	@"increment/decrement: amount = %f, flipped = %d",
+	amount, _contentView ? _contentView->_rFlags.flipped_view : 0);
+
+      point.y = clipViewBounds.origin.y + amount;
     }
 
+  /* scrollToPoint: will call reflectScrolledClipView:, which will
+   * update rules, headers, and scrollers.  */
+  [_contentView scrollToPoint: point];
+}
+
+/**
+ * Scrolls the receiver by simply invoking scrollPageUp:
+ */
+- (void) pageUp: (id)sender
+{
+  [self scrollPageUp: sender];
+}
+
+/*
+ * This code is based on _doScroll: and still may need some tuning. 
+ */
+- (void) scrollPageUp: (id)sender
+{
+  NSRect  clipViewBounds;
+  NSPoint point;
+  float   amount;
+
+  if (_contentView == nil)
+    {
+      clipViewBounds = NSZeroRect;
+    }
+  else
+    {
+      clipViewBounds = [_contentView bounds];
+    }
+  point = clipViewBounds.origin;
+  /*
+   * Take verticalPageScroll into accout, but try to make sure
+   * that amount is never negative (ie do not scroll backwards.)
+   *
+   * FIXME: It seems _doScroll and scrollWheel: should also take
+   * care not to do negative scrolling.
+   */
+  amount = clipViewBounds.size.height - _vPageScroll;
+  amount = (amount < 0) ? 0 : amount;
+
   if (_contentView != nil && !_contentView->_rFlags.flipped_view)
     {
-      /* If view is flipped reverse the scroll direction */
       amount = -amount;
     }
-  NSDebugLog (@"increment/decrement: amount = %f, flipped = %d",
-    amount, _contentView ? _contentView->_rFlags.flipped_view : 0);
-
-  point.y = clipViewBounds.origin.y + amount;
-  
-  if (_hasHeaderView)
-    {
-      NSPoint scrollTo;
-
-      scrollTo = [_headerClipView bounds].origin;
-      scrollTo.x += point.x - clipViewBounds.origin.x;
-      [_headerClipView scrollToPoint: scrollTo];
-    }
+  point.y = clipViewBounds.origin.y - amount;
   [_contentView scrollToPoint: point];
+}
 
-  if (_rulersVisible == YES)
+/**
+ * Scrolls the receiver by simply invoking scrollPageUp:
+ */
+- (void) pageDown: (id)sender
+{
+  [self scrollPageDown: sender];
+}
+
+/*
+ * This code is based on _doScroll:. and still may need some tuning.
+ */
+- (void) scrollPageDown: (id)sender
+{
+  NSRect  clipViewBounds;
+  NSPoint point;
+  float   amount;
+
+  if (_contentView == nil)
     {
-      if (_hasHorizRuler)
-	{
-	  [_horizRuler setNeedsDisplay: YES];
-	}
-      if (_hasVertRuler)
-	{
-	  [_vertRuler setNeedsDisplay: YES];
-	}
+      clipViewBounds = NSZeroRect;
     }
-
+  else
+    {
+      clipViewBounds = [_contentView bounds];
+    }
+  point = clipViewBounds.origin;
+  /*
+   * Take verticalPageScroll into accout, but try to make sure
+   * that amount is never negativ (ie do not scroll backwards.)
+   *
+   * FIXME: It seems _doScroll and scrollWheel: should also take
+   * care not to do negative scrolling.
+   */
+  amount = clipViewBounds.size.height - _vPageScroll;
+  amount = (amount < 0) ? 0 : amount;
+  if (_contentView != nil && !_contentView->_rFlags.flipped_view)
+    {
+      amount = -amount;
+    }
+  point.y = clipViewBounds.origin.y + amount;
+  [_contentView scrollToPoint: point];
 }
 
 - (void) _doScroll: (NSScroller*)scroller
@@ -367,7 +457,7 @@ static float scrollerWidth;
     }
   point = clipViewBounds.origin;
 
-  NSDebugLog (@"_doScroll: float value = %f", floatValue);
+  NSDebugLLog (@"NSScrollView", @"_doScroll: float value = %f", floatValue);
 
   /* do nothing if scroller is unknown */
   if (scroller != _horizScroller && scroller != _vertScroller)
@@ -426,7 +516,8 @@ static float scrollerWidth;
 	      /* If view is flipped reverse the scroll direction */
 	      amount = -amount;
 	    }
-	  NSDebugLog (@"increment/decrement: amount = %f, flipped = %d",
+	  NSDebugLLog (@"NSScrollView", 
+		       @"increment/decrement: amount = %f, flipped = %d",
 	    amount, _contentView ? _contentView->_rFlags.flipped_view : 0);
 	  point.y = clipViewBounds.origin.y + amount;
 	}
@@ -449,28 +540,9 @@ static float scrollerWidth;
 	}
     }
 
-  if (_hasHeaderView)
-    {
-      NSPoint scrollTo;
-
-      scrollTo = [_headerClipView bounds].origin;
-      scrollTo.x += point.x - clipViewBounds.origin.x;
-      [_headerClipView scrollToPoint: scrollTo];
-    }
+  /* scrollToPoint will call reflectScrollerClipView, and that will
+   * update scrollers, rulers and headers */
   [_contentView scrollToPoint: point];
-
-  if (_rulersVisible == YES)
-    {
-      if (_hasHorizRuler)
-	{
-	  [_horizRuler setNeedsDisplay: YES];
-	}
-      if (_hasVertRuler)
-	{
-	  [_vertRuler setNeedsDisplay: YES];
-	}
-    }
-  
 }
 
 - (void) reflectScrolledClipView: (NSClipView *)aClipView
@@ -486,7 +558,7 @@ static float scrollerWidth;
       return;
     }
 
-  NSDebugLog (@"reflectScrolledClipView:");
+  NSDebugLLog (@"NSScrollView", @"reflectScrolledClipView:");
 
   if (_contentView)
     {
@@ -549,16 +621,25 @@ static float scrollerWidth;
       
       headerClipViewOrigin = [_headerClipView bounds].origin;
 
-      // If needed, scroll the headerview too
+      /* If needed, scroll the headerview too.  */
       if (headerClipViewOrigin.x != clipViewBounds.origin.x)
 	{
 	  headerClipViewOrigin.x = clipViewBounds.origin.x;
-	  headerClipViewOrigin = [_headerClipView constrainScrollPoint: 
-						    headerClipViewOrigin];
 	  [_headerClipView scrollToPoint: headerClipViewOrigin];
 	}
     }
-  
+
+  if (_rulersVisible == YES)
+    {
+      if (_hasHorizRuler)
+	{
+	  [_horizRuler setNeedsDisplay: YES];
+	}
+      if (_hasVertRuler)
+	{
+	  [_vertRuler setNeedsDisplay: YES];
+       }
+    }
 }
 
 - (void) setHorizontalRulerView: (NSRulerView *)aRulerView
@@ -818,27 +899,13 @@ static float scrollerWidth;
     }
 
   [_contentView setFrame: contentRect];
+  [self setNeedsDisplay: YES];
 }
 
 - (void) drawRect: (NSRect)rect
 {
-  NSGraphicsContext	*ctxt = GSCurrentContext();
-  float horizLinePosition, horizLineLength = _bounds.size.width;
-  NSSize border = _sizeForBorderType(_borderType);
-  float headerViewHeight = 0;
-
-  if (_hasHeaderView == YES)
-    {
-      headerViewHeight = [[_headerClipView documentView] frame].size.height;
-    }
-  if ((_hasCornerView == YES) && (_hasHorizScroller == YES) 
-      && (headerViewHeight == 0))
-    {
-      headerViewHeight = [[(NSTableView *)[_contentView documentView] 
-					  cornerView] frame].size.height;
-    }
+  NSGraphicsContext *ctxt = GSCurrentContext();
   
-  DPSgsave(ctxt);
   switch (_borderType)
     {
       case NSNoBorder:
@@ -858,46 +925,35 @@ static float scrollerWidth;
 	break;
     }
 
-  horizLinePosition = border.width;
-
+  [[NSColor controlDarkShadowColor] set];
   DPSsetlinewidth(ctxt, 1);
-  DPSsetgray(ctxt, 0);
+
   if (_hasVertScroller)
     {
-      horizLinePosition = scrollerWidth + border.width;
-      horizLineLength -= scrollerWidth + 2 * border.width;
-      DPSmoveto(ctxt, horizLinePosition, border.height);
-      if (_rFlags.flipped_view)
-	{
-	  DPSrmoveto(ctxt, 0, headerViewHeight);
-	}
-      DPSrlineto(ctxt, 0, _bounds.size.height - headerViewHeight 
-		 - 2 * border.height - 1);
+      DPSmoveto(ctxt, [_vertScroller frame].origin.x + scrollerWidth, 
+		[_vertScroller frame].origin.y - 1);
+      DPSrlineto(ctxt, 0, [_vertScroller frame].size.height + 1);
       DPSstroke(ctxt);
-      if ((_hasHeaderView == YES) && (_hasCornerView == NO)) 
-	{
-	  float yStart = border.height + headerViewHeight - 1;
-
-	  if (_rFlags.flipped_view == NO)
-	    yStart = _bounds.size.height - yStart;
-	  DPSmoveto(ctxt, horizLinePosition, yStart); 
-	  DPSlineto(ctxt, border.width, yStart);
-	  DPSstroke(ctxt);
-	}
     }
 
   if (_hasHorizScroller)
     {
-      float ypos = scrollerWidth + border.height + 1;
+      float ypos;
+      float scrollerY = [_horizScroller frame].origin.y;
 
       if (_rFlags.flipped_view)
-	ypos = _bounds.size.height - ypos;
-      DPSmoveto(ctxt, horizLinePosition, ypos);
-      DPSrlineto(ctxt, horizLineLength - 1, 0);
+	{
+	  ypos = scrollerY - 1;
+	}
+      else
+	{
+	  ypos = scrollerY + scrollerWidth + 1;
+	}
+
+      DPSmoveto(ctxt, [_horizScroller frame].origin.x - 1, ypos);
+      DPSrlineto(ctxt, [_horizScroller frame].size.width + 1, 0);
       DPSstroke(ctxt);
     }
-
-  DPSgrestore(ctxt);
 }
 
 - (NSRect) documentVisibleRect
@@ -928,6 +984,7 @@ static float scrollerWidth;
 - (void) setBorderType: (NSBorderType)borderType
 {
   _borderType = borderType;
+  [self tile];
 }
 
 - (void) setDocumentView: (NSView *)aView
@@ -1143,8 +1200,7 @@ static float scrollerWidth;
 - (void) encodeWithCoder: (NSCoder*)aCoder
 {
   [super encodeWithCoder: aCoder];
-
-  NSDebugLLog(@"NSScrollView", @"NSScrollView: start encoding\n");
+      
   [aCoder encodeObject: _contentView];
   [aCoder encodeValueOfObjCType: @encode(NSBorderType) at: &_borderType];
   [aCoder encodeValueOfObjCType: @encode(BOOL) at: &_scrollsDynamically];
@@ -1170,45 +1226,95 @@ static float scrollerWidth;
   if (_hasVertRuler)
     [aCoder encodeObject: _vertRuler];
 
+  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &_hasHeaderView];
+  if (_hasHeaderView)
+    [aCoder encodeObject: _headerClipView];
+
+  [aCoder encodeValueOfObjCType: @encode(BOOL) at: &_hasCornerView];
+
   /* We do not need to encode headerview, cornerview stuff */
-  NSDebugLLog(@"NSScrollView", @"NSScrollView: finish encoding\n");
 }
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  [super initWithCoder: aDecoder];
+  self = [super initWithCoder: aDecoder];
 
-  NSDebugLLog(@"NSScrollView", @"NSScrollView: start decoding\n");
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_contentView];
-  [aDecoder decodeValueOfObjCType: @encode(NSBorderType) at: &_borderType];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_scrollsDynamically];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_rulersVisible];
-  [aDecoder decodeValueOfObjCType: @encode(float) at: &_hLineScroll];
-  [aDecoder decodeValueOfObjCType: @encode(float) at: &_hPageScroll];
-  [aDecoder decodeValueOfObjCType: @encode(float) at: &_vLineScroll];
-  [aDecoder decodeValueOfObjCType: @encode(float) at: &_vPageScroll];
-
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasHorizScroller];
-  if (_hasHorizScroller)
-    [aDecoder decodeValueOfObjCType: @encode(id) at: &_horizScroller];
-
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasVertScroller];
-  if (_hasVertScroller)
-    [aDecoder decodeValueOfObjCType: @encode(id) at: &_vertScroller];
-
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasHorizRuler];
-  if (_hasHorizRuler)
-    [aDecoder decodeValueOfObjCType: @encode(id) at: &_horizRuler];
-
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasVertRuler];
-  if (_hasVertRuler)
-    [aDecoder decodeValueOfObjCType: @encode(id) at: &_vertRuler];
-  
-  /* This recreates all the info about headerView, cornerView, etc */
-  [self setDocumentView: [_contentView documentView]];
-  [self tile];
-
-  NSDebugLLog(@"NSScrollView", @"NSScrollView: finish decoding\n");
+  if ([aDecoder allowsKeyedCoding])
+    {
+      NSClipView *content = [aDecoder decodeObjectForKey: @"NSContentView"];
+      NSScroller *hScroller = [aDecoder decodeObjectForKey: @"NSHScroller"];
+      NSScroller *vScroller = [aDecoder decodeObjectForKey: @"NSVScroller"];
+      
+      if (content != nil)
+        {
+	  [self setContentView: content];
+	}
+      if (hScroller != nil)
+        {
+	  [self setHorizontalScroller: hScroller];
+	}
+      if (vScroller != nil)
+        {
+	  [self setVerticalScroller: vScroller];
+	}
+      if ([aDecoder containsValueForKey: @"NSsFlags"])
+        {
+	  //int flags = [aDecoder decodeIntForKey: @"NSsFlags"];
+	  // FIXME
+	}
+    }
+  else
+    {
+      int version = [aDecoder versionForClassName: 
+				  @"NSScrollView"];
+      NSDebugLLog(@"NSScrollView", @"NSScrollView: start decoding\n");
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_contentView];
+      [aDecoder decodeValueOfObjCType: @encode(NSBorderType) at: &_borderType];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_scrollsDynamically];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_rulersVisible];
+      [aDecoder decodeValueOfObjCType: @encode(float) at: &_hLineScroll];
+      [aDecoder decodeValueOfObjCType: @encode(float) at: &_hPageScroll];
+      [aDecoder decodeValueOfObjCType: @encode(float) at: &_vLineScroll];
+      [aDecoder decodeValueOfObjCType: @encode(float) at: &_vPageScroll];
+      
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasHorizScroller];
+      if (_hasHorizScroller)
+	[aDecoder decodeValueOfObjCType: @encode(id) at: &_horizScroller];
+      
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasVertScroller];
+      if (_hasVertScroller)
+	[aDecoder decodeValueOfObjCType: @encode(id) at: &_vertScroller];
+      
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasHorizRuler];
+      if (_hasHorizRuler)
+	[aDecoder decodeValueOfObjCType: @encode(id) at: &_horizRuler];
+      
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasVertRuler];
+      if (_hasVertRuler)
+	[aDecoder decodeValueOfObjCType: @encode(id) at: &_vertRuler];
+      
+      if (version == 2)
+        {
+	  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasHeaderView];
+	  if (_hasHeaderView)
+	    [aDecoder decodeValueOfObjCType: @encode(id) at: &_headerClipView];
+      
+	  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_hasCornerView];
+	}
+      else if (version == 1)
+        {
+	  /* This recreates all the info about headerView, cornerView, etc */
+	  [self setDocumentView: [_contentView documentView]];
+	}
+      else
+        {
+	  NSLog(@"unknown NSScrollView version (%d)", version);
+	  DESTROY(self);
+	}
+      [self tile];
+      
+      NSDebugLLog(@"NSScrollView", @"NSScrollView: finish decoding\n");
+    }
 
   return self;
 }
