@@ -31,15 +31,16 @@
 
 #include <Foundation/NSDate.h>
 #include <Foundation/NSRunLoop.h>
+#include <Foundation/NSDebug.h>
 
-#include <AppKit/NSScroller.h>
-#include <AppKit/NSScrollView.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSButtonCell.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSGraphics.h>
+#include "AppKit/NSScroller.h"
+#include "AppKit/NSScrollView.h"
+#include "AppKit/NSWindow.h"
+#include "AppKit/NSButtonCell.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSGraphics.h"
 
 
 @implementation NSScroller
@@ -87,6 +88,11 @@ static NSColor *scrollBarColor = nil;
   return YES;
 }
 
+- (BOOL) acceptsFirstResponder
+{
+  return NO;
+}
+
 - (NSScrollArrowPosition) arrowsPosition
 {
   return _arrowsPosition;
@@ -147,36 +153,69 @@ static NSColor *scrollBarColor = nil;
 {
   self = [super initWithCoder: aDecoder];
 
-  if (_frame.size.width > _frame.size.height)
+  if ([aDecoder allowsKeyedCoding])
     {
-      _isHorizontal = YES;
+      NSString *action = [aDecoder decodeObjectForKey: @"NSAction"];
+      id target = [aDecoder decodeObjectForKey: @"NSTarget"];
+      float value = 0.0;
+      float percent = 0.0;
+      int flags;
+      
+      if (action != nil)
+        {
+	  [self setAction: NSSelectorFromString(action)];
+	}
+      [self setTarget: target];
+      
+      if ([aDecoder containsValueForKey: @"NSCurValue"])
+        {
+	  value = [aDecoder decodeFloatForKey: @"NSCurValue"];
+	}
+      if ([aDecoder containsValueForKey: @"NSPercent"])
+        {
+	  percent = [aDecoder decodeFloatForKey: @"NSPercent"];
+	}
+      [self setFloatValue: value knobProportion: percent];
+
+      if ([aDecoder containsValueForKey: @"NSsFlags"])
+        {
+	  flags = [aDecoder decodeIntForKey: @"NSsFlags"];
+	  // FIXME
+	}
     }
   else
     {
-      _isHorizontal = NO;
+      if (_frame.size.width > _frame.size.height)
+        {
+	  _isHorizontal = YES;
+	}
+      else
+        {
+	  _isHorizontal = NO;
+	}
+
+      if (_isHorizontal)
+        {
+	  _floatValue = 0.0;
+	}
+      else
+        {
+	  _floatValue = 1.0;
+	}
+      
+      _hitPart = NSScrollerNoPart;
+      
+      [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &_arrowsPosition];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_isEnabled];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_target];
+      // Undo RETAIN by decoder
+      TEST_RELEASE(_target);
+      [aDecoder decodeValueOfObjCType: @encode(SEL) at: &_action];
+
+      [self drawParts];
+      [self checkSpaceForParts];
     }
 
-  if (_isHorizontal)
-    {
-      _floatValue = 0.0;
-    }
-  else
-    {
-      _floatValue = 1.0;
-    }
-
-  _hitPart = NSScrollerNoPart;
-
-  [aDecoder decodeValueOfObjCType: @encode(unsigned int) at: &_arrowsPosition];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_isEnabled];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_target];
-  // Undo RETAIN by decoder
-  TEST_RELEASE(_target);
-  [aDecoder decodeValueOfObjCType: @encode(SEL) at: &_action];
-
-  [self drawParts];
-  [self checkSpaceForParts];
-
   return self;
 }
 
@@ -236,40 +275,36 @@ static NSColor *scrollBarColor = nil;
     return;
 
   upCell = [NSButtonCell new];
-  [upCell setHighlightsBy: NSChangeBackgroundCellMask|NSContentsCellMask];
+  [upCell setHighlightsBy: NSChangeBackgroundCellMask];
   [upCell setImage: [NSImage imageNamed: @"common_ArrowUp"]];
-  [upCell setAlternateImage: [NSImage imageNamed: @"common_ArrowUpH"]];
   [upCell setImagePosition: NSImageOnly];
   [upCell setContinuous: YES];
   [upCell sendActionOn: (NSLeftMouseDownMask | NSPeriodicMask)];
-  [upCell setPeriodicDelay: 0.3 interval: 0.04];
+  [upCell setPeriodicDelay: 0.3 interval: 0.03];
 
   downCell = [NSButtonCell new];
-  [downCell setHighlightsBy: NSChangeBackgroundCellMask|NSContentsCellMask];
+  [downCell setHighlightsBy: NSChangeBackgroundCellMask];
   [downCell setImage: [NSImage imageNamed: @"common_ArrowDown"]];
-  [downCell setAlternateImage: [NSImage imageNamed: @"common_ArrowDownH"]];
   [downCell setImagePosition: NSImageOnly];
   [downCell setContinuous: YES];
   [downCell sendActionOn: (NSLeftMouseDownMask | NSPeriodicMask)];
-  [downCell setPeriodicDelay: 0.3 interval: 0.04];
+  [downCell setPeriodicDelay: 0.3 interval: 0.03];
 
   leftCell = [NSButtonCell new];
-  [leftCell setHighlightsBy: NSChangeBackgroundCellMask|NSContentsCellMask];
+  [leftCell setHighlightsBy: NSChangeBackgroundCellMask];
   [leftCell setImage: [NSImage imageNamed: @"common_ArrowLeft"]];
-  [leftCell setAlternateImage: [NSImage imageNamed: @"common_ArrowLeftH"]];
   [leftCell setImagePosition: NSImageOnly];
   [leftCell setContinuous: YES];
   [leftCell sendActionOn: (NSLeftMouseDownMask | NSPeriodicMask)];
-  [leftCell setPeriodicDelay: 0.3 interval: 0.04];
+  [leftCell setPeriodicDelay: 0.3 interval: 0.03];
 
   rightCell = [NSButtonCell new];
-  [rightCell setHighlightsBy: NSChangeBackgroundCellMask|NSContentsCellMask];
+  [rightCell setHighlightsBy: NSChangeBackgroundCellMask];
   [rightCell setImage: [NSImage imageNamed: @"common_ArrowRight"]];
-  [rightCell setAlternateImage: [NSImage imageNamed: @"common_ArrowRightH"]];
   [rightCell setImagePosition: NSImageOnly];
   [rightCell setContinuous: YES];
   [rightCell sendActionOn: (NSLeftMouseDownMask | NSPeriodicMask)];
-  [rightCell setPeriodicDelay: 0.3 interval: 0.04];
+  [rightCell setPeriodicDelay: 0.3 interval: 0.03];
 
   knobCell = [NSButtonCell new];
   [knobCell setButtonType: NSMomentaryChangeButton];
@@ -299,31 +334,42 @@ static NSColor *scrollBarColor = nil;
 {
   NSSize	frameSize = _frame.size;
   float		size = (_isHorizontal ? frameSize.width : frameSize.height);
-  float		scrollerWidth = [isa scrollerWidth];
 
   if (_arrowsPosition == NSScrollerArrowsNone)
     {
-      if (size >= scrollerWidth + 2)
-	_usableParts = NSAllScrollerParts;
+      if (size >= buttonsWidth + 3)
+	{
+	  _usableParts = NSAllScrollerParts;
+	}
       else
-	_usableParts = NSNoScrollerParts;
+	{
+	  _usableParts = NSNoScrollerParts;
+	}
     }
   else
     {
-      if (size >= 3 * scrollerWidth + 4)
-	_usableParts = NSAllScrollerParts;
-      else if (size >= 2 * scrollerWidth + 3)
-	_usableParts = NSOnlyScrollerArrows;
+      if (size >= 4 /* spacing */ + 1 /* min. scroll area */ + buttonsWidth * 3)
+	{
+	  _usableParts = NSAllScrollerParts;
+	}
+      else if (size >= 3 /* spacing */ + buttonsWidth * 2)
+	{
+	  _usableParts = NSOnlyScrollerArrows;
+	}
       else
-	_usableParts = NSNoScrollerParts;
+	{
+	  _usableParts = NSNoScrollerParts;
+	}
     }
 }
 
 - (void) setEnabled: (BOOL)flag
 {
   if (_isEnabled == flag)
-    return;
-
+    {
+      return;
+    }
+  
   _isEnabled = flag;
   _cacheValid = NO;
   [self setNeedsDisplay: YES];
@@ -332,8 +378,10 @@ static NSColor *scrollBarColor = nil;
 - (void) setArrowsPosition: (NSScrollArrowPosition)where
 {
   if (_arrowsPosition == where)
-    return;
-
+    {
+      return;
+    }
+  
   _arrowsPosition = where;
   _cacheValid = NO;
   [self setNeedsDisplay: YES];
@@ -382,6 +430,9 @@ static NSColor *scrollBarColor = nil;
       _knobProportion = ratio;
     }
 
+  /* Make sure we mark ourselves as needing redisplay.  */
+  _floatValue = -1;
+  
   [self setFloatValue: aFloat];
 }
 
@@ -403,10 +454,17 @@ static NSColor *scrollBarColor = nil;
 
   [super setFrame: frameRect];
 
-  if (_isHorizontal)
-    _arrowsPosition = NSScrollerArrowsMinEnd;
-  else
-    _arrowsPosition = NSScrollerArrowsMaxEnd;
+  if (_arrowsPosition != NSScrollerArrowsNone)
+    {
+      if (_isHorizontal)
+	{
+	  _arrowsPosition = NSScrollerArrowsMinEnd;
+	}
+      else
+	{
+	  _arrowsPosition = NSScrollerArrowsMaxEnd;
+	}
+    }
 
   _hitPart = NSScrollerNoPart;
   _cacheValid = NO;
@@ -428,6 +486,10 @@ static NSColor *scrollBarColor = nil;
    */
   NSRect rect;
 
+  /* thePoint is in window's coordinate system; convert it to
+   * our own coordinate system.  */
+  thePoint = [self convertPoint: thePoint fromView: nil];
+
   if (thePoint.x <= 0 || thePoint.x >= _frame.size.width
     || thePoint.y <= 0 || thePoint.y >= _frame.size.height)
     return NSScrollerNoPart;
@@ -497,9 +559,7 @@ static NSColor *scrollBarColor = nil;
 
 - (void) mouseDown: (NSEvent*)theEvent
 {
-  NSPoint location = [self convertPoint: [theEvent locationInWindow]
-			       fromView: nil];
-
+  NSPoint location = [theEvent locationInWindow];
   _hitPart = [self testPart: location];
   [self _setTargetAndActionToCells];
 
@@ -508,7 +568,7 @@ static NSColor *scrollBarColor = nil;
       case NSScrollerIncrementLine:
       case NSScrollerDecrementLine:
 	/*
-	 * A hit on a scroller button should be a page meovement
+	 * A hit on a scroller button should be a page movement
 	 * if the alt key is pressed.
 	 */
 	if ([theEvent modifierFlags] & NSAlternateKeyMask)
@@ -535,10 +595,14 @@ static NSColor *scrollBarColor = nil;
 
       case NSScrollerKnobSlot:
 	{
-	  float floatValue = [self _floatValueForMousePoint: location];
-
-	  [self setFloatValue: floatValue];
-	  [self sendAction: _action to: _target];
+	  float floatValue = [self _floatValueForMousePoint: 
+				     [self convertPoint: location
+					   fromView: nil]];
+	  if (floatValue != _floatValue)
+	    {
+	      [self setFloatValue: floatValue];
+	      [self sendAction: _action to: _target];
+	    }
 	  [self trackKnob: theEvent];
 	  break;
 	}
@@ -553,23 +617,21 @@ static NSColor *scrollBarColor = nil;
 - (void) trackKnob: (NSEvent*)theEvent
 {
   unsigned int eventMask = NSLeftMouseDownMask | NSLeftMouseUpMask
-			  | NSLeftMouseDraggedMask | NSMouseMovedMask
-			  | NSPeriodicMask;
+			  | NSLeftMouseDraggedMask | NSFlagsChangedMask;
   NSPoint	point;
-  NSPoint	apoint;
   float		lastPosition;
   float		newPosition;
   float		floatValue;
   float		offset;
-  NSDate	*theDistantFuture = [NSDate distantFuture];
-  NSEventType	eventType;
+  float		initialOffset;
+  NSEvent	*presentEvent = theEvent;
+  NSEventType	eventType = [theEvent type];
   NSRect	knobRect;
   unsigned	flags = [theEvent modifierFlags];
 
   knobRect = [self rectForPart: NSScrollerKnob];
 
-  apoint = [theEvent locationInWindow];
-  point = [self convertPoint: apoint fromView: nil];
+  point = [self convertPoint: [theEvent locationInWindow] fromView: nil];
   if (_isHorizontal)
     {
       lastPosition = NSMidX(knobRect);
@@ -581,132 +643,153 @@ static NSColor *scrollBarColor = nil;
       offset = lastPosition - point.y;
     }
 
+  initialOffset = offset; /* Save the initial offset value */
   _hitPart = NSScrollerKnob;
-  /*
-   * set periodic events rate to achieve max of ~30fps
-   */
-  [NSEvent startPeriodicEventsAfterDelay: 0.02 withPeriod: 0.03];
-  [[NSRunLoop currentRunLoop] limitDateForMode: NSEventTrackingRunLoopMode];
 
-  while ((eventType = [theEvent type]) != NSLeftMouseUp)
+  do
     {
-      if (eventType != NSPeriodic)
-	{
-	  apoint = [theEvent locationInWindow];
-	  flags = [theEvent modifierFlags];
-	}
-      else
-	{
-	  point = [self convertPoint: apoint fromView: nil];
-          if (_isHorizontal)
-	    newPosition = point.x + offset;
-          else
-	    newPosition = point.y + offset;
+       /* Inner loop that gets and (quickly) handles all events that have
+          already arrived. */
+       while (theEvent && eventType != NSLeftMouseUp)
+         {
+           /* Note the event here. Don't do any expensive handling. */
+	   if (eventType == NSFlagsChanged)
+             flags = [theEvent modifierFlags];
+	   presentEvent = theEvent;
 
-          if (newPosition != lastPosition)
-            {
-              if (flags & NSAlternateKeyMask)
-	        {
-	          float	diff;
+           theEvent = [NSApp nextEventMatchingMask: eventMask
+                         untilDate: [NSDate distantPast] /* Only get events that have already arrived. */
+                         inMode: NSEventTrackingRunLoopMode
+                         dequeue: YES];
+	   eventType = [theEvent type];
+         }
 
-	          diff = newPosition - lastPosition;
-	          diff = diff * 3 / 4;
-	          offset -= diff;
-	          newPosition -= diff;
-	        }
+       /* 
+        * No more events right now. Do expensive handling, like drawing, 
+	* here. 
+	*/
+       point = [self convertPoint: [presentEvent locationInWindow] 
+			 fromView: nil];
 
-              // only one coordinate (X or Y) is used to compute floatValue.
-              point = NSMakePoint(newPosition, newPosition);
-	      floatValue = [self _floatValueForMousePoint: point];
+       if (_isHorizontal)
+         newPosition = point.x + offset;
+       else
+	 newPosition = point.y + offset;
 
-	      [self setFloatValue: floatValue];
-	      [self sendAction: _action to: _target];
+       if (newPosition != lastPosition)
+         {
+           if (flags & NSAlternateKeyMask)
+	     {
+	       float	diff;
 
-	      lastPosition = newPosition;
-            }
-	}
+	       diff = newPosition - lastPosition;
+	       diff = diff * 3 / 4;
+	       offset -= diff;
+	       newPosition -= diff;
+	     }
+	   else /* Ok, we are no longer doing slow scrolling, lets go back 
+		   to our original offset. */
+	     {
+	       offset = initialOffset;
+	     }
 
-      theEvent = [NSApp nextEventMatchingMask: eventMask
-				    untilDate: theDistantFuture
-				       inMode: NSEventTrackingRunLoopMode
-				      dequeue: YES];
-    }
-  [NSEvent stopPeriodicEvents];
+           // only one coordinate (X or Y) is used to compute floatValue.
+           point = NSMakePoint(newPosition, newPosition);
+	   floatValue = [self _floatValueForMousePoint: point];
+
+	   if (floatValue != _floatValue)
+	     {
+	       [self setFloatValue: floatValue];
+	       [self sendAction: _action to: _target];
+	     }
+	      
+	     lastPosition = newPosition;
+         }
+
+       /* 
+	* If our current event is actually the mouse up (perhaps the inner 
+	* loop got to this point) we want to update with the last info and 
+	* then quit.
+	*/
+       if (eventType == NSLeftMouseUp)
+         break;
+
+       /* Get the next event, blocking if necessary. */
+       theEvent = [NSApp nextEventMatchingMask: eventMask
+                     untilDate: nil /* No limit, block until we get an event. */
+                     inMode: NSEventTrackingRunLoopMode
+                     dequeue: YES];
+       eventType = [theEvent type];
+  } while (eventType != NSLeftMouseUp);
 }
 
 - (void) trackScrollButtons: (NSEvent*)theEvent
 {
-  NSApplication	*theApp = [NSApplication sharedApplication];
-  unsigned int	eventMask = NSLeftMouseDownMask | NSLeftMouseUpMask |
-			  NSLeftMouseDraggedMask | NSMouseMovedMask;
-  NSPoint	location;
-  BOOL		shouldReturn = NO;
   id		theCell = nil;
   NSRect	rect;
 
+  [self lockFocus];
+
   NSDebugLog (@"trackScrollButtons");
-  do
+
+  _hitPart = [self testPart: [theEvent locationInWindow]];
+  rect = [self rectForPart: _hitPart];
+
+  /*
+   * A hit on a scroller button should be a page movement
+   * if the alt key is pressed.
+   */
+  switch (_hitPart)
     {
-      location = [self convertPoint: [theEvent locationInWindow]fromView: nil];
-      _hitPart = [self testPart: location];
-      rect = [self rectForPart: _hitPart];
-
-      /*
-       * A hit on a scroller button should be a page meovement
-       * if the alt key is pressed.
-       */
-      switch (_hitPart)
-	{
-	  case NSScrollerIncrementLine:
-	    if ([theEvent modifierFlags] & NSAlternateKeyMask)
-	      {
-		_hitPart = NSScrollerIncrementPage;
-	      }
-	    /* Fall through to next case */
-	  case NSScrollerIncrementPage:
-	    theCell = (_isHorizontal ? rightCell : downCell);
-	    break;
-
-	  case NSScrollerDecrementLine:
-	    if ([theEvent modifierFlags] & NSAlternateKeyMask)
-	      {
-		_hitPart = NSScrollerDecrementPage;
-	      }
-	    /* Fall through to next case */
-	  case NSScrollerDecrementPage:
-	    theCell = (_isHorizontal ? leftCell : upCell);
-	    break;
-
-	  default:
-	    theCell = nil;
-	    break;
-	}
-
-      if (theCell)
-	{
-	  [theCell highlight: YES withFrame: rect inView: self];
-	  [_window flushWindow];
-
-	  NSDebugLog (@"tracking cell %x", theCell);
-
-	  shouldReturn = [theCell trackMouse: theEvent
-				      inRect: rect
-				      ofView: self
-				untilMouseUp: YES];
-
-	  [theCell highlight: NO withFrame: rect inView: self];
-	  [_window flushWindow];
-	}
-
-      if (shouldReturn)
+      case NSScrollerIncrementLine:
+        if ([theEvent modifierFlags] & NSAlternateKeyMask)
+	  {
+	    _hitPart = NSScrollerIncrementPage;
+	  }
+	/* Fall through to next case */
+      case NSScrollerIncrementPage:
+	theCell = (_isHorizontal ? rightCell : downCell);
 	break;
 
-      theEvent = [theApp nextEventMatchingMask: eventMask
-				     untilDate: [NSDate distantFuture]
-					inMode: NSEventTrackingRunLoopMode
-				       dequeue: YES];
+      case NSScrollerDecrementLine:
+	if ([theEvent modifierFlags] & NSAlternateKeyMask)
+	  {
+	    _hitPart = NSScrollerDecrementPage;
+	  }
+	/* Fall through to next case */
+      case NSScrollerDecrementPage:
+	theCell = (_isHorizontal ? leftCell : upCell);
+	break;
+
+      default:
+	theCell = nil;
+	break;
     }
-  while ([theEvent type] != NSLeftMouseUp);
+
+  /*
+   * If we don't find a cell this has been all for naught, but we 
+   * shouldn't ever be in that situation.
+   */
+  if (theCell)
+    {
+      [theCell highlight: YES withFrame: rect inView: self];
+      [_window flushWindow];
+
+      NSDebugLog (@"tracking cell %x", theCell);
+
+      /*
+       * The "tracking" in this method actually takes place within 
+       * NSCell's trackMouse: method. 
+       */
+      [theCell trackMouse: theEvent
+		   inRect: rect
+		   ofView: self
+	     untilMouseUp: YES];
+
+      [theCell highlight: NO withFrame: rect inView: self];
+      [_window flushWindow];
+    }
+  [self unlockFocus];
 
   NSDebugLog (@"return from trackScrollButtons");
 }
@@ -752,7 +835,7 @@ static NSColor *scrollBarColor = nil;
 		? NSScrollerIncrementLine : NSScrollerDecrementLine)];
   id theCell = nil;
 
-  NSDebugLog (@"position of %s cell is (%f, %f)",
+  NSDebugLLog (@"NSScroller", @"position of %s cell is (%f, %f)",
 	(whichButton == NSScrollerIncrementArrow ? "increment" : "decrement"),
 	rect.origin.x, rect.origin.y);
 
@@ -766,7 +849,8 @@ static NSColor *scrollBarColor = nil;
 	break;
     }
 
-  [theCell drawWithFrame: rect inView: self];
+  [theCell setHighlighted: flag];
+  [theCell drawWithFrame: rect  inView: self];
 }
 
 - (void) drawKnob
@@ -783,7 +867,7 @@ static NSColor *scrollBarColor = nil;
       rect = [self rectForPart: NSScrollerKnobSlot];
     }
 
-  [scrollBarColor set];
+    [scrollBarColor set];
   NSRectFill (rect);
 }
 
@@ -813,7 +897,7 @@ static NSColor *scrollBarColor = nil;
   float width, height;
   float buttonsSize = 2 * buttonsWidth + 2;
   NSUsableScrollerParts usableParts;
-										  /*
+  /*
    * If the scroller is disabled then the scroller buttons and the
    * knob are not displayed at all.
    */
@@ -860,7 +944,7 @@ static NSColor *scrollBarColor = nil;
 
 	  /* calc the slot Height */
 	  slotHeight = height - (_arrowsPosition == NSScrollerArrowsNone
-	    ?  0 : buttonsSize);
+				 ?  0 : buttonsSize);
 	  knobHeight = _knobProportion * slotHeight;
 	  knobHeight = (float)floor(knobHeight);
 	  if (knobHeight < buttonsWidth)
@@ -872,7 +956,8 @@ static NSColor *scrollBarColor = nil;
 
 	  /* calc actual position */
 	  y += knobPosition + (_arrowsPosition == NSScrollerArrowsMaxEnd
-	    || _arrowsPosition == NSScrollerArrowsNone ?  0 : buttonsSize);
+			       || _arrowsPosition == NSScrollerArrowsNone 
+			       ?  0 : buttonsSize);
 	  height = knobHeight;
 	  width = buttonsWidth;
 	  break;
