@@ -5,8 +5,8 @@
 
    Copyright (C) 2001, 1996 Free Software Foundation, Inc.
 
-   Author: Gregory John Casamento <borgheron@yahoo.com>
-   Date: 2001
+   Author: Gregory John Casamento <greg_casamento@yahoo.com>
+   Date: 2001,2003
 
    Author: Scott Christley <scottc@net-community.com>
    Date: 1996
@@ -29,7 +29,7 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSArray.h>
 #include <Foundation/NSSet.h>
 #include <Foundation/NSDictionary.h>
@@ -37,16 +37,21 @@
 #include <Foundation/NSBundle.h>
 #include <Foundation/NSUtilities.h>
 #include <Foundation/NSValue.h>
-#include <AppKit/NSNibLoading.h>
-#include <AppKit/NSSpellChecker.h>
-#include <AppKit/NSSpellServer.h>
-#include <AppKit/GSServicesManager.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSTextField.h>
-#include <AppKit/NSMatrix.h>
-#include <AppKit/NSBrowser.h>
-#include <AppKit/NSBrowserCell.h>
-#include <AppKit/NSPopUpButton.h>
+#include "AppKit/NSNibLoading.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSSpellChecker.h"
+#include "AppKit/NSSpellServer.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSTextField.h"
+#include "AppKit/NSMatrix.h"
+#include "AppKit/NSBrowser.h"
+#include "AppKit/NSBrowserCell.h"
+#include "AppKit/NSPopUpButton.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSWindow.h"
+#include "AppKit/NSPanel.h"
+#include "GSGuiPrivate.h"
+#include "GNUstepGUI/GSServicesManager.h"
 
 // prototype for function to create name for server
 NSString *GSSpellServerName(NSString *checkerDictionary, NSString *language);
@@ -111,9 +116,8 @@ NSString *GSSpellServerName(NSString *checkerDictionary, NSString *language);
   if (proxy == nil)
     {
       NSRunAlertPanel(nil,
-	[NSString stringWithFormat:
-	    @"Failed to contact spell checker for language '%@'", language],
-	@"Continue", nil, nil);
+	@"Failed to contact spell checker for language '%@'",
+	@"Continue", nil, nil, language);
     }
   else
     {
@@ -252,9 +256,10 @@ static int __documentTag = 0;
 - init
 {
   NSArray *userLanguages = [NSUserDefaults userLanguages];  
+  NSString *panel = nil;
 
   // Set the language to the default for the user.
-  [super init];
+  self = [super init];
   _language = [userLanguages objectAtIndex: 0];
   _wrapFlag = NO;
   _position = 0;
@@ -264,11 +269,13 @@ static int __documentTag = 0;
   _ignoredWords = [NSMutableDictionary new];
 
   // Load the gmodel file
-  if(![NSBundle loadNibFile: @"SpellPanel.gmodel"
+  panel = [GSGuiBundle() pathForResource: @"GSSpellPanel" ofType: @"gorm"
+		      inDirectory: nil];
+  if(![NSBundle loadNibFile: panel
 	  externalNameTable: [NSDictionary dictionaryWithObject: self forKey: @"NSOwner"]
 	           withZone: [self zone]])
     {
-      NSLog(@"NIB file load failed for SpellPanel");
+      NSLog(@"Model file load failed for GSSpellPanel");
       return nil;
     }
 
@@ -280,6 +287,7 @@ static int __documentTag = 0;
 {
   RELEASE(_ignoredWords);
   RELEASE(_serverProxy);
+  [super dealloc];
 }
 
 //
