@@ -2,7 +2,7 @@
 
    <abstract>Allows multiple views to share a region in a window</abstract>
 
-   Copyright (C) 1996, 1998, 1999, 2000, 2001 Free Software Foundation, Inc.
+   Copyright (C) 1996, 1998, 1999, 2000, 2001, 2004 Free Software Foundation, Inc.
 
    Author: Robert Vasvari <vrobi@ddrummer.com>
    Date: Jul 1998
@@ -32,24 +32,32 @@
 */
 
 #include <math.h>
-#include <Foundation/NSCoder.h>
 #include <Foundation/NSArray.h>
+#include <Foundation/NSCoder.h>
+#include <Foundation/NSDebug.h>
 #include <Foundation/NSException.h>
 #include <Foundation/NSGeometry.h>
-#include <Foundation/NSString.h>
 #include <Foundation/NSNotification.h>
 #include <Foundation/NSRunLoop.h>
-#include <Foundation/NSDebug.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSEvent.h>
-#include <AppKit/NSSplitView.h>
+#include <Foundation/NSString.h>
 
-#include "AppKit/DPSOperators.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSColor.h"
+#include "AppKit/NSEvent.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSSplitView.h"
+#include "AppKit/NSWindow.h"
+
+static NSNotificationCenter *nc = nil;
 
 @implementation NSSplitView
 
++ (void) initialize
+{
+  nc = [NSNotificationCenter defaultCenter];
+}
+
 /*
  * Instance methods
  */
@@ -75,6 +83,13 @@
   RELEASE(_backgroundColor);
   RELEASE(_dividerColor);
   RELEASE(_dimpleImage);
+
+  if (_delegate != nil)
+    {
+      [nc removeObserver: _delegate  name: nil  object: self];
+      _delegate = nil;
+    }
+
   [super dealloc];
 }
 
@@ -101,11 +116,9 @@
   unsigned	int eventMask = NSLeftMouseUpMask | NSLeftMouseDraggedMask;
   /* YES if delegate implements splitView:constrainSplitPosition:ofSubviewAt:*/
   BOOL          delegateConstrains = NO;
-  NSNotificationCenter	*nc = [NSNotificationCenter defaultCenter];
-  SEL           constrainSel = 
-    @selector(splitView:constrainSplitPosition:ofSubviewAt:);
-  IMP           constrainImp;
-    
+  SEL constrainSel = @selector(splitView:constrainSplitPosition:ofSubviewAt:);
+  typedef float (*floatIMP)(id, SEL, id, float, int);
+  floatIMP constrainImp = NULL;
 
   /*  if there are less the two subviews, there is nothing to do */
   if (count < 2)
@@ -124,17 +137,19 @@
       r = [v frame];
       /* if the click is inside of a subview, return.  this should
 	 happen only if a subview has leaked a mouse down to next
-	 responder */
+	 responder
+	 */
       if (NSPointInRect(p, r))
 	{
-	  NSLog(@"NSSplitView got mouseDown in subview area");
+	  NSDebugLLog(@"NSSplitView",
+	    @"NSSplitView got mouseDown in subview area");
 	  return;
 	}
       if (_isVertical == NO)
         {
 	  if (NSMinY(r) >= p.y)
             {
-	      offset = i;
+	      offset = i - 1;
 
 	      /* get the enclosing rect for the two views */
 	      if (prev != nil)
@@ -163,7 +178,7 @@
         {
 	  if (NSMinX(r) >= p.x)
             {
-	      offset = i;
+	      offset = i - 1;
 
 	      /* get the enclosing rect for the two views */
 	      if (prev != nil)
@@ -275,7 +290,7 @@
 
   if (delegateConstrains)
     {
-      constrainImp = [_delegate methodForSelector: constrainSel];
+      constrainImp = (floatIMP)[_delegate methodForSelector: constrainSel];
     }
 
   // user is moving the knob loop until left mouse up
@@ -286,21 +301,13 @@
 	{
 	  if (_isVertical)
 	    {
-	      /*
-	      p.x = [_delegate splitView: self  constrainSplitPosition: p.x
-			       ofSubviewAt: offset];
-	      */
-	      (*constrainImp)(_delegate, constrainSel, self,
-			      p.x, offset);
+	      p.x = (*constrainImp)(_delegate, constrainSel, self,
+				    p.x, offset);
 	    }
 	  else
 	    {
-	      /*
-	      p.y = [_delegate splitView: self  constrainSplitPosition: p.y
-			       ofSubviewAt: offset];
-	      */
-	      (*constrainImp)(_delegate, constrainSel, self,
-			      p.y, offset);
+	      p.y = (*constrainImp)(_delegate, constrainSel, self,
+				    p.y, offset);
 	    }
 	}
       
@@ -332,8 +339,8 @@
 	}
       if (NSEqualRects(r, oldRect) == NO)
 	{
-	  NSDebugLog(@"drawing divider at x: %d, y: %d, w: %d, h: %d\n",
-	    (int)NSMinX(r), (int)NSMinY(r), (int)NSWidth(r), (int)NSHeight(r));
+	  NSDebugLLog(@"NSSplitView", @"drawing divider at %@\n",
+		      NSStringFromRect(r));
 	  [_dividerColor set];
 	
 	    
@@ -345,13 +352,7 @@
 		      || (NSMaxY(r) < NSMinY(oldRect)))
 		    // the two rects don't intersect
 		    {
-		      DPScompositerect(GSCurrentContext(), 
-				       NSMinX(oldRect), 
-				       NSMinY(oldRect), 
-				       NSWidth(oldRect), 
-				       NSHeight(oldRect), 
-				       NSCompositeHighlight);
-		      //		      NSHighlightRect(oldRect);
+		      NSHighlightRect(oldRect);
 		      NSHighlightRect(r);
 		    }
 		  else
@@ -373,14 +374,7 @@
 			  onRect.size.height = 
 			    NSMaxY(r) - NSMaxY(oldRect);
 
-			  //			  NSHighlightRect(onRect);
-			  DPScompositerect(GSCurrentContext(), 
-					   NSMinX(onRect), 
-					   NSMinY(onRect), 
-					   NSWidth(onRect), 
-					   NSHeight(onRect), 
-					   NSCompositeHighlight);
-
+			  NSHighlightRect(onRect);
 			  NSHighlightRect(offRect);
 
 			  //NSLog(@"on : %@", NSStringFromRect(onRect));
@@ -404,14 +398,7 @@
 			  onRect.size.height = 
 			    NSMinY(oldRect) - NSMinY(r);
 
-			  //			  NSHighlightRect(onRect);
-			  DPScompositerect(GSCurrentContext(), 
-					   NSMinX(onRect), 
-					   NSMinY(onRect), 
-					   NSWidth(onRect), 
-					   NSHeight(onRect), 
-					   NSCompositeHighlight);
-
+			  NSHighlightRect(onRect);
 			  NSHighlightRect(offRect);
 
 			  //NSLog(@"on : %@", NSStringFromRect(onRect));
@@ -427,12 +414,7 @@
 		      || (NSMaxX(r) < NSMinX(oldRect)))
 		    // the two rects don't intersect
 		    {
-		      DPScompositerect(GSCurrentContext(), 
-				       NSMinX(oldRect), 
-				       NSMinY(oldRect), 
-				       NSWidth(oldRect), 
-				       NSHeight(oldRect), 
-				       NSCompositeHighlight);
+		      NSHighlightRect (oldRect);
 		      NSHighlightRect(r);
 		    }
 		  else
@@ -454,13 +436,7 @@
 			  onRect.size.width = 
 			    NSMaxX(r) - NSMaxX(oldRect);
 
-			  DPScompositerect(GSCurrentContext(), 
-					   NSMinX(onRect), 
-					   NSMinY(onRect), 
-					   NSWidth(onRect), 
-					   NSHeight(onRect), 
-					   NSCompositeHighlight);
-			  //			  NSHighlightRect(onRect);
+			  NSHighlightRect(onRect);
 			  NSHighlightRect(offRect);
 			}
 		      else
@@ -479,14 +455,7 @@
 			  onRect.size.width = 
 			    NSMinX(oldRect) - NSMinX(r);
 
-			  DPScompositerect(GSCurrentContext(), 
-					   NSMinX(onRect), 
-					   NSMinY(onRect), 
-					   NSWidth(onRect), 
-					   NSHeight(onRect), 
-					   NSCompositeHighlight);
-
-			  //			  NSHighlightRect(onRect);
+			  NSHighlightRect(onRect);
 			  NSHighlightRect(offRect);
 			}
 		    }
@@ -497,6 +466,7 @@
 	    {
 	      NSHighlightRect(r);
 	    }
+	  [_window flushWindow];
 	  /*
 	  if (lit == YES)
 	    {
@@ -574,7 +544,7 @@
 	}
     }
   [prev setFrame: r];
-  NSDebugLog(@"drawing PREV at x: %d, y: %d, w: %d, h: %d\n",
+  NSDebugLLog(@"NSSplitView", @"drawing PREV at x: %d, y: %d, w: %d, h: %d\n",
 	     (int)NSMinX(r),(int)NSMinY(r),(int)NSWidth(r),(int)NSHeight(r));
 
   r1 = [v frame];
@@ -605,7 +575,7 @@
 	}
     }
   [v setFrame: r1];
-  NSDebugLog(@"drawing LAST at x: %d, y: %d, w: %d, h: %d\n",
+  NSDebugLLog(@"NSSplitView", @"drawing LAST at x: %d, y: %d, w: %d, h: %d\n",
 	     (int)NSMinX(r1),(int)NSMinY(r1),(int)NSWidth(r1),
 	     (int)NSHeight(r1));
 
@@ -636,7 +606,6 @@
 
 - (void) adjustSubviews
 {
-  NSNotificationCenter	*nc = [NSNotificationCenter defaultCenter];
   NSArray	*subs = [self subviews];
   unsigned	count = [subs count];
   NSView	*views[count];
@@ -837,13 +806,13 @@ static inline NSPoint centerSizeInRect(NSSize innerSize, NSRect outerRect)
   return YES;
 }
 
-- (void) resizeWithOldSuperviewSize: (NSSize)oldSize
+- (void) resizeSubviewsWithOldSize: (NSSize) oldSize
 {
-  [super resizeWithOldSuperviewSize: oldSize];
   [self _adjustSubviews: oldSize];
   [_window invalidateCursorRectsForView: self];
 }
 
+
 - (void) displayIfNeededInRectIgnoringOpacity: (NSRect)aRect
 {
   if (_window == nil)
@@ -853,8 +822,8 @@ static inline NSPoint centerSizeInRect(NSSize innerSize, NSRect outerRect)
   
   if (_never_displayed_before == YES)
     {
-      [self _adjustSubviews: _frame.size];
       _never_displayed_before = NO;
+      [self _adjustSubviews: _frame.size];
     }
 
   [super displayIfNeededInRectIgnoringOpacity: aRect];
@@ -869,8 +838,8 @@ static inline NSPoint centerSizeInRect(NSSize innerSize, NSRect outerRect)
   
   if (_never_displayed_before == YES)
     {
-      [self _adjustSubviews: _frame.size];
       _never_displayed_before = NO;
+      [self _adjustSubviews: _frame.size];
     }
 
   [super displayRectIgnoringOpacity: aRect];
@@ -883,8 +852,6 @@ static inline NSPoint centerSizeInRect(NSSize innerSize, NSRect outerRect)
 
 - (void) setDelegate: (id)anObject
 {
-  NSNotificationCenter* nc = [NSNotificationCenter defaultCenter];
-
   if (_delegate)
     {
       [nc removeObserver: _delegate  name: nil  object: self];
@@ -931,32 +898,34 @@ static inline NSPoint centerSizeInRect(NSSize innerSize, NSRect outerRect)
 - (id) initWithCoder: (NSCoder *)aDecoder
 {
   self = [super initWithCoder: aDecoder];
+ 
+  if ([aDecoder allowsKeyedCoding])
+    {
+      if ([aDecoder containsValueForKey: @"NSIsVertical"])
+        {
+	  [self setVertical: [aDecoder decodeBoolForKey: @"NSIsVertical"]];
+	}
+    }
+  else
+    {
+      // Decode objects that we don't retain.
+      [self setDelegate: [aDecoder decodeObject]];
 
-  /*
-   *	Decode objects that we don't retain.
-   */
-  [self setDelegate: [aDecoder decodeObject]];
+      // Decode objects that we do retain.
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_dimpleImage];
+      if (_dimpleImage == nil)
+	ASSIGN(_dimpleImage, [NSImage imageNamed: @"common_Dimple.tiff"]);
 
-  /*
-   *	Decode objects that we do retain.
-   */
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_dimpleImage];
-  if (_dimpleImage == nil)
-    ASSIGN(_dimpleImage, [NSImage imageNamed: @"common_Dimple.tiff"]);
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_backgroundColor];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_dividerColor];
 
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_backgroundColor];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_dividerColor];
+      // Decode non-object data.
+      [aDecoder decodeValueOfObjCType: @encode(int) at: &_draggedBarWidth];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_isVertical];
 
-  /*
-   *	Decode non-object data.
-   */
-  [aDecoder decodeValueOfObjCType: @encode(int) at: &_draggedBarWidth];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_isVertical];
-
-  /*
-   *
-   */
-  _dividerWidth = [self dividerThickness];
+      _dividerWidth = [self dividerThickness];
+      _never_displayed_before = YES;
+    }
 
   return self;
 }
