@@ -25,16 +25,16 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <AppKit/NSColor.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSForm.h>
-#include <AppKit/NSMatrix.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSTabView.h>
-#include <AppKit/NSTabViewItem.h>
-#include <AppKit/PSOperators.h>
+#include "AppKit/NSColor.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSForm.h"
+#include "AppKit/NSMatrix.h"
+#include "AppKit/NSWindow.h"
+#include "AppKit/NSTabView.h"
+#include "AppKit/NSTabViewItem.h"
+#include "AppKit/PSOperators.h"
 
 @implementation NSTabView
 
@@ -147,7 +147,7 @@
 
 - (void) selectNextTabViewItem: (id)sender
 {
-  if (_selected_item + 1 < [_items count])
+  if ((unsigned)(_selected_item + 1) < [_items count])
     [self selectTabViewItemAtIndex: _selected_item+1];
 }
 
@@ -159,7 +159,7 @@
 
 - (NSTabViewItem*) selectedTabViewItem
 {
-  if (_selected_item == NSNotFound)
+  if (_selected_item == NSNotFound || [_items count] == 0)
     return nil;
   return [_items objectAtIndex: _selected_item];
 }
@@ -227,6 +227,12 @@
     [self selectTabViewItem: [_items objectAtIndex: index]];
 }
 
+- (void) selectTabViewItemWithIdentifier:(id)identifier 
+{
+  int index = [self indexOfTabViewItemWithIdentifier: identifier];
+  [self selectTabViewItemAtIndex: index];
+}
+
 - (void) takeSelectedTabViewItemFromSender: (id)sender
 {
   int	index = -1;
@@ -436,7 +442,7 @@
 	      r.size.height = 15 + iFlex;
 
 	      DPSsetlinewidth(ctxt,1);
-	      DPSsetgray(ctxt,1);
+	      DPSsetgray(ctxt, NSWhite);
 	      DPSmoveto(ctxt, r.origin.x, r.origin.y-1);
 	      DPSrlineto(ctxt, r.size.width, 0);
 	      DPSstroke(ctxt);      
@@ -494,7 +500,7 @@
 	      iFlex = 0;
 
 	      DPSsetlinewidth(ctxt,1);
-	      DPSsetgray(ctxt,1);
+	      DPSsetgray(ctxt, NSWhite);
 	      DPSmoveto(ctxt, r.origin.x, r.origin.y - 1);
 	      DPSrlineto(ctxt, r.size.width, 0);
 	      DPSstroke(ctxt);      
@@ -562,7 +568,7 @@
 	      r.size.height = 15;
 	      
 	      DPSsetlinewidth(ctxt,1);
-	      DPSsetgray(ctxt,1);
+	      DPSsetgray(ctxt, NSWhite);
 	      DPSmoveto(ctxt, r.origin.x, r.origin.y+16);
 	      DPSrlineto(ctxt, r.size.width, 0);
 	      DPSstroke(ctxt);      
@@ -610,7 +616,7 @@
 	      r.size.height = 15;
 
 	      DPSsetlinewidth(ctxt,1);
-	      DPSsetgray(ctxt,1);
+	      DPSsetgray(ctxt, NSWhite);
 	      DPSmoveto(ctxt, r.origin.x, r.origin.y+16);
 	      DPSrlineto(ctxt, r.size.width, 0);
 	      DPSstroke(ctxt);      
@@ -677,24 +683,30 @@
 }
 
 
-- (NSControlSize)controlSize
+- (NSControlSize) controlSize
 {
   // FIXME
   return NSRegularControlSize;
 }
 
-- (void)setControlSize:(NSControlSize)size
+/**
+ * Not implemented.
+ */
+- (void) setControlSize: (NSControlSize)controlSize
 {
   // FIXME 
 }
 
-- (NSControlTint)controlTint
+- (NSControlTint) controlTint
 {
   // FIXME
   return NSDefaultControlTint;
 }
 
-- (void)setControlTint:(NSControlTint)tint
+/**
+ * Not implemented.
+ */
+- (void) setControlTint: (NSControlTint)controlTint
 {
   // FIXME 
 }
@@ -716,16 +728,56 @@
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  [super initWithCoder: aDecoder];
-
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_items];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_font];
-  [aDecoder decodeValueOfObjCType: @encode(NSTabViewType) at: &_type];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_draws_background];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_truncated_label];
-  _delegate = [aDecoder decodeObject];
-  [aDecoder decodeValueOfObjCType: "i" at: &_selected_item];
+  self = [super initWithCoder: aDecoder];
 
+  if ([aDecoder allowsKeyedCoding])
+    {
+      if ([aDecoder containsValueForKey: @"NSAllowTruncatedLabels"])
+        {
+	  [self setAllowsTruncatedLabels: [aDecoder decodeBoolForKey: 
+							@"NSAllowTruncatedLabels"]];
+	}
+      if ([aDecoder containsValueForKey: @"NSDrawsBackground"])
+        {
+	  [self setDrawsBackground: [aDecoder decodeBoolForKey: 
+						  @"NSDrawsBackground"]];
+	}
+      if ([aDecoder containsValueForKey: @"NSFont"])
+        {
+	  [self setFont: [aDecoder decodeObjectForKey: @"NSFont"]];
+	}
+      if ([aDecoder containsValueForKey: @"NSTabViewItems"])
+        {
+	  NSArray *items = [aDecoder decodeObjectForKey: @"NSTabViewItems"];
+	  NSEnumerator *enumerator = [items objectEnumerator];
+	  NSTabViewItem *item;
+	  
+	  while ((item = [enumerator nextObject]) != nil)
+	    {
+	      [self addTabViewItem: item];
+	    }
+	}
+      if ([aDecoder containsValueForKey: @"NSSelectedTabViewItem"])
+        {
+	  [self selectTabViewItem: [aDecoder decodeObjectForKey: 
+						 @"NSSelectedTabViewItem"]];
+	}
+      if ([aDecoder containsValueForKey: @"NSTvFlags"])
+        {
+	    //int flags = [aDecoder decodeObjectForKey: @"NSTvFlags"]];
+	}
+    }
+  else
+    {
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_items];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_font];
+      [aDecoder decodeValueOfObjCType: @encode(NSTabViewType) at: &_type];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_draws_background];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &_truncated_label];
+      _delegate = [aDecoder decodeObject];
+      [aDecoder decodeValueOfObjCType: "i" at: &_selected_item];
+      _selected = [_items objectAtIndex: _selected_item];
+    }
   return self;
 }
 @end
