@@ -29,18 +29,18 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 
 #include <Foundation/NSFormatter.h>
 #include <Foundation/NSNotification.h>
 #include <Foundation/NSString.h>
+#include <Foundation/NSValue.h>
 
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSCursor.h>
-#include <AppKit/NSTextField.h>
-#include <AppKit/NSTextFieldCell.h>
-#include <AppKit/NSWindow.h>
-
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSCursor.h"
+#include "AppKit/NSTextField.h"
+#include "AppKit/NSTextFieldCell.h"
+#include "AppKit/NSWindow.h"
 static NSNotificationCenter *nc;
 
 /*
@@ -94,6 +94,16 @@ static Class textFieldCellClass;
   return self;
 }
 
+- (void) dealloc
+{
+  if (_delegate != nil)
+    {
+      [nc removeObserver: _delegate  name: nil  object: self];
+      _delegate = nil;
+    }
+
+  [super dealloc];
+}
 
 //
 // Setting User Access to Text
@@ -160,12 +170,12 @@ static Class textFieldCellClass;
 //
 - (id) nextText
 {
-  return _nextKeyView;
+  return [self nextKeyView];
 }
 
 - (id) previousText
 {
-  return _previousKeyView;
+  return [self previousKeyView];
 }
 
 - (void) setNextText: (id)anObject
@@ -274,50 +284,87 @@ static Class textFieldCellClass;
 //
 - (void) mouseDown: (NSEvent*)theEvent
 {
-  NSText *t;
-  
   if ([self isSelectable] == NO)
-    return;
-
-  // This could happen if someone pressed the mouse 
-  // on the borders
-  if (_text_object)
-    return;
-
-  t = [_window fieldEditor: YES forObject: self];
-
-  if ([t superview] != nil)
     {
-      if ([t resignFirstResponder] == NO)
-	{
-	  if ([_window makeFirstResponder: _window] == NO)
-	    return;
-	}
+      [super mouseDown: theEvent];
+      return;
     }
 
+  /* NB: If we're receiving this click from the NSWindow, we expect
+     _text_object to never be nil here, since NSWindow makes the
+     NSTextField the first responder (which invokes its
+     -becomeFirstResponder:, which invokes its -selectText:, which, if
+     it is selectable, sets up the _text_object, then makes the
+     _text_object first responder!) before calling its -mouseDown:.
+     Only the first click should go via here; further clicks will be
+     sent directly by the NSWindow to the _text_object.
+  */
+  if (_text_object)
+    {
+      [_text_object mouseDown: theEvent];
+      return;
+    }
+  else
+    {
+      /* I suppose you could get here in subclasses which override
+       * -becomeFirstResponder not to select the text.  In that case,
+       * we set up the _text_object manually to start editing here.
+       */
 
-  //  [NSCursor hide];
-  
-  _text_object = [_cell setUpFieldEditorAttributes: t];
-  [_cell editWithFrame: _bounds
-	inView: self
-	editor: _text_object
-	delegate: self
-	event: theEvent];
+      /* Make sure we have first responder status when we start edit.
+       * This does nothing if we are already first responder; but
+       * (important!) it implicitly should free the fieldEditor if it
+       * was in use by another control.
+       */
+      if ([_window makeFirstResponder: self])
+	{
+	  NSText *t = [_window fieldEditor: YES forObject: self];
+	  
+	  if ([t superview] != nil)
+	    {
+	      /* Can't take the field editor ... give up.  */
+	      return;
+	    }
+
+	  _text_object = [_cell setUpFieldEditorAttributes: t];
+	  [_cell editWithFrame: _bounds
+		 inView: self
+		 editor: _text_object
+		 delegate: self
+		 event: theEvent];
+	}
+    }
 }
+
 - (BOOL) acceptsFirstMouse: (NSEvent *)aEvent
 {
-  return YES;
+  return [self isEditable];
 }
 
 - (BOOL) acceptsFirstResponder
 {
-  return [self isSelectable];
+  // we do not accept first responder if there is already a 
+  // _text_object, else it would make the _text_object resign
+  // and end editing
+  return (_text_object == nil) && [self isSelectable];
 }
 
 - (BOOL) becomeFirstResponder
 {
-  return [self isSelectable];
+  if ([self acceptsFirstResponder])
+    {
+      [self selectText: self];
+      return YES;
+    }
+  else
+    {
+      return NO;
+    }
+}
+
+-(BOOL) needsPanelToBecomeKey
+{
+  return [self isEditable];
 }
 
 - (BOOL) abortEditing
@@ -452,14 +499,16 @@ static Class textFieldCellClass;
 
   [self validateEditing];
 
+  [_cell endEditing: [aNotification object]];
+
+  _text_object = nil;
+
   d = [NSDictionary dictionaryWithObject: [aNotification object] 
 		    forKey: @"NSFieldEditor"];
   [nc postNotificationName: NSControlTextDidEndEditingNotification
       object: self
       userInfo: d];
 
-  [_cell endEditing: [aNotification object]];
-
   textMovement = [[aNotification userInfo] objectForKey: @"NSTextMovement"];
   if (textMovement)
     {
@@ -486,7 +535,6 @@ static Class textFieldCellClass;
 	  break;
 	}
     }
-  _text_object = nil;
 }
 
 - (BOOL) textShouldBeginEditing: (NSText *)textObject
@@ -502,9 +550,9 @@ static Class textFieldCellClass;
     return YES;
 }
 
-- (BOOL) textShouldEndEditing: (NSText *)aTextObject
+- (BOOL) textShouldEndEditing: (NSText*)textObject
 {
-  if ([_cell isEntryAcceptable: [aTextObject text]] == NO)
+  if ([_cell isEntryAcceptable: [textObject text]] == NO)
     {
       [self sendAction: _error_action to: [self target]];
       return NO;
@@ -513,16 +561,14 @@ static Class textFieldCellClass;
   if ([_delegate respondsToSelector: 
 		   @selector(control:textShouldEndEditing:)])
     {
-      if ([_delegate control: self 
-		     textShouldEndEditing: aTextObject] == NO)
+      if ([_delegate control: self textShouldEndEditing: textObject] == NO)
 	{
 	  NSBeep ();
 	  return NO;
 	}
     }
 
-  if ([_delegate respondsToSelector: 
-		   @selector(control:isValidObject:)] == YES)
+  if ([_delegate respondsToSelector: @selector(control:isValidObject:)] == YES)
     {
       NSFormatter *formatter;
       id newObjectValue;
@@ -530,12 +576,13 @@ static Class textFieldCellClass;
       formatter = [_cell formatter];
       
       if ([formatter getObjectValue: &newObjectValue 
-		     forString: [_text_object text] 
-		     errorDescription: NULL] == YES)
+			  forString: [_text_object text] 
+		   errorDescription: NULL] == YES)
 	{
-	  if ([_delegate control: self
-			 isValidObject: newObjectValue] == NO)
-	    return NO;
+	  if ([_delegate control: self isValidObject: newObjectValue] == NO)
+	    {
+	      return NO;
+	    }
 	}
     }
 
@@ -584,10 +631,16 @@ static Class textFieldCellClass;
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  [super initWithCoder: aDecoder];
+  self = [super initWithCoder: aDecoder];
 
-  [self setDelegate: [aDecoder decodeObject]];
-  [aDecoder decodeValueOfObjCType: @encode(SEL) at: &_error_action];
+  if ([aDecoder allowsKeyedCoding])
+    {
+    }
+  else
+    {
+      [self setDelegate: [aDecoder decodeObject]];
+      [aDecoder decodeValueOfObjCType: @encode(SEL) at: &_error_action];
+    }
   _text_object = nil;
 
   return self;
