@@ -27,12 +27,14 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
-#include <AppKit/NSColor.h>
-#include <AppKit/NSFont.h>
-#include <AppKit/NSGraphics.h>
-#include <AppKit/NSTextFieldCell.h>
-#include <AppKit/NSText.h>
+#include "config.h"
+#include <Foundation/NSNotification.h>
+#include "AppKit/NSColor.h"
+#include "AppKit/NSControl.h"
+#include "AppKit/NSFont.h"
+#include "AppKit/NSGraphics.h"
+#include "AppKit/NSTextFieldCell.h"
+#include "AppKit/NSText.h"
 
 static NSColor	*bgCol;
 static NSColor	*txtCol;
@@ -119,6 +121,9 @@ static NSColor	*txtCol;
 {
   ASSIGN (_background_color, aColor);
   _textfieldcell_is_opaque = [self _isOpaque];
+  if (_control_view)
+    if ([_control_view isKindOfClass: [NSControl class]])
+      [(NSControl *)_control_view updateCell: self];
 }
 
 - (NSColor *) backgroundColor
@@ -130,6 +135,9 @@ static NSColor	*txtCol;
 {
   _textfieldcell_draws_background = flag;
   _textfieldcell_is_opaque = [self _isOpaque];
+  if (_control_view)
+    if ([_control_view isKindOfClass: [NSControl class]])
+      [(NSControl *)_control_view updateCell: self];
 }
 
 - (BOOL) drawsBackground
@@ -140,6 +148,9 @@ static NSColor	*txtCol;
 - (void) setTextColor: (NSColor *)aColor
 {
   ASSIGN (_text_color, aColor);
+  if (_control_view)
+    if ([_control_view isKindOfClass: [NSControl class]])
+      [(NSControl *)_control_view updateCell: self];
 }
 
 - (NSColor *) textColor
@@ -160,10 +171,8 @@ static NSColor	*txtCol;
 {
   if (_textfieldcell_draws_background)
     {
-      [controlView lockFocus];
       [_background_color set];
       NSRectFill ([self drawingRectForBounds: cellFrame]);
-      [controlView unlockFocus];
     }
   [super drawInteriorWithFrame: cellFrame inView: controlView];
 }
@@ -189,14 +198,28 @@ static NSColor	*txtCol;
 
 - (id) initWithCoder: (NSCoder*)aDecoder
 {
-  BOOL tmp;
-  [super initWithCoder: aDecoder];
+  self = [super initWithCoder: aDecoder];
 
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_background_color];
-  [aDecoder decodeValueOfObjCType: @encode(id) at: &_text_color];
-  [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &tmp];
-  _textfieldcell_draws_background = tmp;
-  _textfieldcell_is_opaque = [self _isOpaque];
+  if ([aDecoder allowsKeyedCoding])
+    {
+      [self setBackgroundColor: [aDecoder decodeObjectForKey: @"NSBackgroundColor"]];
+      [self setTextColor: [aDecoder decodeObjectForKey: @"NSTextColor"]];
+      if ([aDecoder containsValueForKey: @"NSDrawsBackground"])
+        {
+	  [self setDrawsBackground: [aDecoder decodeBoolForKey: 
+						  @"NSDrawsBackground"]];
+	}
+    }
+  else
+    {
+      BOOL tmp;
+
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_background_color];
+      [aDecoder decodeValueOfObjCType: @encode(id) at: &_text_color];
+      [aDecoder decodeValueOfObjCType: @encode(BOOL) at: &tmp];
+      _textfieldcell_draws_background = tmp;
+      _textfieldcell_is_opaque = [self _isOpaque];
+    }
 
   return self;
 }
