@@ -29,7 +29,7 @@
    59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */ 
 
-#include <gnustep/gui/config.h>
+#include "config.h"
 #include <Foundation/NSBundle.h>
 #include <Foundation/NSDictionary.h>
 #include <Foundation/NSHost.h>
@@ -44,16 +44,27 @@
 #include <Foundation/NSConnection.h>
 #include <Foundation/NSDebug.h>
 #include <Foundation/NSURL.h>
-#include <AppKit/NSWorkspace.h>
-#include <AppKit/NSApplication.h>
-#include <AppKit/NSImage.h>
-#include <AppKit/NSView.h>
-#include <AppKit/NSWindow.h>
-#include <AppKit/NSScreen.h>
-#include <AppKit/GSServicesManager.h>
+#include <Foundation/NSValue.h>
+#include "AppKit/NSWorkspace.h"
+#include "AppKit/NSApplication.h"
+#include "AppKit/NSImage.h"
+#include "AppKit/NSView.h"
+#include "AppKit/NSPanel.h"
+#include "AppKit/NSWindow.h"
+#include "AppKit/NSScreen.h"
+#include "GNUstepGUI/GSServicesManager.h"
+#include "GNUstepGUI/GSDisplayServer.h"
 
 #define PosixExecutePermission	(0111)
 
+static NSImage	*folderImage = nil;
+static NSImage	*homeImage = nil;
+static NSImage	*multipleFiles = nil;
+static NSImage	*rootImage = nil;
+static NSImage	*unknownApplication = nil;
+static NSImage	*unknownTool = nil;
+
+
 static NSString	*GSWorkspaceNotification = @"GSWorkspaceNotification";
 
 @interface	_GSWorkspaceCenter: NSNotificationCenter
@@ -172,9 +183,9 @@ static NSString	*GSWorkspaceNotification = @"GSWorkspaceNotification";
 - (NSImage*) _extIconForApp: (NSString*)appName info: (NSDictionary*)extInfo;
 - (NSImage*) _getImageWithName: (NSString*)name
 		     alternate: (NSString*)alternate;
-- (NSImage*) folderImage;
 - (NSImage*) unknownFiletypeImage;
-- (NSImage*) rootImage;
+- (NSImage*) _saveImageFor: (NSString*)iconPath;
+- (NSString*) thumbnailForFile: (NSString *)file;
 - (NSImage*) _iconForExtension: (NSString*)ext;
 - (BOOL) _extension: (NSString*)ext
                role: (NSString*)role
@@ -188,6 +199,100 @@ static NSString	*GSWorkspaceNotification = @"GSWorkspaceNotification";
 
 @end
 
+
+/**
+ * <p>The NSWorkspace class gathers together a large number of capabilities
+ * needed for workspace management.
+ * </p>
+ * <p>The make_services tool examines all applications (anything with a
+ * .app, .debug, or .profile suffix) in the system, local, and user Apps
+ * directories, and caches information about the services each app
+ * provides (extracted from the Info-gnustep.plist file in each application).
+ * </p>
+ * <p>In addition to the cache of services information, it builds a cache of
+ * information about all known applications (including information about file
+ * types they handle).
+ * </p>
+ * <p>NSWorkspace reads the cache and uses it to determine which application
+ * to use to open a document and which icon to use to represent that document.
+ * </p>
+ * <p>The NSWorkspace API has been extended to provide methods for
+ * finding/setting the preferred icon/application for a particular file
+ * type. NSWorkspace will use the 'best' icon/application available.
+ * </p>
+ * <p>To determine the executable to launch, if there was an
+ * Info-gnustep.plist/Info.plist in the app wrapper and it had an
+ * NSExecutable field - it uses that name. Otherwise, it tries to use
+ * the name of the app - eg. foo.app/foo <br />
+ * The executable is launched by NSTask, which handles the addition
+ * of machine/os/library path components as necessary.
+ * </p>
+ * <p>To determine the icon for a file, it uses the value from the
+ * cache of icons for the file extension, or use an 'unknown' icon.
+ * </p>
+ * <p>To determine the icon for a folder, if the folder has a '.app',
+ * '.debug' or '.profile' extension - the Info-gnustep.plist file
+ * is examined for an 'NSIcon' value and NSWorkspace tries to use that.
+ * If there is no value specified - it tries foo.app/foo.tiff' or
+ * 'foo.app/.dir.tiff'
+ * </p>
+ * <p>If the folder was not an application wrapper, it just tries
+ * the .dir.tiff file.
+ * </p>
+ * <p>If no icon was available, it uses a default folder icon or a
+ * special icon for the root directory.
+ * </p>
+ * <p>The information about what file types an app can handle needs
+ * to be stored in Info-gnustep.plist in an array keyed on the name
+ * <em>NSTypes</em>, within which each value is a dictionary.<br />
+ * </p>
+ * <p>In the NSTypes fields, NSWorkspace uses NSIcon (the icon to use
+ * for the type) NSUnixExtensions (a list of file extensions
+ * corresponding to the type) and NSRole (what the app can do with
+ * documents of this type <em>Editor</em>, <em>Viewer</em>,
+ * or <em>None</em>). In the AppList cache, make_services
+ * generates a dictionary, keyed by file extension, whose values are
+ * the dictionaries containing the NSTypes dictionaries of each
+ * of the apps that handle the extension.  The NSWorkspace class
+ * makes use of this cache at runtime.
+ * </p>
+ * <p>If the Info-gnustep.plist of an application says that it
+ * can open files with a particular extension, then when NSWorkspace
+ * is asked to open such a file it will attempt to send an
+ * -application:openFile: message to the application (which must be
+ * handled by the applications delegate).  If the application is not
+ * running, NSWorkspace will instead attempt to launch the application
+ * passing the filename to open after a '-GSFilePath' flag
+ * in the command line arguments.  For a GNUstep application, the
+ * application will recognize this and invoke the -application:openFile:
+ * method passing it the file name.
+ * </p>
+ * <p>This command line argument mechanism provides a way for non-gnustep
+ * applications to be used to open files simply by provideing a wrapper
+ * for them containing the appropriate Info-gnustep.plist.<br />
+ * For instance - you could set up xv.app to contain a shellscript 'xv'
+ * that would start the real xv binary passing it a file to open if the
+ * '-GSFilePath' argument was given. The Info-gnustep.plist file could look
+ * like this:
+ * </p>
+ * <example>
+ * 
+ * {
+ *   NSExecutable = "xv";
+ *   NSIcon = "xv.tiff";
+ *   NSTypes = (
+ *     {
+ *       NSIcon = "tiff.tiff";
+ *       NSUnixExtensions = ( tiff, tif );
+ *     },
+ *     {
+ *       NSIcon = "xbm.tiff";
+ *       NSUnixExtensions = ( xbm );
+ *     }
+ *   );
+ * }
+ * </example>
+ */
 @implementation	NSWorkspace
 
 static NSWorkspace		*sharedWorkspace = nil;
@@ -223,43 +328,56 @@ static NSString			*_rootPath = @"/";
 	}
 
       beenHere = YES;
-      service = [[NSSearchPathForDirectoriesInDomains(
-	  NSUserDirectory, NSUserDomainMask, YES) objectAtIndex: 0]
-		    stringByAppendingPathComponent: @"Services"];
 
-      /*
-       *	Load file extension preferences.
-       */
-      extPrefPath = [service
-			stringByAppendingPathComponent: @".GNUstepExtPrefs"];
-      RETAIN(extPrefPath);
-      if ([mgr isReadableFileAtPath: extPrefPath] == YES)
+      NS_DURING
 	{
-	  data = [NSData dataWithContentsOfFile: extPrefPath];
-	  if (data)
+	  service = [[NSSearchPathForDirectoriesInDomains(NSUserDirectory, 
+							  NSUserDomainMask, 
+							  YES) 
+							 objectAtIndex: 0]
+		      stringByAppendingPathComponent: @"Library/Services"];
+	  
+	  /*
+	   *	Load file extension preferences.
+	   */
+	  extPrefPath = [service
+			  stringByAppendingPathComponent: @".GNUstepExtPrefs"];
+	  RETAIN(extPrefPath);
+	  if ([mgr isReadableFileAtPath: extPrefPath] == YES)
 	    {
-	      dict = [NSDeserializer deserializePropertyListFromData: data
-						   mutableContainers: NO];
-	      extPreferences = RETAIN(dict);
+	      data = [NSData dataWithContentsOfFile: extPrefPath];
+	      if (data)
+		{
+		  dict = [NSDeserializer deserializePropertyListFromData: data
+					 mutableContainers: NO];
+		  extPreferences = RETAIN(dict);
+		}
+	    }
+	  
+	  /*
+	   *	Load cached application information.
+	   */
+	  appListPath = [service
+			  stringByAppendingPathComponent: @".GNUstepAppList"];
+	  RETAIN(appListPath);
+	  if ([mgr isReadableFileAtPath: appListPath] == YES)
+	    {
+	      data = [NSData dataWithContentsOfFile: appListPath];
+	      if (data)
+		{
+		  dict = [NSDeserializer deserializePropertyListFromData: data
+					 mutableContainers: NO];
+		  applications = RETAIN(dict);
+		}
 	    }
 	}
-
-      /*
-       *	Load cached application information.
-       */
-      appListPath = [service
-			stringByAppendingPathComponent: @".GNUstepAppList"];
-      RETAIN(appListPath);
-      if ([mgr isReadableFileAtPath: appListPath] == YES)
+      NS_HANDLER
 	{
-	  data = [NSData dataWithContentsOfFile: appListPath];
-	  if (data)
-	    {
-	      dict = [NSDeserializer deserializePropertyListFromData: data
-						   mutableContainers: NO];
-	      applications = RETAIN(dict);
-	    }
+	  [gnustep_global_lock unlock];
+	  [localException raise];
 	}
+      NS_ENDHANDLER
+
       [gnustep_global_lock unlock];
     }
 }
@@ -312,7 +430,7 @@ static NSString			*_rootPath = @"/";
     selector: @selector(noteUserDefaultsChanged)
     name: NSUserDefaultsDidChangeNotification
     object: nil];
-  
+
   _workspaceCenter = [_GSWorkspaceCenter new];
   _iconMap = [NSMutableDictionary new];
   _launched = [NSMutableDictionary new];
@@ -469,14 +587,18 @@ static NSString			*_rootPath = @"/";
   id app = [self _workspaceApplication];
 
   if (app == nil)
-    return NO;
+    {
+      return NO;
+    }
   else
-    // Send the request on to the Workspace application
-    return [app performFileOperation: operation
-		       source: source
-		  destination: destination
-		        files: files
-			  tag: tag];
+    {
+      // Send the request on to the Workspace application
+      return [app performFileOperation: operation
+				source: source
+			   destination: destination
+				 files: files
+				   tag: tag];
+    }
 }
 
 - (BOOL) selectFile: (NSString*)fullPath
@@ -485,31 +607,84 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   id app = [self _workspaceApplication];
 
   if (app == nil)
-    return NO;
+    {
+      return NO;
+    }
   else
-    // Send the request on to the Workspace application
-    return [app selectFile: fullPath
-		inFileViewerRootedAtPath: rootFullpath];
+    {
+      // Send the request on to the Workspace application
+      return [app selectFile: fullPath
+	inFileViewerRootedAtPath: rootFullpath];
+    }
 }
 
-/*
- * Requesting Information about Files
+/**
+ * Given an application name, return the full path for that application.<br />
+ * This method looks for the application in standard locations, and if not
+ * found there, according to MacOS-X documentation, returns nil.<br />
+ * If the supplied application name is an absolute path, returns that path
+ * irrespective of whether such an application exists or not.  This is
+ * <em>not</em> the docmented debavior in the MacOS-X documentation, but is
+ * the MacOS-X implemented behavior.<br />
+ * If the appName has an extension, it is used, otherwise in GNUstep
+ * the standard app, debug, and profile extensions * are tried.<br />
  */
 - (NSString*) fullPathForApplication: (NSString*)appName
 {
-  NSString      *last = [appName lastPathComponent];
+  NSString	*base;
+  NSString	*path;
+  NSString	*ext;
 
-  if ([appName isEqual: last])
+  if ([appName length] == 0)
     {
-      NSString  *ext = [appName pathExtension];
-
-      if (ext == nil)
-        {
-          appName = [appName stringByAppendingPathExtension: @"app"];
-        }
-      return [applications objectForKey: appName];
+      return nil;
     }
-  return nil;
+  if ([[appName lastPathComponent] isEqual: appName] == NO)
+    {
+      if ([appName isAbsolutePath] == YES)
+	{
+	  return appName;		// MacOS-X implementation behavior.
+	}
+      /*
+       * Relative path ... get standarized absolute path
+       */
+      path = [[NSFileManager defaultManager] currentDirectoryPath];
+      appName = [path stringByAppendingPathComponent: appName];
+      appName = [appName stringByStandardizingPath];
+    }
+  base = [appName stringByDeletingLastPathComponent];
+  appName = [appName lastPathComponent];
+  ext = [appName pathExtension];
+  if ([ext length] == 0) // no extension, let's find one
+    {
+      path = [appName stringByAppendingPathExtension: @"app"];
+      path = [applications objectForKey: path];
+      if (path == nil)
+	{
+	  path = [appName stringByAppendingPathExtension: @"debug"];
+	  path = [applications objectForKey: path];
+	}
+      if (path == nil)
+	{
+	  path = [appName stringByAppendingPathExtension: @"profile"];
+	  path = [applications objectForKey: path];
+	}
+    }
+  else
+    {
+      path = [applications objectForKey: appName];
+    }
+
+  /*
+   * If the original name included a path, check that the located name
+   * matches it.  If it doesn't we return nil as MacOS-X does.
+   */
+  if ([base length] > 0
+    && [base isEqual: [path stringByDeletingLastPathComponent]] == NO)
+    {
+      path = nil;
+    }
+  return path;
 }
 
 - (BOOL) getFileSystemInfoForPath: (NSString*)fullPath
@@ -523,6 +698,31 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   return NO;
 }
 
+/**
+ * This method gets information about the file at fullPath and
+ * returns YES on success, NO if the named file could not be
+ * found.<br />
+ * On success, the name of the preferred application for opening
+ * the file is returned in *appName, or nil if there is no known
+ * application to open it.<br />
+ * The returned value in *type describes the file using one of
+ * the following constants.
+ * <deflist>
+ *   <term>NSPlainFileType</term>
+ *   <desc>
+ *     A plain file or a directory that some application
+ *     claims to be able to open like a file.
+ *   </desc>
+ *   <term>NSDirectoryFileType</term>
+ *   <desc>An untyped directory</desc>
+ *   <term>NSApplicationFileType</term>
+ *   <desc>A GNUstep application</desc>
+ *   <term>NSFilesystemFileType</term>
+ *   <desc>A file system mount point</desc>
+ *   <term>NSShellCommandFileType</term>
+ *   <desc>Executable shell command</desc>
+ * </deflist>
+ */
 - (BOOL) getInfoForFile: (NSString*)fullPath
 	    application: (NSString **)appName
 		   type: (NSString **)type
@@ -533,59 +733,63 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   NSString	*extension = [fullPath pathExtension];
 
   attributes = [fm fileAttributesAtPath: fullPath traverseLink: YES];
+
   if (attributes != nil)
     {
+      *appName = [self getBestAppInRole: nil forExtension: extension];
       fileType = [attributes fileType];
       if ([fileType isEqualToString: NSFileTypeRegular])
-        {
-          if ([attributes filePosixPermissions] & PosixExecutePermission)
-            {
-              *type = NSShellCommandFileType;
-              *appName = nil;
-            }
-          else
-            {
-              *type = NSPlainFileType;
-              *appName = [self getBestAppInRole: nil forExtension: extension];
-            }
-        }
+	{
+	  if ([attributes filePosixPermissions] & PosixExecutePermission)
+	    {
+	      *type = NSShellCommandFileType;
+	    }
+	  else
+	    {
+	      *type = NSPlainFileType;
+	    }
+	}
       else if ([fileType isEqualToString: NSFileTypeDirectory])
-        {
-          if ([extension isEqualToString: @"app"]
+	{
+	  if ([extension isEqualToString: @"app"]
 	    || [extension isEqualToString: @"debug"]
 	    || [extension isEqualToString: @"profile"])
-            {
-              *type = NSApplicationFileType;
-              *appName = nil;
-            }
-          else if ([extension isEqualToString: @"bundle"])
-            {
-              *type = NSPlainFileType;
-              *appName = nil;
-            }
-          // the idea here is that if the parent directory's fileSystemNumber
-          // differs, this must be a filesystem mount point
-          else if ([[fm fileAttributesAtPath:
+	    {
+	      *type = NSApplicationFileType;
+	    }
+	  else if ([extension isEqualToString: @"bundle"])
+	    {
+	      *type = NSPlainFileType;
+	    }
+	  else if (*appName != nil && [extension length] > 0)
+	    {
+	      *type = NSPlainFileType;
+	    }
+	  /*
+	   * The idea here is that if the parent directory's
+	   * fileSystemNumber differs, this must be a filesystem
+	   * mount point.
+	   */
+	  else if ([[fm fileAttributesAtPath:
 	    [fullPath stringByDeletingLastPathComponent]
 	    traverseLink: YES] fileSystemNumber]
 	    != [attributes fileSystemNumber])
-            {
-              *type = NSFilesystemFileType;
-              *appName = nil;
-            }
-          else
-            {
-              *type = NSDirectoryFileType;
-              *appName = nil;
-            }
-        }
+	    {
+	      *type = NSFilesystemFileType;
+	    }
+	  else
+	    {
+	      *type = NSDirectoryFileType;
+	    }
+	}
       else
-        {
-          // this catches sockets, character special, block special, and
-          // unknown file types
-          *type = NSPlainFileType;
-          *appName = nil;
-        }
+	{
+	  /*
+	   * This catches sockets, character special, block special,
+	   * and unknown file types
+	   */
+	  *type = NSPlainFileType;
+	}
       return YES;
     }
   else
@@ -594,15 +798,16 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
     }
 }
 
-- (NSImage*) iconForFile: (NSString*)aPath
+- (NSImage*) iconForFile: (NSString*)fullPath
 {
   NSImage	*image = nil;
-  NSString	*pathExtension = [[aPath pathExtension] lowercaseString];
+  NSString	*pathExtension = [[fullPath pathExtension] lowercaseString];
 
-  if ([self isFilePackageAtPath: aPath])
+  if ([self isFilePackageAtPath: fullPath])
     {
       NSFileManager	*mgr = [NSFileManager defaultManager];
       NSString		*iconPath = nil;
+      BOOL		isApplication = NO;
       
       if ([pathExtension isEqualToString: @"app"]
 	|| [pathExtension isEqualToString: @"debug"]
@@ -610,7 +815,8 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	{
 	  NSBundle	*bundle;
 
-	  bundle = [NSBundle bundleWithPath: aPath];
+	  isApplication = YES;
+	  bundle = [NSBundle bundleWithPath: fullPath];
 	  iconPath = [[bundle infoDictionary] objectForKey: @"NSIcon"];
 	  if (iconPath && [iconPath isAbsolutePath] == NO)
 	    {
@@ -624,7 +830,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	       */
 	      if (iconPath == nil)
 		{
-		  iconPath = [aPath stringByAppendingPathComponent: file];
+		  iconPath = [fullPath stringByAppendingPathComponent: file];
 		  if ([mgr isReadableFileAtPath: iconPath] == NO)
 		    {
 		      iconPath = nil;
@@ -639,12 +845,17 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	    {
 	      NSString	*str;
 
-	      str = [[aPath lastPathComponent] stringByDeletingPathExtension];
-	      iconPath = [aPath stringByAppendingPathComponent: str];
+	      str = [fullPath lastPathComponent];
+	      str = [str stringByDeletingPathExtension];
+	      iconPath = [fullPath stringByAppendingPathComponent: str];
 	      iconPath = [iconPath stringByAppendingPathExtension: @"tiff"];
 	      if ([mgr isReadableFileAtPath: iconPath] == NO)
 		{
 		  iconPath = nil;
+		  /*
+		   * Just use the appropriate icon for the path extension
+		   */
+		  image = [self _iconForExtension: pathExtension];
 		}
 	    }
 	}
@@ -655,7 +866,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
        */
       if (iconPath == nil)
 	{
-	  iconPath = [aPath stringByAppendingPathComponent: @".dir.tiff"];
+	  iconPath = [fullPath stringByAppendingPathComponent: @".dir.tiff"];
 	  if ([mgr isReadableFileAtPath: iconPath] == NO)
 	    {
 	      iconPath = nil;
@@ -664,16 +875,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 
       if (iconPath != nil)
 	{
-	  NS_DURING
-	    {
-	      image = [[NSImage alloc] initWithContentsOfFile: iconPath];
-	      AUTORELEASE(image);
-	    }
-	  NS_HANDLER
-	    {
-	      NSLog(@"BAD TIFF FILE '%@'", iconPath);
-	    }
-	  NS_ENDHANDLER
+	  image = [self _saveImageFor: iconPath];
 	}
 
       if (image == nil)
@@ -681,10 +883,36 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	  image = [self _iconForExtension: pathExtension];
 	  if (image == nil || image == [self unknownFiletypeImage])
 	    {
-	      if ([aPath isEqual: _rootPath])
-		image = [self rootImage];
+	      if ([fullPath isEqual: _rootPath])
+		{
+		  if (rootImage == nil)
+		    {
+		      rootImage = RETAIN([self _getImageWithName:
+			@"Root_PC.tiff"
+			alternate: @"common_Root_PC.tiff"]);
+		    }
+
+		  image = rootImage;
+		}
+	      else if ([fullPath isEqual: NSHomeDirectory ()])
+		{
+		  if (homeImage == nil)
+		    {
+		      homeImage = RETAIN([self _getImageWithName:
+			@"HomeDirectory.tiff"
+			alternate: @"common_HomeDirectory.tiff"]);
+		    }
+		  image = homeImage;
+		}
 	      else
-		image = [self folderImage];
+		{
+		  if (folderImage == nil)
+		    {
+		      folderImage = RETAIN([self _getImageWithName:
+			@"Folder.tiff" alternate: @"common_Folder.tiff"]);
+		    }
+		  image = folderImage;
+		}
 	    }
 	}
     }
@@ -692,7 +920,44 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
     {
       NSDebugLog(@"pathExtension is '%@'", pathExtension);
 
+      if ([[NSUserDefaults standardUserDefaults] boolForKey: 
+	      @"GSUseFreedesktopThumbnails"])
+        {
+	  /* This image will be 128x128 pixels as oposed to the 48x48 
+	     of other GNUstep icons or the 32x32 of the specification */  
+	  image = [self _saveImageFor: [self thumbnailForFile: fullPath]];
+	  if (image != nil)
+	    {
+	      return image;
+	    }
+	}
+
       image = [self _iconForExtension: pathExtension];
+      if (image == nil || image == [self unknownFiletypeImage])
+	{
+	  NSFileManager	*mgr;
+
+	  mgr = [NSFileManager defaultManager];
+	  if ([mgr isExecutableFileAtPath: fullPath] == YES)
+	    {
+	      NSDictionary	*attributes;
+	      NSString		*fileType;
+
+	      attributes = [mgr fileAttributesAtPath: fullPath
+					traverseLink: YES];
+	      fileType = [attributes objectForKey: NSFileType];
+	      if ([fileType isEqual: NSFileTypeRegular] == YES)
+		{
+		  if (unknownTool == nil)
+		    {
+		      unknownTool = RETAIN([self _getImageWithName:
+			@"UnknownTool.tiff" alternate:
+			@"common_UnknownTool.tiff"]);
+		    }
+		  image = unknownTool;
+		}
+	    }
+	}
     }
 
   if (image == nil)
@@ -705,8 +970,6 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 
 - (NSImage*) iconForFiles: (NSArray*)pathArray
 {
-  static NSImage *multipleFiles = nil;
-
   if ([pathArray count] == 1)
     {
       return [self iconForFile: [pathArray objectAtIndex: 0]];
@@ -740,7 +1003,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   return NO;
 }
 
-/*
+/**
  * Tracking Changes to the File System
  */
 - (BOOL) fileSystemChanged
@@ -762,7 +1025,8 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 }
 
 /**
- * Updates Registered Services and File Types
+ * Updates Registered Services, File Types, and other information about any
+ * applications installed in the standard locations.
  */
 - (void) findApplications
 {
@@ -820,8 +1084,9 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   [_iconMap removeAllObjects];
 }
 
-/*
- * Launching and Manipulating Applications	
+/**
+ * Instructs all the other running applications to hide themselves.
+ * <em>not yet implemented</em>
  */
 - (void) hideOtherApplications
 {
@@ -840,7 +1105,35 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 }
 
 /**
- * Launches the specified application (unless it is alreeady running)
+ * <p>Launches the specified application (unless it is already running).<br />
+ * If the autolaunch flag is yes, sets the autolaunch user default for the
+ * newly launched application, so that applications which understand the
+ * concept of being autolaunched at system startup time can modify their
+ * behavior appropriately.
+ * </p>
+ * <p>Sends an NSWorkspaceWillLaunchApplicationNotification before it
+ * actually attempts to launch the application (this is not sent if the
+ * application is already running).
+ * </p>
+ * <p>The application sends an NSWorkspaceDidlLaunchApplicationNotification
+ * on completion of launching.  This is not sent if the application is already
+ * running, or if it fails to complete its startup.
+ * </p>
+ * <p>Returns NO if the application cannot be launched (eg. it does not exist
+ * or the binary is not executable).
+ * </p>
+ * <p>Returns YES if the application was already running or of it was launched
+ * (this does not necessarily mean that the application succeeded in starting
+ * up fully).
+ * </p>
+ * <p>Once an application has fully started up, you should be able to connect
+ * to it using [NSConnection+rootProxyForConnectionWithRegisteredName:host:]
+ * passing the application name (normally the filesystem name excluding path
+ * and file extension) and an empty host name.  This will let you communicate
+ * with the the [NSApplication-delegate] of the launched application, and you
+ * can generally use this as a test of whether an application is running
+ * correctly.
+ * </p>
  */
 - (BOOL) launchApplication: (NSString*)appName
 		  showIcon: (BOOL)showIcon
@@ -851,12 +1144,56 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   app = [self _connectApplication: appName];
   if (app == nil)
     {
-      return [self _launchApplication: appName arguments: nil];
+      NSArray	*args = nil;
+
+      if (autolaunch == YES)
+	{
+	  args = [NSArray arrayWithObjects: @"-autolaunch", @"YES", nil];
+	}
+      return [self _launchApplication: appName arguments: args];
     }
 
   return YES;
 }
 
+/**
+ * Returns a description of the currently active application, containing
+ * the name (NSApplicationName), path (NSApplicationPath) and process
+ * identifier (NSApplicationProcessIdentifier).
+ */
+- (NSDictionary*) activeApplication
+{
+  NSProcessInfo *processInfo = [NSProcessInfo processInfo];
+
+  return [NSDictionary dictionaryWithObjectsAndKeys:
+    [processInfo processName], @"NSApplicationName",
+    [[NSBundle mainBundle] bundlePath], @"NSApplicationPath",
+    [NSNumber numberWithInt: [processInfo processIdentifier]], 
+		   @"NSApplicationProcessIdentifier",
+    nil];
+}
+
+/**
+ * Returns an array listing all the applications know to have been
+ * launched.
+ */
+- (NSArray*) launchedApplications
+{
+  NSDictionary		*dict;
+  NSString		*app;
+  NSMutableArray	*apps = [NSMutableArray array];
+  NSEnumerator		*enumerator = [_launched keyEnumerator];
+
+  while ((app = [enumerator nextObject]) != nil)
+    {
+      dict = [NSDictionary dictionaryWithObject: app
+					 forKey: @"NSApplicationName"];
+      [apps addObject: dict];
+    }
+  
+  return apps;
+}
+
 /*
  * Unmounting a Device	
  */
@@ -1002,7 +1339,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	       from: (NSPoint)fromPoint
 		 to: (NSPoint)toPoint
 {
-  [GSCurrentContext() _slideImage: image from: fromPoint to: toPoint];
+  [GSCurrentServer() slideImage: image from: fromPoint to: toPoint];
 }
 
 /*
@@ -1074,38 +1411,57 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 }
 
 /**
- * Returns the application bundle for the named application.
+ * Returns the application bundle for the named application. Accepts
+ * either a full path to an app or just the name. The extension (.app,
+ * .debug, .profile) is optional, but if provided it will be used.<br />
+ * Returns nil if the specified app does not exist as requested.
  */
 - (NSBundle*) bundleForApp: (NSString*)appName
 {
-  NSString	*path;
+  if ([appName length] == 0)
+    {
+      return nil;
+    }
+  if ([[appName lastPathComponent] isEqual: appName]) // it's a name
+    {
+      appName = [self fullPathForApplication: appName];
+    }
+  else
+    {
+      NSFileManager	*fm;
+      NSString		*ext;
+      BOOL		flag;
 
+      fm = [NSFileManager defaultManager];
+      ext = [appName pathExtension];
+      if ([ext length] == 0) // no extension, let's find one
+	{
+	  NSString	*path;
+
+	  path = [appName stringByAppendingPathExtension: @"app"];
+	  if ([fm fileExistsAtPath: path isDirectory: &flag] == NO
+	    || flag == NO)
+	    {
+	      path = [appName stringByAppendingPathExtension: @"debug"];
+	      if ([fm fileExistsAtPath: path isDirectory: &flag] == NO
+		|| flag == NO)
+		{
+		  path = [appName stringByAppendingPathExtension: @"profile"];
+		}
+	    }
+	  appName = path;
+	}
+      if ([fm fileExistsAtPath: appName isDirectory: &flag] == NO
+	|| flag == NO)
+	{
+	  appName = nil;
+	}
+    }
   if (appName == nil)
     {
       return nil;
     }
-  path = appName;
-  appName = [path lastPathComponent];
-  if ([appName isEqual: path])
-    {
-      path = [self fullPathForApplication: appName];
-      appName = [[path lastPathComponent] stringByDeletingPathExtension];
-    }
-  else if ([appName pathExtension] == nil)
-    {
-      path = [path stringByAppendingPathExtension: @"app"];
-    }
-  else
-    {
-      appName = [[path lastPathComponent] stringByDeletingPathExtension];
-    }
-
-  if (path == nil)
-    {
-      return nil;
-    }
-
-  return [NSBundle bundleWithPath: path];
+  return [NSBundle bundleWithPath: appName];
 }
 
 /**
@@ -1128,7 +1484,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
       iconPath = [[bundle bundlePath] stringByAppendingPathComponent: iconPath];
     }
   
-  return AUTORELEASE([[NSImage alloc] initWithContentsOfFile: iconPath]);
+  return [self _saveImageFor: iconPath];
 }
 
 /**
@@ -1172,7 +1528,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	}
       else
 	{
-	  path = [path stringByAppendingFormat: @"/%@", file];
+	  path = [path stringByAppendingPathComponent: file];
 	}
     }
 
@@ -1288,13 +1644,14 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	   */
 	  if (iconPath == nil)
 	    {
-	      iconPath = [[bundle bundlePath] stringByAppendingPathComponent: file];
+	      iconPath = [[bundle bundlePath]
+		stringByAppendingPathComponent: file];
 	    }
 	  file = iconPath;
 	}
       if ([[NSFileManager defaultManager] isReadableFileAtPath: file] == YES)
 	{
-	  return AUTORELEASE([[NSImage alloc] initWithContentsOfFile: file]);
+	  return [self _saveImageFor: file];
 	}
     }
   return nil;
@@ -1311,20 +1668,6 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   return image;
 }
 
-/** Returns the default icon to display for a directory */
-- (NSImage*) folderImage
-{
-  static NSImage *image = nil;
-
-  if (image == nil)
-    {
-      image = RETAIN([self _getImageWithName: @"Folder.tiff"
-				   alternate: @"common_Folder.tiff"]);
-    }
-
-  return image;
-}
-
 /** Returns the default icon to display for a directory */
 - (NSImage*) unknownFiletypeImage
 {
@@ -1339,18 +1682,53 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   return image;
 }
 
-/** Returns the default icon to display for a directory */
-- (NSImage*) rootImage
+/** Try to create the image in an exception handling context */
+- (NSImage*) _saveImageFor: (NSString*)iconPath
 {
-  static NSImage *image = nil;
+  NSImage *tmp = nil;
 
-  if (image == nil)
+  NS_DURING
     {
-      image = RETAIN([self _getImageWithName: @"Root_PC.tiff"
-				   alternate: @"common_Root_PC.tiff"]);
+      tmp = [[NSImage alloc] initWithContentsOfFile: iconPath];
+      if (tmp != nil)
+        {
+	  AUTORELEASE(tmp);
+	}
+    }
+  NS_HANDLER
+    {
+      NSLog(@"BAD TIFF FILE '%@'", iconPath);
+    }
+  NS_ENDHANDLER
+
+  return tmp;
+}
+
+/** Returns the freedesktop thumbnail file name for a given file name */
+- (NSString*) thumbnailForFile: (NSString *)file
+{
+  NSString *absolute;
+  NSString *digest;
+  NSString *thumbnail;
+
+  absolute = [[NSURL fileURLWithPath: [file stringByStandardizingPath]] 
+		 absoluteString];
+  // This compensates for a feature we have in NSURL, that is there to have 
+  // MacOSX compatibility.
+  if ([absolute hasPrefix:  @"file://localhost/"])
+    {
+      absolute = [@"file:///" stringByAppendingString: 
+		       [absolute substringWithRange: 
+				     NSMakeRange(17, [absolute length] - 17)]];
     }
 
-  return image;
+  // FIXME: Not sure which encoding to use here. 
+  digest = [[[[absolute dataUsingEncoding: NSASCIIStringEncoding]
+		 md5Digest] hexadecimalRepresentation] lowercaseString];
+  thumbnail = [@"~/.thumbnails/normal" stringByAppendingPathComponent: 
+		    [digest stringByAppendingPathExtension: @"png"]];
+
+  return [thumbnail stringByStandardizingPath];
 }
 
 - (NSImage*) _iconForExtension: (NSString*)ext
@@ -1379,8 +1757,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
       iconPath = [prefs objectForKey: @"Icon"];
       if (iconPath)
 	{
-	  icon = [[NSImage alloc] initWithContentsOfFile: iconPath];
-	  AUTORELEASE(icon);
+	  icon = [self _saveImageFor: iconPath];
 	}
 
       if (icon == nil && (extInfo = [self infoForExtension: ext]) != nil)
@@ -1425,7 +1802,22 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
        */
       if (icon == nil)
 	{
-	  icon = [self unknownFiletypeImage];
+	  if ([ext isEqualToString: @"app"] == YES
+	    || [ext isEqualToString: @"debug"] == YES
+	    || [ext isEqualToString: @"profile"] == YES)
+	    {
+	      if (unknownApplication == nil)
+		{
+		  unknownApplication = RETAIN([self _getImageWithName:
+		    @"UnknownApplication.tiff" alternate:
+		    @"common_UnknownApplication.tiff"]);
+		}
+	      icon = unknownApplication;
+	    }
+	  else
+	    {
+	      icon = [self unknownFiletypeImage];
+	    }
 	}
 
       /*
@@ -1643,9 +2035,11 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
        * App being launched, send
        * NSWorkspaceWillLaunchApplicationNotification
        */
-      userinfo = [NSDictionary dictionaryWithObject:
-	[[appName lastPathComponent] stringByDeletingPathExtension]
-	forKey: @"NSApplicationName"];
+      userinfo = [NSDictionary dictionaryWithObjectsAndKeys:
+	[[appName lastPathComponent] stringByDeletingPathExtension], 
+			       @"NSApplicationName",
+	appName, @"NSApplicationPath",
+	nil];
       [_workspaceCenter
 	postNotificationName: NSWorkspaceWillLaunchApplicationNotification
 	object: self
@@ -1670,10 +2064,9 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   NSString	*host;
   NSString	*port;
   NSDate	*when = nil;
-  BOOL		done = NO;
   id		app = nil;
 
-  while (done == NO)
+  while (app == nil)
     {
       host = [[NSUserDefaults standardUserDefaults] stringForKey: @"NSHost"];
       if (host == nil)
@@ -1690,7 +2083,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	      host = @"";
 	    }
 	}
-      port = [appName stringByDeletingPathExtension];
+      port = [[appName lastPathComponent] stringByDeletingPathExtension];
       /*
        *	Try to contact a running application.
        */
@@ -1706,51 +2099,53 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
 	}
       NS_ENDHANDLER
 
-      done = YES;
       if (app == nil)
 	{
 	  NSTask	*task = [_launched objectForKey: appName];
+	  NSDate	*limit;
 
-	  if (task != nil && [task isRunning] == YES)
+	  if (task == nil || [task isRunning] == NO)
 	    {
-	      if (when == nil)
+	      if (task != nil)	// Not running
 		{
-		  when = [[NSDate alloc] init];
-		  done = NO;
+		  [_launched removeObjectForKey: appName];
 		}
-	      else if ([when timeIntervalSinceNow] > 5.0)
+	      break;		// Need to launch the app
+	    }
+
+	  if (when == nil)
+	    {
+	      when = [[NSDate alloc] init];
+	    }
+	  else if ([when timeIntervalSinceNow] < -5.0)
+	    {
+	      int		result;
+
+	      DESTROY(when);
+	      result = NSRunAlertPanel(appName,
+		@"Application seems to have hung",
+		@"Continue", @"Terminate", @"Wait");
+
+	      if (result == NSAlertDefaultReturn)
 		{
-		  int		result;
-
-		  DESTROY(when);
-		  result = NSRunAlertPanel(appName,
-		    @"Application seems to have hung",
-		    @"Continue", @"Terminate", @"Wait");
-
-		  if (result == NSAlertDefaultReturn)
-		    {
-		      done = YES;
-		    }
-		  else if (result == NSAlertOtherReturn)
-		    {
-		      done = NO;
-		    }
-		  else
-		    {
-		      [task terminate];
-		      [_launched removeObjectForKey: appName];
-		      done = YES;
-		    }
+		  break;		// Finished without app
 		}
-	      if (done == NO)
+	      else if (result == NSAlertOtherReturn)
 		{
-		  NSDate	*limit;
-
-		  limit = [[NSDate alloc] initWithTimeIntervalSinceNow: 0.5];
-		  [[NSRunLoop currentRunLoop] runUntilDate: limit];
-		  RELEASE(limit);
+		  // Continue to wait for app startup.
+		}
+	      else
+		{
+		  [task terminate];
+		  [_launched removeObjectForKey: appName];
+		  break;		// Terminate hung app
 		}
 	    }
+
+	  // Give it another 0.5 of a second to start up.
+	  limit = [[NSDate alloc] initWithTimeIntervalSinceNow: 0.5];
+	  [[NSRunLoop currentRunLoop] runUntilDate: limit];
+	  RELEASE(limit);
 	}
     }
   TEST_RELEASE(when);
@@ -1767,7 +2162,7 @@ inFileViewerRootedAtPath: (NSString*)rootFullpath
   appName = [defs stringForKey: @"GSWorkspaceApplication"];
   if (appName == nil)
     {
-      appName = @"GSWorkspace";
+      appName = @"GWorkspace";
     }
 
   app = [self _connectApplication: appName];
