@@ -56,8 +56,8 @@
  * OF THIS SOFTWARE.
  */
 
-#include <gnustep/gui/config.h>
-#include <gnustep/gui/nsimage-tiff.h>
+#include "config.h"
+#include "nsimage-tiff.h"
 #include <Foundation/NSArray.h>
 #include <Foundation/NSString.h>
 #include <Foundation/NSUtilities.h>
@@ -268,7 +268,8 @@ NSTiffGetInfo(int imageNumber, TIFF* image)
   memset(info, 0, sizeof(NSTiffInfo));
   if (imageNumber >= 0)
     {
-      TIFFSetDirectory(image, imageNumber);
+      if (TIFFSetDirectory(image, imageNumber) == 0)
+	return NULL;
       info->imageNumber = imageNumber;
     }
   
@@ -307,7 +308,7 @@ NSTiffGetInfo(int imageNumber, TIFF* image)
 	default:
 	  TIFFError(TIFFFileName(image),
 		    "Missing needed \"PhotometricInterpretation\" tag");
-	  return (0);
+	  return NULL;
 	}
       TIFFError(TIFFFileName(image),
 		"No \"PhotometricInterpretation\" tag, assuming %s\n",
@@ -320,7 +321,6 @@ NSTiffGetInfo(int imageNumber, TIFF* image)
 #define READ_SCANLINE(sample)				\
   if (TIFFReadScanline(image, buf, row, sample) != 1)	\
     {							\
-      NSLog(@"Tiff bad data read on line %d", row);	\
       error = 1;					\
       break;						\
     }
@@ -335,7 +335,7 @@ int
 NSTiffRead(TIFF* image, NSTiffInfo* info, char* data)
 {
   int     i;
-  int     row, col;
+  unsigned int row, col;
   int     maxval;
   int	  error = 0;
   uint8* outP;
@@ -425,7 +425,7 @@ NSTiffRead(TIFF* image, NSTiffInfo* info, char* data)
 	}
       break;
     default:
-      NSLog(@"Tiff can't read photometric %d", info->photoInterp);
+      NSLog(@"Tiff: reading photometric %d not supported", info->photoInterp);
       error = 1;
       break;
     }
@@ -436,7 +436,6 @@ NSTiffRead(TIFF* image, NSTiffInfo* info, char* data)
 
 #define WRITE_SCANLINE(sample) \
 	if (TIFFWriteScanline(image, buf, row, sample) != 1) { \
-	    NSLog(@"Tiff bad data write on line %d\n", row); \
 	    error = 1; \
 	    break; \
 	}
@@ -447,7 +446,7 @@ NSTiffWrite(TIFF* image, NSTiffInfo* info, char* data)
   tdata_t	buf = (tdata_t)data;
   uint16        sample_info[2];
   int		i;
-  int		row;
+  unsigned int 	row;
   int           error = 0;
 
   TIFFSetField(image, TIFFTAG_IMAGEWIDTH, info->width);
@@ -519,13 +518,13 @@ NSTiffWrite(TIFF* image, NSTiffInfo* info, char* data)
 	break;
 
       default:
-	NSLog(@"Tiff can't write photometric %d for image %s", 
+	NSLog(@"Tiff: photometric %d for image %s not supported", 
 	      info->photoInterp, TIFFFileName(image));
 	return -1;
 	break;
     }
     
-  return 0;
+  return error;
 }
 
 /*------------------------------------------------------------------------*/
@@ -536,7 +535,7 @@ NSTiffWrite(TIFF* image, NSTiffInfo* info, char* data)
 static int
 CheckAndCorrectColormap(NSTiffColormap* map)
 {
-  register int i;
+  register unsigned int i;
 
   for (i = 0; i < map->size; i++)
     if ((map->red[i] > 255)||(map->green[i] > 255)||(map->blue[i] > 255))
